/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-device.c,v 1.9.8.2 2003/10/13 12:27:11 krishmnc Exp $
 *
 * iscsi-device.c
 *
 *     Determine if a Linux SCSI device is an iSCSI device.
 * 
 * May 2001, sferris
 * 
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

typedef unsigned char u8;

#include <scsi/scsi.h>
#include <scsi/scsi_ioctl.h>

/* if the device name on the command line exists and is an iSCSI device,
 * then output a description of the iSCSI target to stdout, and exit 0.
 * if the device name might be an iSCSI device, exit 1.
 * if the device name definately is not an iSCSI device, exit 2.
 */

#define ISCSI_PROC_DIR "/proc/scsi/iscsi"

int
main(int argc, char *argv[])
{
	char devname[PATH_MAX + 1];
	char realdev[PATH_MAX + 1];
	char buffer[4097];
	char *device = argv[1];
	int host = -1;
	int channel = -1;
	int id = -1;
	int lun = -1;
	int iscsi_channel = 0;
	int iscsi_id = 0;
	int iscsi_lun = 0;
	FILE *proc_hba;
	char *line;
	int fd;
	u_long arg[2];

	if (argc > 2) {
		fprintf(stderr,
			"iscsi-device: error, too many arguments, only 1 "
			"device name is allowed.\n");
	}
	if (argc != 2) {
		fprintf(stderr, "\nUsage: iscsi-device devicename\n\n");
		fprintf(stderr,
			"If the device name refers to an iSCSI device, "
			"iscsi-device\n");
		fprintf(stderr,
			"  returns 0 after writing the following information "
			"to stdout:\n");
		fprintf(stderr,
			"\tthe resolved devicename\n\tiscsi host number\n\t"
			"scsi channel number\n\ttarget id number\n\t"
			"LUN number\n");
		fprintf(stderr,
			"\ttarget ip address\n\ttarget port\n\tiscsi "
			"target id\n\tiSCSI target name\n\n");
		fprintf(stderr,
			"If the device name does may be an iSCSI device "
			"not yet ready, iscsi-device\n");
		fprintf(stderr,
			"  returns error code 1, and may write a message "
			"to stderr.\n\n\n");
		fprintf(stderr,
			"If the device name definately does not refer to "
			"an iSCSI device, iscsi-device\n");
		fprintf(stderr,
			"  returns an error code greater than 1, and may "
			"write a message to stderr.\n\n\n");
		exit(2);
	}

	/* get the (host,channel,id,lun) quadruple for this device, if possible          */
	if ((fd = open(device, O_RDONLY)) >= 0) {

		/* try to resolve any symlinks out of the name */
		if (realpath(device, realdev)) {
			realdev[sizeof (realdev) - 1] = '\0';
			device = realdev;
		}

		memset(&arg, 0, sizeof (arg));
		/* try to get the host,channel,id,port quadruple */
		if (ioctl(fd, SCSI_IOCTL_GET_IDLUN, &arg) == 0) {
			id = arg[0] & 0xff;
			lun = (arg[0] >> 8) & 0xff;
			channel = (arg[0] >> 16) & 0xff;
			host = (arg[0] >> 24) & 0xff;
#ifdef SCSI_IOCTL_GET_BUS_NUMBER
			/* 2.4 kernels give us all the info we need with that 
			 * ioctl. 2.2 kernels have another ioctl to get the 
			 * host number, and set the host above to something 
			 * useless to us.  If this 2nd ioctl fails, assume it's
			 * a 2.4 kernel, and use the host number we already have
			 */
			memset(&arg, 0, sizeof (arg));
			if (ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &arg) == 0) {
				host = arg[0];
			}
#endif
		} else {
			fprintf(stderr,
				"iscsi-device: ioctl SCSI_IOCTL_GET_IDLUN "
				"%s failed, %s\n", device, strerror(errno));
			exit(1);
		}
	} else {
		fprintf(stderr, "iscsi-device: couldn't open %s, %s\n", device,
			strerror(errno));
		exit(1);
	}

	/* see if the device node's host (HBA number) is an iSCSI HBA */
	sprintf(devname, "%s/%d", ISCSI_PROC_DIR, host);
	if ((proc_hba = fopen(devname, "r"))) {
		/* read every line, ignoring comments and blank lines */
		while ((line = fgets(buffer, sizeof (buffer), proc_hba))) {
			while (isspace(*line))
				line++;
			if (*line == '#' || *line == '\0')
				continue;

			if (sscanf
			    (line, "%d %d %d", &iscsi_channel, &iscsi_id,
			     &iscsi_lun) == 3) {
				/* For now we assume there is only one host, 
				 * so that the bus number maps directly to the 
				 * channel number.  This breaks if we ever 
				 * have multiple HBAs.
				 */
				if (channel == iscsi_channel &&
				    id == iscsi_id && lun == iscsi_lun) {
					/* found it. Print the device 
					 * and /proc line 
					 */
					printf("%s: %s\n", device, line);
					exit(0);
				}

			}
		}
		fclose(proc_hba);

		/* if /proc/scsi/iscsi has a file for the device node's host,
		 * then we know it's an iSCSI device, even if we can't get any
		 * additional info about it.
		 */
		exit(0);
	}

	/* didn't find a match */
	fprintf(stderr, "iscsi-device: %s is not an iSCSI device\n", device);
	exit(2);
}
