/*
 *
 * iscsi-ls.c - User level utility to display iscsi devices.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <sched.h>
#include <time.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include "iscsi-version.h"
#include "iscsi-common.h"
#include "iscsi-ioctl.h"

int
main(int argc, char **argv)
{
	int count = 1, c, target_id = 0, options = 0, ctlfd, rc, j, bus_id = 0;
	char version[256];
	struct target_info target_details;

#define BUS_SPECIFIC 	1
#define TARGET_SPECIFIC 2
#define TEXT_KEYS       4
#define LUN_INFO        8

	if (argc > 6) {
		printf
		    ("\nUsage: iscsi-ls [-b busId] [-t targetId] [-c] [-l]\n");
		printf("-b : Details regarding the specific bus\n");
		printf("-t : Details regarding the specific target\n");
		printf("-c : iSCSI target configuration parameters\n");
		printf("-l : Lun details available on the target\n");
		return -1;
	}

	while (count < argc) {
		if ((!strcmp(argv[count], "-c")) || (!strcmp(argv[count], "-l"))
		    || (!strcmp(argv[count], "-lc"))
		    || (!strcmp(argv[count], "-cl"))) {
			count++;
			if (count == argc)
				break;
			if ((!strcmp(argv[count], "-c"))
			    || (!strcmp(argv[count], "-l"))
			    || (!strcmp(argv[count], "-t"))
			    || (!strcmp(argv[count], "-lt"))
			    || (!strcmp(argv[count], "-ct"))) {
				break;
			} else {
				printf
				    ("\nUsage: iscsi-ls [-b busId] "
				     "[-t targetId] [-c] [-l]\n");
				printf
				    ("-b : Details regarding the "
				     "specific bus\n");
				printf
				    ("-t : Details regarding the specific "
				     "target\n");
				printf
				    ("-c : iSCSI target configuration "
				     "parameters\n");
				printf
				    ("-l : Lun details available on "
				     "the target\n");
				return -1;
			}
		}
		count++;
	}

	while ((c = getopt(argc, argv, "lct:b:")) >= 0) {
		switch (c) {
		case 'b':
			bus_id = atoi(optarg);
			options |= BUS_SPECIFIC;
			break;
		case 't':
			target_id = atoi(optarg);
			if (target_id < 0 || target_id >= ISCSI_MAX_TARGETS) {
				printf("Target ids should be between 0-%d\n",
				       ISCSI_MAX_TARGETS - 1);
				return -1;
			}
			options |= TARGET_SPECIFIC;
			break;
		case 'c':
			options |= TEXT_KEYS;
			break;
		case 'l':
			options |= LUN_INFO;
			break;
		default:
			printf
			    ("\nUsage: iscsi-ls [-b busId] [-t targetId] "
			     "[-c] [-l]\n");
			printf("-b : Details regarding the specific bus\n");
			printf("-t : Details regarding the specific target\n");
			printf("-c : iSCSI target configuration parameters\n");
			printf("-l : Lun details available on the target\n");
			return -1;
		}
	}

	ctlfd = open("/dev/iscsictl", O_RDONLY);

	if (ctlfd < 0) {
		printf
		    ("\n\n#######################################"
		     "########################################\n");
		printf("\niSCSI driver is not loaded\n");
		printf
		    ("\n\n#######################################"
		     "########################################\n\n\n");
		return -1;
	}
#if defined(DRIVER_INTERNAL_VERSION) && (DRIVER_INTERNAL_VERSION > 0)
	sprintf(version, "Cisco iSCSI Driver Version ... %d.%d.%d.%d%s (%s )",
		DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
		DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
		DRIVER_EXTRAVERSION, ISCSI_DATE);
#else
	sprintf(version, "Cisco iSCSI Driver Version ... %d.%d.%d%s (%s )",
		DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
		DRIVER_PATCH_VERSION, DRIVER_EXTRAVERSION, ISCSI_DATE);
#endif

	printf
	    ("**************************************"
	     "*****************************************\n");
	printf("\t%s\n", version);
	printf
	    ("**************************************"
	     "*****************************************\n");

	for (; target_id < ISCSI_MAX_TARGETS; target_id++) {
		struct timeval tv;
		struct iscsi_portal_list iscsi_ls_portal_list;
		/* Get details on this target_id */
		memset(&target_details, 0, sizeof (target_details));
		target_details.target_id = target_id;
		rc = ioctl(ctlfd, ISCSI_LS_TARGET_INFO, &target_details);
		if (rc) {
			fprintf(stderr, "The iscsi driver is inoperative\n");
			return -1;
		}

		if (options & BUS_SPECIFIC) {
			if (bus_id == target_details.channel) {
				options |= LUN_INFO;
			} else {
				printf("No bus with bus id %d\n", bus_id);
				printf
				    ("****************************************"
				     "*************************************\n");
				break;
			}
		}

		/* This target_id does not have a corresponding session.
		 * proceed with next.
		 */
		if (strcmp(target_details.target_name, "") == 0) {
			if (options & TARGET_SPECIFIC) {
				printf("No target with target id %d\n",
				       target_id);
				printf
				    ("****************************************"
				     "*************************************\n");
				break;
			} else {
				continue;
			}
		}

		printf("TARGET NAME             : %s\n",
		       target_details.target_name);
		printf("TARGET ALIAS            : %s\n",
		       target_details.target_alias);
		printf("HOST NO                 : %d \n",
		       target_details.host_no);
		printf("BUS NO                  : %d \n",
		       target_details.channel);
		printf("TARGET ID               : %d \n",
		       target_details.target_id);
		printf("DISCOVERY ADDRESS       : %u.%u.%u.%u:%d\n",
		       target_details.session_data.addr[0],
		       target_details.session_data.addr[1],
		       target_details.session_data.addr[2],
		       target_details.session_data.addr[3],
		       target_details.session_data.port);
		if (!target_details.session_data.session_alive) {
			gettimeofday(&tv, NULL);
			tv.tv_sec -=
			    target_details.session_data.session_drop_time;
			printf("SESSION DROPPED AT      : %s",
			       ctime((time_t *) & (tv.tv_sec)));
		} else {
			gettimeofday(&tv, NULL);
			tv.tv_sec -=
			    target_details.session_data.establishment_time;
			printf("SESSION ESTABLISHED AT  : %s",
			       ctime((time_t *) & (tv.tv_sec)));
		}
		printf("NO. OF PORTALS          : %d \n",
		       target_details.num_portals);
		memset(&iscsi_ls_portal_list, 0,
		       sizeof (iscsi_ls_portal_list));
		iscsi_ls_portal_list.target_id = target_details.target_id;
		iscsi_ls_portal_list.portals =
		    (struct iscsi_portal_info *) malloc(target_details.
							num_portals *
							sizeof (struct
								iscsi_portal_info));
		rc = ioctl(ctlfd, ISCSI_LS_PORTAL_INFO, &iscsi_ls_portal_list);
		if (rc) {
			fprintf(stderr, "The iscsi driver is inoperative\n");
			return -1;
		}

		for (j = 0; j < target_details.num_portals; j++) {
			printf("TARGET ADDRESS %d        : %u.%u.%u.%u:%d,%d\n",
			       (j + 1),
			       iscsi_ls_portal_list.portals[j].ip_address[0],
			       iscsi_ls_portal_list.portals[j].ip_address[1],
			       iscsi_ls_portal_list.portals[j].ip_address[2],
			       iscsi_ls_portal_list.portals[j].ip_address[3],
			       iscsi_ls_portal_list.portals[j].port,
			       iscsi_ls_portal_list.portals[j].tag);
		}
		free(iscsi_ls_portal_list.portals);
		printf
		    ("SESSION ID              : ISID %02x%02x%02x%02x%02x%02x "
		     "TSID %02x\n",
		     target_details.session_data.isid[0],
		     target_details.session_data.isid[1],
		     target_details.session_data.isid[2],
		     target_details.session_data.isid[3],
		     target_details.session_data.isid[4],
		     target_details.session_data.isid[5],
		     target_details.session_data.tsid);

		if (options & TEXT_KEYS) {
			printf("\nTARGET CONFIGURATION KEYS :\n");
			printf("------------------------- \n");
			printf("INITIAL R2T             : %s\n",
			       (target_details.session_data.
				InitialR2T) ? "Yes" : "No");
			printf("IMMEDIATE DATA          : %s\n",
			       (target_details.session_data.
				ImmediateData) ? "Yes" : "No");
			printf("HEADERDIGESTS           : %s\n",
			       (target_details.session_data.
				DataDigest) ? "CRC32" : "NOT SUPPORTED");
			printf("DATA DIGESTS            : %s\n",
			       (target_details.session_data.
				DataDigest) ? "CRC32" : "NOT SUPPORTED");
			printf("FIRSTBURSTLENGTH        : %d\n",
			       target_details.session_data.FirstBurstLength);
			printf("MAXBURSTLENGTH          : %d\n",
			       target_details.session_data.MaxBurstLength);
			printf("MAXRECVDATASEGMENTLENGTH: initiator - %d\n",
			       target_details.session_data.
			       MaxRecvDataSegmentLength);
			printf("                          target    - %d\n",
			       target_details.session_data.
			       MaxXmitDataSegmentLength);
			printf("LOGIN TIMEOUT           : %d\n",
			       target_details.session_data.login_timeout);
			printf("AUTH TIMEOUT            : %d\n",
			       target_details.session_data.auth_timeout);
			printf("ACTIVE TIMEOUT          : %d\n",
			       target_details.session_data.active_timeout);
			printf("IDLE TIMEOUT            : %d\n",
			       target_details.session_data.idle_timeout);
			printf("PING TIMEOUT            : %d\n",
			       target_details.session_data.ping_timeout);
		}

		if (options & LUN_INFO) {
			FILE *fp;
			char *rdbufp, rdbuf[1024];

			fp = fopen("/proc/scsi/scsi", "r");
			if (!fp) {
				perror("\nCannot open /proc/scsi/scsi\n");
				return -1;
			}

			printf("\nDEVICE DETAILS :\n");
			printf("-------------- \n");
			do {
				int target_id;
				rdbufp = fgets(rdbuf, sizeof (rdbuf), fp);
				if (rdbufp && (*rdbufp != '#')) {
					while (isspace(*rdbufp)) {
						rdbufp += 1;
					}
					if (!strncasecmp(rdbufp, "Attached", 8))
						continue;
					if (!strncasecmp
					    (rdbufp, "Host: scsi", 10)) {
						int channel_id;
						int host_id;
						char *cp = &rdbufp[10];
						sscanf(cp, "%d", &host_id);
						if (host_id !=
						    target_details.host_no) {
							continue;
						}

						cp = &rdbufp[22];
						sscanf(cp, "%d", &channel_id);
						if (channel_id !=
						    target_details.channel) {
							continue;
						}

						cp = &rdbufp[28];
						if (sscanf
						    (cp, "%d", &target_id)) {
							if (target_details.
							    target_id ==
							    target_id) {
								int lun_id;
								cp = &rdbufp
								    [36];
								sscanf(cp, "%d",
								       &lun_id);
								printf
								    ("LUN ID : "
								     "%d\n",
								     lun_id);
								rdbufp =
								    fgets(rdbuf,
									  sizeof
									  (rdbuf),
									  fp);
								printf("%s",
								       rdbufp);
								rdbufp =
								    fgets(rdbuf,
									  sizeof
									  (rdbuf),
									  fp);
								printf("%s",
								       rdbufp);
							}
						}
					}
				}
			} while (rdbufp);
			fclose(fp);
		}
		printf
		    ("************************************************"
		     "*******************************\n");
		if (options & TARGET_SPECIFIC) {
			break;
		}
	}
	close(ctlfd);
	return 0;
}
