/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-probe.c,v 1.21.6.12 2003/10/29 15:10:43 krishmnc Exp $
 *
 */

#include <linux/config.h>
#include <linux/version.h>

#include <linux/sched.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/dirent.h>
#include <linux/proc_fs.h>
#include <linux/blkdev.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/config.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/timer.h>
# include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <scsi/sg.h>

/* these are from $(TOPDIR)/drivers/scsi, not $(TOPDIR)/include */
#include <scsi.h>
#include <hosts.h>

#include "iscsi-common.h"
#include "iscsi-ioctl.h"
#include "iscsi.h"
#include "iscsi-kernel-event.h"

/* LUN probing needs to be serialized across all HBA's, to keep a 
 * somewhat sane ordering 
 */

DECLARE_MUTEX(iscsi_lun_probe_mutex);
spinlock_t iscsi_lun_probe_lock = SPIN_LOCK_UNLOCKED;
static struct list_head iscsi_lun_probe_order = {&iscsi_lun_probe_order,
						 &iscsi_lun_probe_order};
static struct iscsi_session *iscsi_currently_probing = NULL;
static volatile int iscsi_next_probe = 0;
volatile unsigned long iscsi_lun_probe_start = 0;

/* caller must hold iscsi_lun_probe_lock */
static int
enqueue_lun_probe(struct iscsi_session *session)
{
	struct iscsi_session *sess_ptr;
	if (!list_empty(&session->probe_order_link)) {
		DEBUG_INIT
		    ("iSCSI: session for bus_id id %d target_id %d "
		     "already queued for LUN probing\n",
		     session->iscsi_bus, session->target_id);
		return 0;
	}

	list_for_each_entry(sess_ptr,&iscsi_lun_probe_order,
			    probe_order_link) {
		if (session->probe_order < sess_ptr->probe_order) {
			list_add(&session->probe_order_link,
				 sess_ptr->probe_order_link.prev);
			break;
		} 
	}
	if (&(sess_ptr->probe_order_link) == &(iscsi_lun_probe_order)) {
		list_add_tail(&session->probe_order_link,
			      &iscsi_lun_probe_order);
	}
	
	return 1;
}

/* caller must hold iscsi_lun_probe_lock */
static void
dequeue_lun_probe(struct iscsi_session *session)
{
	if (iscsi_currently_probing == session) {
		/* the timer may have tried to start us 
		 * probing just before we gave up 
		 */
		iscsi_currently_probing = NULL;
	} else {
		list_del_init(&session->probe_order_link);
	}
}

static int
wait_for_probe_order(struct iscsi_session *session)
{
	spin_lock(&iscsi_lun_probe_lock);
	if ((iscsi_currently_probing == session)  
	    || !(list_empty(&session->probe_order_link))) {
		/* we're already probing or queued to be probed, 
		 * ignore the 2nd probe request 
		 */
		DEBUG_INIT
		    ("iSCSI: session for bus %d target %d to %s ignoring "
		     "duplicate probe request\n",
		     session->iscsi_bus, session->target_id, session->log_name);
		spin_unlock(&iscsi_lun_probe_lock);
		return 0;
	} else if ((iscsi_currently_probing == NULL)
		   && (session->probe_order <= iscsi_next_probe)) {
		/* if there's no LUN being probed, and our 
		 * probe_order can go now, start probing 
		 */
		DEBUG_INIT
		    ("iSCSI: session for bus %d target %d to %s, "
		     "probe_order %d <= next %d, not waiting\n",
		     session->iscsi_bus, session->target_id, session->log_name,
		     session->probe_order, iscsi_next_probe);
		iscsi_currently_probing = session;

		/* let the timer know another session became ready for 
		 * LUN probing. 
		 */
		iscsi_lun_probe_start = (jiffies + (3 * HZ));
		if (iscsi_lun_probe_start == 0)
			iscsi_lun_probe_start = 1;
		smp_mb();

		spin_unlock(&iscsi_lun_probe_lock);
		return 1;
	} else if (enqueue_lun_probe(session)) {
		/* otherwise queue up based on our probe order */

		/* tell the timer when to start the LUN probing, 
		 * to handle gaps in the probe_order 
		 */
		iscsi_lun_probe_start =
		    (jiffies + (3 * HZ)) ? (jiffies + (3 * HZ)) : 1;
		smp_mb();
		DEBUG_INIT
		    ("iSCSI: queued session for bus %d target %d for "
		     "LUN probing, probe_order %d, probe_start at %lu\n",
		     session->iscsi_bus, session->target_id,
		     session->probe_order, iscsi_lun_probe_start);

		spin_unlock(&iscsi_lun_probe_lock);

		/* and wait for either the timer or the currently 
		 * probing session to wake us up 
		 */
		if (down_interruptible(&session->probe_sem)) {
			logmsg(AS_INFO,
			       "iSCSI: session for bus %d target %d to %s "
			       "interrupted while waiting to probe LUNs\n",
			       session->iscsi_bus, session->target_id,
			       session->log_name);
			/* give up and take ourselves out of the lun 
			 * probing data structures 
			 */
			spin_lock(&iscsi_lun_probe_lock);
			dequeue_lun_probe(session);
			spin_unlock(&iscsi_lun_probe_lock);
			return 0;
		}

		/* give up if the session is terminating */
		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			logmsg(AS_INFO,
			       "iSCSI: session for bus %d target %d to %s "
			       "terminated while waiting to probe LUNs\n",
			       session->iscsi_bus, session->target_id,
			       session->log_name);
			/* give up and take ourselves out of the 
			 * lun probing data structures 
			 */
			spin_lock(&iscsi_lun_probe_lock);
			dequeue_lun_probe(session);
			spin_unlock(&iscsi_lun_probe_lock);
			return 0;
		}
#ifdef DEBUG
		/* we should be out of the queue, and in iscsi_currently_probing
		 */
		spin_lock(&iscsi_lun_probe_lock);
		if (iscsi_currently_probing != session)
			printk
			    ("iSCSI: bug - currently probing should be (bus %d "
			     "target %d) , not (bus %d target %d\n",
			     session->iscsi_bus, session->target_id,
			     iscsi_currently_probing->iscsi_bus,
			     iscsi_currently_probing->target_id);
		spin_unlock(&iscsi_lun_probe_lock);
#endif
		DEBUG_INIT
		    ("iSCSI: wait_for_probe_order (bus %d target %d) "
		     "returning 1\n",
		     session->iscsi_bus, session->target_id);
		return 1;
	}

	/* silently fail, since the enqueue attempt will have logged 
	 * any detailed messages needed 
	 */
	spin_unlock(&iscsi_lun_probe_lock);
	return 0;
}

/* caller must hold iscsi_lun_probe_lock */
static void
start_next_lun_probe(void)
{
	if (iscsi_currently_probing) {
		printk
		    ("iSCSI: bug - start_next_lun_probe called while "
		     "currently probing on (bus %d target %d at %lu\n",
		     iscsi_currently_probing->iscsi_bus,
		     iscsi_currently_probing->target_id, jiffies);
	} else if ((!list_empty(&iscsi_lun_probe_order))) {
		/* pop one off the queue, and tell it to start probing */
	
		iscsi_currently_probing = list_entry(iscsi_lun_probe_order.next,
						     struct iscsi_session,
						     probe_order_link);
		list_del_init(&iscsi_currently_probing->probe_order_link);

		/* skip over any gaps in the probe order */
		if (iscsi_next_probe < iscsi_currently_probing->probe_order) {
			DEBUG_INIT
			    ("iSCSI: LUN probe_order skipping from %d to %d\n",
			     iscsi_next_probe,
			     iscsi_currently_probing->probe_order);
			iscsi_next_probe = iscsi_currently_probing->probe_order;
			smp_mb();
		}

		/* wake up the ioctl which is waiting to do a probe */
		DEBUG_INIT
		    ("iSCSI: starting LUN probe for session "
		     "(bus %d target %d) to %s\n",
		     iscsi_currently_probing->iscsi_bus,
		     iscsi_currently_probing->target_id, 
		     iscsi_currently_probing->log_name);
		up(&iscsi_currently_probing->probe_sem);
	} else {
		/* if there is nothing else queued, then we don't need 
		 * the timer to keep checking, and we want to reset the 
		 * probe order so that future LUN probes get queued,
		 * and maintain the proper relative order amonst themselves, 
		 * even if the global order may have been lost.
		 */
		DEBUG_INIT
		    ("iSCSI: start_next_lun_probe has nothing to start, "
		     "resetting next LUN probe from %d to 0 at %lu\n",
		     iscsi_next_probe, jiffies);
		iscsi_lun_probe_start = 0;
		iscsi_next_probe = 0;
		smp_mb();
	}
}

void
iscsi_possibly_start_lun_probing(void)
{
	spin_lock(&iscsi_lun_probe_lock);
	if (iscsi_currently_probing == NULL) {
		/* if we're not probing already, make sure we start */
		DEBUG_INIT("iSCSI: timer starting LUN probing at %lu\n",
			   jiffies);
		start_next_lun_probe();
	}
	spin_unlock(&iscsi_lun_probe_lock);
}

static void
iscsi_probe_finished(struct iscsi_session *session)
{
	spin_lock(&iscsi_lun_probe_lock);
	if (iscsi_currently_probing == session) {
		iscsi_currently_probing = NULL;
		DEBUG_INIT
		    ("iSCSI: session (bus %d target %d) to %s "
		     "finished probing LUNs at %lu\n",
		     session->iscsi_bus, session->target_id, session->log_name,
		     jiffies);

		/* continue through the probe order */
		if (iscsi_next_probe == session->probe_order)
			iscsi_next_probe++;

		/* and possibly start another session probing */
		if (list_empty(&iscsi_lun_probe_order)) {
			/* nothing is queued, reset LUN probing */
			DEBUG_INIT
			    ("iSCSI: probe_finished has nothing to start, "
			     "resetting next LUN probe from %d to 0 at %lu\n",
			     iscsi_next_probe, jiffies);
			iscsi_next_probe = 0;
			iscsi_lun_probe_start = 0;
			smp_mb();
		} else
		    if (((list_entry(iscsi_lun_probe_order.next,
				     struct iscsi_session, probe_order_link))
				    ->probe_order <= iscsi_next_probe)
			|| (iscsi_lun_probe_start
			    && time_before_eq(iscsi_lun_probe_start,
					      jiffies))) {
			/* next in order is up, or the timer has 
			 * expired, start probing 
			 */
			start_next_lun_probe();
		} else {
			DEBUG_INIT
			    ("iSCSI: iscsi_probe_finished can't "
			     "start_next_lun_probe at %lu, next %d, head %p "
			     "(%d), tail %p (%d), current %p, start time %lu\n",
			     jiffies, iscsi_next_probe,
			     iscsi_lun_probe_order.next,
	  		     list_empty(&iscsi_lun_probe_order)?  
			     (list_entry(iscsi_lun_probe_order.next,
					 struct iscsi_session,probe_order_link))
					 -> probe_order : -1, 
			     iscsi_lun_probe_order.prev,
			     list_empty(&iscsi_lun_probe_order) ?
			     (list_entry(iscsi_lun_probe_order.prev,
					 struct iscsi_session,probe_order_link))
					 ->probe_order : -1,
			     iscsi_currently_probing, iscsi_lun_probe_start);
		}
	} else {
		/* should be impossible */
		printk
		    ("iSCSI: bug - session (bus %d target %d) in "
		     "iscsi_probe_finished, but currently probing "
		     "(bus %d target %d)\n",
		     session->iscsi_bus, session->target_id,
		     iscsi_currently_probing->iscsi_bus,
		     iscsi_currently_probing->target_id);
	}
	spin_unlock(&iscsi_lun_probe_lock);
}

void
iscsi_remove_luns(struct iscsi_session *session)
{
	struct remove_lun rd;
	struct rsp_status *rs;
	int rc;

	rd.host_no = session->host_no;
	rd.channel = session->channel;
	rd.target_id = session->target_id;
	rd.lun = -1;
	rd.target_link_dir = session->target_link_dir;
	memcpy(rd.luns_activated, session->luns_activated,
		sizeof (session->luns_activated));

	/* try to release the kernel's SCSI device structures for every LUN */
	down(&iscsi_lun_probe_mutex);

	rs = kmalloc(sizeof (struct rsp_status), GFP_KERNEL);
	if (!rs) {
		printk
		    ("iSCSI: session (host %d channel %d target %d) "
		     "iscsi_remove_luns couldn't allocate "
		     "a response structure\n",
		     session->host_no, session->channel, session->target_id);
		return;
	}

	send_request(REMOVE_LUN, &rd, rs);
	wait_for_response();
	rc = rs->status;
	kfree(rs);

	up(&iscsi_lun_probe_mutex);
}

void
iscsi_remove_lun_complete(struct iscsi_session *session, int lun_id)
{
	
	struct remove_lun rd;
	struct rsp_status *rs;
	int rc;

	down(&iscsi_lun_probe_mutex);
	memcpy(rd.luns_activated, session->luns_activated,
	       sizeof (session->luns_activated));
	rd.host_no = session->host_no;
	rd.channel = session->channel;
	rd.target_id = session->target_id;
	rd.lun = lun_id;
	rd.target_link_dir = session->target_link_dir;

	rs = kmalloc(sizeof (struct rsp_status), GFP_KERNEL);
	if (!rs) {
		printk
		    ("iSCSI: session (host %d channel %d target %d) "
		     "iscsi_remove_lun_complete couldn't allocate "
		     "a response structure\n",
		     session->host_no, session->channel, session->target_id);
		return;
	}

	send_request(REMOVE_LUN, &rd, rs);
	wait_for_response();
	rc = rs->status;
	kfree(rs);

	up(&iscsi_lun_probe_mutex);
}


/* compute the intersection of the LUNS detected and configured, 
 * and probe each LUN 
 */
void
iscsi_probe_luns(struct iscsi_session *session, uint32_t * lun_bitmap,
		 struct scsi_device_info *device_info)
{

	struct probe_devices pd;
	struct rsp_status *rs;
	int rc;

	/* try wait for our turn to probe, to keep the device node 
	 * ordering as repeatable as possible 
	 */
	DEBUG_INIT
	    ("iSCSI: session %p to %s waiting to probe LUNs at %lu, "
	     "probe order %d\n",
	     session, session->log_name, jiffies, session->probe_order);

	if (!wait_for_probe_order(session)) {
		DEBUG_INIT
		    ("iSCSI: session %p to %s couldn't probe LUNs, "
		     "error waiting for probe order\n",
		     session, session->log_name);
		return;
	}

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		printk
		    ("iSCSI: session %p to %s terminated while waiting "
		     "to probe LUNs\n",
		     session, session->log_name);
		goto done;
	}
	if (signal_pending(current)) {
		printk
		    ("iSCSI: session %p ioctl killed while waiting to "
		     "probe LUNs\n",
		     session);
		goto done;
	}

	/* make sure we're the only driver process trying 
	 * to add or remove LUNs 
	 */
	if (down_interruptible(&iscsi_lun_probe_mutex)) {
		printk
		    ("iSCSI: session %p to %s interrupted while probing LUNs\n",
		     session, session->log_name);
		goto done;
	}

	/* need to set the host's max_channel, max_id, max_lun, since we
	 * zero them in iscsi_detect in order to disable the scan that
	 * occurs during scsi_register_host.  
	 */
	session->hba->host->max_id = ISCSI_MAX_TARGET_IDS_PER_BUS - 1;
	session->hba->host->max_lun = ISCSI_MAX_LUNS_PER_TARGET - 1;
	session->hba->host->max_channel = ISCSI_MAX_CHANNELS_PER_HBA - 1;	/* convert from count to index */
	smp_mb();

	DEBUG_INIT
	    ("iSCSI: probing LUNs for session %p to %s at %lu, "
	     "probe_order %d at %lu\n",
	     session, session->log_name, jiffies, session->probe_order,
	     jiffies);
	if (test_bit(SESSION_TERMINATING, &session->control_bits))
		goto give_up;
	if (signal_pending(current))
		goto give_up;

	memcpy(pd.luns_found, session->luns_found,
	       sizeof (session->luns_found));
	memcpy(pd.luns_detected, session->luns_detected,
	       sizeof (session->luns_detected));
	memcpy(pd.luns_activated, session->luns_activated,
	       sizeof (session->luns_activated));
	pd.host_no = session->host_no;
	pd.channel = session->channel;
	pd.target_id = session->target_id;
	pd.target_link_dir = session->target_link_dir;
	pd.lun_bitmap = lun_bitmap;
	pd.max_sd_devices = device_info->max_sd_devices;
	pd.max_sg_devices = device_info->max_sg_devices;
	pd.max_sr_devices = device_info->max_sr_devices;
	pd.max_st_devices = device_info->max_st_devices;

	rs = kmalloc(sizeof (struct rsp_status), GFP_KERNEL);
	if (!rs) {
		printk
		    ("iSCSI: session (host %d channel %d target %d) "
		     "iscsi_probe_luns couldn't allocate "
		     "a response structure\n",
		     session->host_no, session->channel, session->target_id);
		return;
	}

	send_request(PROBE_LUNS, &pd, rs);
	wait_for_response();
	rc = rs->status;
	kfree(rs);

      give_up:
	up(&iscsi_lun_probe_mutex);

      done:
	/* clean up after wait_for_probe_order, and possibly start the next 
	 * session probing 
	 */
	iscsi_probe_finished(session);
}

struct iscsi_cmnd {
	struct semaphore done_sem;
	Scsi_Cmnd sc;
	unsigned int bufflen;
	uint8_t buffer[1];
};

/* callback function for Scsi_Cmnd's generated by the iSCSI driver itself */
void
iscsi_done(Scsi_Cmnd * sc)
{
	struct iscsi_cmnd *c = (struct iscsi_cmnd *) sc->buffer;

	up(&c->done_sem);
}

static int
iscsi_do_cmnd(struct iscsi_session *session, struct iscsi_cmnd *c,
	      unsigned int attempts_allowed)
{
	Scsi_Cmnd *sc = NULL;
	int queue_attempts = 0;

	if (c->sc.device->host) {
		DEBUG_FLOW
		    ("iSCSI: iscsi_do_cmnd %p to (%u %u %u %u), Cmd 0x%02x, "
		     "%u retries, buffer %p, bufflen %u\n",
		     c, c->sc.device->host->host_no, c->sc.device->channel,
		     c->sc.device->id, c->sc.device->lun, c->sc.cmnd[0],
		     attempts_allowed, c->sc.request_buffer,
		     c->sc.request_bufflen);
	} else {
		printk
		    ("iSCSI: session (bus %d target %d) iscsi_do_cmnd %p, "
		     "buffer %p, bufflen %u, host %p\n",
		     session->iscsi_bus, session->target_id, c,
		     c->sc.request_buffer, c->sc.request_bufflen,
		     c->sc.device->host);
		return 0;
	}
	if (!c->sc.request_buffer)
		return 0;
	if (!c->sc.request_bufflen)
		return 0;

	sc = &(c->sc);
	sc->retries = -1;
	sc->allowed = attempts_allowed;

      retry:
	while (++sc->retries < sc->allowed) {
		if (signal_pending(current))
			return 0;
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			return 0;

		sc->result = 0;
		memset(sc->sense_buffer, 0, sizeof (sc->sense_buffer));
		memset(c->buffer, 0, c->bufflen);

		/* try to queue the command */
		queue_attempts = 0;
		for (;;) {
			sema_init(&c->done_sem, 0);
			smp_mb();

			if (signal_pending(current))
				return 0;
			if (test_bit
			    (SESSION_TERMINATING, &session->control_bits))
				return 0;

			DEBUG_INIT
			    ("iSCSI: detect_luns queueing %p to session "
			     "(bus %d target %d) at %lu\n",
			     sc, session->iscsi_bus, session->target_id,
			     jiffies);

			/* give up eventually, in case the replacement timeout 
			 * is in effect. we don't want to loop forever trying 
			 * to queue to a session that may never accept 
			 * commands.
			 */
			if (iscsi_queue(session, sc, iscsi_done)) {
				break;
			} else if (queue_attempts++ >= 500) {
				/* give up after 10 seconds */
				return 0;
			}

			/* command not queued, wait a bit and try again */
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(MSECS_TO_JIFFIES(20));
		}

		DEBUG_QUEUE
		    ("iSCSI: session (bus %d target %d) queued iscsi_cmnd %p, "
		     "buffer %p, bufflen %u, scsi_done %p\n",
		     session->iscsi_bus, session->target_id, c,
		     c->sc.request_buffer, c->sc.request_bufflen,
		     c->sc.scsi_done);

		/* wait til either the command completes, or we get signalled. 
		 */
		if (down_interruptible(&c->done_sem)) {
			/* if we got signalled, squash the command and give up 
			 */
			iscsi_squash_cmnd(session, sc);
			return 0;
		}

		DEBUG_QUEUE
		    ("iSCSI: session (bus %d target %d) hba %p host %p "
		     "woken up by iscsi_cmnd %p, buffer %p, bufflen %u\n",
		     session->iscsi_bus, session->target_id, session->hba,
		     session->hba->host, c, c->sc.request_buffer,
		     c->sc.request_bufflen);

		/* the command completed, check the result and decide if it 
		 * needs to be retried. 
		 */
		DEBUG_FLOW
		    ("iSCSI: session (bus %d target %d) iscsi cmnd %p to "
		     "(%u %u %u %u), Cmd 0x%02x, "
		     "host byte 0x%x, SCSI status 0x%x, residual %u\n",
		     session->iscsi_bus, session->target_id, c,
		     c->sc.device->host->host_no, c->sc.device->channel,
		     c->sc.device->id, c->sc.device->lun, c->sc.cmnd[0],
		     (sc->result >> 24) & 0xFF, sc->result & 0xFF, sc->resid);

		/* check the host byte */
		switch (host_byte(sc->result)) {
		case DID_OK:
			/* no problems so far */
			break;
		case DID_NO_CONNECT:
			/* give up, we can't talk to the device */
			printk
			    ("iSCSI: failing iscsi cmnd %p to (%u %u %u %u), "
			     "Cmd 0x%02x, host byte 0x%x, SCSI status 0x%x, "
			     "residual %u\n",
			     c, c->sc.device->host->host_no, 
			     c->sc.device->channel,
			     c->sc.device->id, c->sc.device->lun, c->sc.cmnd[0],
			     (sc->result >> 24) & 0xFF, sc->result & 0xFF,
			     sc->resid);
			return 0;
		case DID_ERROR:
		case DID_SOFT_ERROR:
		case DID_ABORT:
		case DID_BUS_BUSY:
		case DID_PARITY:
		case DID_TIME_OUT:
		case DID_RESET:
		default:
			if (LOG_ENABLED(ISCSI_LOG_INIT))
				printk
				    ("iSCSI: iscsi cmnd %p to (%u %u %u %u), "
				     "Cmd 0x%02x, host byte 0x%x, "
				     "SCSI status 0x%x, residual %u\n",
				     c, c->sc.device->host->host_no,
				     c->sc.device->channel, c->sc.device->id,
				     c->sc.device->lun, c->sc.cmnd[0], 
				     (sc->result >> 24) & 0xFF,
				     sc->result & 0xFF, sc->resid);

			/* some sort of problem, possibly retry */
			goto retry;
		}

		/* check the SCSI status byte.  Note, Linux values 
		 * are right-shifted once compared to the SCSI spec 
		 */
		switch (status_byte(sc->result)) {
		case GOOD:
		case COMMAND_TERMINATED:

			/* make sure we got enough of a response */
			if (sc->resid
			    && ((iscsi_expected_data_length(sc) - sc->resid) <
				sc->underflow)) {
				/* try again */
				if (LOG_ENABLED(ISCSI_LOG_INIT))
					printk
					    ("iSCSI: iscsi cmnd %p to "
					     "(%u %u %u %u), Cmd 0x%02x, "
					     "residual %u, retrying to get "
					     "%u bytes desired\n",
					     c, c->sc.device->host->host_no,
					     c->sc.device->channel, 
					     c->sc.device->id,
					     c->sc.device->lun, c->sc.cmnd[0],
					     sc->resid, sc->underflow);
				goto retry;
			}

			/* all done */
			return 1;
		case BUSY:	/* device is busy, try again later */
		case QUEUE_FULL:	/* tagged queuing device has a full 
					 * queue, wait a bit and try again. 
					 */
			sc->allowed++;
			if (sc->allowed > 100) {
				printk
				    ("iSCSI: iscsi cmnd %p to (%u %u %u %u), "
				     "Cmd 0x%02x, SCSI status 0x%x, "
				     "out of retries\n",
				     c, c->sc.device->host->host_no,
				     c->sc.device->channel, c->sc.device->id,
				     c->sc.device->lun, c->sc.cmnd[0],
				     sc->result & 0xFF);
				return 0;
			}
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(MSECS_TO_JIFFIES(20));
			goto retry;
		case CONDITION_GOOD:
		case INTERMEDIATE_GOOD:
		case INTERMEDIATE_C_GOOD:
			/* we should never get the linked command return codes 
			 */
		case RESERVATION_CONFLICT:
			/* this is probably never going to happen for 
			 * INQUIRY or REPORT_LUNS, but retry if it does 
			 */
			printk
			    ("iSCSI: session (bus %d target %d) "
			     "iscsi_do_cmnd %p SCSI status 0x%x at %lu, "
			     "retrying\n",
			     session->iscsi_bus, session->target_id, c,
			     sc->result & 0xFF, jiffies);
			goto retry;
		case CHECK_CONDITION:
			/* look at the sense.  If it's illegal request, 
			 * don't bother retrying the command 
			 */
			if ((sc->sense_buffer[0] & 0x70) == 0x70) {
				switch (SENSE_KEY(sc->sense_buffer)) {
				case ILLEGAL_REQUEST:
					printk
					    ("iSCSI: iscsi cmnd %p to "
					     "(%u %u %u %u), Cmd 0x%02x, "
					     "illegal request\n",
					     c, c->sc.device->host->host_no,
					     c->sc.device->channel,
					     c->sc.device->id,
					     c->sc.device->lun, c->sc.cmnd[0]);
					return 0;
				default:
					/* possibly retry */
					if (LOG_ENABLED(ISCSI_LOG_INIT))
						printk
						    ("iSCSI: iscsi cmnd %p to "
						     "(%u %u %u %u), Cmd 0x%02x"
						     " with sense, retrying\n",
						     c, 
						     c->sc.device->host->host_no,
						     c->sc.device->channel,
						     c->sc.device->id,
						     c->sc.device->lun,
						     c->sc.cmnd[0]);
					goto retry;
				}
			}
			goto retry;
		default:
			printk
			    ("iSCSI: session (bus %d target %d) iscsi_do_cmnd "
			     "%p unexpected SCSI status 0x%x at %lu\n",
			     session->iscsi_bus, session->target_id, c,
			     sc->result & 0xFF, jiffies);
			return 0;
		}
	}

	if (LOG_ENABLED(ISCSI_LOG_INIT))
		printk
		    ("iSCSI: session (bus %d target %d) iscsi_do_cmnd %p "
		     "SCSI status 0x%x, out of retries at %lu\n",
		     session->iscsi_bus, session->target_id, c,
		     sc->result & 0xFF, jiffies);

	return 0;
}

void
send_tur(struct iscsi_session *session)
{
	struct iscsi_cmnd *c = NULL;
	Scsi_Cmnd *sc = NULL;
	size_t cmd_size = sizeof (struct iscsi_cmnd);
	unsigned int bufflen = 255;

	cmd_size += bufflen;

	c = kmalloc(cmd_size, GFP_KERNEL);
	if (!c) {
		printk
		    ("iSCSI: session (bus %d target %d) send_tur "
		     "couldn't allocate a Scsi_Cmnd\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	/* initialize */
	memset(c, 0, cmd_size);
	sema_init(&c->done_sem, 0);
	c->sc.device = kmalloc(sizeof(struct scsi_device),GFP_KERNEL);
	if (!c->sc.device) {
		printk
		    ("iSCSI: session (bus %d target %d) send_tur "
		     "couldn't allocate a scsi device structure\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	memset(c->sc.device,0,sizeof(struct scsi_device));
	c->bufflen = bufflen;
	DEBUG_ALLOC
	    ("iSCSI: session (bus %d target %d) hba %p host %p allocated "
	     "iscsi cmnd %p, size %d, buffer %p, bufflen %u, end %p\n",
	     session->iscsi_bus, session->target_id, session->hba,
	     session->hba->host, c, cmd_size, c->buffer, c->bufflen,
	     c->buffer + c->bufflen);

	/* fill in the basic required info in the Scsi_Cmnd */
	sc = &(c->sc);
	/* FIXME: Who will allocate space for device */
	sc->device->host = session->hba->host;
	sc->device->channel = session->channel;
	sc->device->id = session->target_id;
	sc->device->lun = 0;
	sc->use_sg = 0;
	sc->request_buffer = c->buffer;
	sc->request_bufflen = c->bufflen;
	sc->scsi_done = iscsi_done;
	sc->timeout_per_command = 30 * HZ;
	sc->resid = 0;
	sc->underflow = 8;
	init_timer(&sc->eh_timeout);
	/* save a pointer to the iscsi_cmnd in the Scsi_Cmnd, so that 
	 * iscsi_done can use it 
	 */

	sc->buffer = (void *) c;
	{
		if (signal_pending(current)) {
			DEBUG_INIT
			    ("iSCSI: session (bus %d target %d) send_tur "
			     "aborted by signal\n",
			     session->iscsi_bus, session->target_id);
			goto done;
		}
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			goto done;

		sc->cmd_len = 6;
		memset(sc->cmnd, 0, sizeof (sc->cmnd));
		sc->cmnd[0] = TEST_UNIT_READY;
		sc->cmnd[1] = 0;
		sc->cmnd[2] = 0;
		sc->cmnd[3] = 0;
		sc->cmnd[4] = 0;
		sc->cmnd[5] = 0;

		smp_mb();
		if (iscsi_do_cmnd(session, c, 6)) {
		} else {
			printk
			    ("iSCSI: session (bus %d target %d) Received "
			     "a sense for a TEST UNIT READY\n",
			     session->iscsi_bus, session->target_id);
		}
	}

      done:
	kfree(c);
}

void
reinitialize_disk(struct iscsi_session *session)
{
	struct iscsi_cmnd *c = NULL;
	Scsi_Cmnd *sc = NULL;
	size_t cmd_size = sizeof (struct iscsi_cmnd);
	unsigned int bufflen = 255;

	cmd_size += bufflen;

	c = kmalloc(cmd_size, GFP_KERNEL);
	if (!c) {
		printk
		    ("iSCSI: session (bus %d target %d) reinitialize_disk "
		     "couldn't allocate a Scsi_Cmnd\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	/* initialize */
	memset(c, 0, cmd_size);
	sema_init(&c->done_sem, 0);
	c->sc.device = kmalloc(sizeof(struct scsi_device),GFP_KERNEL);
	if (!c->sc.device) {
		printk
		    ("iSCSI: session (bus %d target %d) reinitialize_disk "
		     "couldn't allocate a scsi device structure\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	memset(c->sc.device,0,sizeof(struct scsi_device));
	c->bufflen = bufflen;
	DEBUG_ALLOC
	    ("iSCSI: session (bus %d target %d) hba %p host %p allocated "
	     "iscsi cmnd %p, size %d, buffer %p, bufflen %u, end %p\n",
	     session->iscsi_bus, session->target_id, session->hba,
	     session->hba->host, c, cmd_size, c->buffer, c->bufflen,
	     c->buffer + c->bufflen);

	/* fill in the basic required info in the Scsi_Cmnd */
	sc = &(c->sc);
	sc->device->host = session->hba->host;
	sc->device->channel = session->channel;
	sc->device->id = session->target_id;
	sc->device->lun = 0;
	sc->use_sg = 0;
	sc->request_buffer = c->buffer;
	sc->request_bufflen = c->bufflen;
	sc->scsi_done = iscsi_done;
	sc->timeout_per_command = 30 * HZ;
	sc->resid = 0;
	sc->underflow = 8;
	init_timer(&sc->eh_timeout);
	/* save a pointer to the iscsi_cmnd in the Scsi_Cmnd, so that 
	 * iscsi_done can use it 
	 */

	sc->buffer = (void *) c;
	{
		if (signal_pending(current)) {
			DEBUG_INIT
			    ("iSCSI: session (bus %d target %d) "
			     "reinitialize_disk aborted by signal\n",
			     session->iscsi_bus, session->target_id);
			goto done;
		}
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			goto done;

		sc->cmd_len = 6;
		memset(sc->cmnd, 0, sizeof (sc->cmnd));
		sc->cmnd[0] = START_STOP;
		sc->cmnd[1] = 0;
		sc->cmnd[1] |= 1;
		sc->cmnd[2] = 0;
		sc->cmnd[3] = 0;
		sc->cmnd[4] = 1;
		sc->cmnd[5] = 0;

		smp_mb();
		if (iscsi_do_cmnd(session, c, 6)) {
		} else {
			printk
			    ("\niSCSI:session (bus %d target %d) Received "
			     "a sense for a START STOP\n",
			     session->iscsi_bus, session->target_id);
		}
	}

      done:
	kfree(c);
}

static void
make_report_luns(Scsi_Cmnd * sc, uint32_t max_entries)
{
	uint32_t length = 8 + (max_entries * 8);	
				/* 8 byte header plus 8 bytes per LUN */

	sc->cmd_len = 10;
	sc->request_bufflen = length;
	sc->underflow = 8;	/* need at least the length */

	sc->resid = 0;

	/* CDB */
	memset(sc->cmnd, 0, sizeof (sc->cmnd));
	sc->cmnd[0] = REPORT_LUNS;
	sc->cmnd[1] = 0;
	sc->cmnd[2] = 0;	/* either reserved or select report 
				 * in various versions of SCSI-3 
				 */
	sc->cmnd[3] = 0;
	sc->cmnd[4] = 0;
	sc->cmnd[5] = 0;
	sc->cmnd[6] = (length >> 24) & 0xFF;
	sc->cmnd[7] = (length >> 16) & 0xFF;
	sc->cmnd[8] = (length >> 8) & 0xFF;
	sc->cmnd[9] = (length) & 0xFF;
}

static void
make_inquiry(Scsi_Cmnd * sc, int lun0_scsi_level)
{
	sc->cmd_len = 6;
	sc->request_bufflen = 255;
	if (sc->device->lun == 0)
		sc->underflow = 3;	/* we need at least the peripheral code 
					 * and SCSI version 
					 */
	else
		sc->underflow = 1;	/* we need at least the peripheral code 
					 */

	sc->resid = 0;

	memset(sc->cmnd, 0, sizeof (sc->cmnd));
	sc->cmnd[0] = INQUIRY;
	if (lun0_scsi_level >= 0x3)
		sc->cmnd[1] = 0;	/* reserved in SCSI-3 and higher */
	else
		sc->cmnd[1] = (sc->device->lun << 5) & 0xe0;

	sc->cmnd[2] = 0;
	sc->cmnd[3] = 0;
	sc->cmnd[4] = 255;	/* length */
	sc->cmnd[5] = 0;
}

/* scan for LUNs */
void
iscsi_detect_luns(struct iscsi_session *session)
{
	int l;
	struct iscsi_cmnd *c = NULL;
	Scsi_Cmnd *sc = NULL;
	int lun0_scsi_level = 0;
	size_t cmd_size = sizeof (struct iscsi_cmnd);
	unsigned int bufflen = 0;
	uint32_t last_luns = 0;
	uint32_t luns = 32;	/* start small to avoid bugs in 
				 * REPORT_LUNS handling 
				 */
	int report_luns_failed = 0;


	memset(session->luns_found, 0, sizeof (session->luns_found));

	/* need enough buffer space for replies to INQUIRY and REPORT_LUNS */
	if ((8 + (ISCSI_MAX_LUN * 8)) < 255)
		bufflen = 255;
	else
		bufflen = (ISCSI_MAX_LUN * 8) + 8;

	cmd_size += bufflen;

	c = kmalloc(cmd_size, GFP_KERNEL);
	if (!c) {
		printk
		    ("iSCSI: session (bus %d target %d) iscsi_detect_luns "
		     "couldn't allocate a Scsi_Cmnd\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	/* initialize */
	memset(c, 0, cmd_size);
	sema_init(&c->done_sem, 0);
	c->sc.device = kmalloc(sizeof(struct scsi_device),GFP_KERNEL);
	if (!c->sc.device) {
		printk
		    ("iSCSI: session (bus %d target %d) iscsi_detect_luns "
		     "couldn't allocate a scsi device structure\n",
		     session->iscsi_bus, session->target_id);
		return;
	}

	memset(c->sc.device,0,sizeof(struct scsi_device));
	memset(c->sc.device,0,sizeof(struct scsi_device));
	c->bufflen = bufflen;
	DEBUG_ALLOC
	    ("iSCSI: session (bus %d target %d) hba %p host %p allocated "
	     "iscsi cmnd %p, size %d, buffer %p, bufflen %u, end %p\n",
	     session->iscsi_bus, session->target_id, session->hba,
	     session->hba->host, c, cmd_size, c->buffer, c->bufflen,
	     c->buffer + c->bufflen);

	/* fill in the basic required info in the Scsi_Cmnd */
	sc = &(c->sc);
	sc->device->host = session->hba->host;
	sc->device->channel = session->channel;
	sc->device->id = session->target_id;
	sc->device->lun = 0;
	sc->use_sg = 0;
	sc->request_buffer = c->buffer;
	sc->request_bufflen = c->bufflen;
	sc->scsi_done = iscsi_done;
	sc->timeout_per_command = 30 * HZ;
	init_timer(&sc->eh_timeout);
	/* save a pointer to the iscsi_cmnd in the Scsi_Cmnd, so that 
	 * iscsi_done can use it 
	 */
	sc->buffer = (void *) c;

	do {
		if (signal_pending(current)) {
			DEBUG_INIT
			    ("iSCSI: session (bus %d target %d) detect LUNs "
			     "aborted by signal\n",
			     session->iscsi_bus, session->target_id);
			goto done;
		}
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			goto done;

		/* send a REPORT_LUNS to LUN 0.  If it works, we know the LUNs. 
		 */
		last_luns = luns;
		make_report_luns(sc, luns);
		smp_mb();
		
		if (iscsi_do_cmnd(session, c, 6)) {
			uint8_t *lun_list = c->buffer + 8;
			int luns_listed;
			uint32_t length = 0;

			/* get the list length the target has */
			length = c->buffer[0] << 24;
			length |= c->buffer[1] << 16;
			length |= c->buffer[2] << 8;
			length |= c->buffer[3];

			if (length < 8) {
				/* odd, assume REPORT_LUNS is broken, 
				 * fall back to doing INQUIRY 
				 */
				DEBUG_INIT
				    ("iSCSI: session (bus %d target %d) "
				     "REPORT_LUNS length 0, falling back "
				     "to INQUIRY\n",
				     session->iscsi_bus, session->target_id);
				report_luns_failed = 1;
				break;
			}

			/* figure out how many luns we were told about this time
			 */
			if ((length / 8U) < luns)
				luns_listed = length / 8U;
			else
				luns_listed = luns;

			/* loop until we run out of data, or out of buffer */
			for (l = 0; l < luns_listed; l++) {
				int address_method = (lun_list[0] & 0xc0) >> 6;
				int lun;

				if (LOG_ENABLED(ISCSI_LOG_LOGIN)
				    || LOG_ENABLED(ISCSI_LOG_INIT))
					printk
					    ("iSCSI: session  (%u %u %u *) "
					     "REPORT_LUNS[%d] = %02x %02x %02x "
					     "%02x %02x %02x %02x %02x\n",
					     session->host_no,
					     session->channel,
					     session->target_id, l, lun_list[0],
					     lun_list[1], lun_list[2],
					     lun_list[3], lun_list[4],
					     lun_list[5], lun_list[6],
					     lun_list[7]);

				switch (address_method) {
				case 0x0:{
						/* single-level LUN if bus id 
						 * is 0, else peripheral 
						 * device addressing 
						 */
						lun = lun_list[1];
						set_bit(lun,
							session->luns_detected);
						/* This is useful while 
						 * checking for deleted luns 
						 */
						set_bit(lun,
							session->luns_found);
						break;
					}
				case 0x1:{
						/* flat-space addressing */
						lun = lun_list[1];
						set_bit(lun,
							session->luns_detected);
						/* This is useful while 
						 * checking for deleted luns 
						 */
						set_bit(lun,
							session->luns_found);
						break;
					}
				case 0x2:{
						/* logical unit addressing 
						 * method 
						 */
						lun = lun_list[1] & 0x1F;
						set_bit(lun,
							session->luns_detected);
						/* This is useful while checking
						 * for deleted luns 
						 */
						set_bit(lun,
							session->luns_found);
						break;
					}
				case 0x3:{
					/* extended logical unit addressing 
					 * method is too complicated for us 
					 * to want to deal with 
					 */
						printk
						    ("iSCSI: session "
						     "(%u %u %u *) REPORT_LUNS"
						     "[%d] with extended LU "
						     "address method 0x%x "
						     "ignored\n",
						     session->host_no,
						     session->channel,
						     session->target_id, l,
						     address_method);
						break;
					}
				default:
					printk
					    ("iSCSI: session (%u %u %u *) "
					     "REPORT_LUNS[%d] with unknown "
					     "address method 0x%x ignored\n",
					     session->host_no,
					     session->channel,
					     session->target_id, l,
					     address_method);
					break;
				}

				/* next LUN in the list */
				lun_list += 8;
			}

			/* decide how many luns to ask for on the next 
			 * iteration, if there is one 
			 */
			luns = length / 8U;
			if (luns > ISCSI_MAX_LUN) {
				/* we only have buffer space for so many LUNs */
				luns = ISCSI_MAX_LUN;
				printk
				    ("iSCSI: session (bus %d target %d) "
				     "REPORT_LUNS length %u (%u entries) "
				     "truncated to %u (%u entries)\n",
				     session->iscsi_bus, session->target_id,
				     length, (length / 8) - 1, (luns + 1) * 8U,
				     luns);
			}

		} else {
			/* REPORT_LUNS failed, fall back to doing INQUIRY */
			DEBUG_INIT
			    ("iSCSI: session (bus %d target %d) REPORT_LUNS "
			     "failed, falling back to INQUIRY\n",
			     session->iscsi_bus, session->target_id);
			report_luns_failed = 1;
			break;
		}

	} while (luns > last_luns);

	if (signal_pending(current)) {
		DEBUG_INIT
		    ("iSCSI: session (bus %d target %d) detect LUNs "
		     "aborted by signal\n",
		     session->iscsi_bus, session->target_id);
		goto done;
	}

	if (report_luns_failed) {
		/* if REPORT_LUNS failed, then either it's a SCSI-2 device
		 * that doesn't understand the command, or it's a SCSI-3
		 * device that only has one LUN and decided not to implement
		 * REPORT_LUNS.  In either case, we're safe just probing LUNs
		 * 0-7 with INQUIRY, since SCSI-2 can't have more than 8 LUNs,
		 * and SCSI-3 should do REPORT_LUNS if it has more than 1 LUN.
		 */
		for (l = 0; l < 8; l++) {
			sc->device->lun = l;
			sc->request_buffer = c->buffer;
			make_inquiry(sc, lun0_scsi_level);

			/* we'll make a note of the LUN when the rx thread 
			 * receives the response.
			 * No need to do it again here.
			 */
			if (iscsi_do_cmnd(session, c, 6)) {
				/* we do need to record the SCSI level so we 
				 * can build inquiries properly though 
				 */
				if (l == 0) {
					lun0_scsi_level = c->buffer[2] & 0x07;
					if (LOG_ENABLED(ISCSI_LOG_INIT))
						printk
						    ("iSCSI: session "
						     "(%u %u %u %u) is SCSI "
						     "level %d\n",
						     sc->device->host->host_no,
						     sc->device->channel, 
						     sc->device->id,
						     sc->device->lun,
						     lun0_scsi_level);
				}
			} else {
				/* just assume there's no LUN */
			}

			if (test_bit
			    (SESSION_TERMINATING, &session->control_bits))
				break;
			if (signal_pending(current))
				break;
		}
	}

      done:
	DEBUG_ALLOC
	    ("iSCSI: session (bus %d target %d) hba %p host %p "
	     "kfree iscsi cmnd %p, bufflen %u\n",
	     session->iscsi_bus, session->target_id, session->hba,
	     session->hba->host, c, c->bufflen);
	kfree(c->sc.device);
	kfree(c);
}

int
iscsi_reset_lun_probing(void)
{
	int ret = 0;

	spin_lock(&iscsi_lun_probe_lock);
	if ((iscsi_currently_probing == NULL)  
	    && (list_empty(&iscsi_lun_probe_order))) {
		/* if we're not currently probing, reset */
		DEBUG_INIT ("iSCSI: reset LUN probing at %lu\n", jiffies);
		iscsi_next_probe = 0;
		iscsi_lun_probe_start = 0;
		smp_mb();
		ret = 1;
	} else {
		DEBUG_INIT
		    ("iSCSI: failed to reset LUN probing at %lu,"
                     " currently probing %p\n",jiffies, 
		     iscsi_currently_probing);
	}
	spin_unlock(&iscsi_lun_probe_lock);

	return ret;
}
