/*
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Kenneth W. Preslan <kpreslan@sistina.com>
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/slab.h>
#include <linux/spinlock.h>

#include "global.h"

#include "lm_interface.h"
#include "nolock.h"

MODULE_DESCRIPTION("Nolock Module");
MODULE_AUTHOR("The OpenGFS Project");
MODULE_LICENSE("GPL");


/**
 * nolock_mount - mount a nolock lockspace
 * @table_name: the name of the space to mount
 * @host_data: host specific data
 * @cb: the callback
 * @fsdata: argument to pass back with the callback
 * @jid: the journal to mount
 *
 * Returns: a lockspace on successful mount or NULL on failure
 */

static lm_lockspace_t *
nolock_mount(char *table_name, char *host_data,
	     lm_callback_t cb, lm_fsdata_t * fsdata,
	     unsigned int *jid, unsigned int *first)
{
	struct nolock_space *lsp;
	char *c;
	int i;

	//MOD_INC_USE_COUNT;

	lsp = kmalloc(sizeof(struct nolock_space), GFP_KERNEL);
	if (!lsp)
		goto fail;

	spin_lock_init(&lsp->lock);
	for (i = 0; i < HASH_SIZE; i++)
		INIT_LIST_HEAD(&lsp->hash[i]);

	/*
	 * If there is a "jid=" in the hostdata, return that jid.
	 * Otherwise, return zero.
	 */
	c = strstr(host_data, "jid=");
	if (c) {
		c += 4;
		*jid = simple_strtoul(c,NULL,10);
	} else
		*jid = 0;

	*first = TRUE;
	return (lm_lockspace_t *)lsp;

fail:
	//MOD_DEC_USE_COUNT;
	return NULL;
}

/**
 * nolock_others_may_mount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

static void
nolock_others_may_mount(lm_lockspace_t * lockspace)
{
	return;
}

/**
 * clear_hash_bucket - clear a hash bucket
 * @head: the bucket
 *
 */

static void
clear_hash_bucket(struct list_head *head)
{
	struct nolock_lock *lk;

	while (!list_empty(head)) {
		lk = list_entry(head->next, struct nolock_lock, list);

		if (!(lk->lvb.lvb_flags & LM_LVB_PERM)) {
			printk(KERN_WARNING "nolock: "
			    "found floating lock: count=%d, ln_type=%u, "
			    "ln_number=%Lu, lvb_count=%d, lvb_flags=%d",
			     atomic_read(&lk->count),
			     lk->name.ln_type,
			     lk->name.ln_number,
			     atomic_read(&lk->lvb_count),
			     lk->lvb.lvb_flags);
		}

		list_del(&lk->list);
		kfree(lk);
	}
}

/**
 * clear_lockspace - clear a lock space
 * lsp: the lock space
 *
 */

static void
clear_lockspace(struct nolock_space * lsp)
{
	unsigned int i;

	spin_lock(&lsp->lock);
	for (i = 0; i < HASH_SIZE; i++)
		clear_hash_bucket(&lsp->hash[i]);
	spin_unlock(&lsp->lock);
}

/**
 * nolock_unmount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

static void
nolock_unmount(lm_lockspace_t * lockspace)
{
	struct nolock_space *lsp = (struct nolock_space *)lockspace;

	clear_lockspace(lsp);
	kfree(lsp);

	//MOD_DEC_USE_COUNT;
}

/**
 * nolock_hash - hash a lockname
 * @ln: the lockname
 *
 */

static __inline__ unsigned
nolock_hash(lm_lockname_t * ln)
{
	unsigned hash = ln->ln_type;
	hash ^= (unsigned) ln->ln_number;
	hash ^= (unsigned) (ln->ln_number >> 32);
	hash ^= (unsigned) (ln->ln_number >> 16);
	return hash;
}

/**
 * __find_lock - Find a lock in a given bucket
 * @head: the head of the bucket
 * @hash: the lock name's hash
 * @name: the lock name
 *
 * Returns: the lock or NULL
 */

static struct nolock_lock *
__find_lock(struct list_head * head, unsigned hash, lm_lockname_t * name)
{
	struct list_head *tmp;
	struct nolock_lock *lk;

	tmp = head;
	for (;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		lk = list_entry(tmp, struct nolock_lock, list);

		if (lk->hash != hash)
			continue;
		if (lm_name_equal(&lk->name, name))
			return lk;
	}

	return NULL;
}

/**
 * new_lock - Allocate a new lock
 * @lsp: the lock space
 * @hash: the lock name's hash
 * @name: the lock name
 *
 * Returns: the lock or NULL
 */

static struct nolock_lock *
new_lock(struct nolock_space * lsp, unsigned hash, lm_lockname_t * name)
{
	struct nolock_lock *lk;

	lk = kmalloc(sizeof(struct nolock_lock), GFP_NOFS);
	if (!lk)
		return NULL;
	memset(lk, 0, sizeof(struct nolock_lock));

	lk->hash = hash;
	lk->name = *name;
	atomic_set(&lk->count, 0);

	lk->state = LM_ST_UNLOCKED;

	atomic_set(&lk->lvb_count, 0);
	lk->lvb.lvb_flags = 0;
	lk->lvb.lvb_length = NOLOCK_LVB_SIZE;
	lk->lvb.lvb_data = lk->lvb_data;

	lk->ls = lsp;
	return lk;
}

/**
 * nolock_get_lock - get a lm_lock_t given a descripton of the lock
 * @lockspace: the lockspace the lock lives in
 * @name: the name of the lock
 *
 * Returns: A lock structure or NULL on failure
 */

static lm_lock_t *
nolock_get_lock(lm_lockspace_t * lockspace, lm_lockname_t * name)
{
	struct nolock_space *lsp = (struct nolock_space *)lockspace;
	unsigned int hash = nolock_hash(name);
	struct list_head *lh = &lsp->hash[hash & HASH_MASK];
	struct nolock_lock *lk, *nlk;

	spin_lock(&lsp->lock);
	lk = __find_lock(lh, hash, name);
	if (lk)
		atomic_inc(&lk->count);
	spin_unlock(&lsp->lock);

	if (lk)
		return lk;
	nlk = new_lock(lsp, hash, name);
	if (!nlk)
		return NULL;

	spin_lock(&lsp->lock);
	lk = __find_lock(lh, hash, name);
	if (!lk) {
		list_add(&nlk->list, lh);
		lk = nlk;
		nlk = NULL;
	}

	atomic_inc(&lk->count);
	spin_unlock(&lsp->lock);

	if (nlk)
		kfree(nlk);
	return lk;
}

/**
 * nolock_put_lock - get rid of a lock structure
 * @lock: the lock to throw away
 *
 */

static void
nolock_put_lock(lm_lock_t * lock)
{
	struct nolock_lock *lk = (struct nolock_lock *) lock;
	struct nolock_space *ls = lk->ls;

	NOLOCK_ASSERT(ls);	/*  Keep compiler from bitching on a UP system.  */
	NOLOCK_ASSERT(atomic_read(&lk->count));
	NOLOCK_ASSERT(lk->state == LM_ST_UNLOCKED);
	NOLOCK_ASSERT(!atomic_read(&lk->lvb_count));

	if (atomic_dec_and_test(&lk->count)) {
		spin_lock(&ls->lock);
		if (!atomic_read(&lk->count)
		    && !(lk->lvb.lvb_flags & LM_LVB_PERM)) {
			list_del(&lk->list);
			kfree(lk);
		}
		spin_unlock(&ls->lock);
	}
}

/**
 * nolock_lock - acquire a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 * @req_state: the requested state
 * @flags: modifier flags
 *
 * Returns: A bitmap of LM_OUT_* on success, -EXXX on failure
 */

static int
nolock_lock(lm_lock_t * lock, unsigned int cur_state, unsigned int req_state,
	    unsigned int flags)
{
	struct nolock_lock *lk = (struct nolock_lock *) lock;

	NOLOCK_ASSERT(atomic_read(&lk->count));
	NOLOCK_ASSERT(lk->state == cur_state);
	NOLOCK_ASSERT(req_state != LM_ST_UNLOCKED);

	lk->state = req_state;

	if (atomic_read(&lk->lvb_count)) {
		if (cur_state == LM_ST_EXCLUSIVE)
			memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);
		else
			memcpy(lk->lvb_data, lk->lvb_data_net, NOLOCK_LVB_SIZE);
	}

	return req_state;
}

/**
 * nolock_unlock - unlock a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
nolock_unlock(lm_lock_t * lock, unsigned int cur_state)
{
	struct nolock_lock *lk = (struct nolock_lock *) lock;

	NOLOCK_ASSERT(atomic_read(&lk->count));
	NOLOCK_ASSERT(lk->state == cur_state);
	NOLOCK_ASSERT(cur_state != LM_ST_UNLOCKED);

	lk->state = LM_ST_UNLOCKED;

	if (atomic_read(&lk->lvb_count) && cur_state == LM_ST_EXCLUSIVE)
		memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);

	return 0;
}

/**
 * nolock_reset - Move the lock into an unlocked state (if it isn't already)
 * @lock: the lock to reset
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
nolock_reset(lm_lock_t * lock)
{
	NOLOCK_ASSERT(0);
	return 0;
}

/**
 * nolock_cancel - cancel a request on a lock
 * @lock: the lock to cancel request for
 *
 */

static void
nolock_cancel(lm_lock_t * lock, int on)
{
	return;
}

/**
 * nolock_hold_lvb - hold on to a lock value block
 * @lock: the lock the LVB is associated with
 *
 * Returns: the LVB
 */

static lm_lvb_t *
nolock_hold_lvb(lm_lock_t * lock)
{
	struct nolock_lock *lk = (struct nolock_lock *)lock;

	NOLOCK_ASSERT(atomic_read(&lk->count));

	atomic_inc(&lk->count);
	atomic_inc(&lk->lvb_count);

	return &lk->lvb;
}

/**
 * nolock_unhold_lvb - release a LVB
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

static void
nolock_unhold_lvb(lm_lock_t * lock, lm_lvb_t * lvb)
{
	struct nolock_lock *lk = (struct nolock_lock *)lock;

	NOLOCK_ASSERT(atomic_read(&lk->count));
	NOLOCK_ASSERT(atomic_read(&lk->lvb_count));

	if (atomic_dec_and_test(&lk->lvb_count)
	    && !(lk->lvb.lvb_flags & LM_LVB_PERM)) {
		memset(lk->lvb_data, 0, NOLOCK_LVB_SIZE);
		memset(lk->lvb_data_net, 0, NOLOCK_LVB_SIZE);
	}

	atomic_dec(&lk->count);
}

/**
 * nolock_sync_lvb - sync out the value of a lvb
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

static void
nolock_sync_lvb(lm_lock_t * lock, lm_lvb_t * lvb)
{
	struct nolock_lock *lk = (struct nolock_lock *)lock;

	NOLOCK_ASSERT(atomic_read(&lk->count));
	NOLOCK_ASSERT(atomic_read(&lk->lvb_count));
	NOLOCK_ASSERT(lk->state == LM_ST_EXCLUSIVE);

	memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);
}

/**
 * nolock_reset_exp - reset the expired locks for a given jid
 * @lockspace: the lockspace
 * @jid: the jid
 *
 */

static void
nolock_reset_exp(lm_lockspace_t * lockspace, unsigned int jid)
{
	return;
}

struct lm_lockops nolock_ops = {
	proto_name:		"nolock",
	local_fs:		TRUE,
	mount:			nolock_mount,
	others_may_mount:	nolock_others_may_mount,
	unmount:		nolock_unmount,
	get_lock:		nolock_get_lock,
	put_lock:		nolock_put_lock,
	lock:			nolock_lock,
	unlock:			nolock_unlock,
	reset:			nolock_reset,
	cancel:			nolock_cancel,
	hold_lvb:		nolock_hold_lvb,
	unhold_lvb:		nolock_unhold_lvb,
	sync_lvb:		nolock_sync_lvb,
	reset_exp:		nolock_reset_exp
};

static int __init
init_nolock(void)
{
	return lm_register_proto(&nolock_ops);
}

static void __exit
exit_nolock(void)
{
	lm_unregister_proto(&nolock_ops);
}

module_init(init_nolock);
module_exit(exit_nolock);
