/*
 *
 *    Connectix USB QuickCam VC Video Camera driver
 *
 *    (C) Copyright 2001 De Marchi Daniele
 *    (C) Copyright 2004 Terry Mohan
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of the
 *    License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Jan 2001 This driver develop started on the linux 
 *            kernel 2.4.0.
 *
 */

#include <linux/config.h>
#include <linux/version.h>

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0))
#include <linux/wrapper.h>
#endif
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/parport.h>
#include <linux/delay.h>

#include "qcamvc.h"

//*#define _QCAMVC_DEBUG_
//#define CONFIG_VIDEO_QCAMVC_PP_MODULE
//#undef CONFIG_PROC_FS

static int video_nr = -1;

#ifdef MODULE
MODULE_PARM(video_nr,"i");
MODULE_AUTHOR("De Marchi Daniele <demarchidaniele@libero.it>");
MODULE_DESCRIPTION("V4L-driver for QuickCam VC cameras");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("video");
#endif

#ifdef CONFIG_VIDEO_QCAMVC_PP
extern int qcamvc_pp_init(void);
#endif
#ifdef CONFIG_VIDEO_QCAMVC_USB
extern int qcamvc_usb_init(void);
#endif

static int		qcamvc_vopen(struct inode *inode, struct file *file);
static int		qcamvc_vrelease(struct inode *inode, struct file *file);
static int		qcamvc_vioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static ssize_t	qcamvc_vread(struct file *file, char *data, size_t count, loff_t *ppos);
static int		qcamvc_vmmap(struct file *file, struct vm_area_struct *vma);
static int		qcamvc_do_ioctl(struct inode *inode, struct file *file, unsigned int cmd, void *arg);
static void		qcamvc_parse(struct qcamvc_data *qcamvc, unsigned char *inbuf, unsigned char *outbuf);


static struct file_operations qcamvc_fops =
{
	.owner		= THIS_MODULE,
	.open		= qcamvc_vopen,
	.release	= qcamvc_vrelease,
	.ioctl		= qcamvc_vioctl,
	.llseek		= no_llseek,
	.read		= qcamvc_vread,
	.mmap		= qcamvc_vmmap,
};

static struct video_device qcamvc_template =
{
	.name		= "UNSET",
	.type		= VID_TYPE_CAPTURE,
	.hardware	= VID_HARDWARE_QCAM_C,
	.fops		= &qcamvc_fops,
	.minor		= -1,
};


/**********************************************************************
 *
 * Memory management
 *
 **********************************************************************/

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the area.
 */
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1);
	ret = __pa(kva);
	return ret;
}

static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size);
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

static int allocate_frame_buf(struct qcamvc_data *qcamvc)
{
	int i;
	
	if (qcamvc->frame_buf)
		return 0; //already allocated

	qcamvc->frame_buf = rvmalloc(QUICKCAM_NUMFRAMES * QUICKCAM_MAX_SIZE);
	if (!qcamvc->frame_buf)
		return -ENOBUFS;
	
	for (i = 0; i < QUICKCAM_NUMFRAMES; i++)
		qcamvc->frame[i].data = qcamvc->frame_buf + i * QUICKCAM_MAX_SIZE;
	
	return 0;
}

static int free_frame_buf(struct qcamvc_data *qcamvc)
{
	int i;
	
	if (!qcamvc->frame_buf)
		return 0; //already unallocated
	
	rvfree(qcamvc->frame_buf, QUICKCAM_NUMFRAMES * QUICKCAM_MAX_SIZE);
	qcamvc->frame_buf = 0;
	for (i=0; i < QUICKCAM_NUMFRAMES; i++)
		qcamvc->frame[i].data = NULL;
		
	return 0;
}

static void inline free_frames(struct qcamvc_frame frame[QUICKCAM_NUMFRAMES])
{
	int i;
	
	for (i=0; i < QUICKCAM_NUMFRAMES; i++)
		frame[i].state = FRAME_UNUSED;

}

static int qcamvc_vmmap(struct file *file, struct vm_area_struct *vma)
{
	struct video_device *dev = file->private_data;
	unsigned long start = vma->vm_start;
	unsigned long size  = vma->vm_end - vma->vm_start;
	unsigned long page, pos;
	struct qcamvc_data *cam = dev->priv;
	int retval;

	if (!cam || !cam->ops)
		return -ENODEV;
	
printk("qcamvc_mmap: %ld\n", size);

	if (size > QUICKCAM_NUMFRAMES*QUICKCAM_MAX_SIZE)
		return -EINVAL;

	if (!cam || !cam->ops)
		return -ENODEV;
	
	if (down_interruptible(&cam->busy_lock))
		return -EINTR;

	if (!cam->frame_buf) {	
		if ((retval = allocate_frame_buf(cam))) {
			up(&cam->busy_lock);
			return retval;
		}
	}

	pos = (unsigned long)(cam->frame_buf);
	while (size > 0) {
		page = kvirt_to_pa(pos);
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)) {
			up(&cam->busy_lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	DBG("qcamvc_mmap: %ld\n", size);
	up(&cam->busy_lock);

	return 0;
}

/* kernel thread function to read image from camera */
static int fetch_frame(void *data)
{
	struct qcamvc_data *cam = (struct qcamvc_data*)data;
	unsigned long oldjif, rate, diff;

//cam->frame[cam->curframe].state = FRAME_DONE;
//return 0;

//printk(">> cam->curframe=%i\n",cam->curframe);
	unsigned char *tempbuf = kmalloc(cam->packet_len, GFP_KERNEL);
	//unsigned char *tempbuf = rvmalloc(cam->packet_len);
//printk("__packetlen=%i\n",cam->packet_len);
	cam->frame[cam->curframe].state = FRAME_GRABBING;

	oldjif = jiffies;
	
	cam->ops->qcamvc_stream_read(cam->lowlevel_data, tempbuf, cam->packet_len);
	qcamvc_parse(cam, tempbuf, cam->frame[cam->curframe].data);
	//cam->ops->qcamvc_stream_read(cam->lowlevel_data, cam->frame_buf, cam->packet_len);
	//qcamvc_parse(cam, cam->frame_buf, cam->frame[cam->curframe].data, 0);
	rate = cam->packet_len * HZ / 1024;
	diff = jiffies-oldjif;
	cam->transfer_rate = diff==0 ? rate : rate/diff;
	
	//memcpy(qcamvc->frame[qcamvc->curframe].data, tempbuf, qcamvc->packet_len);
	kfree(tempbuf);	
	
	//rvfree(tempbuf, cam->packet_len);
	cam->frame[cam->curframe].state = FRAME_DONE;
	
	return 0;
}

static int capture_frame(struct qcamvc_data *cam, struct video_mmap *vm)
{
	if (!cam->frame_buf) {
		int err;
		if ((err = allocate_frame_buf(cam)))
			return err;
	}

	cam->curframe = vm->frame;
	cam->frame[cam->curframe].state = FRAME_READY;

	return fetch_frame(cam);
}

static int match_videosize( int width, int height )
{
	/* return the best match */

	if (width>=352 || height>=288)
		return R352x288; /* CIF */
	if (width>=320 || height>=240)
		return R320x240; /* SIF */
	if (width>=176 || height>=144)
		return R176x144; /* QCIF */

	return R160x120; /* QSIF */
}


/*****************************************************************************/
/*                                                                           */
/* QuickCam Video CODEC                                                      */
/*                                                                           */
/*****************************************************************************/
static void hue_reg(struct qcamvc_data *qcamvc)
{
	float a,b;
	int cnt;
	
	if (qcamvc->hue_calc == 1) return;
	
	a = (qcamvc->blue_hue * 1.12) / 128;
	b = (qcamvc->red_hue * 1.39999999999) / 128;  
	
	for (cnt = 0 ; cnt < 0xff ; cnt++)
	{
		qcamvc->table_a[cnt] = (float) cnt * b;
		qcamvc->table_b[cnt] = (float) cnt * a;
	
		if(qcamvc->table_a[cnt] > 0xff) qcamvc->table_a[cnt] = 0xff; if(qcamvc->table_a[cnt] < 0) qcamvc->table_a[cnt] = 0;
		if(qcamvc->table_b[cnt] > 0xff) qcamvc->table_b[cnt] = 0xff; if(qcamvc->table_b[cnt] < 0) qcamvc->table_b[cnt] = 0;
	}
	
	a = qcamvc->santrast / 100;
	
	qcamvc->VAR_A = a * 1.731999999999999;
	qcamvc->VAR_B = a * 0.336;
	qcamvc->VAR_C = a * 0.697999999999999;
	qcamvc->VAR_D = a * 1.370999999999999;
	
	qcamvc->hue_calc = 1;
}

static void get_pixel(struct qcamvc_data *qcamvc, float pix_a, float pix_b, unsigned char pix_c, \
	      unsigned char pix_d, float pix_e, \
	      unsigned char *pred, unsigned char *pgreen, unsigned char *pblue)
{
  
	float a,b;
	float blue,green,red;
	
	a = pix_b - qcamvc->table_a[pix_c];
	b = qcamvc->table_b[pix_d] - pix_e;
	
	if(pix_a >= 0) {if(pix_a > 0xff) {pix_a = 0xff;}} else {pix_a = 0;}
	if(a <= 112)   {if(a < -112) {a = -112;}} else {a = 112;}
	if(b <= 112)   {if(b < -112) {b = -112;}} else {b = 112;}
	
	blue =  ( b * qcamvc->VAR_A ) +  pix_a;
	green = ( a * qcamvc->VAR_C ) - (( b * qcamvc->VAR_B) - pix_a );
	red =   ( a * qcamvc->VAR_D ) + pix_a;
	
	/*blue =  pix_a;
	green = pix_b;
	red =   pix_c;*/

	if(red >= 0) {if(red > 0xff) red = 0xff;} else {red = 0;}
	if(green >= 0) {if(green > 0xff) green = 0xff;} else {green = 0;}  
	if(blue >= 0) {if(blue > 0xff) blue = 0xff;} else {blue = 0;}

	/* swap red & blue for V4L, BGR24 */
	if (qcamvc->mmap_kludge)
	{
		*pred   = (unsigned char) blue;
		*pgreen = (unsigned char) green;
		*pblue  = (unsigned char) red;
	}
	else
	{
		*pred   = (unsigned char) red;
		*pgreen = (unsigned char) green;
		*pblue  = (unsigned char) blue;
	}
  
}

#if 1
static void codec(struct qcamvc_data *qcamvc, unsigned char *inbuf, unsigned char *outbuf)
{
	int width=qcamvc->width;
	int height=qcamvc->height;
	unsigned char red=0, green=0, blue=0;
	int x,y;
	unsigned char pix1, pix2, pix3, pix4;
	unsigned char *pointer1, *pointer2;
	//unsigned char *table = (unsigned char *) kmalloc(width * height, GFP_KERNEL);
	unsigned char *pointer_last_line =(unsigned char *) kmalloc(width, GFP_KERNEL);
	
	/* what's this unused code supposed to have been meant for?
	
	pointer1 = inbuf;
	pointer2 = table;
	
	for (y = 0 ; y < height ; y ++ )
	{
		for (x = width - 1 ; x >= 0 ; x--)
		{
			*(pointer2 + x) = *(pointer1);
			pointer1 ++;
		}
		pointer2 = pointer2 + width;
	}
	*/	
  
	pointer1 = inbuf;
	pointer2 = (outbuf + qcamvc->width * qcamvc->height * 3) - 1;
	if (qcamvc->hue_calc == 0 ) hue_reg(qcamvc);
	memcpy( pointer_last_line, pointer1, width);
	
	for(y = 0 ; y < height ; y++)
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		if (current->need_resched)
			schedule ();
#else
		cond_resched();
#endif
		if((y & 1) == 0)
		{
			for (x = 0 ; x < width ; x++)
			{
				if((x & 1) != 0)
				{
					pix1 = (( *(width + pointer1)) + (*(x + pointer_last_line))) / 2;
					pix2 = (((*(width + pointer1 + 1)) + (*(x + pointer_last_line)) )) / 2;
					pix3 = *pointer1;
					pix4 = *(pointer1 + 1);
				}
				else
				{	  
					pix1 = ((*(width + pointer1 + 1)) + (*(x + 1 + pointer_last_line))) / 2;
					pix2 = ((*(width + pointer1)) + (*(x + pointer_last_line))) / 2;
					pix3 = *(pointer1 + 1);
					pix4 = *pointer1;
				}	
				get_pixel(qcamvc, (pix4 + pix3) / 2,  pix4, pix3, pix1, pix2, &blue, &green, &red);
				(unsigned char)*(pointer2)     = blue;  
				(unsigned char)*(pointer2 - 1) = green; 
				(unsigned char)*(pointer2 - 2) = red;
				pointer2-=3;
				pointer1++;
			}      
		}
		else
		{
			
			for (x = 0 ; x < width ; x++)
			{
				if((x & 1) != 0)
				{
					pix3 = (*(width + pointer1) + *(x + pointer_last_line)) / 2;
					pix4 = (*(width + pointer1 + 1) + *(x + 1 + pointer_last_line)) / 2;
					pix1 = *pointer1 & 0xff;
					pix2 = *(pointer1 + 1);
				}
				else
				{	  
					pix3 = (*(width + pointer1 + 1) + *(x + 1 + pointer_last_line)) / 2;
					pix4 = (*(width + pointer1)     + *(x +     pointer_last_line)) / 2;
					pix1 = *(pointer1 + 1);
					pix2 = *pointer1;
				}	
				get_pixel(qcamvc, (pix1 + pix2) / 2,  pix4, pix3, pix1, pix2, &blue, &green, &red);
				(unsigned char)*(pointer2)     = blue;  
				(unsigned char)*(pointer2 - 1) = green; 
				(unsigned char)*(pointer2 - 2) = red;
				pointer2-=3;
				pointer1++;
			}                  
		}
		memcpy(pointer_last_line, pointer1 - width, width);
	} 
	//kfree(table);
	kfree(pointer_last_line);
}
#endif

static void qcamvc_set_topmost_row(struct qcamvc_data *qcamvc,  unsigned char val)
{
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_TOPMOST_ROW, &val, sizeof(val));
}

static void qcamvc_set_bottommost_row(struct qcamvc_data *qcamvc,  unsigned char val)
{
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_BOTTOMMOST_ROW, &val, sizeof(val));
}

static void qcamvc_set_leftmost_column(struct qcamvc_data *qcamvc,  unsigned char val)
{
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_LEFTMOST_COLUMN, &val, sizeof(val));
}

static void qcamvc_set_rightmost_column(struct qcamvc_data *qcamvc,  unsigned char val)
{
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_RIGHTMOST_COLUMN, &val, sizeof(val));
}

static void qcamvc_set_exposure(struct qcamvc_data *qcamvc,  unsigned char val)
{
	if (!qcamvc || !qcamvc->lowlevel_data)
		return;

	unsigned char buffer[2];
	GET_VALUES(buffer,val);
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_EXPOSURE, buffer, sizeof(buffer));
	if(qcamvc->streaming) qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_GET_FRAME, NULL, 0);
}

static void qcamvc_set_brightness(struct qcamvc_data *qcamvc,  unsigned char val)
{
	if (!qcamvc || !qcamvc->lowlevel_data)
		return;
	
	short i,j;
	unsigned char buffer[21]=
		{
		0x58, 0xd8, 0x58, 0xd8, 0x58,
		0xd8, 0x58, 0xd8, 0x58, 0xd8,
		0x58, 0xd8, 0x58, 0xd8, 0x58,
		0xd8, 0x58, 0xd8, 0x58, 0xd8, 0x5c
		};
	
	for(i=1, j=18 ; i<=0x80 ; i*=2 , j-=2)
	{
		if( val & i)
		{
			buffer[j] = buffer[j] | 1;
			buffer[j+1] = buffer[j+1] | 1;
		}
	}
	
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_BRIGHTNESS, buffer, sizeof(buffer));
	if(qcamvc->streaming) qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_GET_FRAME, NULL, 0);
}

static int bpc6_init(struct qcamvc_data *qcamvc)
{
	int ret=0;
	long i,q;
	unsigned char buffer[0x40];
	unsigned char *bigbuffer;
	
	FIRST_ROW; LAST_ROW;
	FIRST_COL; LAST_COL;
	MUL_FACTOR;
	
	bigbuffer=kmalloc(BIGBUFFER_SIZE , GFP_KERNEL);
	if (bigbuffer==NULL) 
	{
		printk("QuickCam VC: Init bigbuffer kmalloc error.\n");
		return -1;
	}
	
	for(i=0; i<BIGBUFFER_SIZE ; i+=2)
	{
		q=i;
		*(bigbuffer + i ) = (unsigned char )(q >> 11);
		*(bigbuffer + i + 1) = (unsigned char) 0xc0;
	}

	buffer[0]=0x78; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x58; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	buffer[0]=0x18; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	
	qcamvc_set_topmost_row     (qcamvc, 0x0e);
	qcamvc_set_leftmost_column (qcamvc, 0x0f);
	qcamvc_set_bottommost_row  (qcamvc, 0x86);
	qcamvc_set_rightmost_column(qcamvc, 0xaf);
	
	qcamvc_set_brightness(qcamvc, qcamvc->brightness);
	
	buffer[0]=0x7f;buffer[1]=0x81;
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x02, buffer, 2);
	
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0d, bigbuffer, BIGBUFFER_SIZE);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	
	buffer[0]=0x01; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x78; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x58; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);

	qcamvc_set_brightness(qcamvc, qcamvc->brightness);
	qcamvc_set_exposure(qcamvc, qcamvc->exposure);
	
	buffer[0]=0x11; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);
	buffer[0]=0x11; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);
	
	qcamvc_set_topmost_row     (qcamvc, firstrow[qcamvc->video_size]);
	qcamvc_set_leftmost_column (qcamvc, firstcolumn[qcamvc->video_size]);
	qcamvc_set_bottommost_row  (qcamvc, lastrow[qcamvc->video_size]);
	qcamvc_set_rightmost_column(qcamvc, lastcolumn[qcamvc->video_size]);

	buffer[0] = mulfactor[qcamvc->video_size]; 
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);

	kfree(bigbuffer);
	
	return ret;
}

static int bpc8_init(struct qcamvc_data *qcamvc)
{
	FIRST_ROW; LAST_ROW;
	FIRST_COL; LAST_COL;
	MUL_FACTOR;
	int q,i;
	unsigned char *bigbuffer;
	unsigned char buffer[0x40];
	//unsigned char a=0xff;
	
	bigbuffer = kmalloc(BIGBUFFER_SIZE , GFP_KERNEL);
	if (bigbuffer==NULL) 
	{
		printk("QuickCam VC: Init bigbuffer kmalloc error.\n");
		return -1;
	}
	
	for(i=0; i<BIGBUFFER_SIZE ; i+=2)
	{    
		q=i;
		*(bigbuffer + i)   = 0;
		*(bigbuffer + i +1) = (unsigned char )(q >> 11);
	}
	
	buffer[0]=0x78; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x58; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	buffer[0]=0x18; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
//#ifdef CAM_DEBUG
#if 0
	qcamvc->ops->qcamvc_get_reg(qcamvc->lowlevel_data, 0x8f, buffer, sizeof(buffer));
	printk("QuickCam VC:8bpp init bpc8 0x8f, len=%d, ",sizeof(buffer));
	for (i=0 ; i < sizeof(buffer) ;  i++) printk (" 0x%02x ", buffer[i]);
	printk("\n");
	
	qcamvc->ops->qcamvc_get_reg(qcamvc->lowlevel_data, 0x80, buffer, sizeof(buffer));
	printk("QuickCam VC:Set init bcp8 0x80, len=%d, ",sizeof(buffer));
	for (i=0 ; i < sizeof(buffer) ;  i++) printk (" 0x%02x ", buffer[i]);
	printk("\n");
#endif
	printk("a\n");
	
	qcamvc_set_topmost_row     (qcamvc, 0x0e);
	qcamvc_set_leftmost_column (qcamvc, 0x0f);
	qcamvc_set_bottommost_row  (qcamvc, 0x86);
	qcamvc_set_rightmost_column(qcamvc, 0xaf);
	printk("b\n");
	qcamvc_set_brightness(qcamvc, 0x80);
	printk("c\n");
	buffer[0]=0xaf;buffer[1]=0xaf;
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x02, buffer, 2);
	printk("d\n");
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	printk("e\n");
  
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0d, bigbuffer, BIGBUFFER_SIZE);
	/*printk("QuickCam VC:8bpc bigbuffer write, len=%d, ",sizeof(buffer));
	for (i=1 ; i < BIGBUFFER_SIZE ;  i+=2) 
	{
		if (a != *(bigbuffer + i)) 
		{
			printk (" 0x%02x ", *(bigbuffer + i)); a=*(bigbuffer + i);
		}
	}
	printk("\n");
	*/
	buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	
	/*qcamvc->ops->qcamvc_get_reg(qcamvc->lowlevel_data, 0x8d, bigbuffer, BIGBUFFER_SIZE);
	printk("QuickCam VC:8bpc bigbuffer read, len=%d, ",sizeof(buffer));
	for (i=1 ; i < BIGBUFFER_SIZE ;  i+=2) 
	{
		if (a != *(bigbuffer + i)) 
		{
			printk (" 0x%02x ", *(bigbuffer + i)); a=*(bigbuffer + i);
		}
	}
	printk("\n");*/

	//buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0e, buffer, 1);
	//buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	//buffer[0]=0x78; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	//buffer[0]=0x00; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_SET_MISC, buffer, 1);
	//buffer[0]=0x58; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0f, buffer, 1);
	
	//qcamvc_set_brightness(qcamvc, qcamvc->brightness);
	//qcamvc_set_exposure(qcamvc, qcamvc->exposure);
	
	//buffer[0]=0x11; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);
	//buffer[0]=0x11; qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);

	qcamvc_set_topmost_row     (qcamvc, firstrow[qcamvc->video_size]);
	qcamvc_set_leftmost_column (qcamvc, firstcolumn[qcamvc->video_size]);
	qcamvc_set_bottommost_row  (qcamvc, lastrow[qcamvc->video_size]);
	qcamvc_set_rightmost_column(qcamvc, lastcolumn[qcamvc->video_size]);
	
	buffer[0] = mulfactor[qcamvc->video_size]; 
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, 0x0a, buffer, 1);
	
	printk("h\n");
	kfree(bigbuffer);
	
	return 0;
}

static void put_cam(struct qcamvc_camera_ops* ops)
{
	if (ops->owner)
		module_put(ops->owner);
}

static int qcamvc_camera_init(struct qcamvc_data *qcamvc)
{
	FIRST_ROW; LAST_ROW;
	FIRST_COL; LAST_COL;
	MUL_FACTOR; ISO_LEN;
	int ret=-1;

	switch(qcamvc->bpc)
	{
		case bpc6:
		{
		ret = bpc6_init(qcamvc);
		break;
		}   
		
		case bpc8:
		{
		ret = bpc8_init(qcamvc);
		break;
		}
	}

	qcamvc->packet_len = iso_len[qcamvc->bpc][qcamvc->video_size];
	qcamvc->height = lastrow[qcamvc->video_size] -    firstrow[qcamvc->video_size];
	qcamvc->width = lastcolumn[qcamvc->video_size] - firstcolumn[qcamvc->video_size];
	
	if (!(mulfactor[qcamvc->video_size] & 0x02)) 
	{
		qcamvc->height*=2;
		qcamvc->width*=2;
	}
	qcamvc->ops->qcamvc_set_reg(qcamvc->lowlevel_data, QCAM_VC_GET_FRAME, NULL, 0);
	qcamvc->streaming=1;
	
	qcamvc->hue_calc=0;

#ifdef DBG_CAM
	printk("QuickCam VC:Init at %dx%d\n", qcamvc->width, qcamvc->height);
#endif
	return ret;
}

static void bpc6_parse(struct qcamvc_data *qcamvc, unsigned char *inbuf, unsigned char *outbuf)
{
	size_t i,q;
	unsigned char *temp=kmalloc(qcamvc->width*qcamvc->height, GFP_KERNEL);
	for( i = q = 0 ; i < qcamvc->width * qcamvc->height; i+=4,q+=3 )
	{
		if ((q + 2) < qcamvc->packet_len) *(temp+i+3) = *((inbuf+2) + q + 2) & 0xfc;
		if ((q + 2) < qcamvc->packet_len) *(temp+i+2) = (((*((inbuf+2) + q + 1) >> 4) & 0x0f)|((*((inbuf+2)+ q + 2) <<4) & 0x30))<<2;
		if ((q + 1) < qcamvc->packet_len) *(temp+i+1) = (((*((inbuf+2) + q)>>6)&0x03)|((*((inbuf+2)+ q + 1)&0x0f)<<2))<<2;
		if (q < qcamvc->packet_len)       *(temp+i+0) = (*((inbuf+2) + q)&0x3f)<<2;

	}
	codec(qcamvc, temp, outbuf);
	//memcpy(outbuf,temp,qcamvc->width*qcamvc->height);
	kfree(temp);
}

static void bpc8_parse(struct qcamvc_data *qcamvc, unsigned char *inbuf, unsigned char *outbuf)
{
	unsigned char *temp=kmalloc(qcamvc->width*qcamvc->height, GFP_KERNEL);
	memcpy(outbuf, inbuf, qcamvc->packet_len);
	kfree(temp);
}

static void qcamvc_parse(struct qcamvc_data *qcamvc, unsigned char *inbuf, unsigned char *outbuf)
{
	switch(qcamvc->bpc)
	{
		case bpc6:
		{
		bpc6_parse(qcamvc, inbuf, outbuf);
		break;
		}   
		
		case bpc8:
		{
		bpc8_parse(qcamvc, inbuf, outbuf);
		break;
		}
		
	}
}

static int qcamvc_set_res(struct qcamvc_data *qcamvc, int res)
{
	if (res == qcamvc->video_size)
		return -1;
	
	qcamvc->video_size = res;
	qcamvc_camera_init(qcamvc);
	
	return qcamvc->video_size;
}

/****************************************************************************
 *
 * /proc interface
 *
 ***************************************************************************/

#ifdef CONFIG_PROC_FS

static struct proc_dir_entry *qcamvc_proc_root = NULL;

#define YES_NO(x) ((x) ? "yes" : "no")

static int qcamvc_read_proc(char *page, char **start, off_t off, int count,
			   int *eof, void *data)
{
	char *out = page;
	int len;
	struct qcamvc_data *qcamvc = data;
	
	// Stay under PAGE_SIZE or bad things will happen :(
	out+=sprintf(out, "driver_version   : %s\n", QCAMVCVERSION);
	out+=sprintf(out, "model            : %s\n", qcamvc->vcap.name);
	out+=sprintf(out, "max resolution   : %dx%d\n", qcamvc->vcap.maxwidth, qcamvc->vcap.maxheight);
	out+=sprintf(out, "min resolution   : %dx%d\n\n", qcamvc->vcap.minwidth, qcamvc->vcap.minheight);
	out+=sprintf(out, "Settings\n");
	out+=sprintf(out, "resolution       : %dx%d\n", qcamvc->width, qcamvc->height);
	out+=sprintf(out, "bit per component: %d\n", (qcamvc->bpc==bpc6)? 6 : 8);
	out+=sprintf(out, "brightness       : %d\n", qcamvc->brightness);
	out+=sprintf(out, "exposure         : %d\n", (int)qcamvc->exposure);
	out+=sprintf(out, "red hue          : %d\n", (int)qcamvc->red_hue);
	out+=sprintf(out, "blue hue         : %d\n", (int)qcamvc->blue_hue);
	out+=sprintf(out, "santrast         : %d\n", (int)qcamvc->santrast);
	out+=sprintf(out, "transfer_rate    : %dkB/s\n", qcamvc->transfer_rate);
	
	out+=sprintf(out, "\n");
	
	len = out - page;
	len -= off;
	if (len < count)
	{
		*eof = 1;
		if (len <= 0) return 0;
	}
	else
	{
		len = count;
	}
	
	*start = page + off;
	
	return len;	
}

static void setparam (struct qcamvc_data *qcamvc, char *name, char *sval)
{
	if (down_interruptible(&qcamvc->busy_lock))
		return;
	
	unsigned int val;
	
	if (!qcamvc || !sval) return;
	if (!isdigit(sval[0])) return;

	val = simple_strtoul(sval, NULL, 0);

#define MATCH(x) (strcmp(name, x)==0)
	
	if (MATCH("brightness") && (val <= 255))
	{
		qcamvc->brightness = val;
		qcamvc_set_brightness(qcamvc,qcamvc->brightness);
	}
	else if (MATCH("exposure") && (val <= 255))
	{		
		qcamvc->exposure = val;
		qcamvc_set_exposure(qcamvc,qcamvc->exposure);
	}
	else if (MATCH("contrast") && (val <= 255))
	{
		qcamvc->blue_hue = val;
		qcamvc->hue_calc=0;
	}
	else if (MATCH("hue") && (val <= 255))
	{
		qcamvc->red_hue = val;
		qcamvc->hue_calc=0;
	}
	else if (MATCH("colour") && (val <= 255))
	{
		qcamvc->santrast = val;
		qcamvc->hue_calc=0;
	}

	printk("Setting: %s=%i\n", name, val);

#undef MATCH
 
	up(&qcamvc->busy_lock);
}

/* copied from zoran_procfs */
static int qcamvc_write_proc(struct file *file, const char *buffer,
                           unsigned long count, void *data)
{
	char *string, *sp;
	char *line, *ldelim, *varname, *svar, *tdelim;

	struct qcamvc_data* qcamvc = (struct qcamvc_data*) data;

	string = sp = vmalloc(count + 1);
	if (!string) {
		printk("write_proc: can not allocate memory\n");
		return -ENOMEM;
	}
	if (copy_from_user(string, buffer, count)) {
		vfree (string);
		return -EFAULT;
	}
	string[count] = 0;

	ldelim = " \t\n";
	tdelim = "=";
	line = strpbrk(sp, ldelim);
	while (line) {
		*line = 0;
		svar = strpbrk(sp, tdelim);
		if (svar) {
			*svar = 0;
			varname = sp;
			svar++;
			setparam(qcamvc, varname, svar);
		}
		sp = line + 1;
		line = strpbrk(sp, ldelim);
	}
	vfree(string);

	return count;
}

static void create_proc_qcamvc_cam(struct qcamvc_data *qcamvc)
{
	char name[8];
	struct proc_dir_entry *ent;
	
	if (!qcamvc_proc_root || !qcamvc)
		return;

	printk(KERN_INFO "QuickCam VC: Creating a camera entry in /proc/qcamvc\n");
	//printk("QuickCam VC: QuickCam dev struct in create_proc_qcamvc %p\n",qcamvc);

	sprintf(name, "video%d", qcamvc->vdev->minor);
	
	ent = create_proc_entry(name, S_IFREG|S_IRUGO|S_IWUSR, qcamvc_proc_root);
	if (!ent)
	{
		printk(KERN_ERR "QuickCam VC: couldn't create /proc/qcamvc/%s\n", name);
		return;
	}

	ent->data = qcamvc;
	ent->read_proc = qcamvc_read_proc;
	ent->write_proc = qcamvc_write_proc;

	//FIXME: size for VC?
	ent->size = 3736 + 189;
	qcamvc->proc_entry = ent;
}

static void destroy_proc_qcamvc_cam(struct qcamvc_data *qcamvc)
{
	char name[8];
	
	if (!qcamvc || !qcamvc->proc_entry)
		return;
	
	sprintf(name, "video%d", qcamvc->vdev->minor);
	remove_proc_entry(name, qcamvc_proc_root);
	qcamvc->proc_entry = NULL;
}

static void proc_qcamvc_create(void)
{
	qcamvc_proc_root = create_proc_entry("qcamvc", S_IFDIR, 0);

	if (qcamvc_proc_root)
		qcamvc_proc_root->owner = THIS_MODULE;
	else
		LOG("Unable to initialise /proc/qcamvc\n");
}

static void __exit proc_qcamvc_destroy(void)
{
	remove_proc_entry("qcamvc", 0);
}
#endif /* CONFIG_PROC_FS && CONFIG_VIDEO_PROC_FS */

/*****************************************************************************/
/*                                                                           */
/* Video 4 Linux                                                             */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

static int qcamvc_vopen(struct inode *inode, struct file *file)
{
	struct video_device *dev = video_devdata(file);
	struct qcamvc_data *qcamvc = (struct qcamvc_data *)dev->priv;
  
#ifdef DBG_CAM
	printk("QuickCam qcamvc_vopen:\n");
#endif

	if (!qcamvc) {
		printk("Internal error, qcamvc_data not found!\n");
		return -ENODEV;
	}

	if (qcamvc->open_count > 0) {
		printk("Camera already open\n");
		return -EBUSY;
	}

	if (!try_module_get(qcamvc->ops->owner))
	{
		printk("qcamvc_open: try_module_get() FAILED\n");
		return -ENODEV;
	}

	/* Set the frames */
	qcamvc->curframe = 0;
	unsigned int i;
	for (i = 0; i < QUICKCAM_NUMFRAMES; i++)
	{
		qcamvc->frame[i].state = FRAME_UNUSED;
		qcamvc->frame[i].data = NULL;
	}

	if (qcamvc->ops->qcamvc_open(qcamvc->lowlevel_data))
		return 0;

	/* Set ownership of /proc/cpia/videoX to current user */
	if(qcamvc->proc_entry)
		qcamvc->proc_entry->uid = current->uid;

	file->private_data = dev;
	qcamvc->open_count++;
	
	return 0;
}

static int qcamvc_vrelease(struct inode *inode, struct file *file)
{
	struct video_device *dev = video_devdata(file);
	struct qcamvc_data *qcamvc = (struct qcamvc_data *)dev->priv;

	if (qcamvc->ops)
	{
		/* Return ownership of /proc/qcamvc/videoX to root */
		if(qcamvc->proc_entry)
			qcamvc->proc_entry->uid = 0;
	
		/* close cpia */
		qcamvc->ops->qcamvc_close(qcamvc->lowlevel_data);

		put_cam(qcamvc->ops);
	}

	if (--qcamvc->open_count == 0)
	{
		if (qcamvc->frame_buf)
			free_frame_buf(qcamvc);

		if (!qcamvc->ops)
			kfree(qcamvc);
	}
	
	file->private_data = NULL;

	return 0;
}

static ssize_t qcamvc_vread(struct file *file, char *data, size_t count, loff_t *ppos)
{ 
	struct video_device *dev = file->private_data;
	struct qcamvc_data *qcamvc = dev->priv;
	unsigned long oldjif, rate, diff;
	unsigned char *tempbuf;

	if (down_interruptible(&qcamvc->busy_lock))
		return -EINTR;
	
	if (!data || !count || !qcamvc->ops)
		return -EINVAL;
	
	tempbuf = kmalloc(qcamvc->packet_len, GFP_KERNEL);
	//tempbuf = rvmalloc(qcamvc->packet_len);
	
	if (!tempbuf)
	{
		up(&qcamvc->busy_lock);
		return -EINVAL;
	}

#ifdef DBG_CAM
	printk("USB QuickCam quickcam_read count=%ld\n",count);
#endif

	oldjif = jiffies;
	
	qcamvc->ops->qcamvc_stream_read(qcamvc->lowlevel_data, tempbuf, qcamvc->packet_len);
	qcamvc_parse(qcamvc, tempbuf, data);
	
	rate = qcamvc->packet_len * HZ / 1024;
	diff = jiffies-oldjif;
	qcamvc->transfer_rate = diff==0 ? rate : rate/diff;
	
	kfree(tempbuf);
	//rvfree(tempbuf, qcamvc->packet_len);
	up(&qcamvc->busy_lock);
	
	return count;
}

static int qcamvc_vioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, qcamvc_do_ioctl);
}

static int qcamvc_do_ioctl(struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
	struct video_device *dev = file->private_data;
	struct qcamvc_data *qcamvc = dev->priv;

	int retval = 0;

	if (!qcamvc || !qcamvc->ops)
		return -ENODEV;

	//if (down_interruptible(&qcamvc->busy_lock))
	//	return -EINTR;

	switch(cmd)
	{
      
	/* Return capability */
	case VIDIOCGCAP:
	{
		struct video_capability *cap = arg;
		memset(cap,0,sizeof(*cap));

		strcpy(cap->name, qcamvc->vcap.name);
		cap->type = qcamvc->vcap.type;
		cap->channels = qcamvc->vcap.channels;
		cap->audios = qcamvc->vcap.audios;
		cap->maxwidth = qcamvc->vcap.maxwidth;
		cap->maxheight = qcamvc->vcap.maxheight;
		cap->minwidth = qcamvc->vcap.minwidth;
		cap->minheight = qcamvc->vcap.minheight;
		break;
	}
      
	/* Video channel ioctl */
	case VIDIOCGCHAN:
	{
		struct video_channel *v = arg;
		if (v->channel != 0)
		{
			retval = -EINVAL;
			break;
		}
		
		v->channel = qcamvc->vchan.channel;
		strcpy(v->name, qcamvc->vchan.name);
		v->tuners = qcamvc->vchan.tuners;
		v->flags = qcamvc->vchan.flags;
		v->type = qcamvc->vchan.type;
		v->norm = 0;
		break;
	}
      
	case VIDIOCSCHAN:
	{
		struct video_channel *v = arg;
		if (v->channel != 0)
			retval = -EINVAL;
		break;
	}
      
	/* Video picture ioctl */
	case VIDIOCGPICT:
	{
		struct video_picture *pic = arg;
		*pic = qcamvc->vpic;
		break;
	}
      
	case VIDIOCSPICT:
	{
		struct video_picture *vp = arg;
		
		/*if (vp->depth != 24 || vp->palette != VIDEO_PALETTE_RGB24)
		{
			retval = -EINVAL;
			break;
		}*/
		
		if (vp->brightness >> 8 != qcamvc->brightness)
		{
			down(&qcamvc->busy_lock);
			qcamvc->brightness = vp->brightness >> 8;
			qcamvc_set_brightness(qcamvc,qcamvc->brightness);
			up(&qcamvc->busy_lock);
		}

		if (vp->whiteness >> 8 != qcamvc->exposure)
		{
			down(&qcamvc->busy_lock);
			qcamvc->exposure = vp->whiteness >> 8;
			qcamvc_set_exposure(qcamvc,qcamvc->exposure);
			up(&qcamvc->busy_lock);
		}

		if (vp->contrast != qcamvc->vpic.contrast)
		{
			qcamvc->blue_hue = vp->contrast >> 8;
			qcamvc->hue_calc=0;
		}

		if (vp->hue != qcamvc->vpic.hue)
		{
			qcamvc->red_hue = vp->hue >> 8;
			qcamvc->hue_calc=0;
		}

		if (vp->colour  != qcamvc->vpic.colour)
		{
			qcamvc->santrast = vp->colour >> 8;
			qcamvc->hue_calc=0;
		}

		qcamvc->vpic = *vp;
		break;
	}
      
	/* Capture window ioctl */
	case VIDIOCSWIN:
	{
		struct video_window *vw = arg;
		int video_size = qcamvc->video_size;
//printk("VIDIOCSWIN\n");
		if(vw->flags)
		{
			printk("1\n");
			retval = -EINVAL;
			break;
		}
   
		if(vw->clipcount)
		{
			printk("2\n");
			retval = -EINVAL;
			break;
		}
		       
		if(vw->height<MIN_HEIGHT || vw->height>MAX_HEIGHT)
		{
			printk("Want to set %dx%d 3\n",vw->width, vw->height);
			retval = -EINVAL;
			break;
		}
 
		if(vw->width<MIN_WIDTH || vw->width>MAX_WIDTH)
		{
			printk("Want to set %dx%d 4\n",vw->width, vw->height);
			retval = -EINVAL;
			break;
		}
		
		video_size = match_videosize(vw->width, vw->height);
		if (video_size != qcamvc->video_size)
		{
			down(&qcamvc->busy_lock);
			qcamvc_set_res(qcamvc, video_size);
			up(&qcamvc->busy_lock);
		}
		break;
	}
      
	case VIDIOCGWIN:
	{
//printk("VIDIOCGWIN\n");
		struct video_window *vw = arg;
		memset(vw, 0, sizeof(struct video_window));
		vw->width  = qcamvc->width;
		vw->height = qcamvc->height;
		vw->flags  = 0;
		break;
	}

	/* mmap interface */
	case VIDIOCGMBUF:
	{
		struct video_mbuf *vm = arg;
		int i;

		DBG("VIDIOCGMBUF\n");
		memset(vm, 0, sizeof(*vm));
		vm->size = QUICKCAM_MAX_SIZE*QUICKCAM_NUMFRAMES;
		vm->frames = QUICKCAM_NUMFRAMES;
		for (i = 0; i < QUICKCAM_NUMFRAMES; i++)
			vm->offsets[i] = QUICKCAM_MAX_SIZE * i;
		break;
	}
      
	case VIDIOCMCAPTURE:
	{
		struct video_mmap *vm = arg;
		int video_size;

//printk("VIDIOCMCAPTURE: %d / %d / %dx%d\n", vm->format, vm->frame, vm->width, vm->height);
		if (vm->frame<0||vm->frame>=QUICKCAM_NUMFRAMES)
		{
			retval = -EINVAL;
			break;
		}

		qcamvc->vpic.palette = vm->format;
		switch(vm->format)
		{
		case VIDEO_PALETTE_GREY:
			qcamvc->vpic.depth=8;
			break;
		case VIDEO_PALETTE_RGB555:
		case VIDEO_PALETTE_RGB565:
		case VIDEO_PALETTE_YUV422:
		case VIDEO_PALETTE_YUYV:
		case VIDEO_PALETTE_UYVY:
			qcamvc->vpic.depth = 16;
			break;
		case VIDEO_PALETTE_RGB24:
			qcamvc->vpic.depth = 24;
			break;
		case VIDEO_PALETTE_RGB32:
			qcamvc->vpic.depth = 32;
			break;
		default:
			//retval = -EINVAL;
			break;
		}
		if (retval)
			break;

		video_size = match_videosize(vm->width, vm->height);
		if (video_size != qcamvc->video_size)
		{
			qcamvc->video_size = video_size;
printk("size changing\n");
			//memset(&qcamvc->vc, 0, sizeof(qcamvc->vc));
			
			down(&qcamvc->busy_lock);
			qcamvc_set_res(qcamvc, video_size);
			up(&qcamvc->busy_lock);
		}
		
		down(&qcamvc->busy_lock);
		qcamvc->mmap_kludge = 1;
		retval = capture_frame(qcamvc, vm);
		qcamvc->mmap_kludge = 0;
		up(&qcamvc->busy_lock);
		break;
	}

	case VIDIOCSYNC:
	{
		int *frame = arg;
//printk("VIDIOCSYNC frame=%i\n",*frame);
		if (*frame<0 || *frame >= QUICKCAM_NUMFRAMES )
		{
			retval = -EINVAL;
			break;
		}
	
		switch (qcamvc->frame[*frame].state) 
		{
			case FRAME_UNUSED:
			case FRAME_READY:
			case FRAME_GRABBING:
			{
				printk("sync to unused frame %d\n", *frame);
				retval = -EINVAL;
				break;
			}
			case FRAME_DONE:
	    	qcamvc->frame[*frame].state = FRAME_UNUSED;
			break;
			
			default:
			retval = -EINVAL;
			break;
		}
		break;
	}

	case VIDIOCGUNIT:
	{
		struct video_unit *vu = arg;

		DBG("VIDIOCGUNIT\n");

		vu->video    = qcamvc->vdev->minor;
		vu->vbi      = VIDEO_NO_UNIT;
		vu->radio    = VIDEO_NO_UNIT;
		vu->audio    = VIDEO_NO_UNIT;
		vu->teletext = VIDEO_NO_UNIT;

		break;
	}
      
	/* pointless to implement overlay with this camera */
	case VIDIOCCAPTURE:
	case VIDIOCGFBUF:
	case VIDIOCSFBUF:
	case VIDIOCKEY:
	/* tuner interface - we have none */
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
	/* audio interface - we have none */
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		retval = -EINVAL;
		break;
	
	/* ***  v4l2  *** ************************************************ */
	/*case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *cap = arg;

		if (0 == v4l2)
			retval = -EINVAL;
		strcpy(cap->driver,"qcamvc");
		strlcpy(cap->card,qcamvc->vcap.name,sizeof(cap->card));
		cap->bus_info = NULL;
		cap->version = QCAMVC_VERSION_CODE;
		cap->capabilities = 
			V4L2_CAP_VIDEO_CAPTURE |
			V4L2_CAP_READWRITE | 
			V4L2_CAP_STREAMING;
		break;
	}*/
	
	default:
		printk("Unknown IOCTL cmd: %u\n", cmd);
		retval = -ENOIOCTLCMD;
		break;
    }
	
	//up(&qcamvc->busy_lock);
	return retval;
}


/*****************************************************************************/
/*                                                                           */
/* QuickCam Video Camera layer                                               */
/* This layer want be indipendent from the camera structure (PP/USB)         */
/*                                                                           */
/*****************************************************************************/
static int qcamvc_video_init(struct qcamvc_data *qcamvc, char *camtype)
{
	/* Set the video capability */
	memset(&qcamvc->vcap, 0, sizeof(struct video_capability));
	
	strcpy(qcamvc->vcap.name, camtype);
	qcamvc->vcap.type      = VID_TYPE_CAPTURE;
	qcamvc->vcap.channels  = 1;
	qcamvc->vcap.audios    = 0;
	qcamvc->vcap.maxwidth  = MAX_WIDTH;
	qcamvc->vcap.maxheight = MAX_HEIGHT;
	qcamvc->vcap.minwidth  = MIN_WIDTH;
	qcamvc->vcap.minheight = MIN_HEIGHT;
	
	/* Set the video channel structure */
	memset(&qcamvc->vchan,  0, sizeof(struct video_channel));
	qcamvc->vchan.flags   = 0;
	qcamvc->vchan.tuners  = 0;
	qcamvc->vchan.channel = 0;
	qcamvc->vchan.type    = VIDEO_TYPE_CAMERA;
	strcpy(qcamvc->vchan.name, "Camera");
	
	/* Set the video picture structure */
	memset(&qcamvc->vpic,  0, sizeof(struct video_picture));
	
	qcamvc->bpc			= bpc6;
	qcamvc->video_size	= R352x288;	
	qcamvc->width		= 352;
	qcamvc->height		= 288;
	
	qcamvc->brightness	= 175;
	qcamvc->exposure	= 150;
	qcamvc->hue_calc	= 0;
	qcamvc->blue_hue	= 132;
	qcamvc->red_hue		= 52;
	qcamvc->santrast	= 50;
	qcamvc->contrast	= 128;

	qcamvc->vpic.colour     = qcamvc->santrast   * 0xff;
	qcamvc->vpic.hue        = qcamvc->red_hue    * 0xff;
	qcamvc->vpic.contrast   = qcamvc->blue_hue   * 0xff;
	qcamvc->vpic.brightness = qcamvc->brightness * 0xff;
	qcamvc->vpic.whiteness  = qcamvc->exposure   * 0xff;
	qcamvc->vpic.depth      = 24;
	qcamvc->vpic.palette    = VIDEO_PALETTE_RGB24;
	
	qcamvc->transfer_rate = 0;
	qcamvc->curframe = 0;
	
	qcamvc->frame_buf = NULL;

	int i;
	for (i = 0; i < QUICKCAM_NUMFRAMES; i++)
	{
		qcamvc->frame[i].state = FRAME_UNUSED;
		qcamvc->frame[i].data = NULL;
	}
  
	qcamvc->vdev = video_device_alloc();
	if (qcamvc->vdev == NULL)
	{
		printk(KERN_ERR "QuickCam VC: video_device_alloc() returned NULL!");
		return -1;
	}
	*qcamvc->vdev = qcamvc_template;
	strcpy(qcamvc->vdev->name, qcamvc->vcap.name);
	qcamvc->vdev->minor = -1;
	qcamvc->vdev->dev = NULL;
	qcamvc->vdev->release = video_device_release;
	qcamvc->vdev->priv = qcamvc;

	return 0;
}

struct qcamvc_data *qcamvc_register_camera(struct qcamvc_camera_ops *ops, void *lowlevel, char *camtype)
{
	struct qcamvc_data *qcamvc;

	if ((qcamvc = kmalloc(sizeof(struct qcamvc_data), GFP_KERNEL)) == NULL)
	{
		printk(KERN_ERR "QuickCam VC: Error allocating memory in qcamvc_register_camera\n");
		return NULL;
	}

	memset(qcamvc, 0, sizeof(struct qcamvc_data));
	qcamvc->ops = ops;
	init_MUTEX(&qcamvc->busy_lock);
	
	if (qcamvc_video_init(qcamvc, camtype) < 0)
	{
		kfree(qcamvc);
		printk(KERN_ERR "QuickCam VC: qcamvc_video_init failed.\n"); 
		return NULL;
	}

	qcamvc->lowlevel_data = lowlevel;
	qcamvc->open_count = 0;
	qcamvc->mmap_kludge = 0;
	
	if (video_register_device(qcamvc->vdev, VFL_TYPE_GRABBER, video_nr) == -1)
	{
		kfree(qcamvc);
		printk(KERN_ERR "QuickCam VC: video_register_device failed.\n"); 
		return NULL;
	}

	if (qcamvc->ops->qcamvc_open(qcamvc->lowlevel_data))
		return qcamvc;

	qcamvc_camera_init(qcamvc);
	
	qcamvc->ops->qcamvc_close(qcamvc->lowlevel_data);

#ifdef CONFIG_PROC_FS
	create_proc_qcamvc_cam(qcamvc);
#endif
	
	return qcamvc;
}

void qcamvc_unregister_camera(struct qcamvc_data *qcamvc)
{
#ifdef CONFIG_PROC_FS
	printk(KERN_INFO "QuickCam VC: destroying /proc/qcamvc/video%d\n", qcamvc->vdev->minor);
	destroy_proc_qcamvc_cam(qcamvc);
#endif	

	printk(KERN_INFO "QuickCam VC: unregistering video...\n");
	
	if (qcamvc->vdev)
	{
		if (-1 != qcamvc->vdev->minor)
			video_unregister_device(qcamvc->vdev);
		else
			video_device_release(qcamvc->vdev);
		qcamvc->vdev = NULL;
	}

	if (qcamvc->open_count)
	{
		put_cam(qcamvc->ops);
		printk(KERN_INFO "QuickCam VC: camera open -- setting ops to NULL\n");
		qcamvc->ops = NULL;
	}

	if (!qcamvc->open_count)
	{
		printk(KERN_INFO "QuickCam VC: freeing camera\n");
		kfree(qcamvc);
	}

}

/*****************************************************************************/
/*                                                                           */
/* QuickCam Video Camera layer                                               */
/* Module subs                                                               */
/*                                                                           */
/*****************************************************************************/

static int __init qcamvc_driver_init(void)
{
	printk("QuickCam VC: (C) 2001 by De Marchi Daniele, <demarchidaniele@libero.it>\n");
	printk("QuickCam VC: v4l level driver version %s registered.\n",QCAMVCVERSION);
//#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
#ifdef CONFIG_PROC_FS
	proc_qcamvc_create();
#endif
	
#ifdef CONFIG_KMOD
#ifdef CONFIG_VIDEO_QCAMVC_PP_MODULE
	request_module("qcamvc_pp");
#endif

#ifdef CONFIG_VIDEO_QCAMVC_USB_MODULE
	request_module("qcamvc_usb");
#endif
#endif	/* CONFIG_KMOD */
	
#ifdef CONFIG_VIDEO_QCAMVC_PP
	qcamvc_pp_init();
#endif
#ifdef CONFIG_VIDEO_QCAMVC_USB
	qcamvc_usb_init();
#endif

	return 0;
}

static void __exit qcamvc_driver_exit(void)
{
#ifdef CONFIG_PROC_FS
	proc_qcamvc_destroy();
#endif
	printk(KERN_INFO "QuickCam VC: v4l driver unregistered\n");
}

module_init(qcamvc_driver_init);
module_exit(qcamvc_driver_exit);

/* Exported symbols for modules. */
EXPORT_SYMBOL(qcamvc_register_camera);
EXPORT_SYMBOL(qcamvc_unregister_camera);

