/* ------------------------------------------------------------------------- */
/* ch7003b.c I2C driver for Chrontel CH7003B video encoder chips	     */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 2000 Ferenc Bakonyi <fero@drama.obuda.kando.hu>
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/videodev.h>
#include <linux/video_encoder.h>

#include "rivatv.h"	/* HACK for register dump! */



/*----------------------------------------------------------------------
 *	CH7003B I2C 'driver' driver
 *----------------------------------------------------------------------*/

static int ch7003b_attach(struct i2c_adapter *adap);
static int ch7003b_detach(struct i2c_client *client);
static int ch7003b_command(struct i2c_client *client, 
		unsigned int cmd, void *arg);
static void ch7003b_inc_use(struct i2c_client *client);
static void ch7003b_dec_use(struct i2c_client *client);

/* Possible CH7003B addresses: 0x75, 0x76 */
static unsigned short normal_i2c[] = {0x75, 0x76, I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};

/* Magic definition of all other I2C variables and things */
I2C_CLIENT_INSMOD;

/* Number of CH7003B registers */
#define NR_REGISTER	43

/* CH7003B instance */
struct ch7003b {
	unsigned char reg[NR_REGISTER];	/* local copy of ch7003b's registers*/
};

/*  For registering the I2C 'driver' driver */
static struct i2c_driver i2c_driver_ch7003b = {
	name:		"CH7003B",
	id:		I2C_DRIVERID_CH7003,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	ch7003b_attach,
	detach_client:	ch7003b_detach,
	command:	ch7003b_command,
	inc_use:	ch7003b_inc_use,
	dec_use:	ch7003b_dec_use
};

/* Template for new clients */
static struct i2c_client ch7003b_client_template = {
	name:		"CH7003B",
	id:		-1,
	flags:		0,
	addr:		0,
	adapter:	NULL,
	driver:		&i2c_driver_ch7003b,
	data:		NULL /* struct ch7003b *data */
};

/* Unique ID allocation */
static int ch7003b_id = 0;



/* Helper functions */
#if 0
/* Read from a register */
static int ch7003b_read(struct i2c_client *client, unsigned char subaddr)
{
	return i2c_smbus_read_byte_data(client, subaddr);
}

/* Write to a register */
static int ch7003b_write(struct i2c_client *client, unsigned char subaddr, unsigned char data)
{
	struct ch7003b *dev = (struct ch7003b *)(client->data);

	dev->reg[subaddr] = data;
	return i2c_smbus_write_byte_data(client, subaddr, data);
}
#endif
#ifdef READ_REGISTERS
/* Read all registers */
static int ch7003b_read_block(struct i2c_client *client, unsigned char *data)
{
	int i;

	for (i = 0 ; i < NR_REGISTER ; i++)
    		data[i] = i2c_smbus_read_byte_data(client, i);
	return NR_REGISTER;
}

/* Write to a register block */
static int ch7003b_write_block(struct i2c_client *client, unsigned char *data, unsigned int len)
{
	struct ch7003b *dev = (struct ch7003b *)(client->data);
	int i, err;

	if (len > NR_REGISTER) len = NR_REGISTER;
	for (i = 0 ; i < len ; i++) {
		if ((err = i2c_smbus_write_byte_data(client, i, data[i])))
			return err;
    		dev->reg[i] = data[i];
	}
	return 0;
}
#endif

/* I2C driver functions */

/* Called when a 'CH7003B like' device found */
static int ch7003b_detect(struct i2c_adapter *adap, int addr, unsigned short flags, int kind)
{
	int err = 0;
	struct i2c_client *client;
	struct ch7003b *encoder;
	int version;

	if (!i2c_check_functionality(adap, I2C_FUNC_SMBUS_READ_BYTE_DATA | 
	        		          I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
		goto err_out;

	client = kmalloc(sizeof(*client), GFP_KERNEL);
	if (client == NULL) {
		err = -ENOMEM;
		goto err_out;
	}
	
	memcpy(client, &ch7003b_client_template, sizeof(*client));
	client->adapter = adap;
        client->addr = addr;
        client->id = ch7003b_id++;

	if (kind < 0) {
		version = i2c_smbus_read_byte_data(client, 0x25);
		if (version == -1)
			goto err_out_kfree_client;
		if (version == 0x02) {
			printk("ch7003b: video encoder chip found. Chip version: %#x\n", version);
		} else {
			printk("ch7003b: unknown CH700x chip found. (Maybe 7005?) Chip version: %#x\n", version);
			goto err_out_kfree_client;
		}
	} else {
		printk("ch7003b: detection skipped\n");
	}
	
	encoder = kmalloc(sizeof(struct ch7003b), GFP_KERNEL);
	if (encoder == NULL) {
		err = -ENOMEM;
		goto err_out_kfree_client;
	}

	memset(encoder, 0, sizeof(struct ch7003b));
	client->data = encoder;

	if ((err = i2c_attach_client(client)))
		goto err_out_kfree_encoder;

	return 0;
	
err_out_kfree_encoder:
	kfree(encoder);
err_out_kfree_client:
	kfree(client);
err_out:
	return err;
}

/* Called when a new I2C bus found */
static int ch7003b_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, ch7003b_detect);
}

/* Called on exit */
static int ch7003b_detach(struct i2c_client *client)
{
	int err;

	if ((err = i2c_detach_client(client)))
		return err;
	kfree(client->data);
	kfree(client);
	return 0;
}

/* Interface to the world */
static int ch7003b_command(struct i2c_client *client,
			     unsigned int cmd, void *arg)
{
	switch (cmd) {
#ifdef READ_REGISTERS
		case READ_REGISTERS: /* for debugging */
		{
			return ch7003b_read_block(client, arg); /* bytes read */
		}
		case WRITE_REGISTERS: /* for debugging */
		{
			ch7003b_write_block(client, arg, NR_REGISTER);
			return NR_REGISTER;
		}
		case GET_NR_OF_REGISTERS: /* for debugging */
		{
			return NR_REGISTER;
		}
#endif /* READ_REGISTERS */
		case ENCODER_GET_CAPABILITIES:
		case ENCODER_SET_NORM:
		case ENCODER_SET_INPUT:
		case ENCODER_SET_OUTPUT:
		case ENCODER_ENABLE_OUTPUT:
		default:
			return -EINVAL;
	}
	return 0;
}

static void ch7003b_inc_use(struct i2c_client *client)
{
	MOD_INC_USE_COUNT;
}

static void ch7003b_dec_use(struct i2c_client *client)
{
	MOD_DEC_USE_COUNT;
}



/*---------------------------------------------------------------------- 
 *	Modularization
 *----------------------------------------------------------------------*/

MODULE_AUTHOR("Ferenc Bakonyi <fero@drama.obuda.kando.hu>");
MODULE_DESCRIPTION("CH7003B driver");
MODULE_LICENSE("GPL");

int __init init_ch7003b(void) 
{
	return i2c_add_driver(&i2c_driver_ch7003b);
}

void __exit cleanup_ch7003b(void) 
{
	i2c_del_driver(&i2c_driver_ch7003b);
}

module_init(init_ch7003b);
module_exit(cleanup_ch7003b);
