/*
 * cpia_pp CPiA Parallel Port driver
 *
 * Supports CPiA based parallel port Video Camera's.
 *
 * (C) Copyright 1999 Bas Huisman <bhuism@cs.utwente.nl>
 * (C) Copyright 1999-2000 Scott J. Bertin <sbertin@securenym.net>,
 * (C) Copyright 1999-2000 Peter Pregler <Peter_Pregler@email.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* define _CPIA_DEBUG_ for verbose debug output (see cpia.h) */
/* #define _CPIA_DEBUG_  1 */  

#undef CPIA_PARPORT_DMA
#define CPIA_PARPORT_DMA 1   /* comment this out to remove ECP DMA support */
#include <linux/config.h>
#include <linux/version.h>

#include <linux/module.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/parport.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <asm/dma.h>
#include <linux/parport_pc.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,3))
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#else
#include <linux/kmod.h>
#endif

#include "cpia.h"

static int cpia_pp_open(void *privdata, int alt);
static int cpia_pp_registerCallback(void *privdata, void (*cb) (void *cbdata),
                                    void *cbdata);
static int cpia_pp_transferCmd(void *privdata, u8 *command, u8 *data);
static int cpia_pp_streamStart(void *privdata);
static int cpia_pp_streamStop(void *privdata);
static int cpia_pp_streamRead(void *privdata, u8 *buffer, int noblock);
static int cpia_pp_close(void *privdata);
static int cpia_pp_upload_mode(void *privdata);

#define ABOUT "Parallel port driver for Vision CPiA based cameras"

#define PACKET_LENGTH  8

/* Magic numbers for defining port-device mappings */
#define PPCPIA_PARPORT_UNSPEC -4
#define PPCPIA_PARPORT_AUTO -3
#define PPCPIA_PARPORT_OFF -2
#define PPCPIA_PARPORT_NONE -1

#ifdef MODULE
static int parport_nr[PARPORT_MAX] = {[0 ... PARPORT_MAX - 1] = PPCPIA_PARPORT_UNSPEC};
static char *parport[PARPORT_MAX] = {NULL,};

MODULE_AUTHOR("B. Huisman <bhuism@cs.utwente.nl> & Peter Pregler <Peter_Pregler@email.com>");
MODULE_DESCRIPTION("Parallel port driver for Vision CPiA based cameras");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10))
MODULE_LICENSE("GPL");
#endif

MODULE_PARM(parport, "1-" __MODULE_STRING(PARPORT_MAX) "s");
MODULE_PARM_DESC(parport, "'auto' or a list of parallel port numbers. Just like lp.");
#else
static int parport_nr[PARPORT_MAX] __initdata =
	{[0 ... PARPORT_MAX - 1] = PPCPIA_PARPORT_UNSPEC};
static int parport_ptr = 0;
#endif


#define NIBBLE_UPLOAD 0x00
#define ECP_UPLOAD    0x01

struct pp_cam_entry {
	struct pardevice *pdev;
	struct parport *port;
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0))
	struct tq_struct cb_task;
#else
	struct work_struct cb_task;
#endif
	int open_count;
	wait_queue_head_t wq_stream;
	/* image state flags */
	int image_ready;	/* we got an interrupt */
	int image_complete;	/* we have seen 4 EOI */

	int streaming; /* we are in streaming mode */
	int stream_irq;
	int upload_mode;  /* IEEE1284 mode used for uploads  */
};

static struct cpia_camera_ops cpia_pp_ops = 
{
	cpia_pp_open,
	cpia_pp_registerCallback,
	cpia_pp_transferCmd,
	cpia_pp_streamStart,
	cpia_pp_streamStop,
	cpia_pp_streamRead,
	cpia_pp_close,
	CAM_TYPE_PP,
	cpia_pp_upload_mode,
	1,
	THIS_MODULE
};

static LIST_HEAD(cam_list);
static spinlock_t cam_list_lock_pp;

/* FIXME */
static void cpia_parport_enable_irq( struct parport *port ) {
	parport_enable_irq(port);
	mdelay(10);
	return;
}

static void cpia_parport_disable_irq( struct parport *port ) {
	parport_disable_irq(port);
	mdelay(10);
	return;
}

/****************************************************************************
 *
 *  Low level parport functions for nibble and ECP DMA uploads
 *
 ***************************************************************************/
/* Special CPiA PPC modes: These are invoked by using the 1284 Extensibilty
 * Link Flag during negotiation */  
#define UPLOAD_FLAG  0x08
#define NIBBLE_TRANSFER 0x01
#define ECP_TRANSFER 0x03

#define CPIA_UPLOAD_CHUNKSIZE 4096       /* for non-ECP hardware uploads */
/* FIXME: what chunksize for non-DMA ECP hardware (PIO) ? */
#define CPIA_FIFO_UPLOAD_CHUNKSIZE 4096  /* for ECP PIO hardware uploads */

/*  CPiA nonstandard "Nibble" mode (no nDataAvail signal after each byte). */
/* The standard kernel parport_ieee1284_read_nibble() fails with the CPiA... */


static size_t cpia_read_nibble (struct parport *port, 
			 void *buffer, size_t len, 
			 int flags)
{

  /* adapted verbatim, with one change, from parport_ieee1284_read_nibble()
     in drivers/parport/ieee1284-ops.c, by Tim Waugh */

	unsigned char *buf = buffer;
	int i;
	unsigned char byte = 0;
	
	len *= 2; /* in nibbles */
	for (i=0; i < len; i++) {
		unsigned char nibble;

		/* The CPiA firmware suppresses the use of nDataAvail (nFault LO)
		 * after every second nibble to signal that more
		 * data is available.  (the total number of Bytes that
		 * should be sent is known; if too few are received, an error
		 * will be recorded after a timeout).  
		 * This is incompatible with parport_ieee1284_read_nibble(),
		 * which expects to find nFault LO after every second nibble.
		 */

		/* Solution: modify cpia_read_nibble to only check for 
		 * nDataAvail before the first nibble is sent.
		 */

		/* Does the error line indicate end of data? */
		if (((i /*& 1*/) == 0) &&
		    (parport_read_status(port) & PARPORT_STATUS_ERROR)) {
			port->physport->ieee1284.phase = IEEE1284_PH_HBUSY_DNA;
				DBG("%s: No more nibble data (%d bytes)\n",
				port->name, i/2);

			/* Go to reverse idle phase. */
			parport_frob_control (port,
					      PARPORT_CONTROL_AUTOFD,
					      PARPORT_CONTROL_AUTOFD);
			port->physport->ieee1284.phase = IEEE1284_PH_REV_IDLE;
			break;
		}

		/* Event 7: Set nAutoFd low. */
		parport_frob_control (port,
				      PARPORT_CONTROL_AUTOFD,
				      PARPORT_CONTROL_AUTOFD);

		/* Event 9: nAck goes low. */
		port->ieee1284.phase = IEEE1284_PH_REV_DATA;
		if (parport_wait_peripheral (port,
					     PARPORT_STATUS_ACK, 0)) {
			/* Timeout -- no more data? */
				 DBG("%s: Nibble timeout at event 9 (%d bytes)\n",
				 port->name, i/2);
			parport_frob_control (port, PARPORT_CONTROL_AUTOFD, 0);
			break;
		}


		/* Read a nibble. */
		nibble = parport_read_status (port) >> 3;
		nibble &= ~8;
		if ((nibble & 0x10) == 0)
			nibble |= 8;
		nibble &= 0xf;

		/* Event 10: Set nAutoFd high. */
		parport_frob_control (port, PARPORT_CONTROL_AUTOFD, 0);

		/* Event 11: nAck goes high. */
		if (parport_wait_peripheral (port,
					     PARPORT_STATUS_ACK,
					     PARPORT_STATUS_ACK)) {
			/* Timeout -- no more data? */
			DBG("%s: Nibble timeout at event 11\n",
				 port->name);
			break;
		}

		if (i & 1) {
			/* Second nibble */
			byte |= nibble << 4;
			*buf++ = byte;
		} else 
			byte = nibble;
	}

	i /= 2; /* i is now in bytes */

	if (i == len) {
		/* Read the last nibble without checking data avail. */
		port = port->physport;
		if (parport_read_status (port) & PARPORT_STATUS_ERROR)
			port->ieee1284.phase = IEEE1284_PH_HBUSY_DNA;
		else
			port->ieee1284.phase = IEEE1284_PH_HBUSY_DAVAIL;
	}

	return i;
}

/* CPiA nonstandard "Nibble Stream" mode (2 nibbles per cycle, instead of 1)
 * (See CPiA Data sheet p. 31) 
 * 
 * "Nibble Stream" mode used by CPiA for uploads to non-ECP ports is a 
 * nonstandard variant of nibble mode which allows the same (mediocre) 
 * data flow of 8 bits per cycle as software-enabled ECP by TRISTATE-capable 
 * parallel ports, but works also for  non-TRISTATE-capable ports.
 * (Standard nibble mode only send 4 bits per cycle)
 *
 */

static size_t cpia_read_nibble_stream(struct parport *port, 
			       void *buffer, size_t len, 
			       int flags)
{
	int i;
	unsigned char *buf = buffer;
	int endseen = 0;

	for (i=0; i < len; i++) {
		unsigned char nibble[2], byte = 0;
		int j;

		/* Image Data is complete when 4 consecutive EOI bytes (0xff) are seen */ 
		if (endseen > 3 )
			break;

		/* Event 7: Set nAutoFd low. */
		parport_frob_control (port,
				      PARPORT_CONTROL_AUTOFD,
				      PARPORT_CONTROL_AUTOFD);
		
		/* Event 9: nAck goes low. */
		port->ieee1284.phase = IEEE1284_PH_REV_DATA;
		if (parport_wait_peripheral (port,
					     PARPORT_STATUS_ACK, 0)) {
			/* Timeout -- no more data? */
				 DBG("%s: Nibble timeout at event 9 (%d bytes)\n",
				 port->name, i/2);
			parport_frob_control (port, PARPORT_CONTROL_AUTOFD, 0);
			break;
		}

		/* Read lower nibble */
		nibble[0] = parport_read_status (port) >>3;
		
		/* Event 10: Set nAutoFd high. */
		parport_frob_control (port, PARPORT_CONTROL_AUTOFD, 0);

		/* Event 11: nAck goes high. */
		if (parport_wait_peripheral (port,
					     PARPORT_STATUS_ACK,
					     PARPORT_STATUS_ACK)) {
			/* Timeout -- no more data? */
			DBG("%s: Nibble timeout at event 11\n",
				 port->name);
			break;
		}
		
		/* Read upper nibble */
		nibble[1] = parport_read_status (port) >>3;
		
		/* reassemble the byte */
		for (j = 0; j < 2 ; j++ ) {
			nibble[j] &= ~8;
			if ((nibble[j] & 0x10) == 0)
				nibble[j] |= 8;
			nibble[j] &= 0xf;
		}
		byte = (nibble[0] |(nibble[1] << 4));
		*buf++ = byte;

		if(byte == EOI)
		  endseen++;
		else
		  endseen = 0;
	}
	return i;
}

#ifdef CPIA_PARPORT_DMA
/* forward declarations for IEEE1284 ECP DMA read support  */
static int change_mode(struct parport *, int );
static void frob_econtrol (struct parport *, unsigned char, unsigned char);

/* DMA support here thanks to: Blaise Gassend <blaise@gassend.com> */

/*ECP parport lowlevel IEEE1284 DMA read support: */

#define ECR_PS2 01
#define ECR_ECP 03

/* this an adaptation of code from parport_pc_ecp_read_block_pio() and
 * parport_pc_fifo_write_block_dma() */

static size_t cpia_ecp_read_block_dma (struct parport *port,
				       void *bufv, size_t length, int flags)
{
	int r = 0, endseen = 0;
	unsigned long dmaflag;
	size_t left = length;
	const struct parport_pc_private *priv = port->physport->private_data;
	dma_addr_t dma_addr, dma_handle = 0;
	size_t maxlen = 0x10000; /* max 64k per DMA transfer */
	char *buf = bufv;
	unsigned long start = (unsigned long) buf;
	unsigned long end = (unsigned long) buf + length - 1;
	size_t count = length;
	
	if (length == 0)
		return 0;
	
	/* Special case: a timeout of zero means we cannot call schedule(). */
	if (!port->cad->timeout)
		return parport_ieee1284_ecp_read_data (port, buf,
						       length, flags);
	
	port = port->physport;
	
	if (port->ieee1284.phase != IEEE1284_PH_REV_IDLE) {
		/* change to reverse-idle phase (must be in forward-idle) */
		
		/* Event 38: Set nAutoFd low (also make sure nStrobe is high) */ 
		parport_frob_control (port,
				      PARPORT_CONTROL_STROBE |PARPORT_CONTROL_AUTOFD,
				      PARPORT_CONTROL_AUTOFD);
		parport_pc_data_reverse (port); /* Must be in PS2 mode */
		udelay (5);
		
		/* Event 39: Set nInit low to initiate bus reversal */
		parport_frob_control (port,
				      PARPORT_CONTROL_INIT,
				      0);
		
		/* Event 40: Wait for  nAckReverse (PError) to go low */
		if ((r = parport_wait_peripheral (port, 
						  PARPORT_STATUS_PAPEROUT, 
						  0))){
			printk (KERN_DEBUG "%s: PE timeout Event 40 (%d) "
				"in ecp_read_block_dma\n", port->name, r);
			return 0;
		}

		
		port->ieee1284.phase = IEEE1284_PH_REV_DATA;
		
		parport_frob_control (port,
				      PARPORT_CONTROL_STROBE |PARPORT_CONTROL_AUTOFD,
				      PARPORT_CONTROL_AUTOFD);

		if (change_mode (port, ECR_ECP))      /* ECP FIFO */
			printk (KERN_DEBUG "%s: Warning change_mode ECR_ECP failed\n", 
				port->name);
		
		/* the first byte must be collected manually */

		/* Event 43: Wait for nAck to go low */
		if( parport_wait_peripheral (port, 
					     PARPORT_STATUS_ACK, 
					     0)) {
			/* timed out while reading -- no data */
			printk (KERN_DEBUG "DMA read timed out (initial byte)\n");
			goto out_no_data;
		}
#if 0   /* this is active in parport_pc_ecp_read_block_pio */
		/* read byte */
		*buf++ = inb (DATA (port));
	       	left--;
#endif
		
		/* Event 44: nAutoFd (HostAck) goes high to acknowledge */
		parport_pc_frob_control (port,
					 PARPORT_CONTROL_AUTOFD,
					 0);
		
#if 1  /* this is commented out  in parport_pc_ecp_read_block_pio */
		/* Event 45: Wait for nAck to go high */
		if (parport_wait_peripheral (port, 
					     PARPORT_STATUS_ACK, 
					     PARPORT_STATUS_ACK)) {
			/* timed out while waiting for peripheral to respond to ack */
			printk (KERN_DEBUG "ECP DMA read timed out (waiting for nAck)\n");
			
			/* keep hold of the byte we've got already */
			goto out_no_data;
		}
#endif
	}
	
#if 0   /* active in parport_pc_ecp_read_block_pio */
	/* Event 46: nAutoFd (HostAck) goes low to accept more data */
	parport_pc_frob_control (port,
				 PARPORT_CONTROL_AUTOFD,
				 PARPORT_CONTROL_AUTOFD);

#endif
	
        if (!left) 
		goto out_no_data;
	
	/* now do the DMA transfer (adapted from parport_pc_fifo_write_block_dma) */
	
	if (end < MAX_DMA_ADDRESS) {
		/* If it would cross a 64k boundary, cap it at the end. */
		if ((start ^ end) & ~0xffffUL)
			maxlen = 0x10000 - (start & 0xffff);
		
		dma_addr = dma_handle = pci_map_single(priv->dev, (void *)buf, length,
						       PCI_DMA_FROMDEVICE);
        } else {
		/* above 16 MB we use a bounce buffer as ISA-DMA is not possible */
		maxlen   = PAGE_SIZE;          /* sizeof(priv->dma_buf) */
		dma_addr = priv->dma_handle;
	}
	
	DBG ("DMA buffer is ready\n");
	
	/* We don't want to be interrupted every character. */
	parport_pc_disable_irq (port);
	/* set nErrIntrEn and serviceIntr */
	frob_econtrol (port, (1<<4) | (1<<2), (1<<4) | (1<<2));
	
	/* Reverse mode. */
	parport_pc_data_reverse (port); /* Must be in PS2 mode */

	while (left) {
		int i;
		char *ptr;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		long expire = jiffies + port->physport->cad->timeout;
#else
		unsigned long expire = jiffies + port->physport->cad->timeout;
#endif	
		DBG ("start of while %i left\n", left);
		
	        count = left;
		
		if (count > maxlen)
			count = maxlen;
		
		dmaflag = claim_dma_lock();
		disable_dma(port->dma);
		clear_dma_ff(port->dma);
		set_dma_mode(port->dma, DMA_MODE_READ);
		set_dma_addr(port->dma, dma_addr);
		set_dma_count(port->dma, count);
		
		/* Set DMA mode */
		frob_econtrol (port, 1<<3, 1<<3);
		
		/* Clear serviceIntr */
		frob_econtrol (port, 1<<2, 0);
		
		enable_dma(port->dma);
		release_dma_lock(dmaflag);
		
		/* Wait for interrupt. */
		DBG ("about to wait\n");
	false_alarm:
		if (0 > parport_wait_event (port, HZ))
			break;
		if (!time_before (jiffies, expire)) {
			/* Timed out. */
			DBG("DMA read timed out\n");
			break;
		}
		/* Is serviceIntr set? */
		if (!(inb (ECONTROL (port)) & (1<<2))) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
			if (current->need_resched)
				schedule ();
#else
			cond_resched();
#endif			
			goto false_alarm;
		}
		
		dmaflag = claim_dma_lock();
		disable_dma(port->dma);
		clear_dma_ff(port->dma);
		count -= get_dma_residue(port->dma);
		release_dma_lock(dmaflag);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))		
		if (current->need_resched)
			/* Can't yield the port. */
			schedule ();
#else
		cond_resched(); /* Can't yield the port. */
#endif

		
		/* update for possible DMA residue ! */
		DBG ("read %i in loop\n", count);
		
		if (dma_handle) 
			dma_addr += count;
		else 
			memcpy(buf, priv->dma_buf, count);
		buf  += count;
		left -= count;

		/* (cpia specific): test the last 4 bytes of data for EOI */
		endseen = 1;
		ptr = buf;
		for (i = 0; i < 4; i++) {
			ptr--;
			if(((unsigned char) *ptr ) != EOI) {
				endseen = 0;
				break;
			}
		}
		if(endseen) 
			break;
	}
	
	/* Maybe got here through break, so adjust for DMA residue! */
        if (left && !endseen) {
		DBG("exited on break\n");
		dmaflag = claim_dma_lock();
		disable_dma(port->dma);
		clear_dma_ff(port->dma);
		count -= get_dma_residue(port->dma);
		release_dma_lock(dmaflag);
		
		if (!dma_handle)
			memcpy(buf, priv->dma_buf, count);
		left -= count;
	}
	
	frob_econtrol (port, 1<<2, 1<<2);
	/* Turn off DMA mode */
	frob_econtrol (port, 1<<3, 0);
	
	if (dma_handle)
		pci_unmap_single(priv->dev, dma_handle, length, PCI_DMA_TODEVICE);
	
 out_no_data:	
	if (change_mode (port, ECR_PS2)) /* ECP FIFO */ 
		printk (KERN_DEBUG "%s: Warning change_mode ECR_PS2 failed\n", 
			port->name);
	
	/* Go to forward idle mode to shut the peripheral up (event 47). */
	parport_frob_control (port, PARPORT_CONTROL_INIT | PARPORT_CONTROL_AUTOFD, 
			      PARPORT_CONTROL_INIT | PARPORT_CONTROL_AUTOFD);
	
	/* event 49: PError goes high */
	if( (r = parport_wait_peripheral (port,
					  PARPORT_STATUS_PAPEROUT,
					  PARPORT_STATUS_PAPEROUT)) ) {
		printk (KERN_DEBUG
			"%s: PE timeout FWDIDLE (%d) in ecp_read_block_dma\n",
			port->name, r);
	} else {
		parport_data_forward (port);        
		port->ieee1284.phase = IEEE1284_PH_FWD_IDLE;
	}
	DBG ("returning %i bytes read, endseen = %d\n", length - left, endseen);
	return length - left;
}
#endif /*CPIA_PARPORT_DMA */

static ssize_t cpia_parport_read (struct parport *port, void *buffer, size_t len, int stream)
{
        int mode = port->physport->ieee1284.mode;
        size_t (*fn) (struct parport *, void *, size_t, int);

        /* Ignore the device-ID-request bit and the address bit. */
        mode &= ~(IEEE1284_DEVICEID | IEEE1284_ADDR);

        switch (mode) {
        case IEEE1284_MODE_NIBBLE:

                if(stream){
			DBG("%s: Using cpia nibble stream mode\n", port->name);
			fn = cpia_read_nibble_stream;
		} else {
			DBG("%s: Using cpia nibble mode\n", port->name);
			fn = cpia_read_nibble;
		}
		return (*fn) (port, buffer, len, 0);
		break;

        case IEEE1284_MODE_ECP:
#ifdef CPIA_PARPORT_DMA
		if (stream && port->dma != PARPORT_DMA_NONE){
			fn = cpia_ecp_read_block_dma;
			return (*fn) (port, buffer, len, 0);
			break;
		}
#endif
	default:
		return parport_read(port, buffer, len);
		break;
	}
}

/****************************************************************************
 *
 *  EndTransferMode
 *
 ***************************************************************************/
static void EndTransferMode(struct pp_cam_entry *cam)
{
	parport_negotiate(cam->port, IEEE1284_MODE_COMPAT);
}

/****************************************************************************
 *
 *  ForwardSetup
 *
 ***************************************************************************/
static int ForwardSetup(struct pp_cam_entry *cam)
{
	int retry;
	
	/* The CPiA uses ECP protocol for Downloads from the Host to the camera. 
	 * This will be software-emulated if ECP hardware is not present
	 */

	/* the usual camera maximum response time is 10ms, but after receiving
	 * some commands, it needs up to 40ms. (Data Sheet p. 32)*/

	for(retry = 0; retry < 4; ++retry) {
		if(!parport_negotiate(cam->port, IEEE1284_MODE_ECP)) {
			break;
		}
		mdelay(10);
	}
	if(retry == 4) {
		DBG("Unable to negotiate IEEE1284 ECP Download mode\n");
		return -1;
	}

	return 0;
}

/****************************************************************************
 *
 *  ReverseSetup
 *
 ***************************************************************************/
static int ReverseSetup(struct pp_cam_entry *cam, int extensibility)
{
	int retry, mode;
	int transfer_mode = NIBBLE_TRANSFER;

	/* The CPiA uses either ECP protocol or non-standard variants of
	 * Nibble protocol for Uploads from the camera to the Host. 
	 * ECP protocol will be software-emulated if ECP hardware is not present.
	 * Private CPiA-specific susbstitutes for port->ops->nibble_read_data()
	 * are provided here.
	 */

	switch (cam->upload_mode) {
	case IEEE1284_MODE_ECPSWE:
	case IEEE1284_MODE_ECP:
		transfer_mode = ECP_TRANSFER;
	case IEEE1284_MODE_NIBBLE:   /* for non-ECP non-TRISTATE ports */

		if(!extensibility)
			mode = cam->upload_mode; 
		else 
			mode = UPLOAD_FLAG|transfer_mode|IEEE1284_EXT_LINK;

	/* the usual camera maximum response time is 10ms, but after 
	 * receiving some commands, it needs up to 40ms. */
		
	for(retry = 0; retry < 4; ++retry) {
		if(!parport_negotiate(cam->port, mode)) {
			break;
		}
		mdelay(10);
	}
	if(retry == 4) {
		if(extensibility)
				DBG("Unable to negotiate IEEE1284 extensibility mode\n");
		else
				DBG("Unable to negotiate IEEE1284 Upload mode\n");
		return -1;
	}

		if(extensibility) 
			cam->port->ieee1284.mode = cam->upload_mode;
	return 0;
		break;
	default:
		return -1;
		break;
	}
}

/****************************************************************************
 *
 *  WritePacket
 *
 ***************************************************************************/
static int WritePacket(struct pp_cam_entry *cam, const u8 *packet, size_t size)
{
	int retval=0;
	int size_written;

	if (packet == NULL) {
		return -EINVAL;
	}
	if (ForwardSetup(cam)) {
		DBG("Write failed in setup\n");
		return -EIO;
	}
	size_written = parport_write(cam->port, packet, size);
	if(size_written != size) {
		DBG("Write failed, wrote %d/%d\n", size_written, size);
		retval = -EIO;
	}
	EndTransferMode(cam);
	return retval;
}

/****************************************************************************
 *
 *  ReadPacket
 *
 ***************************************************************************/
static int ReadPacket(struct pp_cam_entry *cam, u8 *packet, size_t size)
{
	int retval = 0, size_read;

	if (packet == NULL) {
		return -EINVAL;
	}
	if (ReverseSetup(cam, 0)) {
		DBG("Read failed in setup\n");
		return -EIO;
	}
	if((size_read = cpia_parport_read(cam->port, packet, size, 0)) != size) {
		DBG("Read failed, read %d/%d\n", size_read, size);
			retval = -EIO;
	}
	EndTransferMode(cam);
	return retval;
}

/****************************************************************************
 *
 *  cpia_pp_streamStart
 *
 ***************************************************************************/
static int cpia_pp_streamStart(void *privdata)
{
	struct pp_cam_entry *cam = privdata;
	DBG("\n");
	cam->streaming=1;
	cam->image_ready=0;
	//if (ReverseSetup(cam,1)) return -EIO;
	if(cam->stream_irq)
		cpia_parport_enable_irq(cam->port);
	return 0;
}

/****************************************************************************
 *
 *  cpia_pp_streamStop
 *
 ***************************************************************************/
static int cpia_pp_streamStop(void *privdata)
{
	struct pp_cam_entry *cam = privdata;

	DBG("\n");
	cam->streaming=0;
	cpia_parport_disable_irq(cam->port);
	//EndTransferMode(cam);
	return 0;
}

/****************************************************************************
 *
 *  cpia_pp_streamRead
 *
 ***************************************************************************/
static int cpia_pp_streamRead(void *privdata, u8 *buffer, int noblock)
{
	struct pp_cam_entry *cam = privdata;
	int read_bytes = 0,  endseen = 0, chunks = 0;
	size_t chunksize = CPIA_UPLOAD_CHUNKSIZE;
	u8 *buf = buffer, *buf_end = buffer + CPIA_MAX_IMAGE_SIZE;

	if(cam == NULL) {
		DBG("Internal driver error: cam is NULL\n");
		return -EINVAL;
	}
	if(buffer == NULL) {
		DBG("Internal driver error: buffer is NULL\n");
		return -EINVAL;
	}
	//if(cam->streaming) DBG("%d / %d\n", cam->image_ready, noblock);
	if( cam->stream_irq ) {
		DBG("%d\n", cam->image_ready);
		cam->image_ready--;
	}
	cam->image_complete=0;
	if (0/*cam->streaming*/) {
		if(!cam->image_ready) {
			if(noblock) return -EWOULDBLOCK;
			interruptible_sleep_on(&cam->wq_stream);
			if( signal_pending(current) ) return -EINTR;
			DBG("%d\n", cam->image_ready);
		}
	} else {
		if (ReverseSetup(cam, 1)) {
			DBG("unable to ReverseSetup\n");
			return -EIO;
		}
	}

	if (cam->upload_mode == IEEE1284_MODE_ECP) {
		chunksize = CPIA_FIFO_UPLOAD_CHUNKSIZE;
#ifdef CPIA_PARPORT_DMA
		if(cam->port->dma != PARPORT_DMA_NONE) 
			chunksize = CPIA_MAX_IMAGE_SIZE;
#endif
	}	     

	/*  If we are using "Nibble Stream" or Software-Emulated ECP 
	 *  protocols to upload data byte-by-byte through the parport.
	 *  we must read the image in small chunks and call schedule
	 *  to avoid eating up all the cpu resources.
	 */

	while (read_bytes < CPIA_MAX_IMAGE_SIZE && endseen < 4) {
		size_t datasize, bufsize;
		int i;
		u8 *pos;

		/* be nice to other processes ! */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))		
		if (current->need_resched)
			schedule();
#else
		cond_resched();
#endif
		
		bufsize = chunksize;
		if(bufsize > (size_t)(buf_end - buf)) 
			bufsize = (size_t)(buf_end - buf);

		datasize = cpia_parport_read(cam->port, buf, bufsize,1); 
		chunks++;

		buf += datasize;
		read_bytes += (int) datasize;

		/* test the last 4 bytes of data for EOI */
		if (read_bytes >= 4) {
			endseen = 1;
			pos = buf;
			for (i = 0; i < 4; i++) {
				pos--;
				if( *pos != EOI) {
					endseen = 0;
			break;
		}
	                }
		}
		if (endseen) {
			cam->image_complete = 1 ;
			break;
		}
	}
	EndTransferMode(cam);

	DBG("read %d bytes in %d chunks of size <= %d, endseen = %d\n", 
	    read_bytes, chunks, chunksize, endseen);
	
	if( endseen ) 
		DBG("EOI at  %d bytes \n", read_bytes);
	else
		DBG("EOI not seen, read  %d bytes\n", read_bytes);
	return cam->image_complete ? read_bytes : -EIO;
	
}

/****************************************************************************
 *
 *  cpia_pp_transferCmd
 *
 ***************************************************************************/
static int cpia_pp_transferCmd(void *privdata, u8 *command, u8 *data)
{
	int err;
	int retval=0;
	int databytes;
	struct pp_cam_entry *cam = privdata;

	if(cam == NULL) {
		DBG("Internal driver error: cam is NULL\n");
		return -EINVAL;
	}
	if(command == NULL) {
		DBG("Internal driver error: command is NULL\n");
		return -EINVAL;
	}
	databytes = (((int)command[7])<<8) | command[6];
	if ((err = WritePacket(cam, command, PACKET_LENGTH)) < 0) {
		DBG("Error writing command\n");
		return err;
	}
	if(command[0] == DATA_IN) {
		u8 buffer[8];
		if(data == NULL) {
			DBG("Internal driver error: data is NULL\n");
			return -EINVAL;
		}
		if((err = ReadPacket(cam, buffer, 8)) < 0) {
			DBG("Error reading command result\n");
                       return err;
		}
		memcpy(data, buffer, databytes);
	} else if(command[0] == DATA_OUT) {
		if(databytes > 0) {
			if(data == NULL) {
				DBG("Internal driver error: data is NULL\n");
				retval = -EINVAL;
			} else {
				if((err=WritePacket(cam, data, databytes)) < 0){
					DBG("Error writing command data\n");
					return err;
				}
			}
		}
	} else {
		DBG("Unexpected first byte of command: %x\n", command[0]);
		retval = -EINVAL;
	}
	return retval;
}

/****************************************************************************
 *
 *  cpia_pp_upload_mode
 *
 ***************************************************************************/
static int cpia_pp_upload_mode(void *privdata )
{
	struct pp_cam_entry *cam = (struct pp_cam_entry *)privdata;

	switch (cam->upload_mode) {
	case IEEE1284_MODE_ECP:
		return 3;
		break;
	case IEEE1284_MODE_ECPSWE:
		return 2;
		break;
	case IEEE1284_MODE_NIBBLE:
		return 1;
		break;
	}
	return 0;
}

/****************************************************************************
 *
 *  cpia_pp_open
 *
 ***************************************************************************/

static int cpia_pp_open(void *privdata , int alt )
{
	struct pp_cam_entry *cam = (struct pp_cam_entry *)privdata;
	
	/* alt has no meaning for parport cameras */

	if (cam == NULL)
		return -EINVAL;
	
	if(cam->open_count == 0) {
		if (parport_claim(cam->pdev)) {
			LOG("cpia_pp: failed to claim parport%d\n",
			    cam->port->portnum);
			return -EBUSY;
		}
		DBG("cpia_pp: claim parport%d\n",cam->port->portnum);
		parport_negotiate(cam->port, IEEE1284_MODE_COMPAT);
		parport_data_forward(cam->port);
		parport_write_control(cam->port, PARPORT_CONTROL_SELECT);
		udelay(50);
		parport_write_control(cam->port,
		                      PARPORT_CONTROL_SELECT
		                      | PARPORT_CONTROL_INIT);
	} else
		return -EBUSY;
	
	++cam->open_count;
	
	return 0;
}

/****************************************************************************
 *
 *  cpia_pp_registerCallback
 *
 ***************************************************************************/
static int cpia_pp_registerCallback(void *privdata, void (*cb)(void *cbdata), void *cbdata)
{
	struct pp_cam_entry *cam = privdata;
	int retval = 0;
	
	if(cam->port->irq != PARPORT_IRQ_NONE) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))		
		cam->cb_task.routine = cb;
		cam->cb_task.data = cbdata;
#else
		INIT_WORK(&cam->cb_task, cb, cbdata);
#endif
	} else {
		retval = -1;
	}
	return retval;
}

/****************************************************************************
 *
 *  cpia_pp_close
 *
 ***************************************************************************/
static int cpia_pp_close(void *privdata)
{

	struct pp_cam_entry *cam = privdata;
	if (--cam->open_count == 0) {
		/* delay to allow parport status lines to reset: 
		 * seems to fix problems when lp is also registered on port(?)*/
		if(cam->port->ops->read_status(cam->port) != 80)	
			mdelay(40);
		DBG("cpia_pp: releasing parport%d\n",cam->port->portnum,status);
		parport_release(cam->pdev);
	}
	return 0;
}

/****************************************************************************
 *
 *  cpia_pp_register
 *
 ***************************************************************************/
static int cpia_pp_register(struct parport *port)
{
	struct pardevice *pdev = NULL;
	struct pp_cam_entry *cam;
	struct cam_data *cpia;


	if (!(port->modes & PARPORT_MODE_PCSPP)) {
		printk ("Unsupported (non-IEEE1284) Parallel Port.\n");
		return -ENODEV;
	}

	cam = kmalloc(sizeof(struct pp_cam_entry), GFP_KERNEL);
	if (cam == NULL) {
		LOG("failed to allocate camera structure\n");
		return -ENOMEM;
	}
	memset(cam,0,sizeof(struct pp_cam_entry));
	
	pdev = parport_register_device(port, "cpia_pp", NULL, NULL,
	                               NULL, 0, cam);

	if (!pdev) {
		LOG("failed to parport_register_device\n");
		kfree(cam);
		return -ENXIO;
	}

	cam->pdev = pdev;
	cam->port = port;
	init_waitqueue_head(&cam->wq_stream);

	cam->streaming = 0;
	cam->stream_irq = 0;

	cam->upload_mode = IEEE1284_MODE_NIBBLE;
	if (port->modes & PARPORT_MODE_TRISTATE)
		cam->upload_mode = IEEE1284_MODE_ECPSWE;
	if (port->modes & PARPORT_MODE_ECP) 
		cam->upload_mode = IEEE1284_MODE_ECP;

	if((cpia = cpia_register_camera(&cpia_pp_ops, cam)) == NULL) {
		LOG("failed to cpia_register_camera\n");
		parport_unregister_device(pdev);
		kfree(cam);
		return -ENXIO;
	}

	/* (for testing:) 
	 * if cam->upload_mode = IEEE1284_MODE_ECPSWE or IEEE1284_NIBBLE
	 * is set here, even if "better" modes are detected, this mode will
	 * be used for uploads from the camera.  
	 */
	//cam->upload_mode = IEEE1284_MODE_NIBBLE;
	//cam->upload_mode = IEEE1284_MODE_ECPSWE;

	switch(cam->upload_mode) {
	case IEEE1284_MODE_ECP:
		if(port->modes & PARPORT_MODE_DMA) {
#ifdef CPIA_PARPORT_DMA
			printk("ECP Parallel Port with DMA detected: "
#else
			printk("ECP Parallel Port (DMA not enabled in cpia_pp): "
#endif
		       "image stream should be fast\n");
		} else {
			printk("ECP Parallel Port without DMA detected: "
		       "image stream should be fast\n");
		}
		break;
	case IEEE1284_MODE_ECPSWE: 
		printk("Bidirectional Parallel Port will emulate ECP mode: "
		       "image stream may be slow\n");
		break;
	case IEEE1284_MODE_NIBBLE: 
	default:
       		printk("Non-bidirectional Parallel Port will use"
		       " \"nibble\" mode: image stream will be slow\n");
	}

	spin_lock( &cam_list_lock_pp );
	list_add( &cpia->cam_data_list, &cam_list );
	spin_unlock( &cam_list_lock_pp );

	return 0;
}

static void cpia_pp_detach (struct parport *port)
{
	struct list_head *tmp;
	struct cam_data *cpia = NULL;
	struct pp_cam_entry *cam;

	spin_lock( &cam_list_lock_pp );
	list_for_each (tmp, &cam_list) {
		cpia = list_entry(tmp, struct cam_data, cam_data_list);
		cam = (struct pp_cam_entry *) cpia->lowlevel_data;
		if (cam && cam->port->number == port->number) {
			list_del(&cpia->cam_data_list);
			break;
		}
		cpia = NULL;
	}
	spin_unlock( &cam_list_lock_pp );			

	if (!cpia) {
		DBG("cpia_pp_detach failed to find cam_data in cam_list\n");
		return;
	}
	
	cam = (struct pp_cam_entry *) cpia->lowlevel_data;	
	cpia_unregister_camera(cpia);
	if(cam->open_count > 0) 
		cpia_pp_close(cam);
	parport_unregister_device(cam->pdev);
	cpia->lowlevel_data = NULL;	
	kfree(cam);
}

static void cpia_pp_attach (struct parport *port)
{
	unsigned int i;

	switch (parport_nr[0])
	{
	case PPCPIA_PARPORT_UNSPEC:
	case PPCPIA_PARPORT_AUTO:
		if (port->probe_info[0].class != PARPORT_CLASS_MEDIA ||
		    port->probe_info[0].cmdset == NULL ||
		    strncmp(port->probe_info[0].cmdset, "CPIA_1", 6) != 0)
			return;

		cpia_pp_register(port);

		break;

	default:
		for (i = 0; i < PARPORT_MAX; ++i) {
			if (port->number == parport_nr[i]) {
				cpia_pp_register(port);
				break;
			}
		}
		break;
	}
}

static struct parport_driver cpia_pp_driver = {
	.name = "cpia_pp",
	.attach = cpia_pp_attach,
	.detach = cpia_pp_detach,
};

int cpia_pp_init(void)
{
	printk(KERN_INFO "%s v%d.%d.%d\n",ABOUT, 
	       CPIA_PP_MAJ_VER,CPIA_PP_MIN_VER,CPIA_PP_PATCH_VER);

	if(parport_nr[0] == PPCPIA_PARPORT_OFF) {
		printk("  disabled\n");
		return 0;
	}
	
	spin_lock_init( &cam_list_lock_pp );

	if (parport_register_driver (&cpia_pp_driver)) {
		LOG ("unable to register with parport\n");
		return -EIO;
	}
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	if (parport[0]) {
		/* The user gave some parameters.  Let's see what they were. */
		if (!strncmp(parport[0], "auto", 4)) {
			parport_nr[0] = PPCPIA_PARPORT_AUTO;
		} else {
			int n;
			for (n = 0; n < PARPORT_MAX && parport[n]; n++) {
				if (!strncmp(parport[n], "none", 4)) {
					parport_nr[n] = PPCPIA_PARPORT_NONE;
				} else {
					char *ep;
					unsigned long r = simple_strtoul(parport[n], &ep, 0);
					if (ep != parport[n]) {
						parport_nr[n] = r;
					} else {
						LOG("bad port specifier `%s'\n", parport[n]);
						return -ENODEV;
					}
				}
			}
		}
	}
	return cpia_pp_init();
}

void cleanup_module(void)
{
	parport_unregister_driver (&cpia_pp_driver);
	return;
}

#else /* !MODULE */

static int __init cpia_pp_setup(char *str)
{
	if (!strncmp(str, "parport", 7)) {
		int n = simple_strtoul(str + 7, NULL, 10);
		if (parport_ptr < PARPORT_MAX) {
			parport_nr[parport_ptr++] = n;
		} else {
			LOG("too many ports, %s ignored.\n", str);
		}
	} else if (!strcmp(str, "auto")) {
		parport_nr[0] = PPCPIA_PARPORT_AUTO;
	} else if (!strcmp(str, "none")) {
		parport_nr[parport_ptr++] = PPCPIA_PARPORT_NONE;
	}

	return 0;
}

__setup("cpia_pp=", cpia_pp_setup);

#endif /* !MODULE */


#ifdef CPIA_PARPORT_DMA
/*-----------------------------------------------------------------------*/
/* stuff below is static stuff 
 * copied from parport_pc.c (should be made public) */

#define DPRINTK(fmt,args...) do {} while(0)
/* ECR modes */
#define ECR_SPP 00
#define ECR_PS2 01
#define ECR_PPF 02
#define ECR_ECP 03
#define ECR_EPP 04
#define ECR_VND 05
#define ECR_TST 06
#define ECR_CNF 07
#define ECR_MODE_MASK 0xe0
#define ECR_WRITE(p,v) frob_econtrol((p),0xff,(v))
#define ECONTROL(p) ((p)->base_hi + 0x2)

/* frob_control, but for ECR */
static void frob_econtrol (struct parport *pb, unsigned char m,
			   unsigned char v)
{
	unsigned char ectr = 0;

	if (m != 0xff)
		ectr = inb (ECONTROL (pb));

	DPRINTK (KERN_DEBUG "frob_econtrol(%02x,%02x): %02x -> %02x\n",
		m, v, ectr, (ectr & ~m) ^ v);

	outb ((ectr & ~m) ^ v, ECONTROL (pb));
}


/* Safely change the mode bits in the ECR 
   Returns:
	    0    : Success
	   -EBUSY: Could not drain FIFO in some finite amount of time,
		   mode not changed!
 */
static int change_mode(struct parport *p, int m)
{
	const struct parport_pc_private *priv = p->physport->private_data;
	unsigned char oecr;
	int mode;

	DPRINTK(KERN_INFO "parport change_mode ECP-ISA to mode 0x%02x\n",m);

	if (!priv->ecr) {
		printk (KERN_DEBUG "change_mode: but there's no ECR!\n");
		return 0;
	}

	/* Bits <7:5> contain the mode. */
	oecr = inb (ECONTROL (p));
	mode = (oecr >> 5) & 0x7;
	if (mode == m) return 0;

	if (mode >= 2 && !(priv->ctr & 0x20)) {
		/* This mode resets the FIFO, so we may
		 * have to wait for it to drain first. */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		long expire = jiffies + p->physport->cad->timeout;
#else
		unsigned long expire = jiffies + p->physport->cad->timeout;
#endif
		int counter;
		switch (mode) {
		case ECR_PPF: /* Parallel Port FIFO mode */
		case ECR_ECP: /* ECP Parallel Port mode */
			/* Busy wait for 200us */
			for (counter = 0; counter < 40; counter++) {
				if (inb (ECONTROL (p)) & 0x01)
					break;
				if (signal_pending (current)) break;
				udelay (5);
			}

			/* Poll slowly. */
			while (!(inb (ECONTROL (p)) & 0x01)) {
				if (time_after_eq (jiffies, expire))
					/* The FIFO is stuck. */
					return -EBUSY;
				__set_current_state (TASK_INTERRUPTIBLE);
				schedule_timeout ((HZ + 99) / 100);
				if (signal_pending (current))
					break;
			}
		}
	}

	if (mode >= 2 && m >= 2) {
		/* We have to go through mode 001 */
		oecr &= ~(7 << 5);
		oecr |= ECR_PS2 << 5;
		ECR_WRITE (p, oecr);
	}

	/* Set the mode. */
	oecr &= ~(7 << 5);
	oecr |= m << 5;
	ECR_WRITE (p, oecr);
	return 0;
}
#endif /* CPIA_PARPORT_DMA */
