/*******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include <linux/802_11.h>
#include "ipw2100.h"
#include "ipw2100_frag.h"

/* Get the msdu: real payload size. */ 
inline u32 ipw2100_msdu_frag_threshold(struct ipw2100_priv *priv)
{
        return priv->frag_threshold - (ETH_ALEN + CRC_LENGTH);
}


static void defrag_timeout(unsigned long arg)
{
	struct defrag_info *info = (struct defrag_info *)arg;
	
	IPW2100_DEBUG_FRAG("Fragmentation timer freeing packet sequence.\n");
	
	if (info->skb) {
		/* Pull this from the live fragment list */
		list_del(&info->list);

		dev_kfree_skb_irq(info->skb);
		info->skb = NULL;
		info->priv->num_of_defrag_rx--;
		info->free_space = 0;

		/* Add back to the free fragment list */
		list_add_tail(&info->list, &info->priv->defrag_free_list);
		
	}
}

static struct defrag_info *defrag_alloc_slot(struct ipw2100_priv *priv)
{
	struct defrag_info *info;

	if (list_empty(&priv->defrag_free_list))
		return NULL;

	info = list_entry(priv->defrag_free_list.next, struct defrag_info, list);

	/* Pull from the free fragment list */
	list_del(priv->defrag_free_list.next);

	/* Add to the live fragment list */
	list_add_tail(&info->list, &priv->defrag_list);
	
	priv->num_of_defrag_rx++;

	return info;
}


static struct defrag_info *defrag_match_info(
	struct ipw2100_priv *priv, struct ipw2100_802_11_header *hdr, 
	u32 seq_num, u32 frag_num)
{
	struct list_head *element;
	struct defrag_info *info = NULL;

	for (element = priv->defrag_list.next;
	     element != &priv->defrag_list;
	     element = element->next) {
		info = list_entry(element, struct defrag_info, list);
		if ((info->seq_num == seq_num) && 
		    (!memcmp(info->skb->data, hdr->address1, ETH_ALEN)) &&
		    (!memcmp(info->skb->data + ETH_ALEN, 
			    hdr->address3, ETH_ALEN)) && 
		    (info->num_of_frags == frag_num))
			return info;
	}

	return NULL;
}

static void defrag_first(struct ipw2100_priv *priv, struct ipw2100_hw_packet *packet, u32 frame_size, u32 seq_num)
{
	struct ipw2100_802_11_header *hdr = &packet->rxp->header;
	struct defrag_info *info = NULL;
		
	IPW2100_DEBUG_FRAG("First fragment, seq_num: %X\n", seq_num);
	
	info = defrag_alloc_slot(priv);

	if (info==NULL) {
		IPW2100_DEBUG_FRAG("No valid slot found for fragmented "
				 "packet, dropped\n");
		priv->wstats.discard.fragment++;	
		dev_kfree_skb_irq(packet->skb);
		return;
	}
	
	info->skb = packet->skb;
	
	skb_put(info->skb, frame_size - IPW_802_11_PAYLOAD_OFFSET);
	
	memcpy(skb_push(info->skb, ETH_ALEN), &hdr->address3, ETH_ALEN);
	memcpy(skb_push(info->skb, ETH_ALEN), &hdr->address1, ETH_ALEN);
	
	info->seq_num = seq_num;
	info->num_of_frags = 1;
	info->free_space = info->skb->end - info->skb->head;
	
	init_timer(&info->defrag_timer);
	info->defrag_timer.expires = jiffies + priv->defrag_timeout;
	info->defrag_timer.data = (unsigned long)info;
	info->defrag_timer.function = defrag_timeout;
	add_timer(&info->defrag_timer);
} 


static struct sk_buff *defrag_cont(struct ipw2100_priv *priv, 
				   struct ipw2100_hw_packet *packet, 
				   u32 frame_size, u32 seq_num, u32 frag_num)
{
	struct defrag_info *info = NULL;
    	u32 length;	
	struct ipw2100_802_11_header *hdr = &packet->rxp->header;
	struct sk_buff *skb = NULL;
	
	IPW2100_DEBUG_FRAG("Non-first fragment, seq_num: %X frag_num: %X\n",
			   seq_num, frag_num);

	length = frame_size - IPW_802_11_PAYLOAD_OFFSET + 
			sizeof(struct ieee_802_11_snap_header);

	info = defrag_match_info(priv, hdr, seq_num, frag_num);
	
	if (info != NULL) {
		IPW2100_DEBUG_FRAG("Found matched defrag_info\n");

		if (info->free_space < length) {
			printk(KERN_WARNING "ipw2100"
			       ": AP sending fragment overrun.  Dropping "
			       "%d frames.\n", info->num_of_frags);
			priv->wstats.discard.fragment++;	

			/* For consistent cleanup, we just force a timeout
			 * expiration */
			del_timer_sync(&info->defrag_timer);
			defrag_timeout((unsigned long)info);

		} else {

			info->num_of_frags++;
			mod_timer(&info->defrag_timer, 
				  jiffies + priv->defrag_timeout);
			
			skb_push(packet->skb, 
				 sizeof(struct ieee_802_11_snap_header));
			memcpy(skb_put(info->skb, length), packet->skb->data, 
			       length);
		
			if ((*hdr).flags == 0) {
				
				IPW2100_DEBUG_FRAG("Last fragment received\n");
				
				del_timer_sync(&info->defrag_timer);
				
				
				/* Pull this from the live fragment list */
				list_del(&info->list);

				skb = info->skb;
				info->skb = NULL;
				priv->num_of_defrag_rx--;
				info->free_space = 0;

				/* Add back to the free fragment list */
				list_add_tail(&info->list, 
					      &info->priv->defrag_free_list);
			}
		}
	} else {
		IPW2100_DEBUG_FRAG("Cannot match this fragment, dropped!\n");
		priv->wstats.discard.fragment++;	
	}

	dev_kfree_skb_irq(packet->skb);

	return skb;
} 

struct sk_buff *ipw2100_defrag_rx(struct ipw2100_priv *priv, 
				  struct ipw2100_hw_packet *packet, 
				  u32 frame_size)
{
	struct sk_buff *skb = NULL;

	struct ipw2100_802_11_header *hdr = &packet->rxp->header;

	u32 seq_num = SEQUENCE_NUMBER(hdr);
	u32 frag_num = FRAGMENT_NUMBER(hdr);

	if (frag_num == 0)
		defrag_first(priv, packet, frame_size, seq_num);
	else 
		skb = defrag_cont(priv, packet, frame_size, seq_num, frag_num);

	return skb;
} 


int ipw2100_defrag_alloc(struct ipw2100_priv *priv)
{
	priv->defrag_pool = (struct defrag_info *)kmalloc(
		MAX_DEFRAG_RX * sizeof(struct defrag_info), 
		GFP_KERNEL);
	if (!priv->defrag_pool)
		return -ENOMEM;

	memset(priv->defrag_pool, 0, MAX_DEFRAG_RX * 
	       sizeof(struct defrag_info));

	return 0;
}

void ipw2100_defrag_free(struct ipw2100_priv *priv)
{
	struct defrag_info *info;
	struct list_head *element;

	if (!priv->defrag_pool)
		return;

	for (element = priv->defrag_list.next;
	     element != &priv->defrag_list;
	     element = element->next) {
		info = list_entry(element, struct defrag_info, list);
		if (info->skb) {
			del_timer_sync(&info->defrag_timer);
			dev_kfree_skb(info->skb);
			info->skb = NULL;
		}
		priv->num_of_defrag_rx--;
		if (priv->num_of_defrag_rx < 0) {
			IPW2100_DEBUG_FRAG("Bad Rx count!\n");
			break;
		}
	}

	kfree(priv->defrag_pool);
	priv->defrag_pool = NULL;
}

void ipw2100_defrag_initialize(struct ipw2100_priv *priv)
{
	int i;

	INIT_LIST_HEAD(&priv->defrag_free_list);
	INIT_LIST_HEAD(&priv->defrag_list);

	priv->num_of_defrag_rx = 0;
	priv->defrag_timeout = DEFAULT_DEFRAG_TIMEOUT;
		
	for(i = 0; i < MAX_DEFRAG_RX; i++) {
		priv->defrag_pool[i].priv = priv;

		/* Free any stale SKBs */
		if (priv->defrag_pool[i].skb) {
			dev_kfree_skb(priv->defrag_pool[i].skb);
			priv->defrag_pool[i].skb = NULL;
		}

		list_add_tail(&priv->defrag_pool[i].list, 
			      &priv->defrag_free_list);
	}
}

