/*******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

/*******************************************************************************

  Changes

  0.33 Luc Saillard
       Added support for wireless stats (/proc/net/wireless)

  Portions of this file are based on the sample_* files provided by Wireless
  Extensions 0.26 package and copyright (c) 1997-2003 Jean Tourrilhes 
  <jt@hpl.hp.com>
  
*******************************************************************************/

/*
 * This file defines the Wireless Extension handlers.  It does not
 * define any methods of hardware manipulation and relies on the
 * functions defined in ipw2100_main to provide the HW interaction.
 * 
 * The exception to this is the use of the ipw2100_get_ordinal() 
 * function used to poll the hardware vs. making unecessary calls.
 *
 */
#include <asm/uaccess.h>

#include "ipw2100.h"
#include "ipw2100_hw.h"
#include "ipw2100_wx.h"

#if WIRELESS_EXT > 12

const long ipw2100_frequencies[] = {  
	2412, 2417, 2422, 2427, 
	2432, 2437, 2442, 2447, 
	2452, 2457, 2462, 2467, 
	2472, 2484  
};


static int ipw2100_wx_get_name(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	strcpy(wrqu->name, "IEEE 802.11b");
	IPW2100_DEBUG_WX("Name: %s\n", wrqu->name);
	return 0;
}


static int ipw2100_wx_set_freq(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra) 
{
	struct ipw2100_priv *priv = dev->priv;
	struct iw_freq *fwrq = &wrqu->freq;
	
	/* we only set channel in adhoc mode */
	if (priv->ctx->port_type != BSS) {
		/* if setting by freq convert to channel */
		if (fwrq->e == 1) {
			if ((fwrq->m >= (int) 2.412e8 &&
			     fwrq->m <= (int) 2.487e8)) {
				int f = fwrq->m / 100000;
				int c = 0;
				
				while ((c < REG_MAX_CHANNEL) &&
				       (f != ipw2100_frequencies[c]))
					c++;
				
				/* hack to fall through */
				fwrq->e = 0;
				fwrq->m = c + 1;
			}
		}

		if (fwrq->e > 0 || fwrq->m > 1000) 
			return -EOPNOTSUPP;
		else /* set the channel */
			return ipw2100_set_channel(priv, fwrq->m, 0);
	}

	IPW2100_DEBUG_WX("SET Freq/Channel -> %d \n", priv->ctx->channel);
	return 0;
}


static int ipw2100_wx_get_freq(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	struct iw_freq *fwrq = &wrqu->freq;
	u32 chan;
	u32 len = sizeof(u32);
	int err;

	/* feed crap if we are not associated */
	if (priv->disabled || priv->ctx->ssid[0] == '\0') {
		fwrq->m = 0;
		fwrq->e = 0;
		return 0;
	}

	err = ipw2100_get_ordinal(priv, IPW_ORD_OUR_FREQ, &chan, &len);
	if (err) {
		IPW2100_DEBUG_WX("failed querying ordinals.\n");
		return err;
	}

	/* we keep it */
	priv->ctx->channel = chan;

#ifdef WEXT_USECHANNELS
	/* return CHANNEL no if requested so */
	fwrq->m = chan;
	fwrq->e = 0;
#else
	/* otherwise return freq */
	fwrq->m = ipw2100_frequencies[chan - 1] * 100000;
	fwrq->e = 1;
#endif	/* WEXT_USECHANNELS */

	IPW2100_DEBUG_WX("GET Freq/Channel -> %d \n", fwrq->m);
	return 0;
}

static int ipw2100_wx_set_mode(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;

	err = ipw2100_set_port_type(
		priv, wrqu->mode == IW_MODE_ADHOC ? IBSS : BSS, 0);
	if (err)
		return err;

	IPW2100_DEBUG_WX("SET Mode -> %s \n",
	       port_type_str[priv->ctx->port_type]);

	return 0;
}

static int ipw2100_wx_get_mode(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	switch (priv->ctx->port_type) {
	case BSS:
		wrqu->mode = IW_MODE_INFRA;
		break;
	case IBSS:
		wrqu->mode = IW_MODE_ADHOC;
		break;
	default:
		/*
		wrqu->mode = IW_MODE_AUTO;
		*/
		wrqu->mode = IW_MODE_INFRA;
	}

	IPW2100_DEBUG_WX("GET Mode -> %s \n",
	       port_type_str[priv->ctx->port_type]);

	return 0;
}

static int ipw2100_wx_get_range(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct iw_range *range = (struct iw_range *)extra;

	wrqu->data.length = sizeof(*range);
	memset(range, 0, sizeof(*range));

	/* Set the Wireless Extension versions */
	range->we_version_compiled	= WIRELESS_EXT;
	range->we_version_source	= 16;

        /* ~5 Mb/s real (802.11b) */
	range->throughput		= 5 * 1000 * 1000;     

	range->max_rts = DEFAULT_RTS_THRESHOLD;
	range->min_frag = MIN_FRAG_THRESHOLD;
	range->max_frag = MAX_FRAG_THRESHOLD;

	/* TODO: See what else we can put in here ... */

	IPW2100_DEBUG_WX("GET Range\n");

	return 0;
}

static int ipw2100_wx_set_wap(struct net_device *dev, 
			      struct iw_request_info *info, 
			      union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;

	static const unsigned char any[] = {
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	};
	static const unsigned char off[] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	// sanity checks
	if (wrqu->ap_addr.sa_family != ARPHRD_ETHER) 
		return -EINVAL;

	if (!memcmp(any, wrqu->ap_addr.sa_data, ETH_ALEN) ||
	    !memcmp(off, wrqu->ap_addr.sa_data, ETH_ALEN)) {
		/* we disable mandatory BSSID association */
		IPW2100_DEBUG_WX("exit - disable mandatory BSSID\n");
		return ipw2100_set_mandatory_bssid(priv, NULL, 0);
	}

	err = ipw2100_set_mandatory_bssid(
		priv, wrqu->ap_addr.sa_data, 0);
	if (err)
		return err;

	IPW2100_DEBUG_WX("SET BSSID -> %02X:%02X:%02X:%02X:%02X:%02X\n",
	       wrqu->ap_addr.sa_data[0] & 0xff,
	       wrqu->ap_addr.sa_data[1] & 0xff,
	       wrqu->ap_addr.sa_data[2] & 0xff,
	       wrqu->ap_addr.sa_data[3] & 0xff,
	       wrqu->ap_addr.sa_data[4] & 0xff, 
	       wrqu->ap_addr.sa_data[5] & 0xff);

	return 0;
}

static int ipw2100_wx_get_wap(struct net_device *dev, 
			      struct iw_request_info *info, 
			      union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int len = ETH_ALEN;

	wrqu->ap_addr.sa_family = ARPHRD_ETHER;

	/* Fetch from the hardware */
	if (priv->disabled ||
	    ipw2100_get_ordinal(priv, IPW_ORD_STAT_ASSN_AP_BSSID, 
				&priv->ctx->bssid,  &len) ||
	    priv->ctx->ssid[0] == '\0') {
		memset(wrqu->ap_addr.sa_data, 0, ETH_ALEN);
		return 0;
	}

	memcpy(wrqu->ap_addr.sa_data, &priv->ctx->bssid, ETH_ALEN);

	IPW2100_DEBUG_WX("GET BSSID -> %02X:%02X:%02X:%02X:%02X:%02X\n",
			 (u8)wrqu->ap_addr.sa_data[0],
			 (u8)wrqu->ap_addr.sa_data[1],
			 (u8)wrqu->ap_addr.sa_data[2],
			 (u8)wrqu->ap_addr.sa_data[3],
			 (u8)wrqu->ap_addr.sa_data[4],
			 (u8)wrqu->ap_addr.sa_data[5]);

	return 0;
}

static int ipw2100_wx_set_essid(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;
	
	if (priv->disabled) {
		IPW2100_DEBUG_WX("Caching ESSID: %s\n", extra);
		strncpy(priv->ctx->ssid, extra, IW_ESSID_MAX_SIZE);
		return 0;
	}

	err = ipw2100_set_essid(priv, extra, 0);
	if (err) {
		IPW2100_DEBUG_WX("SET SSID failed\n");
		return err;
	}

	IPW2100_DEBUG_WX("exit\n");

	return 0;
}

static int ipw2100_wx_get_essid(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	wrqu->data.length = strlen(priv->ctx->ssid) + 1;
	memcpy(extra, priv->ctx->ssid, wrqu->data.length);
	wrqu->data.flags = 1; /* active */

	IPW2100_DEBUG_WX("GET ESSID -> %s\n", extra);
	return 0;
}

static int ipw2100_wx_set_nick(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	if (wrqu->data.length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	wrqu->data.length = min((size_t)wrqu->data.length, sizeof(priv->nick));
	memset(priv->nick, 0, sizeof(priv->nick));
	memcpy(priv->nick, extra,  wrqu->data.length);

	IPW2100_DEBUG_WX("SET Nickname -> %s \n", priv->nick);

	return 0;
}

static int ipw2100_wx_get_nick(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	
	wrqu->data.length = strlen(priv->nick) + 1;
	memcpy(extra, priv->nick, wrqu->data.length);
	wrqu->data.flags = 1; /* active */
	
	IPW2100_DEBUG_WX("GET Nickname -> %s \n", extra);

	return 0;
}

static int ipw2100_wx_get_rate(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int val;
	int len = sizeof(val);
	int err;

	if (priv->disabled || priv->ctx->ssid[0] == '\0') {
		wrqu->bitrate.value = 0;
		return 0;
	}

	err = ipw2100_get_ordinal(priv, IPW_ORD_CURRENT_TX_RATE, &val, &len);
	if (err) {
		IPW2100_DEBUG_WX("failed querying ordinals.\n");
		return err;
	}

	switch (val & TX_RATE_MASK) {
	case TX_RATE_1_MBIT:
		wrqu->bitrate.value = 1000000;
		break;
	case TX_RATE_2_MBIT:
		wrqu->bitrate.value = 2000000;
		break;
	case TX_RATE_5_5_MBIT:
		wrqu->bitrate.value = 5500000;
		break;
	case TX_RATE_11_MBIT:
		wrqu->bitrate.value = 11000000;
		break;
	default:
		wrqu->bitrate.value = 0;
	}

	IPW2100_DEBUG_WX("GET Rate -> %d \n", wrqu->bitrate.value);
	return 0;
}

static int ipw2100_wx_set_rts(struct net_device *dev, 
			      struct iw_request_info *info, 
			      union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;

	if (wrqu->rts.value < 1 || 
	    wrqu->rts.value > 2304 || 
	    wrqu->rts.disabled)
		return -EINVAL;

	err = ipw2100_set_rts_threshold(priv, wrqu->rts.value);
	if (err)
		return err;

	IPW2100_DEBUG_WX("SET RTS Threshold -> %d \n", wrqu->rts.value);

	return 0;
}

static int ipw2100_wx_get_rts(struct net_device *dev, 
			      struct iw_request_info *info, 
			      union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int len = sizeof(wrqu->rts.value);
	int err;

	if (priv->disabled)
		wrqu->rts.value = priv->rts_threshold;
	else {
		err = ipw2100_get_ordinal(priv, IPW_ORD_RTS_THRESHOLD, 
					  &wrqu->rts.value, &len);
		if (err) {
			IPW2100_DEBUG_WX("query ordinal failed.\n");
			return err;
		}
	}

	wrqu->rts.fixed = 1; /* no auto select */
	wrqu->rts.disabled = 0; /* can't be disabled */

	IPW2100_DEBUG_WX("GET RTS Threshold -> %d \n", wrqu->rts.value);

	return 0;
}

static int ipw2100_wx_set_txpow(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	IPW2100_DEBUG_WX("TODO: Power management by wireless extension...\n");

	IPW2100_DEBUG_WX("SET TX Power -> %d \n", wrqu->rts.value);

	return 0;
}

static int ipw2100_wx_get_txpow(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	wrqu->power.value = priv->tx_power;
	wrqu->power.fixed = 1;
	wrqu->power.flags = IW_TXPOW_DBM;
	wrqu->power.disabled = 0;

	IPW2100_DEBUG_WX("GET TX Power -> %d \n", wrqu->power.value);

	return 0;
}

static int ipw2100_wx_set_frag(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;

	if (wrqu->frag.value < MIN_FRAG_THRESHOLD ||
	    wrqu->frag.value > MAX_FRAG_THRESHOLD)
		return -EINVAL;

	err = ipw2100_set_fragmentation_threshold(
		priv, wrqu->frag.value & ~0x1, 0);
	if (err) 
		return err;

	IPW2100_DEBUG_WX("SET Frag Threshold -> %d \n", 
	       wrqu->frag.value & ~0x01);

	return 0;
}

static int ipw2100_wx_get_frag(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
/*	int len = sizeof(wrqu->frag.value);
	int err;

	err = ipw2100_get_ordinal(priv, IPW_ORD_FRAGMENTATION_THRESHOLD,
				  &wrqu->frag.value, &len);
*/
	wrqu->frag.value = ipw2100_msdu_frag_threshold(priv);
	wrqu->frag.fixed = 0;	/* no auto select */
	wrqu->frag.disabled = 
		(wrqu->frag.value == DEFAULT_FRAG_THRESHOLD);

	IPW2100_DEBUG_WX("GET Frag Threshold -> %d \n", wrqu->frag.value);

	return 0;
}

static int ipw2100_wx_set_retry(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	int err;

	if (wrqu->retry.flags & IW_RETRY_LIFETIME || 
	    wrqu->retry.disabled)
		return -EINVAL;

	if (!wrqu->retry.flags & IW_RETRY_LIMIT)
		return 0;

	if (wrqu->retry.flags & IW_RETRY_MIN) {
		err = ipw2100_set_short_retry(priv, wrqu->retry.value);
		if (err)
			return err;

		IPW2100_DEBUG_WX("SET Short Retry Limit -> %d \n",
		       wrqu->retry.value);

		return 0;
	} 

	if (wrqu->retry.flags & IW_RETRY_MAX) {
		err = ipw2100_set_long_retry(priv, wrqu->retry.value);
		if (err)
			return err;

		IPW2100_DEBUG_WX("SET Long Retry Limit -> %d \n",
		       wrqu->retry.value);

		return 0;
	}

	err = ipw2100_set_short_retry(priv, wrqu->retry.value);
	if (err)
		return err;
	err = ipw2100_set_long_retry(priv, wrqu->retry.value);
	if (err)
		return err;

	IPW2100_DEBUG_WX("SET Both Retry Limits -> %d \n", wrqu->retry.value);

	return 0;
}

static int ipw2100_wx_get_retry(struct net_device *dev, 
				struct iw_request_info *info, 
				union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	wrqu->retry.disabled = 0; /* can't be disabled */

	if ((wrqu->retry.flags & IW_RETRY_TYPE) == 
	    IW_RETRY_LIFETIME) 
		return -EINVAL;
	
	if (wrqu->retry.flags & IW_RETRY_MAX) {
		wrqu->retry.flags = IW_RETRY_LIMIT & IW_RETRY_MAX;
		wrqu->retry.value = priv->long_retry_limit;
	} else {
		wrqu->retry.flags =
		    (priv->short_retry_limit !=
		     priv->long_retry_limit) ?
		    IW_RETRY_LIMIT & IW_RETRY_MIN : IW_RETRY_LIMIT;

		wrqu->retry.value = priv->short_retry_limit;
	}

	IPW2100_DEBUG_WX("GET Retry -> %d \n", wrqu->retry.value);

	return 0;
}

static int ipw2100_wx_set_scan(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;

	IPW2100_DEBUG_WX("Initiating scan...\n");
	if (ipw2100_start_scan(priv)) {
		IPW2100_DEBUG_WX("Start scan failed.\n");

		/* TODO: Mark a scan as pending so when hardware initialized
		 *       a scan starts */
	}

	return 0;
}

/*
 *
 * Following logic based on the code in drivers/net/wireless/airo.c from 2.6.1
 *
 */
static inline char *ipw2100_translate_scan(char *start, char *stop, 
					   struct ipw2100_beacon *beacon)
{
	struct iw_event iwe;
	int i;

	/* First entry *MUST* be the AP MAC address */
	iwe.cmd = SIOCGIWAP;
	iwe.u.ap_addr.sa_family = ARPHRD_ETHER; 
	memcpy(iwe.u.ap_addr.sa_data, beacon->bssid, ETH_ALEN);
	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_ADDR_LEN);
	
	/* Remaining entries will be displayed in the order we provide them */

	/* Add the ESSID */
        iwe.u.data.length = beacon->ssid_len;
        if (iwe.u.data.length > 32)
		iwe.u.data.length = 32;
        iwe.cmd = SIOCGIWESSID;
        iwe.u.data.flags = 1;
        start = iwe_stream_add_point(start, stop, &iwe, beacon->ssid);

        /* Add mode */
        iwe.cmd = SIOCGIWMODE;
        if (beacon->capability & (SSID_CAPABILITY_BSS | SSID_CAPABILITY_IBSS)) {
		if (beacon->capability & SSID_CAPABILITY_BSS)
			iwe.u.mode = IW_MODE_MASTER;
		else
			iwe.u.mode = IW_MODE_ADHOC;

		start = iwe_stream_add_event(start, stop, &iwe, IW_EV_UINT_LEN);
	}
	
        /* Add frequency */
	iwe.cmd = SIOCGIWFREQ;
	iwe.u.freq.m = ipw2100_frequencies[beacon->channel - 1] * 100000;
	iwe.u.freq.e = 1;
	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_FREQ_LEN);

	/* Add rates */
	iwe.cmd = SIOCGIWRATE;
	iwe.u.bitrate.fixed = iwe.u.bitrate.disabled = 0;
	for (i = 0; i < beacon->rates_len; i++) {
		iwe.u.bitrate.value = (beacon->rates[i] & 0x7F) * 500000;
		start = iwe_stream_add_event(start, stop, &iwe, 
					     IW_EV_PARAM_LEN);
	}

	/* Add quality statistics */
	iwe.cmd = IWEVQUAL;
	iwe.u.qual.level = beacon->rssi + IPW2100_RSSI_TO_DBM;
	iwe.u.qual.noise = 0;
	iwe.u.qual.qual = 0;
	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_QUAL_LEN);

	/* Add encryption capability */
	iwe.cmd = SIOCGIWENCODE;
	if (beacon->capability & SSID_CAPABILITY_ENCRYPT)
		iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
	else
		iwe.u.data.flags = IW_ENCODE_DISABLED;
	iwe.u.data.length = 0;
	start = iwe_stream_add_point(start, stop, &iwe, beacon->ssid);

	return start;
}

static int ipw2100_wx_get_scan(struct net_device *dev, 
			       struct iw_request_info *info, 
			       union iwreq_data *wrqu, char *extra)
{
	struct ipw2100_priv *priv = dev->priv;
	struct list_head *element;
	unsigned long flags;

	char *ev = extra;
	char *stop = ev + IW_SCAN_MAX_DATA;
	int i;

	IPW2100_DEBUG_WX("enter\n");
	
	spin_lock_irqsave(&priv->low_lock, flags);

	for (element = priv->beacon_list.next, i = 0;
	     element != &priv->beacon_list;
	     element = element->next, i++) 
		ev = ipw2100_translate_scan(
			ev, stop, 
			list_entry(element, struct ipw2100_beacon, list));

	spin_unlock_irqrestore(&priv->low_lock, flags);

	wrqu->data.length = ev -  extra;
	wrqu->data.flags = 0;
	
	IPW2100_DEBUG_WX("exit: %d beacons returned.\n", i);

	return 0;
}


/*
 * Implementation based on code in wireless tools sample_enc v0.26
 */
static int ipw2100_wx_set_encode(struct net_device *dev, 
				 struct iw_request_info *info, 
				 union iwreq_data *wrqu, char *key)
{
	struct ipw2100_priv *priv = dev->priv;
	struct iw_point *erq = &(wrqu->encoding);
	int index = (erq->flags & IW_ENCODE_INDEX) - 1;

	if (erq->length > 0) {
		/* Check the size of the key */
		if (erq->length > MAX_KEY_SIZE)
			return(-EINVAL);

		/* Check the index */
		if ((index < 0) || (index >= MAX_KEYS))
			index = priv->key_current;

		/* Copy the key in the driver */
		memcpy(priv->key_table[index], key, erq->length);
		priv->key_size[index] = erq->length;
		priv->key_on = 1;
	} else {
		/* Do we want to just set the current key ? */
		if ((index >= 0) && (index < MAX_KEYS)) {
			if (priv->key_size[index] > 0) {
				priv->key_current = index;
				priv->key_on = 1;
			}
			else
				return(-EINVAL);
		}
	}

	/* Read the flags */
	if (erq->flags & IW_ENCODE_DISABLED)
		priv->key_on = 0; /* disable encryption */
	if (erq->flags & IW_ENCODE_RESTRICTED)
		priv->key_open = 0; /* disable open mode */
	if (erq->flags & IW_ENCODE_OPEN)
		priv->key_open = 1; /* enable open mode */

	return 0;
}

static int ipw2100_wx_get_encode(struct net_device *dev, 
				 struct iw_request_info *info, 
				 union iwreq_data *wrqu, char *key)
{
	struct ipw2100_priv *priv = dev->priv;
	struct iw_point *erq = &(wrqu->encoding);
	int index = (erq->flags & IW_ENCODE_INDEX) - 1;

	IPW2100_DEBUG_WX("TODO: Implement ipw2100_wx_set_encode()\n");

	/* Set the flags */
	erq->flags = 0;
	if (priv->key_on == 0)
		erq->flags |= IW_ENCODE_DISABLED;
	if (priv->key_open == 0)
		erq->flags |= IW_ENCODE_RESTRICTED;
	else
		erq->flags |= IW_ENCODE_OPEN;
	
	/* Which key do we want */
	if ((index < 0) || (index >= MAX_KEYS))
		index = priv->key_current;
	erq->flags |= index + 1;
	
	/* Copy the key to the user buffer */
	erq->length = priv->key_size[index];
	memcpy(key, priv->key_table[index], priv->key_size[index]);

	return 0;
}

static iw_handler ipw2100_wx_handlers[] =
{
        NULL,                     /* SIOCSIWCOMMIT */
        ipw2100_wx_get_name,      /* SIOCGIWNAME */
        NULL,                     /* SIOCSIWNWID */
        NULL,                     /* SIOCGIWNWID */
        ipw2100_wx_set_freq,      /* SIOCSIWFREQ */
        ipw2100_wx_get_freq,      /* SIOCGIWFREQ */
        ipw2100_wx_set_mode,      /* SIOCSIWMODE */
        ipw2100_wx_get_mode,      /* SIOCGIWMODE */
        NULL,                     /* SIOCSIWSENS */
        NULL,                     /* SIOCGIWSENS */
        NULL,                     /* SIOCSIWRANGE */
        ipw2100_wx_get_range,     /* SIOCGIWRANGE */
        NULL,                     /* SIOCSIWPRIV */
        NULL,                     /* SIOCGIWPRIV */
        NULL,                     /* SIOCSIWSTATS */
        NULL,                     /* SIOCGIWSTATS */
        NULL,                     /* SIOCSIWSPY */
        NULL,                     /* SIOCGIWSPY */
        NULL,                     /* SIOCGIWTHRSPY */
        NULL,                     /* SIOCWIWTHRSPY */
        ipw2100_wx_set_wap,       /* SIOCSIWAP */
        ipw2100_wx_get_wap,       /* SIOCGIWAP */
        NULL,                     /* -- hole -- */
        NULL,                     /* SIOCGIWAPLIST -- depricated */
        ipw2100_wx_set_scan,      /* SIOCSIWSCAN */
        ipw2100_wx_get_scan,      /* SIOCGIWSCAN */
        ipw2100_wx_set_essid,     /* SIOCSIWESSID */
        ipw2100_wx_get_essid,     /* SIOCGIWESSID */
        ipw2100_wx_set_nick,      /* SIOCSIWNICKN */
        ipw2100_wx_get_nick,      /* SIOCGIWNICKN */
        NULL,                     /* -- hole -- */
        NULL,                     /* -- hole -- */
        NULL,                     /* SIOCSIWRATE */
        ipw2100_wx_get_rate,      /* SIOCGIWRATE */
        ipw2100_wx_set_rts,       /* SIOCSIWRTS */
        ipw2100_wx_get_rts,       /* SIOCGIWRTS */
        ipw2100_wx_set_frag,      /* SIOCSIWFRAG */
        ipw2100_wx_get_frag,      /* SIOCGIWFRAG */
        ipw2100_wx_set_txpow,     /* SIOCSIWTXPOW */
        ipw2100_wx_get_txpow,     /* SIOCGIWTXPOW */
        ipw2100_wx_set_retry,     /* SIOCSIWRETRY */
        ipw2100_wx_get_retry,     /* SIOCGIWRETRY */
        ipw2100_wx_set_encode,    /* SIOCSIWENCODE */
        ipw2100_wx_get_encode,    /* SIOCGIWENCODE */
        NULL,                     /* SIOCSIWPOWER */
        NULL,                     /* SIOCGIWPOWER */
};

struct iw_handler_def ipw2100_wx_handler_def = 
{
	.num_standard = sizeof(ipw2100_wx_handlers) / sizeof(iw_handler),
	.num_private = 0,
	.num_private_args = 0,
	.standard = ipw2100_wx_handlers,
	.private = NULL,
	.private_args = NULL,
};

/*
 * Get wireless statistics.
 * Called by /proc/net/wireless
 */
struct iw_statistics *ipw2100_wx_wireless_stats(struct net_device * dev)
{
	struct ipw2100_priv *priv = dev->priv;
	struct iw_statistics *wstats;
	u32 ord_value, ord_len = sizeof(u32);

	if (!priv)
		return (struct iw_statistics *) NULL;

	/* if hw is disabled, then ipw2100_get_ordinal() can't be called.
	 * ipw2100_wx_wireless_stats seems to be called before fw is 
	 * initialized. */
	if (priv->disabled)
		return (struct iw_statistics *) NULL;

	wstats = &priv->wstats;

	if (ipw2100_get_ordinal(priv, IPW_ORD_RSSI_AVG_CURR, 
				&ord_value, &ord_len))
		goto fail_get_ordinal;

        /* FIXME: updated for what ???? */
	wstats->qual.level = ord_value + IPW2100_RSSI_TO_DBM;
	wstats->qual.updated = 7;	

	if (ipw2100_get_ordinal(priv, IPW_ORD_STAT_PERCENT_MISSED_BCNS, 
				&ord_value, &ord_len))
		goto fail_get_ordinal;
        /* FIXME: this is percent and not a # */
	wstats->miss.beacon = ord_value; 

	if (ipw2100_get_ordinal(priv, IPW_ORD_STAT_TX_RETRIES, 
				&ord_value, &ord_len))
		goto fail_get_ordinal;
	wstats->discard.retries = ord_value;

	return wstats;

 fail_get_ordinal:
	IPW2100_DEBUG_WX("failed querying ordinals.\n");

	return (struct iw_statistics *) NULL;
}



#endif /* WIRELESS_EXT > 12 */ 
