// ****************************************************************************
//
//      CDspCommObject.cpp
//
//              Implementation file for EchoGals generic driver DSP interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#ifndef ECHO24_FAMILY
#define ECHOGALS_FAMILY
#endif

#include "Echoaudio.h"

#ifdef DSP_56361
#include "DSP/LoaderDSP.c"
#endif



int CreateCommPage(struct CEchoGals *eg) {
  int i;

  // Check if the compiler added extra padding inside the structure
  if (offsetof(DspCommPage_t, byVmixerLevel) != 0xba0L) {
    ECHO_DEBUGPRINTF(("NewDspComm() - Invalid DspComm structure\n"));
    return(ECHOSTATUS_ARCH_NOT_SUPPORTED);
  }
  if (sizeof(dma_addr_t)!=4) {
    ECHO_DEBUGPRINTF(("NewDspComm() - dma_addr_t must be 32 bits long\n"));
    return(ECHOSTATUS_ARCH_NOT_SUPPORTED);
  }

  // Init all the basic stuff
  eg->BadBoard = TRUE;               // Set TRUE until DSP loaded
  eg->DspCode = NULL;               // Current DSP code not loaded
  eg->DigitalMode = DIGITAL_MODE_NONE;
  eg->InputClock = ECHO_CLOCK_INTERNAL;
  eg->OutputClock = ECHO_CLOCK_WORD;

#ifdef MIDI_SUPPORT
  eg->NextMidiWriteTime = 0;
#endif
  eg->ASICLoaded=0;

  // Create the DSP comm page - this is the area of memory read and written by the DSP via bus mastering
  if (ECHOSTATUS_OK != OsPageAllocate(eg->ossprivate_data, (sizeof(DspCommPage_t)+PAGE_SIZE-1)/PAGE_SIZE, (void **) &eg->DspCommPage, &eg->CommPagePhys) || eg->DspCommPage == NULL) {
    ECHO_DEBUGPRINTF(("CDspCommObject DSP comm page memory allocation failed\n"));
    return(ECHOSTATUS_NO_MEM);
  }

  // Init the comm page
  eg->DspCommPage->dwCommSize = SWAP32(sizeof(DspCommPage_t));       // Size of DSP comm page
  eg->DspCommPage->dwHandshake = 0xffffffff;
  eg->DspCommPage->dwMidiXmitStatus = 0xffffffff;
  for (i=0; i<DSP_MAXAUDIOINPUTS; i++)
    eg->DspCommPage->InLineLevel[i] = 0x00;

  // Set line levels so we don't blast any inputs on startup
  memset(eg->DspCommPage->byMonitors, GENERIC_TO_DSP(ECHOGAIN_MUTED), MONITOR_ARRAY_SIZE);   // Mute all monitors
  memset(eg->DspCommPage->byVmixerLevel, GENERIC_TO_DSP(ECHOGAIN_MUTED), VMIXER_ARRAY_SIZE); // Mute all virtual mixer levels

#ifdef DIGITAL_INPUT_AUTO_MUTE_SUPPORT
  eg->DigitalInAutoMute = TRUE;
#endif
  return(ECHOSTATUS_OK);
}


//===========================================================================
//
// Destructor
//
//===========================================================================

void DeleteCDspComm(struct CEchoGals *eg) {

  // Reset all active pipes (@@ they should already be turned off by DelEchoGals())
  ResetTransport(eg, eg->ActiveMask);

  // Meters off
  SetMetersOn(eg, FALSE);

#ifdef MIDI_SUPPORT
  // MIDI input off
  eg->MidiOnCount = 1;
  SetMidiOn(eg, FALSE);
#endif // MIDI_SUPPORT

  // Go to sleep
  GoComatose(eg);

  // Free the comm page
  if (eg->DspCommPage) {
    if (ECHOSTATUS_OK != OsPageFree(eg->ossprivate_data, (sizeof(DspCommPage_t)+PAGE_SIZE-1)/PAGE_SIZE, eg->DspCommPage, eg->CommPagePhys)) {
      ECHO_DEBUGPRINTF(("DeleteCDspComm: DSP comm page memory free failed\n"));
    }
  }

  ECHO_DEBUGPRINTF(("DeleteCDspComm: DspCommObject is toast!\n"));
}



u32 GetHandshakeFlag(struct CEchoGals *eg) {
  return(SWAP32(eg->DspCommPage->dwHandshake));
}

void ClearHandshake(struct CEchoGals *eg) {
  eg->DspCommPage->dwHandshake=0;
}

u32 GetDspRegister(struct CEchoGals *eg, u32 Index) {
  return(SWAP32(eg->DspRegBase[Index]));
}

void SetDspRegister(struct CEchoGals *eg, u32 Index, u32 Value) {
  eg->DspRegBase[Index]=SWAP32(Value);
}

u32 GetStatusReg(struct CEchoGals *eg) {
  return(SWAP32(eg->DspRegBase[CHI32_STATUS_REG]));
}

u32 GetControlRegister(struct CEchoGals *eg) {
  return(SWAP32(eg->DspCommPage->dwControlReg));
}

void SetControlRegister(struct CEchoGals *eg, u32 ControlRegister) {
  eg->DspCommPage->dwControlReg = SWAP32(ControlRegister);
}



//===========================================================================
//
// Some vector commands involve the DSP reading or writing data to and
// from the comm page; if you send one of these commands to the DSP,
// it will complete the command and then write a non-zero value to
// the dwHandshake field in the comm page.  This function waits for the
// handshake to show up.
//
//===========================================================================

int WaitForHandshake(struct CEchoGals *eg) {
  int i;

  MB;
  // Wait up to three milliseconds for the handshake from the DSP
  for (i=0; i<HANDSHAKE_TIMEOUT; i++) {
    // Look for the handshake value
    if (GetHandshakeFlag(eg))
      break;
    // Give the DSP time to access the comm page
    OsSnooze(2);
  }

  if (i == HANDSHAKE_TIMEOUT) {
    ECHO_DEBUGPRINTF(("WaitForHandshake: Timeout waiting for DSP\n"));
    return(FALSE);
  }
  return(TRUE);
}



//===========================================================================
//
// Much of the interaction between the DSP and the driver is done via vector
// commands; SendVector writes a vector command to the DSP.  Typically,
// this causes the DSP to read or write fields in the comm page.
//
// Returns ECHOSTATUS_OK if sent OK.
//
//===========================================================================

int SendVector(struct CEchoGals *eg, u32 Command) {
  u64 Timeout, CurTime;

  WMB;
  OsGetSystemTime(&CurTime);
  Timeout = CurTime + 100000L;    // 100ms

  // Wait for the "vector busy" bit to be off
  while (CurTime <= Timeout) {
    RMB;
    if (!(GetDspRegister(eg, CHI32_VECTOR_REG) & CHI32_VECTOR_BUSY)) {
      SetDspRegister(eg, CHI32_VECTOR_REG, Command);
      return(ECHOSTATUS_OK);
    }
    OsGetSystemTime(&CurTime);
  }

  ECHO_DEBUGPRINTF(("Punked out on SendVector\n"));
  return(ECHOSTATUS_DSP_TIMEOUT);
}



// Tell DSP to release the hardware interrupt
void AckInt(struct CEchoGals *eg) {
  eg->DspCommPage->wMidiInData[0] = 0;
  WMB;
  SendVector(eg, DSP_VC_ACK_INT);
  WMB;
}




/****************************************************************************

	Firmware loading functions

 ****************************************************************************/


//===========================================================================
//
// This function is used to read back the serial number from the DSP;
// this is triggered by the SET_COMMPAGE_ADDR command.
//
// Only some early Echogals products have serial numbers in the ROM;
// the serial number is not used, but you still need to do this as
// part of the DSP load process.
//
//===========================================================================

int ReadSn(struct CEchoGals *eg) {
  int j;
  u32 Sn[6];
  int Status;

  for (j=0; j<5; j++) {
    Status = Read_DSP(eg, &Sn[j]);
    if (Status != 0) {
      ECHO_DEBUGPRINTF(("Failed to read serial number word %d\n", j));
      return(ECHOSTATUS_DSP_DEAD);
    }
  }
  ECHO_DEBUGPRINTF(("Read serial number %08x %08x %08x %08x %08x\n", Sn[0], Sn[1], Sn[2], Sn[3], Sn[4]));
  return(ECHOSTATUS_OK);
}




//===========================================================================
//
// ASIC status check - some cards have one or two ASICs that need to be
// loaded.  Once that load is complete, this function is called to see if
// the load was successful.
//
// If this load fails, it does not necessarily mean that the hardware is
// defective - the external box may be disconnected or turned off.
//
// This routine sometimes fails for Layla20; for Layla20, the loop runs 5 times
// and succeeds if it wins on three of the loops.
//
//===========================================================================

int CheckAsicStatus(struct CEchoGals *eg) {
#if NUM_ASIC_TESTS!=1
  u32 GoodCt = 0;
#endif
  u32 AsicStatus;
  u32 st;
  int NumTests, i;

  if (!eg->HasASIC) {
    eg->ASICLoaded = TRUE;
    return(TRUE);
  }

  NumTests = NUM_ASIC_TESTS;

  for (i=0; i<NumTests; i++) {
    // Send the vector command
    SendVector(eg, DSP_VC_TEST_ASIC);

    // The DSP will return a value to indicate whether or not the ASIC is currently loaded
    st = Read_DSP(eg, &AsicStatus);
    if (st != ECHOSTATUS_OK) {
      ECHO_DEBUGPRINTF(("CheckAsicStatus: failed on Read_DSP\n"));
      eg->ASICLoaded = FALSE;
      return FALSE;
    }
/*#ifdef ECHO_DEBUG
    if (AsicStatus != ASIC_ALREADY_LOADED && AsicStatus != ASIC_NOT_LOADED) {
      ECHO_DEBUGBREAK();
    }
#endif*/

    if (AsicStatus == ASIC_ALREADY_LOADED) {

#if NUM_ASIC_TESTS==1
      eg->ASICLoaded = TRUE;
      break;
#else
      // For Layla20
      if (++GoodCt == 3) {
        eg->ASICLoaded = TRUE;
        break;
      }
#endif
    } else {
      eg->ASICLoaded = FALSE;
    }
  }

  return(eg->ASICLoaded);
}


//===========================================================================
//
//      Load ASIC code - done after the DSP is loaded
//
//===========================================================================

int LoadASICgeneric(struct CEchoGals *eg, u32 Cmd, u8 *Code, u32 Size) {
  u32 i;

  if (!eg->HasASIC)
    return TRUE;

#ifdef _DEBUG
  u32 dwChecksum = 0;
  u64 StartTime, CurTime;

  OsGetSystemTime(&StartTime);
#endif

  // Send the "Here comes the ASIC" command
  if (Write_DSP(eg, Cmd) != ECHOSTATUS_OK)
    return(FALSE);

  // Write length of ASIC file in bytes
  if (Write_DSP(eg, Size) != ECHOSTATUS_OK)
    return(FALSE);

  for (i=0; i<Size; i++) {
#ifdef _DEBUG
    dwChecksum += Code[i];
#endif

    if (Write_DSP(eg, Code[i]) != ECHOSTATUS_OK) {
      ECHO_DEBUGPRINTF(("failed on Write_DSP\n"));
      return(FALSE);
    }
  }

#ifdef _DEBUG
  OsGetSystemTime(&CurTime);
  ECHO_DEBUGPRINTF(("LoadASIC took %ld usec.\n", (ULONG) (CurTime - StartTime)));
  ECHO_DEBUGPRINTF(("Checksum is 0x%x\n", dwChecksum));
  ECHO_DEBUGPRINTF(("ASIC load OK\n"));
#endif
  ECHO_DEBUGPRINTF(("ASIC loaded\n"));
  return(TRUE);
}



//===========================================================================
//
// InstallResidentLoader
//
// Install the resident loader for 56361 DSPs;  The resident loader
// is on the EPROM on the board for 56301 DSP.
//
// The resident loader is a tiny little program that is used to load
// the real DSP code.
//
//===========================================================================

#ifdef DSP_56361

int InstallResidentLoader(struct CEchoGals *eg) {
  u32 Address;
  int Index;
  int Num;
  int i;
  int dwReturn;
  u16 *pCode;
  u32 dwStatus;
  int fSuccess;
  u64 CurTime, Timeout;

  // 56361 cards only!
  if (DEVICE_ID_56361 != eg->DeviceId)
    return ECHOSTATUS_OK;

  // Look to see if the resident loader is present.  If the resident loader
  // is already installed, host flag 5 will be on.

  dwStatus = GetDspRegister(eg, CHI32_STATUS_REG);
  if (dwStatus & CHI32_STATUS_REG_HF5) {
    ECHO_DEBUGPRINTF(("Resident loader already installed; status is 0x%x\n", dwStatus));
    return ECHOSTATUS_OK;
  }

  // Set DSP format bits for 24 bit mode
  SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) | 0x900);

  //---------------------------------------------------------------------------
  //
  // Loader
  //
  // The DSP code is an array of 16 bit words.  The array is divided up into
  // sections.  The first word of each section is the size in words, followed
  // by the section type.
  //
  // Since DSP addresses and data are 24 bits wide, they each take up two
  // 16 bit words in the array.
  //
  // This is a lot like the other loader loop, but it's not a loop,
  // you don't write the memory type, and you don't write a zero at the end.
  //
  //---------------------------------------------------------------------------

  pCode = pwLoaderDSP;
  
  // Skip the header section; the first word in the array is the size of
  // the first section, so the first real section of code is pointed to
  //  by pCode[0].
  Index = pCode[0];
  
  // Skip the section size, LRS block type, and DSP memory type
  Index += 3;
  
  // Get the number of DSP words to write
  Num = pCode[Index++];
  
  // Get the DSP address for this block; 24 bits, so build from two words
  Address = (pCode[Index] << 16) + pCode[Index + 1];
  Index += 2;
  
  // Write the count to the DSP
  dwReturn = Write_DSP(eg, (u32)Num);
  if (dwReturn != 0) {
    ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write word count!\n"));
    return ECHOSTATUS_DSP_DEAD;
  }
  // Write the DSP address
  dwReturn = Write_DSP(eg, Address);
  if (dwReturn != 0) {
    ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write DSP address!\n"));
    return ECHOSTATUS_DSP_DEAD;
  }

  // Write out this block of code to the DSP
  for (i=0; i<Num; i++) {
    u32 Data;

    Data = (pCode[Index] << 16) + pCode[Index + 1];
    dwReturn = Write_DSP(eg, Data);
    if (dwReturn) {
      ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write DSP code\n"));
      return ECHOSTATUS_DSP_DEAD;
    }
    Index+=2;
  }

  // Wait for flag 5 to come up

  OsGetSystemTime(&CurTime);
  Timeout = CurTime + 10000L;     // 10m.s.
  fSuccess = FALSE;
  do {
    OsSnooze(50);                       // Give the DSP some time; no need to hog the CPU
    dwStatus = GetDspRegister(eg, CHI32_STATUS_REG);
    if (dwStatus & CHI32_STATUS_REG_HF5) {
      fSuccess = TRUE;
      break;
    }
    OsGetSystemTime(&CurTime);
  } while (CurTime < Timeout);

  if (FALSE == fSuccess) {
    ECHO_DEBUGPRINTF(("Resident loader failed to set HF5\n"));
    return ECHOSTATUS_DSP_DEAD;
  }

  ECHO_DEBUGPRINTF(("Resident loader successfully installed\n"));
  return ECHOSTATUS_OK;
}

#endif // DSP_56361


//===========================================================================
//
// LoadDSP
//
// This loads the DSP code.
//
//===========================================================================

int LoadDSP(struct CEchoGals *eg, u16 *pCode) {
  u32 Address;
  int Index;
  int Num;
  int i;
  int err;
  u64 Timeout, CurTime;

  if (eg->DspCode == pCode) {
    ECHO_DEBUGPRINTF(("DSP is already loaded!\n"));
    return(ECHOSTATUS_FIRMWARE_LOADED);
  }
  eg->BadBoard = TRUE;               // Set TRUE until DSP loaded
  eg->DspCode = NULL;               // Current DSP code not loaded
  eg->ASICLoaded = FALSE;            // Loading the DSP code will reset the ASIC

  ECHO_DEBUGPRINTF(("LoadDSP  Set BadBoard to TRUE\n"));

  //
  //      If this board requires a resident loader, install it.
  //
#ifdef DSP_56361
  InstallResidentLoader(eg);
#endif

  // Send software reset command
  if (ECHOSTATUS_OK != SendVector(eg, DSP_VC_RESET)) {
    ECHO_DEBUGPRINTF(("LoadDsp SendVector DSP_VC_RESET failed, Critical Failure\n"));
    return(ECHOSTATUS_DSP_DEAD);
  }
  // Delay 10us
  OsSnooze(10L);

  // Wait 10ms for HF3 to indicate that software reset is complete
  OsGetSystemTime(&CurTime);
  Timeout = CurTime + 10000L;     // 10ms

  while (1) {
    if (GetDspRegister(eg, CHI32_STATUS_REG) & CHI32_STATUS_REG_HF3)
      break;
    OsGetSystemTime(&CurTime);
    if (CurTime > Timeout) {
      ECHO_DEBUGPRINTF(("LoadDSP Timeout waiting for CHI32_STATUS_REG_HF3\n"));
      return(ECHOSTATUS_DSP_TIMEOUT);
    }
  }

  // Set DSP format bits for 24 bit mode now that soft reset is done
  SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) | (u32)0x900);

  //---------------------------------------------------------------------------
  // Main loader loop
  //---------------------------------------------------------------------------
  Index = pCode[0];

  for (;;) {
    int iBlockType;
    int iMemType;

    // Total Block Size
    Index++;

    // Block Type
    iBlockType = pCode[Index];
    if (iBlockType == 4)                // We're finished
      break;

    Index++;

    // Memory Type  P=0,X=1,Y=2
    iMemType = pCode[Index];
    Index++;

    // Block Code Size
    Num = pCode[Index];
    Index++;
    if (Num == 0)                       // We're finished
      break;

    // Start Address
    Address = ((u32)pCode[Index] << 16) + pCode[Index + 1];
    Index += 2;

    err = Write_DSP(eg, (u32)Num);
    if (err) {
      ECHO_DEBUGPRINTF(("LoadDSP: failed to write number of DSP words\n"));
      return ECHOSTATUS_DSP_DEAD;
    }

    err = Write_DSP(eg, Address);
    if (err) {
      ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP address\n"));
      return ECHOSTATUS_DSP_DEAD;
    }

    err = Write_DSP(eg, (u32)iMemType);
    if (err) {
      ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP memory type\n"));
      return ECHOSTATUS_DSP_DEAD;
    }
    // Code
    for (i=0; i<Num; i++) {
      u32 dwData;

      dwData = ((u32)pCode[Index] << 16) + pCode[Index + 1];
      err = Write_DSP(eg, dwData);
      if (err) {
        ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP data\n"));
        return ECHOSTATUS_DSP_DEAD;
      }

      Index += 2;
    }
//              ECHO_DEBUGPRINTF( ("\tEnd Code Block\n") );
  }
  err = Write_DSP(eg, 0);  // We're done!!!
  if (err) {
    ECHO_DEBUGPRINTF(("LoadDSP: Failed to write final zero\n"));
    return ECHOSTATUS_DSP_DEAD;
  }

  // Delay 10us
  OsSnooze(10L);

  OsGetSystemTime(&CurTime);
  Timeout = CurTime + 500000L;    // 1/2 sec. timeout

  while (CurTime <= Timeout) {
    // Wait for flag 4 - indicates that the DSP loaded OK
    if (GetDspRegister(eg, CHI32_STATUS_REG) & CHI32_STATUS_REG_HF4) {
      SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) & ~0x1b00);

      err = Write_DSP(eg, DSP_FNC_SET_COMMPAGE_ADDR);
      if (err) {
        ECHO_DEBUGPRINTF(("LoadDSP: Failed to write DSP_FNC_SET_COMMPAGE_ADDR\n"));
        return ECHOSTATUS_DSP_DEAD;
      }

      err = Write_DSP(eg, eg->CommPagePhys);
      if (err) {
        ECHO_DEBUGPRINTF(("LoadDSP: Failed to write comm page address\n"));
        return ECHOSTATUS_DSP_DEAD;
      }

      /* Get the serial number via slave mode.
       * This is triggered by the SET_COMMPAGE_ADDR command.
       * We don't actually use the serial number but we have to get
       * it as part of the DSP init voodoo. */
      if ((err = ReadSn(eg))) {
        ECHO_DEBUGPRINTF(("LoadDSP: Failed to read serial number\n"));
        return err;
      }

      eg->DspCode = pCode;          // Show which DSP code loaded
      eg->BadBoard = FALSE;          // DSP OK
      ECHO_DEBUGPRINTF(("LoadDSP:  Set BadBoard to FALSE\n"));
      return ECHOSTATUS_OK;
    }
    OsGetSystemTime(&CurTime);
  }

  ECHO_DEBUGPRINTF(("LoadDSP: DSP load timed out waiting for HF4\n"));
  return ECHOSTATUS_DSP_TIMEOUT;
}


//===========================================================================
//
// LoadFirmware takes care of loading the DSP and any ASIC code.
//
//===========================================================================

int LoadFirmware(struct CEchoGals *eg) {
  int st;

  // Try to load the DSP
  if (!eg->DspCodeToLoad || !eg->DspCommPage)
    return(ECHOSTATUS_NO_MEM);

  // See if the ASIC is present and working - only if the DSP is already loaded
  if (eg->DspCode) {
    st = CheckAsicStatus(eg);
    if (st)
      return(ECHOSTATUS_OK);
    // ASIC check failed; force the DSP to reload
    eg->DspCode = NULL;
  }

  // Try and load the DSP
  st = LoadDSP(eg, eg->DspCodeToLoad);
  if (st!=ECHOSTATUS_OK && st!=ECHOSTATUS_FIRMWARE_LOADED)
    return(st);

  // Load the ASIC if the DSP load succeeded; LoadASIC will
  // always return TRUE for cards that don't have an ASIC.
  st = LoadASIC(eg);
  if (!st) {
    st = ECHOSTATUS_ASIC_NOT_LOADED;
  } else {
    // ASIC load was successful
    st = RestoreDspSettings(eg);
  }

  return(st);
}



//===========================================================================
//
// This is called after LoadFirmware to restore old gains, meters on, monitors, etc.
//
//===========================================================================

int RestoreDspSettings(struct CEchoGals *eg) {
  int err;

  ECHO_DEBUGPRINTF(("RestoreDspSettings\n"));

  if (!CheckAsicStatus(eg))
    return(ECHOSTATUS_DSP_DEAD);

  eg->DspCommPage->byGDClockState = GD_CLOCK_UNDEF;          //@@ Gina20/Darla20 only. Should be harmless for other cards.
  eg->DspCommPage->byGDSpdifStatus = GD_SPDIF_STATUS_UNDEF;

  eg->DspCommPage->dwHandshake = 0xffffffff;

#ifdef MIDI_SUPPORT
  eg->NextMidiWriteTime = 0;
#endif

  SetSampleRate(eg, GetSampleRate(eg));
  if (eg->MetersOn)
    SendVector(eg, DSP_VC_METERS_ON);

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  SetInputClock(eg, eg->InputClock);
  SetOutputClock(eg, eg->OutputClock);

  if (UpdateOutputLineLevel(eg))
    return(ECHOSTATUS_DSP_DEAD);

  if (UpdateInputLineLevel(eg))
    return(ECHOSTATUS_DSP_DEAD);
#ifndef EMUL
  if (eg->HasVmixer) {
    if (UpdateVmixerLevel(eg))
      return(ECHOSTATUS_DSP_DEAD);
  }
#endif

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  ClearHandshake(eg);
  err=SendVector(eg, DSP_VC_UPDATE_FLAGS);

  ECHO_DEBUGPRINTF(("RestoreDspSettings done\n"));
  return(err);
}



/****************************************************************************

	DSP utilities

 ****************************************************************************/

//===========================================================================
//
// Write_DSP writes a 32-bit value to the DSP; this is used almost
// exclusively for loading the DSP.
//
//===========================================================================

int Write_DSP(struct CEchoGals *eg, u32 Data) {
  u32 Status;
  u64 CurTime, Timeout;

//      ECHO_DEBUGPRINTF(("Write_DSP\n"));
  OsGetSystemTime(&CurTime);
  Timeout = CurTime + 10000000L;   // 10 sec.
  while (Timeout >= CurTime) {
    MB;
    Status = GetDspRegister(eg, CHI32_STATUS_REG);
    if ((Status & CHI32_STATUS_HOST_WRITE_EMPTY) != 0) {
      SetDspRegister(eg, CHI32_DATA_REG, Data);
    MB;
//                      ECHO_DEBUGPRINTF(("Write DSP: 0x%x", Data));
      return ECHOSTATUS_OK;
    }
    OsGetSystemTime(&CurTime);
  }

  eg->BadBoard = TRUE;       // Set TRUE until DSP re-loaded
  ECHO_DEBUGPRINTF(("Write_DSP: Set BadBoard to TRUE\n"));
  return(ECHOSTATUS_DSP_TIMEOUT);
}



//===========================================================================
//
// Read_DSP reads a 32-bit value from the DSP; this is used almost
// exclusively for loading the DSP and checking the status of the ASIC.
//
//===========================================================================

int Read_DSP(struct CEchoGals *eg, u32 *Data) {
  u32 Status;
  u64 CurTime, Timeout;

//  ECHO_DEBUGPRINTF(("Read_DSP\n"));
  OsGetSystemTime(&CurTime);

  Timeout = CurTime + READ_DSP_TIMEOUT;
  while (Timeout >= CurTime) {
    Status = GetDspRegister(eg, CHI32_STATUS_REG);
    if ((Status & CHI32_STATUS_HOST_READ_FULL) != 0) {
      *Data = GetDspRegister(eg, CHI32_DATA_REG);
//       ECHO_DEBUGPRINTF(("Read DSP: 0x%x\n", *pdwData));
      return ECHOSTATUS_OK;
    }
    OsGetSystemTime(&CurTime);
  }

  eg->BadBoard = TRUE;       // Set TRUE until DSP re-loaded
  ECHO_DEBUGPRINTF(("Read_DSP: Set BadBoard to TRUE\n"));
  return ECHOSTATUS_DSP_TIMEOUT;
}



//===========================================================================
//
// Tell the DSP to go into low-power mode
//
//===========================================================================

int GoComatose(struct CEchoGals *eg) {

  ECHO_DEBUGPRINTF(("GoComatose\n"));
  if (eg->DspCode) {
    // Make LoadFirmware do a complete reload
    eg->DspCode = NULL;
    // Put the DSP to sleep
    return(SendVector(eg, DSP_VC_GO_COMATOSE));
  }
  return(ECHOSTATUS_OK);
}



// Get the current sample rate
u32 GetSampleRate(const struct CEchoGals *eg) {
  return(SWAP32(eg->DspCommPage->dwSampleRate));
}



// Returns what input clocks are currently detected
u32 DspGetInputClockDetect(const struct CEchoGals *eg) {
  return(SWAP32(eg->DspCommPage->dwStatusClocks));
}



void DspMask_SetBit(ChMaskDsp *cm, int bit) {

  if (bit<16)
    cm->DspMask[0] |= SWAP16(1<<bit);
  else
    cm->DspMask[1] |= SWAP16(1<<(bit-16));
//ECHO_DEBUGPRINTF(("setbit  %04x %04x\n",cm->DspMask[0],cm->DspMask[1]));
}



void DspMask_ClearBit(ChMaskDsp *cm, int bit) {

  if (bit<16)
    cm->DspMask[0] &= ~SWAP16(1<<bit);
  else
    cm->DspMask[1] &= ~SWAP16(1<<(bit-16));
//ECHO_DEBUGPRINTF(("clbit  %04x %04x\n",cm->DspMask[0],cm->DspMask[1]));
}



void DspMask_Clear(ChMaskDsp *cm) {

  cm->DspMask[0]=0;
  cm->DspMask[1]=0;
}



void DspMask_OrChannelMask(ChMaskDsp *cm, u32 Mask) {

  cm->DspMask[0] |= SWAP16(Mask&0x0000ffff);
  cm->DspMask[1] |= SWAP16((Mask&0xffff0000)>>16);
//ECHO_DEBUGPRINTF(("orbit  %04x %04x\n",cm->DspMask[0],cm->DspMask[1]));
}



void DspMask_AndChannelMask(ChMaskDsp *cm, u32 Mask) {

  cm->DspMask[0] &= SWAP16(Mask&0x0000ffff);
  cm->DspMask[1] &= SWAP16((Mask&0xffff0000)>>16);
//ECHO_DEBUGPRINTF(("andbit  %04x %04x\n",cm->DspMask[0],cm->DspMask[1]));
}



int DspMask_IsEmpty(ChMaskDsp *cm) {

  if (cm->DspMask[0] || cm->DspMask[1])
    return(0);
  return(1);
}



/****************************************************************************

  Mixer functions

 ****************************************************************************/

// Set the nominal level for an input or output bus (true = -10dBV, false = +4dBu)
int SetNominalLevel(struct CEchoGals *eg, u16 Index, char Consumer) {

  ECHO_DEBUGPRINTF(("SetNominalLevel[%d]=%d\n", (int)Index, (int)Consumer));
  // Wait for the handshake (OK even if ASIC is not loaded)
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Check the channel index
  if (Index < eg->NumBussesOut+eg->NumBussesIn) {
    eg->NominalLevel[Index]=Consumer;
    // Set the nominal bit

    if (Consumer)
      DspMask_SetBit(&eg->DspCommPage->cmdNominalLevel, Index);
    else
      DspMask_ClearBit(&eg->DspCommPage->cmdNominalLevel, Index);

    return(ECHOSTATUS_OK);
  }

  ECHO_DEBUGPRINTF(("SetNominalOutLineLevel: Invalid index %d\n", (int)Index));
  return(ECHOSTATUS_INVALID_CHANNEL);
}



//===========================================================================
//
// Set the gain for a single physical output channel (dB).
//
// Cond: @@ ECHOGAIN_MINOUT <= Gain <= ECHOGAIN_MAXOUT, Channel<eg->NumBussesOut
//
//===========================================================================

int SetOutputGain(struct CEchoGals *eg, u16 Channel, int Gain) {

  if (Gain<ECHOGAIN_MINOUT)
    Gain=ECHOGAIN_MINOUT;
  else if (Gain>ECHOGAIN_MAXOUT)
    Gain=ECHOGAIN_MAXOUT;

  if (Channel<eg->NumBussesOut) {
    if (!WaitForHandshake(eg))
      return(ECHOSTATUS_DSP_DEAD);

    // Save the new value
    eg->OutLevel[Channel] = Gain;
    eg->DspCommPage->OutLineLevel[Channel] = GENERIC_TO_DSP(Gain);

//    ECHO_DEBUGPRINTF(("SetBusOutGain: [%d] = 0x%x\n", Channel, Gain));
    return(ECHOSTATUS_OK);
  }

  ECHO_DEBUGPRINTF(("SetPipeOutGain: Invalid out pipe %d\n", Channel));
  return(ECHOSTATUS_INVALID_CHANNEL);
}



//===========================================================================
//
// Set input bus gain (one unit is 0.5dB !)
//
//===========================================================================

int SetInputGain(struct CEchoGals *eg, u16 Channel, int Gain) {

  if (Gain<ECHOGAIN_MININP)
    Gain=ECHOGAIN_MININP;
  else if (Gain>ECHOGAIN_MAXINP)
    Gain=ECHOGAIN_MAXINP;

  if (Channel>eg->NumBussesIn)
    return(ECHOSTATUS_INVALID_CHANNEL);

  // Wait for the handshake (OK even if ASIC is not loaded)
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Save the new value
  eg->InLevel[Channel]=Gain;
  Gain=GENERIC_TO_DSP(Gain);
  Gain+=GL20_INPUT_GAIN_MAGIC_NUMBER;

  // Put it in the comm page
  eg->DspCommPage->InLineLevel[Channel]=Gain;
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// Set the monitor level from an input bus to an output bus.
//
//===========================================================================

int SetMonitorGain(struct CEchoGals *eg, u16 OutChannel, u16 InChannel, int Gain) {

  if (Gain<ECHOGAIN_MINOUT)
    Gain=ECHOGAIN_MINOUT;
  else if (Gain>ECHOGAIN_MAXOUT)
    Gain=ECHOGAIN_MAXOUT;
  //ECHO_DEBUGPRINTF(("SetMonitorGain[%d][%d] = %d\n", OutChannel, InChannel, Gain));

  // Wait for the offset
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  eg->MonitorLevel[OutChannel][InChannel]=Gain;
  // Write the gain value to the comm page
  eg->DspCommPage->byMonitors[MONITOR_INDEX(OutChannel, InChannel)]=GENERIC_TO_DSP(Gain);

  return(ECHOSTATUS_OK);
}



// Tell the DSP to read and update output, nominal & monitor levels in comm page.
int UpdateOutputLineLevel(struct CEchoGals *eg) {

//  ECHO_DEBUGPRINTF(("UpdateAudioOutLineLevel\n"));
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);
  ClearHandshake(eg);
  WMB;
  return(SendVector(eg, DSP_VC_UPDATE_OUTVOL));
}



// Tell the DSP to read and update input levels in comm page
int UpdateInputLineLevel(struct CEchoGals *eg) {

//  ECHO_DEBUGPRINTF(("UpdateAudioInLineLevel\n"));
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);
  ClearHandshake(eg);
  WMB;
  return(SendVector(eg, DSP_VC_UPDATE_INGAIN));
}



// Tell the DSP to reread the flags from the comm page
int UpdateFlags(struct CEchoGals *eg) {

//  ECHO_DEBUGPRINTF(("UpdateFlags\n"));
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);
  ClearHandshake(eg);
  WMB;
  return(SendVector(eg, DSP_VC_UPDATE_FLAGS));
}



//===========================================================================
//
// SetMetersOn turns the meters on or off.  If meters are turned on, the DSP
// will write the meter and clock detect values to the comm page at about 30Hz
//
//===========================================================================

void SetMetersOn(struct CEchoGals *eg, char On) {
int Index;

  if (On && !eg->MetersOn) {
    SendVector(eg, DSP_VC_METERS_ON);
    eg->MetersOn=1;
  } else if (!On && eg->MetersOn) {
    SendVector(eg, DSP_VC_METERS_OFF);
    eg->MetersOn=0;
    for (Index=0; Index<DSP_MAXPIPES; Index++) {
      eg->DspCommPage->VULevel[Index] = GENERIC_TO_DSP(ECHOGAIN_MUTED);
      eg->DspCommPage->PeakMeter[Index] = GENERIC_TO_DSP(ECHOGAIN_MUTED);
    }
  }
}



//===========================================================================
//
// Fill out an ECHOGALS_METERS struct using the current values in the comm page.
//
// Meters are written to the comm page by the DSP as follows:
//
// Output busses
// Input busses
// Output pipes (vmixer cards only)
//
// This function is designed to match the requirements of ALSA and it assumes
// there are no more than 16 in/out busses or pipes (maximum is ECHO_MAXAUDIO*,
// but currently there are no Echoaudio cards have so many channels. Meters is
// an array of signed chars[3][16][2])
//
//===========================================================================

void GetAudioMeters(struct CEchoGals *eg, long *Meters) {
  int i, m, n;

  m=0;
  n=0;
  for (i=eg->NumBussesOut; i>0; i--, m++) {
    Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->VULevel[m]));
    Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->PeakMeter[m]));
  }
  for (; n<32; n++)
    Meters[n]=0;

  for (i=eg->NumBussesIn; i>0; i--, m++) {
    Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->VULevel[m]));
    Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->PeakMeter[m]));
  }
  for (; n<64; n++)
    Meters[n]=0;

  if (eg->HasVmixer) {
    for (i=eg->NumPipesOut; i>0; i--, m++) {
      Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->VULevel[m]));
      Meters[n++] = DSP_TO_GENERIC(((int)eg->DspCommPage->PeakMeter[m]));
    }
  }
  for (; n<96; n++)
    Meters[n]=0;
}



void DspSetFlags(struct CEchoGals *eg, u32 Flags) {
  u32 fl;

  fl=SWAP32(eg->DspCommPage->dwFlags);
  fl|=Flags;
  eg->DspCommPage->dwFlags=fl;
  if (eg->ASICLoaded && WaitForHandshake(eg))
    UpdateFlags(eg);
}



void DspClearFlags(struct CEchoGals *eg, u32 Flags) {
  u32 fl;

  fl=SWAP32(eg->DspCommPage->dwFlags);
  fl &= ~Flags;
  eg->DspCommPage->dwFlags=fl;
  if (eg->ASICLoaded && WaitForHandshake(eg))
    UpdateFlags(eg);
}




/****************************************************************************

	Transport methods

 ****************************************************************************/

int DspSetFormat(struct CEchoGals *eg, u16 PipeIndex, Audioformat_t *Format, char DitherDigitalInputs) {
  u16 DspFormat;

  DspFormat = DSP_AUDIOFORM_SS_16LE;

  // Look for super-interleave
  if (Format->DataInterleave>2) {
    DspFormat = DSP_AUDIOFORM_SUPER_INTERLEAVE_32LE | Format->DataInterleave;
  } else {
    // For big-endian data, only 32 bit mono->mono samples and 32 bit stereo->stereo are supported
    if (Format->DataAreBigEndian) {
      switch (Format->DataInterleave) {
        case 1:
          DspFormat = DSP_AUDIOFORM_MM_32BE;
          break;
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
        case 2:
          DspFormat = DSP_AUDIOFORM_SS_32BE;
          break;
#endif
      }
    } else {
      // Check for 32 bit little-endian mono->mono case
      if (Format->DataInterleave==1 && Format->BitsPerSample==32 && !Format->MonoToStereo) {
        DspFormat = DSP_AUDIOFORM_MM_32LE;
      } else {
        // Handle the other little-endian formats
        switch (Format->BitsPerSample) {
          case 8:
            if (Format->DataInterleave==2)
              DspFormat = DSP_AUDIOFORM_SS_8;
            else
              DspFormat = DSP_AUDIOFORM_MS_8;           //@@ mono->stereo  You may not want this
            break;
          default:
          case 16:
            // If this is a digital input and the no dither flag is set, use the no-dither format
            if (DitherDigitalInputs && (PipeIndex >= (eg->NumPipesOut + eg->FirstDigitalBusIn))) {
              // 16 bit, no dither
              if (Format->DataInterleave==2)
                DspFormat = DSP_AUDIOFORM_SS_16LE_ND;
              else
                DspFormat = DSP_AUDIOFORM_MS_16LE_ND;   //@@ m->s
            } else {
              // 16 bit with dither
              if (Format->DataInterleave==2)
                DspFormat = DSP_AUDIOFORM_SS_16LE;
              else
                DspFormat = DSP_AUDIOFORM_MS_16LE;      //@@ m->s
            }
            break;
          case 32:
            if (Format->DataInterleave==2)
              DspFormat = DSP_AUDIOFORM_SS_32LE;
            else
              DspFormat = DSP_AUDIOFORM_MS_32LE;
            break;
        }
      }                         // check other little-endian formats
    }                           // not big endian data
  }                             // not super-interleave
  ECHO_DEBUGPRINTF(("SetAudioFormat [%d] = %x\n", PipeIndex, DspFormat));
  eg->DspCommPage->wAudioFormat[PipeIndex] = SWAP16(DspFormat);
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// StartTransport starts transport for a set of pipes
//
//===========================================================================

int StartTransport(struct CEchoGals *eg, u32 ChannelMask, u32 CyclicMask) {
  ECHO_DEBUGPRINTF(("StartTransport %x\n", ChannelMask));

  // Wait for the previous command to complete
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Write the appropriate fields in the comm page
  DspMask_OrChannelMask(&eg->DspCommPage->cmdStart, ChannelMask);
  DspMask_OrChannelMask(&eg->DspCommPage->cmdCyclicBuffer, CyclicMask);

  if (!DspMask_IsEmpty(&eg->DspCommPage->cmdStart)) {
    // Clear the handshake and send the vector command
#if 0
{ int x, y;
  for (y=0; y<4096; y+=32) {
    print("%03x ", y);
    for (x=0; x<32; x++)
      print("%02x ", ((char *)eg->DspCommPage)[y*32+x]);
    print("\n");
  }
}
#endif

    ClearHandshake(eg);
    SendVector(eg, DSP_VC_START_TRANSFER);

    // Wait for transport to start
    if (!WaitForHandshake(eg))
      return(ECHOSTATUS_DSP_DEAD);

    // Keep track of which pipes are transporting
    eg->ActiveMask |= ChannelMask;
    DspMask_Clear(&eg->DspCommPage->cmdStart);

    return(ECHOSTATUS_OK);
  }				// if this monkey is being started

  ECHO_DEBUGPRINTF(("StartTransport: No pipes to start!\n"));
  return(ECHOSTATUS_INVALID_CHANNEL);
}



//===========================================================================
//
// StopTransport pauses transport for a set of pipes
//
//===========================================================================

int StopTransport(struct CEchoGals *eg, u32 ChannelMask) {
  ECHO_DEBUGPRINTF(("StopTransport %x\n", ChannelMask));

  // Wait for the last command to finish
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Write to the comm page
  DspMask_OrChannelMask(&eg->DspCommPage->cmdStop, ChannelMask);
  DspMask_Clear(&eg->DspCommPage->cmdReset);   //@@ useless ?
  if (!DspMask_IsEmpty(&eg->DspCommPage->cmdStop)) {
    // Clear the handshake and send the vector command
    ClearHandshake(eg);
    SendVector(eg, DSP_VC_STOP_TRANSFER);

    // Wait for transport to stop
    if (!WaitForHandshake(eg))
      return(ECHOSTATUS_DSP_DEAD);

    // Keep track of which pipes are transporting
    eg->ActiveMask &= ~ChannelMask;
    DspMask_Clear(&eg->DspCommPage->cmdStop);
    DspMask_Clear(&eg->DspCommPage->cmdReset);

    return(ECHOSTATUS_OK);
  }     // if this monkey is being started

  ECHO_DEBUGPRINTF(("StopTransport: No pipes to stop!\n"));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// ResetTransport resets transport for a set of pipes
//
//===========================================================================

int ResetTransport(struct CEchoGals *eg, u32 ChannelMask) {
  ECHO_DEBUGPRINTF(("ResetTransport %x\n", ChannelMask));

  // Wait for the last command to finish
  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Write to the comm page
  DspMask_OrChannelMask(&eg->DspCommPage->cmdStop, ChannelMask);
  DspMask_OrChannelMask(&eg->DspCommPage->cmdReset, ChannelMask);
  if (!DspMask_IsEmpty(&eg->DspCommPage->cmdReset)) {
    // Clear the handshake and send the vector command
    ClearHandshake(eg);
    SendVector(eg, DSP_VC_STOP_TRANSFER);

    // Wait for transport to stop
    if (!WaitForHandshake(eg))
      return(ECHOSTATUS_DSP_DEAD);

    // Keep track of which pipes are transporting
    eg->ActiveMask &= ~ChannelMask;
    DspMask_Clear(&eg->DspCommPage->cmdStop);
    DspMask_Clear(&eg->DspCommPage->cmdReset);

    return(ECHOSTATUS_OK);
  }     // if this monkey is being started

  ECHO_DEBUGPRINTF(("ResetTransport: No pipes to reset!\n"));
  return(ECHOSTATUS_OK);
}



/*
#ifdef DIGITAL_INPUT_AUTO_MUTE_SUPPORT

//===========================================================================
//
// Digital input auto-mute - Gina24, Layla24, and Mona only
//
//===========================================================================

ECHOSTATUS CDspCommObject::GetDigitalInputAutoMute(struct CEchoGals *eg, BOOL & fAutoMute)
{
  fAutoMute = DigitalInAutoMute;

  ECHO_DEBUGPRINTF(("CDspCommObject::GetDigitalInputAutoMute %d\n", fAutoMute));

  return ECHOSTATUS_OK;
}

ECHOSTATUS CDspCommObject::SetDigitalInputAutoMute(BOOL fAutoMute)
{
  ECHO_DEBUGPRINTF(("CDspCommObject::SetDigitalInputAutoMute %d\n", fAutoMute));

  //
  // Store the flag
  //
  DigitalInAutoMute = fAutoMute;

  //
  // Re-set the input clock to the current value - indirectly causes the 
  // auto-mute flag to be sent to the DSP
  //
  SetInputClock(m_wInputClock);

  return ECHOSTATUS_OK;
}

#endif // DIGITAL_INPUT_AUTO_MUTE_SUPPORT
*/




#ifdef MIDI_SUPPORT

/****************************************************************************

	MIDI

 ****************************************************************************/

//===========================================================================
//
// Send a buffer full of MIDI data to the DSP
//
//===========================================================================
/*
ECHOSTATUS CDspCommObject::WriteMidi(struct CEchoGals *eg, PBYTE pData,	// Ptr to data buffer
				     DWORD dwLength,	// How many bytes to write
				     PDWORD pdwActualCt	// Return how many actually written
  )
{
  ECHOSTATUS Status;
  DWORD dwLengthTemp, dwShift, dwPacked, i;

#ifdef ECHO_DEBUG
  DWORD dwOrgLgth = dwLength;
#endif

  //
  // Send all the MIDI data
  //
  *pdwActualCt = 0;
  Status = ECHOSTATUS_OK;

  while (dwLength > 0) {
    //
    // Wait for the last command to complete
    //
    if (FALSE == WaitForHandshake()) {
      ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi - Timed out waiting for handshake\n"));

      Status = ECHOSTATUS_DSP_DEAD;
      break;
    }
    //
    // See if the DSP is ready to go
    //
    if (0 == (GetDspRegister(CHI32_STATUS_REG) & CHI32_STATUS_REG_HF4)) {
      //
      // MIDI output is full
      //
      break;
    }
    //
    // Pack the MIDI data three bytes at a time
    //
    if (dwLength > 3)
      dwLengthTemp = 3;
    else
      dwLengthTemp = dwLength;

    dwShift = 16;
    dwPacked = 0;
    for (i = 0; i < dwLengthTemp; i++) {
      dwPacked |= (DWORD) * pData++ << dwShift;
      dwShift -= 8;
    }

    dwPacked |= dwLengthTemp << 24;
    dwLength -= dwLengthTemp;

    //
    // Write the data to the DSP
    //
    DspCommPage->dwMIDIOutData = SWAP(dwPacked);
    ClearHandshake();
    SendVector(DSP_VC_MIDI_WRITE);

    ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi - wrote 0x%08lx\n", dwPacked));

    //
    // Update the count
    //
    *pdwActualCt += dwLengthTemp;

  }				// while( dwLength )

  if (0 != *pdwActualCt) {
    //
    // Save the current time - used to detect if MIDI out is currently busy
    //
    m_pOsSupport->OsGetSystemTime(&MidiOutTime);
    // Last time MIDI out occured
  }

  ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi: Actual %ld " "Expected %ld OK\n", *pdwActualCt, dwOrgLgth));

  return Status;

}				// ECHOSTATUS CDspCommObject::WriteMidi


//===========================================================================
//
// Called from the interrupt handler - get a MIDI input byte
//
//===========================================================================

ECHOSTATUS CDspCommObject::ReadMidi(struct CEchoGals *eg, WORD wIndex,	// Buffer index
				    DWORD & dwData	// Return data
  )
{
  if (wIndex >= DSP_MIDI_BUFFER_SIZE - 1)
    return ECHOSTATUS_INVALID_INDEX;

  //
  // Get the data
  //      
  dwData = SWAP(DspCommPage->wMidiInData[wIndex]);

  //
  // Timestamp for the MIDI input activity indicator
  //
  m_pOsSupport->OsGetSystemTime(&MidiInTime);

  return ECHOSTATUS_OK;

}				// ECHOSTATUS CDspCommObject::ReadMidi


ECHOSTATUS CDspCommObject::SetMidiOn(struct CEchoGals *eg, BOOL bOn)
{
  if (bOn) {
    if (0 == MidiOnCount) {
      if (!WaitForHandshake())
	return ECHOSTATUS_DSP_DEAD;

      DspCommPage->dwFlags |= SWAP((DWORD) DSP_FLAG_MIDI_INPUT);

      ClearHandshake();
      SendVector(DSP_VC_UPDATE_FLAGS);
    }
    MidiOnCount++;
  } else {
    if (MidiOnCount == 0)
      return ECHOSTATUS_OK;

    if (0 == --MidiOnCount) {
      if (!WaitForHandshake())
	return ECHOSTATUS_DSP_DEAD;

      DspCommPage->dwFlags &= SWAP((DWORD) ~ DSP_FLAG_MIDI_INPUT);

      ClearHandshake();
      SendVector(DSP_VC_UPDATE_FLAGS);
    }
  }

  return ECHOSTATUS_OK;

}				// ECHOSTATUS CDspCommObject::SetMidiOn
*/

#endif // MIDI_SUPPORT

