// ****************************************************************************
//
//      CGina24DspCommObject.cpp
//
//              Implementation file for Gina24 DSP interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/Gina20DSP.c"


/****************************************************************************

	Construction and destruction

 ****************************************************************************/

//===========================================================================
//
// Constructor
//
//===========================================================================


int NewDspComm(struct CEchoGals *eg) {
  int st;

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // This part is common to all the cards
  if ((st = CreateCommPage(eg)))
    return(st);

  // Write the features of the card
  eg->NumPipesOut = 10;
  eg->NumPipesIn = 4;
  eg->NumBussesOut = 10;
  eg->NumBussesIn = 4;
  eg->FirstDigitalBusOut = 8;
  eg->FirstDigitalBusIn = 2;
  eg->HasVmixer = FALSE;
  eg->NumMidiOut = 0;                        // # MIDI out channels
  eg->NumMidiIn = 0;                         // # MIDI in  channels
  eg->DspCommPage->dwSampleRate = SWAP32((u32)44100);
  eg->DspCodeToLoad=pwGina20DSP;
  eg->SpdifStatus = GD_SPDIF_STATUS_UNDEF;
  eg->ClockState = GD_CLOCK_UNDEF;
  eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF;
  // Since this card has no ASIC, mark it as loaded so everything works OK
  eg->ASICLoaded = TRUE;
  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



// The Gina20 has no ASIC. Just do nothing
int LoadASIC(struct CEchoGals *eg) {

  return(TRUE);
}



//===========================================================================
//
// SetSampleRate
//
// Set the audio sample rate for Gina20/Darla20
//
// For Gina and Darla, three parameters need to be set: the resampler state,
// the clock state, and the S/PDIF output status state.  The idea is that
// these parameters are changed as infrequently as possible.
//
// Resampling is no longer supported.
//
//===========================================================================

u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {
  u8 ClockState, SpdifStatus;

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  // Pick the clock
  if (eg->CardType==GINA && eg->InputClock==ECHO_CLOCK_SPDIF) {
    ClockState = GD_CLOCK_SPDIFIN;
  } else {
    switch (SampleRate) {
      case 44100:
        ClockState = GD_CLOCK_44;
        break;
      case 48000:
        ClockState = GD_CLOCK_48;
        break;
      default:
        ClockState = GD_CLOCK_NOCHANGE;
        break;
    }
  }
  if (eg->ClockState==ClockState)
    ClockState = GD_CLOCK_NOCHANGE;
//@@ottim
  // Select S/PDIF output status
   switch (SampleRate) {
     case 44100:
       SpdifStatus = GD_SPDIF_STATUS_44;
       break;
     case 48000:
       SpdifStatus = GD_SPDIF_STATUS_48;
       break;
     default:
       SpdifStatus = GD_SPDIF_STATUS_NOCHANGE;
       break;
   }
   if (SpdifStatus==eg->SpdifStatus)
     SpdifStatus = GD_SPDIF_STATUS_NOCHANGE;

  // Write the audio state to the comm page
  eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);
  eg->DspCommPage->byGDClockState = ClockState;
  eg->DspCommPage->byGDSpdifStatus = SpdifStatus;
  eg->DspCommPage->byGDResamplerState = 3;         // magic number - should always be 3

  // Send command to DSP
  ClearHandshake(eg);
  SendVector(eg, DSP_VC_SET_GD_AUDIO_STATE);

  // Save the new audio state
  if (ClockState != GD_CLOCK_NOCHANGE)
    eg->ClockState = ClockState;

  if (SpdifStatus != GD_SPDIF_STATUS_NOCHANGE)
    eg->SpdifStatus = SpdifStatus;

  return(SampleRate);
}



//===========================================================================
//
// Set the input clock to internal, S/PDIF, ADAT
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {

  ECHO_DEBUGPRINTF(("SetInputClock:\n"));

  switch (Clock) {
    case ECHO_CLOCK_INTERNAL:
      // Reset the audio state to unknown (just in case)
      eg->ClockState = GD_CLOCK_UNDEF;
      eg->SpdifStatus = GD_SPDIF_STATUS_UNDEF;
      SetSampleRate(eg, GetSampleRate(eg));
      eg->InputClock = Clock;
      ECHO_DEBUGPRINTF(("Set Gina clock to INTERNAL\n"));
      break;
    case ECHO_CLOCK_SPDIF:
      eg->DspCommPage->byGDClockState = GD_CLOCK_SPDIFIN;
      eg->DspCommPage->byGDSpdifStatus = GD_SPDIF_STATUS_NOCHANGE;
      ClearHandshake(eg);
      SendVector(eg, DSP_VC_SET_GD_AUDIO_STATE);
      eg->ClockState = GD_CLOCK_SPDIFIN;
      ECHO_DEBUGPRINTF(("Set Gina20 clock to SPDIF\n"));
      eg->InputClock = Clock;
      break;
    default :
      return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
  }

  return(ECHOSTATUS_OK);
}



// Layla20 only
int SetOutputClock(struct CEchoGals *eg, u16 Clock) {

  return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
}



// This card has no vmixer.
int SetVmixerGain(struct CEchoGals *eg, u16 Output, u16 Pipe, int Gain) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// Gina24, Mona, Layla24 - only
int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {

  return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
}



// This card has no vmixer.
int UpdateVmixerLevel(struct CEchoGals *eg) {
  
  return(ECHOSTATUS_NOT_SUPPORTED);
}



int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {

  ECHO_DEBUGPRINTF(("SetProfessionalSpdif %d\n", Prof));
  if (Prof)
    DspSetFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  else
    DspClearFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  return(ECHOSTATUS_OK);
}



