// ****************************************************************************
//
//              CGina24.cpp
//
//              Implementation file for the CGina24 driver class.
//              Set editor tabs to 3 for your viewing pleasure.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"



//===========================================================================
//
// Constructor and destructor
//
//===========================================================================

struct CEchoGals *NewEchoGals(u16 Device, u16 SubDevice) {
  struct CEchoGals *eg;
  int i;

  ECHO_DEBUGPRINTF(("NewEchoGals - Gina24\n"));
  if ((SubDevice&0xfff0)!=0x0050) {
    ECHO_DEBUGPRINTF(("NewEchoGals() - wrong card type\n"));
    return(0);
  }
  eg=OsAllocateNonPaged(sizeof(struct CEchoGals));
  if (eg) {
    memset(eg, 0, sizeof(struct CEchoGals));
  } else {
    ECHO_DEBUGPRINTF(("NewEchoGals() - memory allocation failed\n"));
    return(0);
  }

  eg->DeviceId=Device;
  eg->SubDeviceId=SubDevice;
  eg->CardType=0x0050;
  eg->BadBoard=1;
  eg->Flags = ECHOGALS_FLAG_BADBOARD | ECHOGALS_FLAG_SPDIF_NODITHER;
  for (i=0; i<ECHO_MAXAUDIOPIPES; i++)
    eg->Pipes[i].BytesPerSample=1;
  return(eg);
}



int InitHw(struct CEchoGals *eg) {
  int Status, i;

  ECHO_DEBUGPRINTF(("InitHw()\n"));
  // Create the DSP comm object
  Status = NewDspComm(eg);
  if (Status) {
    ECHO_DEBUGPRINTF(("InitHw - could not create DSP comm object\n"));
    return(ECHOSTATUS_NO_MEM);
  }

  // Load the DSP and the ASIC on the PCI card
  Status = LoadFirmware(eg);
  if (Status || eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  // Clear the "bad board" flag; set the flags to indicate that Gina24
  // can handle super-interleave and supports the digital input auto-mute.
  eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;
  eg->Flags |= ECHOGALS_ROFLAG_SUPER_INTERLEAVE_OK | ECHOGALS_ROFLAG_DIGITAL_IN_AUTOMUTE;

  // Must call this here after DSP is init to init gains and mutes
  Status = InitLineLevels(eg);
#ifndef EMUL
  if (Status)
    return(Status);
#endif

  // Set professional nominal levels
  for (i=0; i<eg->FirstDigitalBusOut; i++)
    Status=SetNominalLevel(eg, i, FALSE);          // FALSE is +4dBu here

  for (i=0; i<eg->FirstDigitalBusIn; i++)
    Status=SetNominalLevel(eg, eg->NumBussesOut+i, FALSE);

  // Set the digital mode to S/PDIF RCA
  SetDigitalMode(eg, DIGITAL_MODE_SPDIF_RCA);

  // Set the S/PDIF output format to "professional"
  SetProfessionalSpdif(eg, TRUE);

  // Get default sample rate from DSP
  eg->SampleRate = GetSampleRate(eg);

  ECHO_DEBUGPRINTF(("InitHw done\n"));
  return(Status);
}




/****************************************************************************

	Informational methods

 ****************************************************************************/

//===========================================================================
//
// Override GetCapabilities to enumerate unique capabilties for Gina24
//
//===========================================================================
#ifndef EMUL
int GetCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities) {
  int Status, i;

  Status = GetBaseCapabilities(eg, pCapabilities);

  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Add nominal level control to analog ins & outs
  for (i=0; i<eg->FirstDigitalBusOut; i++)
    pCapabilities->dwBusOutCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  for (i=0; i<eg->FirstDigitalBusIn; i++)
    pCapabilities->dwBusInCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  pCapabilities->dwInClockTypes |= ECHO_CLOCK_BIT_SPDIF | ECHO_CLOCK_BIT_ESYNC | ECHO_CLOCK_BIT_ESYNC96 | ECHO_CLOCK_BIT_ADAT;

  return(Status);
}


#else
int GetCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities) {
  int Status, i;

  Status = GetBaseCapabilities(eg, pCapabilities);

  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Add meters & pans to output pipes
  for (i=0; i<eg->NumPipesOut; i++)
    pCapabilities->dwPipeOutCaps[i] |= ECHOCAPS_PEAK_METER | ECHOCAPS_PAN;

  // Add nominal level control to analog ins & outs
  for (i=0; i<eg->FirstDigitalBusOut; i++)
    pCapabilities->dwBusOutCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  for (i=0; i<eg->FirstDigitalBusIn; i++)
    pCapabilities->dwBusInCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  pCapabilities->dwInClockTypes |= ECHO_CLOCK_BIT_SPDIF;

  return(Status);
}
#endif

//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a given sample rate.
//
//===========================================================================

int QueryAudioSampleRate(struct CEchoGals *eg, u32 SampleRate) {

  if (SampleRate !=  8000 &&
      SampleRate != 11025 &&
      SampleRate != 16000 &&
      SampleRate != 22050 &&
      SampleRate != 32000 &&
      SampleRate != 44100 &&
      SampleRate != 48000 &&
      SampleRate != 88200 &&
      SampleRate != 96000) {
    ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be 8,000 Hz, 11,025 Hz, 16,000 Hz, 22,050 Hz, 32,000 Hz, 44,100 Hz, 48,000 Hz, 88,200 Hz or 96,000 Hz\n"));
    return(ECHOSTATUS_BAD_FORMAT);
  }
  if (SampleRate>=88200 && !eg->BadBoard && eg->DigitalMode==DIGITAL_MODE_ADAT) {
    ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate cannot be set to 88,200 Hz or 96,000 Hz in ADAT mode\n"));
    return(ECHOSTATUS_BAD_FORMAT);
  }

  ECHO_DEBUGPRINTF(("QueryAudioSampleRate()\n"));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Gina24 supports S/PDIF, Esync, and ADAT input clocks.
//
//===========================================================================

u32 GetInputClockDetect(const struct CEchoGals *eg) {
  u32 ClocksFromDsp, ClockDetectBits;

  ECHO_DEBUGPRINTF(("GetInputClockDetect\n"));

  // Map the DSP clock detect bits to the generic driver clock detect bits
  ClocksFromDsp = DspGetInputClockDetect(eg);

  ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_SPDIF)
    ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_ADAT)
    ClockDetectBits |= ECHO_CLOCK_BIT_ADAT;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_ESYNC)
    ClockDetectBits |= ECHO_CLOCK_BIT_ESYNC | ECHO_CLOCK_BIT_ESYNC96;

  return(ClockDetectBits);
}

