// ****************************************************************************
//
//      CGina24DspCommObject.cpp
//
//              Implementation file for Gina24 DSP interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/Gina24DSP.c"
#include "DSP/Gina24_361DSP.c"
#include "ASIC/Gina24ASIC.c"
#include "ASIC/Gina24ASIC_361.c"


/****************************************************************************

	Construction and destruction

 ****************************************************************************/

//===========================================================================
//
// Constructor
//
//===========================================================================

int NewDspComm(struct CEchoGals *eg) {
  int st;

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // This part is common to all the cards
  if ((st = CreateCommPage(eg)))
    return(st);

  // Write the features of the card
  eg->NumPipesOut = 16;
  eg->NumPipesIn = 10;
  eg->NumBussesOut = 16;
  eg->NumBussesIn = 10;
  eg->FirstDigitalBusOut = 8;
  eg->FirstDigitalBusIn = 2;
  eg->HasVmixer = FALSE;
  eg->NumMidiOut = 0;                        // # MIDI out channels
  eg->NumMidiIn = 0;                         // # MIDI in  channels
  eg->DspCommPage->dwSampleRate = SWAP32((u32)44100);        // Need this in case we start with ESYNC
  eg->HasASIC = TRUE;

  // Gina24 comes in both '301 and '361 flavors
  if (eg->DeviceId==DEVICE_ID_56361) {
    eg->DspCodeToLoad=pwGina24_361DSP;
    eg->DigitalModes=ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA | ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL | ECHOCAPS_HAS_DIGITAL_MODE_ADAT;
  } else {
    eg->DspCodeToLoad=pwGina24DSP;
    eg->DigitalModes=ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA | ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL | ECHOCAPS_HAS_DIGITAL_MODE_ADAT | ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_CDROM;
  }

  eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF | ECHO_CLOCK_BIT_ESYNC | ECHO_CLOCK_BIT_ESYNC96 | ECHO_CLOCK_BIT_ADAT;

#ifdef EMUL
  eg->NumPipesOut = 8;
  eg->NumPipesIn = 4;
  eg->NumBussesOut = 4;
  eg->NumBussesIn = 4;
  eg->FirstDigitalBusOut = 2;
  eg->FirstDigitalBusIn = 2;
  eg->HasVmixer = TRUE;
  eg->DigitalModes=ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA;
  eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF;
#endif

  eg->DigitalMode=DIGITAL_MODE_SPDIF_RCA;
  eg->ProfessionalSpdif=FALSE;
  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



/****************************************************************************

  Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// Gina24 has an ASIC on the PCI card which must be loaded for anything
// interesting to happen.
//
//===========================================================================

int LoadASIC(struct CEchoGals *eg) {
  u32 ControlReg, dwSize;
  u8 *pbAsic;

  if (eg->ASICLoaded)
    return(TRUE);

  // Give the DSP a few milliseconds to settle down
  OsSnooze(10000);

  // Pick the correct ASIC for '301 or '361 Gina24
  if (DEVICE_ID_56361 == eg->DeviceId) {
    pbAsic = pbGina24ASIC_361;
    dwSize = sizeof(pbGina24ASIC_361);
  } else {
    pbAsic = pbGina24ASIC;
    dwSize = sizeof(pbGina24ASIC);
  }
  if (!LoadASICgeneric(eg, DSP_FNC_LOAD_GINA24_ASIC, pbAsic, dwSize))
    return(FALSE);

  eg->Asic = pbAsic;

  // Now give the new ASIC a little time to set up
  OsSnooze(10000);
  // See if it worked
  CheckAsicStatus(eg);

  // Set up the control register if the load succeeded - 48 kHz, internal clock, S/PDIF RCA mode
  if (eg->ASICLoaded) {
    ControlReg = GML_CONVERTER_ENABLE | GML_48KHZ;
    WriteControlReg(eg, ControlReg);
  }
  ECHO_DEBUGPRINTF(("LoadASIC() done\n"));
  return(eg->ASICLoaded);
}



//===========================================================================
//
// WriteControlReg
//
// Most configuration of Gina24, Layla24, or Mona is
// accomplished by writing the control register.  WriteControlReg
// sends the new control register value to the DSP.
//
//===========================================================================

int WriteControlReg(struct CEchoGals *eg, u32 ControlReg) {

  ECHO_DEBUGPRINTF(("WriteControlReg...\n"));
  if (!eg->ASICLoaded)
    return(ECHOSTATUS_ASIC_NOT_LOADED);

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

#ifdef DIGITAL_INPUT_AUTO_MUTE_SUPPORT
  // Handle the digital input auto-mute
  if (eg->DigitalInAutoMute)
    ControlReg |= GML_DIGITAL_IN_AUTO_MUTE;
  else
    ControlReg &= ~GML_DIGITAL_IN_AUTO_MUTE;
#endif

  // Write the control register
  if (ControlReg != GetControlRegister(eg)) {
    SetControlRegister(eg, ControlReg);
    ECHO_DEBUGPRINTF(("WriteControlReg: 0x%x\n", ControlReg));
    ClearHandshake(eg);
    return(SendVector(eg, DSP_VC_WRITE_CONTROL_REG));
  }
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// SetSampleRate
// 
// Set the audio sample rate for Gina24 - fixme make this common for Gina24 & Mona
//
//===========================================================================

u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {
  u32 ControlReg, NewClock;

  // Only set the clock for internal mode.  If the clock is not set to
  // internal, try and re-set the input clock; this more transparently
  // handles switching between single and double-speed mode
  if (eg->InputClock != ECHO_CLOCK_INTERNAL) {
    ECHO_DEBUGPRINTF(("SetSampleRate: Cannot set sample rate - clock not set to CLK_CLOCKININTERNAL\n"));

    // Save the rate anyhow
    eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);

    // Set the input clock to the current value
    SetInputClock(eg, eg->InputClock);

    return(GetSampleRate(eg));
  }

  // Set the sample rate
  NewClock = 0;

  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_CLOCK_CLEAR_MASK;
  ControlReg &= GML_SPDIF_RATE_CLEAR_MASK;

  switch (SampleRate) {
    case 96000:
      NewClock = GML_96KHZ;
      break;
    case 88200:
      NewClock = GML_88KHZ;
      break;
    case 48000:
      NewClock = GML_48KHZ | GML_SPDIF_SAMPLE_RATE1;
      break;
    case 44100:
      NewClock = GML_44KHZ;
      // Professional mode ?
      if (ControlReg & GML_SPDIF_PRO_MODE)
        NewClock |= GML_SPDIF_SAMPLE_RATE0;
      break;
    case 32000:
      NewClock = GML_32KHZ | GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
      break;
    case 22050:
      NewClock = GML_22KHZ;
      break;
    case 16000:
      NewClock = GML_16KHZ;
      break;
    case 11025:
      NewClock = GML_11KHZ;
      break;
    case 8000:
      NewClock = GML_8KHZ;
      break;
    default:
      ECHO_DEBUGPRINTF(("SetSampleRate: %d invalid!\n", SampleRate));
      return(GetSampleRate(eg));
  }

  ControlReg |= NewClock;

  // Send the new value to the DSP
  if (ECHOSTATUS_OK == WriteControlReg(eg, ControlReg)) {
    eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);
    ECHO_DEBUGPRINTF(("SetSampleRate: %d clock %d\n", SampleRate, NewClock));
  }
  return(GetSampleRate(eg));
}



//===========================================================================
//
// Set the input clock to internal, S/PDIF, ADAT
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {
  char WriteReg;
  u32 ControlReg, SampleRate;

  ECHO_DEBUGPRINTF(("SetInputClock:\n"));

  ControlReg = GetControlRegister(eg);

  // Mask off the clock select bits
  ControlReg &= GML_CLOCK_CLEAR_MASK;

  SampleRate=0;
  WriteReg = TRUE;
  switch (Clock) {
    case ECHO_CLOCK_INTERNAL:
      ECHO_DEBUGPRINTF(("Set Gina24 clock to INTERNAL\n"));
      SampleRate=GetSampleRate(eg);
      // If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
      if (SampleRate<8000 || SampleRate>96000)
        SampleRate = 48000;
      WriteReg = FALSE;
      break;
    case ECHO_CLOCK_SPDIF:
      if (eg->DigitalMode==DIGITAL_MODE_ADAT)
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      ECHO_DEBUGPRINTF(("Set Gina24 clock to SPDIF\n"));
      ControlReg |= GML_SPDIF_CLOCK;
      if (GML_CLOCK_DETECT_BIT_SPDIF96 & GetInputClockDetect(eg))
        ControlReg |= GML_DOUBLE_SPEED_MODE;
      else
        ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    case ECHO_CLOCK_ADAT:
      ECHO_DEBUGPRINTF(("Set Gina24 clock to ADAT\n"));
      if (eg->DigitalMode!=DIGITAL_MODE_ADAT)
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      ControlReg |= GML_ADAT_CLOCK;
      ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    case ECHO_CLOCK_ESYNC:
      ECHO_DEBUGPRINTF(("Set Gina24 clock to ESYNC\n"));
      ControlReg |= GML_ESYNC_CLOCK;
      ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    case ECHO_CLOCK_ESYNC96:
      ECHO_DEBUGPRINTF(("Set Gina24 clock to ESYNC96\n"));
      ControlReg |= GML_ESYNC_CLOCK | GML_DOUBLE_SPEED_MODE;
      break;
    default:
      ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Gina24\n", Clock));
      return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
  }

  // Save the new input clock.
  eg->InputClock = Clock;

  // Write the control reg if that's called for
  if (WriteReg)
    WriteControlReg(eg, ControlReg);

  // Set Gina24 sample rate to something sane if word or superword is being turned off
  if (SampleRate)
    SetSampleRate(eg, SampleRate);

  return(ECHOSTATUS_OK);
}



// Layla20 only
int SetOutputClock(struct CEchoGals *eg, u16 Clock) {

  return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
}



// This card has no vmixer.
int SetVmixerGain(struct CEchoGals *eg, u16 Output, u16 Pipe, int Gain) {
ECHO_DEBUGPRINTF(("SetVmixerGain: pipe %d, out %d = %d\n", (int)Pipe, (int)Output, Gain));
  return(ECHOSTATUS_NOT_SUPPORTED);
}



// This card has no vmixer.
int UpdateVmixerLevel(struct CEchoGals *eg) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



//===========================================================================
//
// SetDigitalMode
//
//===========================================================================

int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {
  u32 ControlReg;

  // Clear the current digital mode
  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_DIGITAL_MODE_CLEAR_MASK;

  // Tweak the control reg
  switch (Mode) {
    case DIGITAL_MODE_SPDIF_OPTICAL:
      ControlReg |= GML_SPDIF_OPTICAL_MODE;
      goto ChkClk;
    case DIGITAL_MODE_SPDIF_CDROM:
      // '361 Gina24 cards do not have the S/PDIF CD-ROM mode
      if (eg->DeviceId==DEVICE_ID_56361)
        return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
      ControlReg |= GML_SPDIF_CDROM_MODE;
      // fall through
    case DIGITAL_MODE_SPDIF_RCA:
ChkClk:
      // If the input clock is set to ADAT, set the input clock to internal and the sample rate to 48 KHz
      if (eg->InputClock==ECHO_CLOCK_ADAT) {
        eg->DspCommPage->dwSampleRate = SWAP32((u32)48000);
        SetInputClock(eg, ECHO_CLOCK_INTERNAL);
      }
      break;
    case DIGITAL_MODE_ADAT:
      // If the input clock is set to S/PDIF, set the input clock to internal and the sample rate to 48 KHz
      if (eg->InputClock==ECHO_CLOCK_SPDIF) {
        eg->DspCommPage->dwSampleRate = SWAP32((u32)48000);
        SetInputClock(eg, ECHO_CLOCK_INTERNAL);
      }
      ControlReg |= GML_ADAT_MODE;
      ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    default:
      ECHO_DEBUGPRINTF(("Digital mode not supported: %d\n", Mode));
      return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
  }

  // Write the control reg
  WriteControlReg(eg, ControlReg);
  eg->DigitalMode = Mode;
  ECHO_DEBUGPRINTF(("SetDigitalMode to %d\n", eg->DigitalMode));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// Set the S/PDIF output format (common for Gina24, Layla24, and Mona)
//
//===========================================================================

int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {
  u32 ControlReg;

  // Clear the current S/PDIF flags
  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_SPDIF_FORMAT_CLEAR_MASK;

  // Set the new S/PDIF flags depending on the mode
  ControlReg |= GML_SPDIF_TWO_CHANNEL | GML_SPDIF_24_BIT | GML_SPDIF_COPY_PERMIT;
  if (Prof) {
    // Professional mode
    ControlReg |= GML_SPDIF_PRO_MODE;

    switch (GetSampleRate(eg)) {
      case 32000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
        break;
      case 44100:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0;
        break;
      case 48000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE1;
        break;
    }
  } else {
    // Consumer mode
    switch (GetSampleRate(eg)) {
      case 32000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
        break;
      case 48000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE1;
        break;
    }
  }

  WriteControlReg(eg, ControlReg);
  eg->ProfessionalSpdif=Prof;
  ECHO_DEBUGPRINTF(("SetProfessionalSpdif to %s\n", (Prof) ? "Professional" : "Consumer"));
  return(ECHOSTATUS_OK);
}
