// ****************************************************************************
//
//      CGina24DspCommObject.cpp
//
//              Implementation file for Gina24 DSP interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/Layla20DSP.c"
#include "ASIC/LaylaASIC.c"



/****************************************************************************

	Construction and destruction

 ****************************************************************************/

//===========================================================================
//
// Constructor
//
//===========================================================================

int NewDspComm(struct CEchoGals *eg) {
  int st;

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // This part is common to all the cards
  if ((st = CreateCommPage(eg)))
    return(st);

  // Write the features of the card
  eg->NumPipesOut = 12;
  eg->NumPipesIn = 10;
  eg->NumBussesOut = 12;
  eg->NumBussesIn = 10;
  eg->FirstDigitalBusOut = 10;
  eg->FirstDigitalBusIn = 8;
  eg->HasVmixer = FALSE;
  eg->NumMidiOut = 1;                        // # MIDI out channels
  eg->NumMidiIn = 1;                         // # MIDI in  channels
  eg->DspCommPage->dwSampleRate = SWAP32((u32)44100);        // Need this in case we start with ESYNC
  eg->HasASIC = TRUE;
  eg->DspCodeToLoad=pwLayla20DSP;
  eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF | ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_SUPER;
  eg->OutClockTypes = ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_SUPER;

  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



/****************************************************************************

  Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// Layla20 has an ASIC in the external box
//
//===========================================================================

int LoadASIC(struct CEchoGals *eg) {
  if (eg->ASICLoaded)
    return(TRUE);

  if (!LoadASICgeneric(eg, DSP_FNC_LOAD_LAYLA_ASIC, pbLaylaASIC, LAYLA_ASIC_SIZE))
    return(FALSE);

  // Check if ASIC is alive and well.
  return(CheckAsicStatus(eg));
}



//===========================================================================
//
// SetSampleRate
//
// Set the audio sample rate for Layla20
//
// Layla is simple; just send it the sampling rate (assuming that the clock mode is correct).
//
//===========================================================================

u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {

  // Only set the clock for internal mode. Do not return failure, simply treat it as a non-event.
  if (eg->InputClock != ECHO_CLOCK_INTERNAL) {
    ECHO_DEBUGPRINTF(("SetSampleRate: Cannot set sample rate - clock not set to CLK_CLOCKININTERNAL\n"));
    eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);
    return(GetSampleRate(eg));
  }

  if (SampleRate<8000 || SampleRate>50000) {
    ECHO_DEBUGPRINTF(("SetSampleRate: %d invalid!\n", SampleRate));
    return(GetSampleRate(eg));
  }

  if (!WaitForHandshake(eg))
    return(0xffffffff);

  eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);

  ClearHandshake(eg);
  SendVector(eg, DSP_VC_SET_LAYLA_SAMPLE_RATE);

  ECHO_DEBUGPRINTF(("SetSampleRate: Layla sample rate changed to %d\n", SampleRate));
  return(SampleRate);
}



//===========================================================================
//
// Set the input clock to internal, S/PDIF, ADAT
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {
  u16 NewClock;
  u32 SampleRate;

  ECHO_DEBUGPRINTF(("SetInputClock:\n"));
  SampleRate=0;
  switch (Clock) {
    case ECHO_CLOCK_INTERNAL:
      ECHO_DEBUGPRINTF(("Set Layla20 clock to INTERNAL\n"));
      SampleRate = GetSampleRate(eg);
      // If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
      if (SampleRate<8000 || SampleRate>50000) {
        SampleRate = 48000;
        eg->DspCommPage->dwSampleRate = SWAP32((u32)48000);
      }
      NewClock = LAYLA20_CLOCK_INTERNAL;
      break;
    case ECHO_CLOCK_SPDIF:
      ECHO_DEBUGPRINTF(("Set Layla20 clock to SPDIF\n"));
      NewClock = LAYLA20_CLOCK_SPDIF;
      break;
    case ECHO_CLOCK_WORD:
      ECHO_DEBUGPRINTF(("Set Layla20 clock to WORD\n"));
      NewClock = LAYLA20_CLOCK_WORD;
      break;
    case ECHO_CLOCK_SUPER:
      ECHO_DEBUGPRINTF(("Set Layla20 clock to SUPER\n"));
      NewClock = LAYLA20_CLOCK_SUPER;
      break;
    default:
      ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Layla24\n", Clock));
      return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
  }
  eg->InputClock = Clock;

  // Send the new clock to the DSP
  eg->DspCommPage->wInputClock = SWAP16(NewClock);
  ClearHandshake(eg);
  SendVector(eg, DSP_VC_UPDATE_CLOCKS);

  if (SampleRate)
    SetSampleRate(eg, SampleRate);

  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// Set new output clock
//
//===========================================================================

int SetOutputClock(struct CEchoGals *eg, u16 Clock) {

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  ECHO_DEBUGPRINTF(("SetOutputClock:\n"));
  eg->DspCommPage->wOutputClock = SWAP16(Clock);
  eg->OutputClock = Clock;

  ClearHandshake(eg);
  return(SendVector(eg, DSP_VC_UPDATE_CLOCKS));
}



// This card has no vmixer.
int SetVmixerGain(struct CEchoGals *eg, u16 Output, u16 Pipe, int Gain) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// This card has no vmixer.
int UpdateVmixerLevel(struct CEchoGals *eg) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// Gina24, Mona, Layla24 - only
int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {

  return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
}



int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {

  ECHO_DEBUGPRINTF(("SetProfessionalSpdif %d\n", Prof));
  if (Prof)
    DspSetFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  else
    DspClearFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  return(ECHOSTATUS_OK);
}



/*
BOOL CLaylaDspCommObject::IsMidiOutActive()
{
        ULONGLONG       ullCurTime;

        m_pOsSupport->OsGetSystemTime( &ullCurTime );
        return( ( ( ullCurTime - MidiOutTime ) > MIDI_ACTIVITY_TIMEOUT_USEC ) ? FALSE : TRUE );
}       // BOOL CLaylaDspCommObject::IsMidiOutActive()
*/
