// ****************************************************************************
//
//		CLayla24.cpp
//
//		Implementation file for the CLayla24 driver class.
//		Set editor tabs to 3 for your viewing pleasure.
//
//		Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//		All rights reserved
//		www.echoaudio.com
//		
//		Permission is hereby granted, free of charge, to any person obtaining a
//		copy of this software and associated documentation files (the
//		"Software"), to deal with the Software without restriction, including
//		without limitation the rights to use, copy, modify, merge, publish,
//		distribute, sublicense, and/or sell copies of the Software, and to
//		permit persons to whom the Software is furnished to do so, subject to
//		the following conditions:
//		
//		- Redistributions of source code must retain the above copyright
//		notice, this list of conditions and the following disclaimers.
//		
//		- Redistributions in binary form must reproduce the above copyright
//		notice, this list of conditions and the following disclaimers in the
//		documentation and/or other materials provided with the distribution.
//		
//		- Neither the name of Echo Digital Audio, nor the names of its
//		contributors may be used to endorse or promote products derived from
//		this Software without specific prior written permission.
//
//		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//		EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//		MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//		IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//		ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//		TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//		SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"


//===========================================================================
//
// Constructor and destructor
//
//===========================================================================

struct CEchoGals *NewEchoGals(u16 Device, u16 SubDevice) {
  struct CEchoGals *eg;
  int i;

  ECHO_DEBUGPRINTF(("NewEchoGals - Layla24\n"));
  if ((SubDevice&0xfff0)!=0x0060) {
    ECHO_DEBUGPRINTF(("NewEchoGals() - wrong card type\n"));
    return(0);
  }
  eg=OsAllocateNonPaged(sizeof(struct CEchoGals));
  if (eg) {
    memset(eg, 0, sizeof(struct CEchoGals));
  } else {
    ECHO_DEBUGPRINTF(("NewEchoGals() - memory allocation failed\n"));
    return(0);
  }

  eg->DeviceId=Device;
  eg->SubDeviceId=SubDevice;
  eg->CardType=0x0060;
  eg->BadBoard=1;
  eg->Flags = ECHOGALS_FLAG_BADBOARD | ECHOGALS_FLAG_SPDIF_NODITHER;
  for (i=0; i<ECHO_MAXAUDIOPIPES; i++)
    eg->Pipes[i].BytesPerSample=1;
  return(eg);
}



int InitHw(struct CEchoGals *eg) {
  int Status, i;

  ECHO_DEBUGPRINTF(("InitHw()\n"));
  // Create the DSP comm object
  Status = NewDspComm(eg);
  if (Status) {
    ECHO_DEBUGPRINTF(("InitHw - could not create DSP comm object\n"));
    return(ECHOSTATUS_NO_MEM);
  }

  // Load the DSP and the ASIC on the PCI card
  Status = LoadFirmware(eg);
  if (Status || eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  // Clear the "bad board" flag; set the flags to indicate that Gina24
  // can handle super-interleave and supports the digital input auto-mute.
  eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;
  eg->Flags |= ECHOGALS_ROFLAG_SUPER_INTERLEAVE_OK | ECHOGALS_ROFLAG_DIGITAL_IN_AUTOMUTE;

  // Must call this here after DSP is init to init gains and mutes
  Status = InitLineLevels(eg);
  if (ECHOSTATUS_OK != Status)
    return(Status);
/*
	//
	// Initialize the MIDI input
	//
	Status = m_MidiIn.Init( this );
	if ( ECHOSTATUS_OK != Status )
		return Status;
*/

  // Set professional nominal levels
  for (i=0; i<eg->FirstDigitalBusOut; i++)
    Status=SetNominalLevel(eg, i, FALSE);          // FALSE is +4dBu here

  for (i=0; i<eg->FirstDigitalBusIn; i++)
    Status=SetNominalLevel(eg, eg->NumBussesOut+i, FALSE);

  // Set the digital mode to S/PDIF RCA
  SetDigitalMode(eg, DIGITAL_MODE_SPDIF_RCA);

  // Set the S/PDIF output format to "professional"
  SetProfessionalSpdif(eg, TRUE);

  // Get default sample rate from DSP
  eg->SampleRate=GetSampleRate(eg);

  ECHO_DEBUGPRINTF(("InitHw done\n"));
  return(Status);
}




/****************************************************************************

	Informational methods

 ****************************************************************************/

//===========================================================================
//
// Override GetCapabilities to enumerate unique capabilties for this card
//
//===========================================================================

int GetCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities) {
  int Status, i;

  Status = GetBaseCapabilities(eg, pCapabilities);

  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Add nominal level control to analog ins & outs
  for (i=0; i<eg->FirstDigitalBusOut; i++)
    pCapabilities->dwBusOutCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  for (i=0; i<eg->FirstDigitalBusIn; i++)
    pCapabilities->dwBusInCaps[i] |= ECHOCAPS_NOMINAL_LEVEL;

  pCapabilities->dwInClockTypes |= ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_SPDIF | ECHO_CLOCK_BIT_ADAT;

  return(Status);
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a
// given sample rate.
//
//===========================================================================

int QueryAudioSampleRate(struct CEchoGals *eg, u32 SampleRate) {

  // Standard sample rates are allowed (like Mona & Layla24)
  switch (SampleRate) {
    case 8000:
    case 11025:
    case 16000:
    case 22050:
    case 32000:
    case 44100:
    case 48000:
      goto qasr_ex;
    case 88200:
    case 96000:
      // Double speed sample rates not allowed in ADAT mode
      if (eg->DigitalMode==DIGITAL_MODE_ADAT) {
        ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be < 50,000 Hz in ADAT mode\n"));
        return(ECHOSTATUS_BAD_FORMAT);
      }
      goto qasr_ex;
  }

  // Any rate from 25000 to 50000 is allowed
  if (SampleRate>=25000 && SampleRate<=50000)
    goto qasr_ex;

  // If not in ADAT mode, any rate from 50,000 to 100,000 is allowed
  if (eg->DigitalMode!=DIGITAL_MODE_ADAT && SampleRate>=50000 && SampleRate<=100000)
    goto qasr_ex;

  ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be >= 50,000 Hz and <= 100,000 Hz and NOT in ADAT mode\n"));
  return(ECHOSTATUS_BAD_FORMAT);

qasr_ex:
  ECHO_DEBUGPRINTF(("QueryAudioSampleRate() %d Hz OK\n", SampleRate));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Layla24 supports S/PDIF, word, and ADAT input clocks.
//
//===========================================================================

u32 GetInputClockDetect(const struct CEchoGals *eg) {
  u32 ClocksFromDsp, ClockDetectBits;

  ECHO_DEBUGPRINTF(("GetInputClockDetect\n"));

  // Map the DSP clock detect bits to the generic driver clock detect bits
  ClocksFromDsp = DspGetInputClockDetect(eg);

  ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_SPDIF)
    ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_ADAT)
    ClockDetectBits |= ECHO_CLOCK_BIT_ADAT;

  if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_WORD)
    ClockDetectBits |= ECHO_CLOCK_BIT_WORD;

  return(ClockDetectBits);
}


