// ****************************************************************************
//
//      CGina24DspCommObject.cpp
//
//              Implementation file for Gina24 DSP interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/MonaDSP.c"
#include "DSP/Mona361DSP.c"

#include "ASIC/Mona1ASIC48.c"
#include "ASIC/Mona1ASIC96.c"
#include "ASIC/Mona1ASIC48_361.c"
#include "ASIC/Mona1ASIC96_361.c"
#include "ASIC/Mona2ASIC.c"


/****************************************************************************

	Construction and destruction

 ****************************************************************************/

//===========================================================================
//
// Constructor
//
//===========================================================================

int NewDspComm(struct CEchoGals *eg) {
  int st;

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // This part is common to all the cards
  if ((st = CreateCommPage(eg)))
    return(st);

  // Write the features of the card
  eg->NumPipesOut = 14;
  eg->NumPipesIn = 12;
  eg->NumBussesOut = 14;
  eg->NumBussesIn = 12;
  eg->FirstDigitalBusOut = 6;
  eg->FirstDigitalBusIn = 4;
  eg->HasVmixer = FALSE;
  eg->NumMidiOut = 0;                        // # MIDI out channels
  eg->NumMidiIn = 0;                         // # MIDI in  channels
  eg->DspCommPage->dwSampleRate = SWAP32((u32)44100);        // Need this in case we start with ESYNC
  eg->HasASIC = TRUE;
  eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF | ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_ADAT;

  // Mona comes in both '301 and '361 flavors
  if (DEVICE_ID_56361==eg->DeviceId)
    eg->DspCodeToLoad=pwMona361DSP;
  else
    eg->DspCodeToLoad=pwMonaDSP;

  eg->DigitalMode=DIGITAL_MODE_SPDIF_RCA;
  eg->ProfessionalSpdif=FALSE;
  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



/****************************************************************************

  Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// Mona has an ASIC on the PCI card and another ASIC in the external box; 
// both need to be loaded.
//
//===========================================================================

int LoadASIC(struct CEchoGals *eg) {
  u32 ControlReg, Size;
  u8 *Asic1;

  if (eg->ASICLoaded)
    return(TRUE);

  OsSnooze(10000);

  if (eg->DeviceId==DEVICE_ID_56361) {
    Asic1 = pbMona1ASIC48_361;
    Size = sizeof(pbMona1ASIC48_361);
  } else {
    Asic1 = pbMona1ASIC48;
    Size = sizeof(pbMona1ASIC48);
  }

  if (!LoadASICgeneric(eg, DSP_FNC_LOAD_MONA_PCI_CARD_ASIC, Asic1, Size))
    return(FALSE);

  eg->Asic = Asic1;
  OsSnooze(10000);

  // Do the external one
  if (!LoadASICgeneric(eg, DSP_FNC_LOAD_MONA_EXTERNAL_ASIC, pbMona2ASIC, sizeof(pbMona2ASIC)))
    return(FALSE);

  OsSnooze(10000);
  CheckAsicStatus(eg);

  // Set up the control register if the load succeeded - 48 kHz, internal clock, S/PDIF RCA mode
  if (eg->ASICLoaded) {
    ControlReg = GML_CONVERTER_ENABLE | GML_48KHZ;
    WriteControlReg(eg, ControlReg);
  }

  return(eg->ASICLoaded);
}



//===========================================================================
//
// Depending on what digital mode you want, Mona needs different ASICs
// loaded.  This function checks the ASIC needed for the new mode and sees
// if it matches the one already loaded.
//
//===========================================================================

int SwitchAsic(struct CEchoGals *eg, u32 Mask96) {
  u8 *AsicNeeded;
  u32 AsicSize;

  // Check the clock detect bits to see if this is
  // a single-speed clock or a double-speed clock; load
  // a new ASIC if necessary.

  if (eg->DeviceId==DEVICE_ID_56361) {
    AsicNeeded = pbMona1ASIC48_361;
    AsicSize = sizeof(pbMona1ASIC48_361);
    if (GetInputClockDetect(eg) & Mask96) {
      AsicNeeded = pbMona1ASIC96_361;
      AsicSize = sizeof(pbMona1ASIC96_361);
    }
  } else {
    AsicNeeded = pbMona1ASIC48;
    AsicSize = sizeof(pbMona1ASIC48);
    if (GetInputClockDetect(eg) & Mask96) {
      AsicNeeded = pbMona1ASIC96;
      AsicSize = sizeof(pbMona1ASIC96);
    }
  }

  if (AsicNeeded != eg->Asic) {
    // Load the desired ASIC
    if (!LoadASICgeneric(eg, DSP_FNC_LOAD_MONA_PCI_CARD_ASIC, AsicNeeded, AsicSize))
      return(FALSE);

    eg->Asic = AsicNeeded;
  }

  return(TRUE);
}


//===========================================================================
//
// WriteControlReg
//
// Most configuration of Gina24, Layla24, or Mona is
// accomplished by writing the control register.  WriteControlReg
// sends the new control register value to the DSP.
//
//===========================================================================

int WriteControlReg(struct CEchoGals *eg, u32 ControlReg) {

  ECHO_DEBUGPRINTF(("WriteControlReg...\n"));
  if (!eg->ASICLoaded)
    return(ECHOSTATUS_ASIC_NOT_LOADED);

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

#ifdef DIGITAL_INPUT_AUTO_MUTE_SUPPORT
  // Handle the digital input auto-mute
  if (eg->DigitalInAutoMute)
    ControlReg |= GML_DIGITAL_IN_AUTO_MUTE;
  else
    ControlReg &= ~GML_DIGITAL_IN_AUTO_MUTE;
#endif

  // Write the control register
  if (ControlReg != GetControlRegister(eg)) {
    SetControlRegister(eg, ControlReg);
    ECHO_DEBUGPRINTF(("WriteControlReg: 0x%x\n", ControlReg));
    ClearHandshake(eg);
    return(SendVector(eg, DSP_VC_WRITE_CONTROL_REG));
  }
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// SetSampleRate
//
// Set the audio sample rate for Mona
//
//===========================================================================

u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {
  u32 AsicSize, ControlReg, NewClock;
  u8 *AsicNeeded;

  // Only set the clock for internal mode.  If the clock is not set to
  // internal, try and re-set the input clock; this more transparently
  // handles switching between single and double-speed mode
  if (eg->InputClock != ECHO_CLOCK_INTERNAL) {
    ECHO_DEBUGPRINTF(("SetSampleRate: Cannot set sample rate - clock not set to CLK_CLOCKININTERNAL\n"));
    
    // Save the rate anyhow
    eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);
    
    // Set the input clock to the current value
    SetInputClock(eg, eg->InputClock);

    return(GetSampleRate(eg));
  }

  // Now, check to see if the required ASIC is loaded
  if (SampleRate >= 88200) {
    if (eg->DigitalMode==DIGITAL_MODE_ADAT)
      return(GetSampleRate(eg));

    if (eg->DeviceId==DEVICE_ID_56361) {
      AsicNeeded = pbMona1ASIC96_361;
      AsicSize = sizeof(pbMona1ASIC96_361);
    } else {
      AsicNeeded = pbMona1ASIC96;
      AsicSize = sizeof(pbMona1ASIC96);
    }
  } else {
    if (eg->DeviceId==DEVICE_ID_56361) {
      AsicNeeded = pbMona1ASIC48_361;
      AsicSize = sizeof(pbMona1ASIC48_361);
    } else {
      AsicNeeded = pbMona1ASIC48;
      AsicSize = sizeof(pbMona1ASIC48);
    }
  }

  if (AsicNeeded != eg->Asic) {
    // Load the desired ASIC
    if (!LoadASICgeneric(eg, DSP_FNC_LOAD_MONA_PCI_CARD_ASIC, AsicNeeded, AsicSize))
      return(GetSampleRate(eg));

    eg->Asic = AsicNeeded;
  }

  // Get the new control register value
  NewClock = 0;

  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_CLOCK_CLEAR_MASK;
  ControlReg &= GML_SPDIF_RATE_CLEAR_MASK;

  switch (SampleRate) {
    case 96000:
      NewClock = GML_96KHZ;
      break;
    case 88200:
      NewClock = GML_88KHZ;
      break;
    case 48000:
      NewClock = GML_48KHZ | GML_SPDIF_SAMPLE_RATE1;
      break;
    case 44100:
      NewClock = GML_44KHZ;
      // Professional mode
      if (ControlReg & GML_SPDIF_PRO_MODE)
        NewClock |= GML_SPDIF_SAMPLE_RATE0;
      break;
    case 32000:
      NewClock = GML_32KHZ | GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
      break;
    case 22050:
      NewClock = GML_22KHZ;
      break;
    case 16000:
      NewClock = GML_16KHZ;
      break;
    case 11025:
      NewClock = GML_11KHZ;
      break;
    case 8000:
      NewClock = GML_8KHZ;
      break;
  }

  ControlReg |= NewClock;

  // Send the new value to the card
  if (ECHOSTATUS_OK == WriteControlReg(eg, ControlReg)) {
    eg->DspCommPage->dwSampleRate = SWAP32(SampleRate);
    ECHO_DEBUGPRINTF(("SetSampleRate: %d clock %d\n", SampleRate, NewClock));
  }

  return(GetSampleRate(eg));

}



//===========================================================================
//
// SetInputClock
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {
  char WriteReg;
  u32 ControlReg, SampleRate;

  ECHO_DEBUGPRINTF(("SetInputClock:\n"));

  ControlReg = GetControlRegister(eg);

  // Mask off the clock select bits
  ControlReg &= GML_CLOCK_CLEAR_MASK;

  SampleRate=0;
  WriteReg = TRUE;
  switch (Clock) {
    case ECHO_CLOCK_INTERNAL:
      ECHO_DEBUGPRINTF(("Set Gina24 clock to INTERNAL\n"));
      SampleRate=GetSampleRate(eg);
      // If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
      if (SampleRate<8000 || SampleRate>96000)
        SampleRate = 48000;
      WriteReg = FALSE;
      break;
    case ECHO_CLOCK_SPDIF:
      if (eg->DigitalMode==DIGITAL_MODE_ADAT)
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      if (!SwitchAsic(eg, GML_CLOCK_DETECT_BIT_SPDIF96))
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      ECHO_DEBUGPRINTF(("Set Mona clock to SPDIF\n"));
      ControlReg |= GML_SPDIF_CLOCK;

      if (GetInputClockDetect(eg) & GML_CLOCK_DETECT_BIT_SPDIF96)
        ControlReg |= GML_DOUBLE_SPEED_MODE;
      else
        ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    case ECHO_CLOCK_WORD:
      ECHO_DEBUGPRINTF(("Set Mona clock to WORD\n"));
      if (!SwitchAsic(eg, GML_CLOCK_DETECT_BIT_WORD96))
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      ControlReg |= GML_WORD_CLOCK;

      if (GetInputClockDetect(eg) & GML_CLOCK_DETECT_BIT_WORD96)
        ControlReg |= GML_DOUBLE_SPEED_MODE;
      else
        ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    case ECHO_CLOCK_ADAT:
      ECHO_DEBUGPRINTF(("Set Mona clock to ADAT\n"));
      if (eg->DigitalMode != DIGITAL_MODE_ADAT)
        return(ECHOSTATUS_CLOCK_NOT_AVAILABLE);
      ControlReg |= GML_ADAT_CLOCK;
      ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
    default :
      ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Mona\n", Clock));
      return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
  }

  // Winner! Save the new input clock.
  eg->InputClock = Clock;

  // Do things according to the flags
  if (WriteReg)
    WriteControlReg(eg, ControlReg);

  // Set Mona sample rate to something sane if word or superword is being turned off
  if (SampleRate)
    SetSampleRate(eg, GetSampleRate(eg));

  return(ECHOSTATUS_OK);
}



// Layla20 only
int SetOutputClock(struct CEchoGals *eg, u16 Clock) {

  return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
}



// This card has no vmixer.
int SetVmixerGain(struct CEchoGals *eg, u16 Output, u16 Pipe, int Gain) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// This card has no vmixer.
int UpdateVmixerLevel(struct CEchoGals *eg) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



//===========================================================================
//
//      Set digital mode
//
//===========================================================================

int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {
  u32 ControlReg;

  // Clear the current digital mode
  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_DIGITAL_MODE_CLEAR_MASK;

  // Tweak the control reg
  switch (Mode) {
    default :
      return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
    case DIGITAL_MODE_SPDIF_OPTICAL:
      ControlReg |= GML_SPDIF_OPTICAL_MODE;
      // fall through
    case DIGITAL_MODE_SPDIF_RCA:
      // If the input clock is set to ADAT, set the input clock to internal and the sample rate to 48 KHz
      if (eg->InputClock==ECHO_CLOCK_ADAT) {
        eg->DspCommPage->dwSampleRate = SWAP32((u32)48000);
        SetInputClock(eg, ECHO_CLOCK_INTERNAL);
      }
      break;
    case DIGITAL_MODE_ADAT:
      // If the input clock is set to S/PDIF, set the input clock to internal and the sample rate to 48 KHz
      if (eg->InputClock==ECHO_CLOCK_SPDIF) {
        eg->DspCommPage->dwSampleRate = SWAP32((u32)48000);
        SetInputClock(eg, ECHO_CLOCK_INTERNAL);
      }
      // If the current ASIC is the 96KHz ASIC, switch the ASIC and set to 48 KHz
      if ((eg->DeviceId==DEVICE_ID_56361 && eg->Asic==pbMona1ASIC96_361) ||
          (eg->DeviceId==DEVICE_ID_56301 &&  eg->Asic==pbMona1ASIC96)) {
        SetSampleRate(eg, 48000);
      }
      ControlReg |= GML_ADAT_MODE;
      ControlReg &= ~GML_DOUBLE_SPEED_MODE;
      break;
  }

  // Write the control reg
  WriteControlReg(eg, ControlReg);

  eg->DigitalMode = Mode;

  ECHO_DEBUGPRINTF(("SetDigitalMode to %d\n", (u32)eg->DigitalMode));
  return ECHOSTATUS_OK;
}



//===========================================================================
//
// Set the S/PDIF output format (common for Gina24, Layla24, and Mona)
//
//===========================================================================

int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {
  u32 ControlReg;

  // Clear the current S/PDIF flags
  ControlReg = GetControlRegister(eg);
  ControlReg &= GML_SPDIF_FORMAT_CLEAR_MASK;

  // Set the new S/PDIF flags depending on the mode
  ControlReg |= GML_SPDIF_TWO_CHANNEL | GML_SPDIF_24_BIT | GML_SPDIF_COPY_PERMIT;
  if (Prof) {
    // Professional mode
    ControlReg |= GML_SPDIF_PRO_MODE;

    switch (GetSampleRate(eg)) {
      case 32000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
        break;
      case 44100:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0;
        break;
      case 48000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE1;
        break;
    }
  } else {
    // Consumer mode
    switch (GetSampleRate(eg)) {
      case 32000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
        break;
      case 48000:
        ControlReg |= GML_SPDIF_SAMPLE_RATE1;
        break;
    }
  }

  WriteControlReg(eg, ControlReg);
  eg->ProfessionalSpdif=Prof;
  ECHO_DEBUGPRINTF(("SetProfessionalSpdif to %s\n", (Prof) ? "Professional" : "Consumer"));
  return(ECHOSTATUS_OK);
}

