// ****************************************************************************
//
//              CEchoGals.cpp
//
//              Implementation file for the CEchoGals driver class.
//              Set editor tabs to 3 for your viewing pleasure.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#ifndef ECHO24_FAMILY
#define ECHOGALS_FAMILY
#endif

#include "Echoaudio.h"



void DelEchoGals(struct CEchoGals *eg) {
  int i;

  ECHO_DEBUGPRINTF(("Del.  open=%x\n", eg->AudioOpenMask));

  // Close all open pipes. We start from lower pipes, so CloseAudio() clears the higher bits if the
  // pipes have more than one channel and it never fails.
  for (i=0; i<ECHO_MAXAUDIOPIPES; i++)
    if (VerifyAudioOpen(eg, i)==ECHOSTATUS_OK)
      FreeAudio(eg, i);

  // Free stuff
  DeleteCDspComm(eg);                    // This powers down the DSP

  ECHO_DEBUGPRINTF(("DelEchoGals: EchoGals is toast!\n"));

  OsFreeNonPaged(eg);
}





//===========================================================================
//
// This method initializes classes to control the mixer controls for
// the busses and pipes.  This is a protected method; it is called
// from each of the CEchoGals derived classes once the DSP is up
// and running.
//
//===========================================================================

int InitLineLevels(struct CEchoGals *eg) {
  int st;
  int i, o;

  ECHO_DEBUGPRINTF(("InitLineLevels\n"));

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  // Mute output busses
  for (i=0; i<eg->NumBussesOut; i++)
    if ((st = SetOutputGain(eg, i, ECHOGAIN_MUTED)))
      return(st);
  if ((st = UpdateOutputLineLevel(eg)))
    return(st);

  // Mute the Vmixer if the card has it
  if (eg->HasVmixer) {
    for (i=eg->NumPipesOut-1; i>=0; i--)
      for (o=eg->NumBussesOut-1; o>=0; o--)
        if ((st = SetVmixerGain(eg, o, i, ECHOGAIN_MUTED)))
          return(st);
    if ((st = UpdateVmixerLevel(eg)))
      return(st);
  }

/*@@ vedere CPipeOutCtrl per la schede com VMixer
  st=m_PipeOutCtrl.Init(this);
  if (ECHOSTATUS_OK != st)
    return(st);
*/

  // Mute the monitor mixer
  for (o=eg->NumBussesOut-1; o>=0; o--)
    for (i=eg->NumBussesIn-1; i>=0; i--)
      if ((st = SetMonitorGain(eg, o, i,ECHOGAIN_MUTED)))
        return(st);
  if ((st = UpdateOutputLineLevel(eg)))
    return(st);

  // Mute input channels @@?
  for (i=0; i<eg->NumBussesIn; i++)
    if ((st = SetInputGain(eg, i, ECHOGAIN_MUTED)))
      return(st);
  if ((st = UpdateInputLineLevel(eg)))
    return(st);

  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// Set the digital mode - currently for Gina24, Layla24, and Mona
//
//===========================================================================

int SetDigitalMode(struct CEchoGals *eg, u8 Mode) {
  int Status;
  u8 PreviousDigitalMode;
  int i, o;

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  if (eg->AudioOpenMask) {
    ECHO_DEBUGPRINTF(("SetDigitalMode()  All audio channels must be closed before changing the digital mode\n"));
    return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
  }

  if (!(eg->DigitalModes & (1<<Mode)))
    return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);

  PreviousDigitalMode = eg->DigitalMode;
  Status = DspSetDigitalMode(eg, Mode);

  // If we successfully changed the digital mode from or to ADAT, then make sure
  // all output, input and monitor levels are updated by the DSP comm object.
  if (Status==ECHOSTATUS_OK && PreviousDigitalMode!=Mode && (PreviousDigitalMode==DIGITAL_MODE_ADAT || Mode==DIGITAL_MODE_ADAT)) {

    for (o=eg->NumBussesOut-1; o>=0; o--)
      for (i=eg->NumBussesIn-1; i>=0; i--)
        SetMonitorGain(eg, o, i, eg->MonitorLevel[o][i]);

/* @@@ ???
    for (wBus = 0; wBus < GetNumBussesOut(); wBus++) {
      for (wPipe = 0; wPipe < GetNumPipesOut(); wPipe++) {
        m_PipeOutCtrl.SetGain(wPipe, wBus, ECHOGAIN_UPDATE, FALSE);
      }
    }
*/

    for (i=eg->NumBussesIn-1; i>=0; i--)
      SetInputGain(eg, i, eg->InLevel[i]);

    for (i=eg->NumBussesOut-1; i>=0; i--)
      SetOutputGain(eg, i, eg->OutLevel[i]);

    // Now set them all at once. Do the output bus _and_ the output pipe in case this is a vmixer card.
    UpdateOutputLineLevel(eg);
    UpdateInputLineLevel(eg);
/*@@@ vmixer
    m_BusOutLineLevels[0].SetGain(ECHOGAIN_UPDATE, TRUE);
    m_PipeOutCtrl.SetGain(0, 0, ECHOGAIN_UPDATE, TRUE);
*/
  }

  return(Status);
}



void SetFlags(struct CEchoGals *eg, u16 Flags) {

  // Mask off the read-only flags so they don't change
  Flags &= ECHOGALS_FLAG_WRITABLE_MASK;
  // Set the flags & mark the flags as changed
  eg->Flags |= Flags;
}



void ClearFlags(struct CEchoGals *eg, u16 Flags) {

  // Mask off the read-only flags so they don't change
  Flags &= ECHOGALS_FLAG_WRITABLE_MASK;
  // Set the flags & mark the flags as changed
  eg->Flags &= ~Flags;
}



/******************************************************************************

 CEchoGals interrupt handler functions

 ******************************************************************************/

//===========================================================================
//
// This isn't the interrupt handler itself; rather, the OS-specific layer
// of the driver has an interrupt handler that calls this function.
//
//===========================================================================

int ServiceIrq(struct CEchoGals *eg, int *fMidiReceived) {

  // Read the DSP status register and see if this DSP generated this interrupt
  *fMidiReceived=FALSE;

  if (GetStatusReg(eg) & CHI32_STATUS_IRQ) {
/*
#ifdef MIDI_SUPPORT

    //
    // If this was a MIDI input interrupt, get the MIDI input data
    //
    DWORD dwMidiInCount;

    pDCO->ReadMidi(0, dwMidiInCount);	// The count is at index 0
    if (0 != dwMidiInCount) {
      m_MidiIn.ServiceIrq();
      fMidiReceived = TRUE;
    }
#endif // MIDI_SUPPORT
*/
    //
    //      Clear the hardware interrupt
    //
    AckInt(eg);
    return(ECHOSTATUS_OK);
  }
  return(ECHOSTATUS_IRQ_NOT_OURS);
}



/****************************************************************************

	CEchoGals informational methods

 ****************************************************************************/

//===========================================================================
//
// GetBaseCapabilities is used by the CEchoGals-derived classes to
// help fill out the ECHOGALS_CAPS struct.
//
//===========================================================================

int GetBaseCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS Caps) {
  int i;

  // Test only for no DSP object.  We can still get capabilities even if board is bad.
/*@@@  dsp=eg->dsp;
  if (!dsp)
    return(ECHOSTATUS_DSP_DEAD);
*/
  memset(Caps, 0, sizeof(ECHOGALS_CAPS));
//  strcpy(Caps->szName, m_szCardInstallName);

  Caps->wCardType = eg->CardType;
  Caps->wNumPipesOut = eg->NumPipesOut;
  Caps->wNumPipesIn = eg->NumPipesIn;
  Caps->wNumBussesOut = eg->NumBussesOut;
  Caps->wNumBussesIn = eg->NumBussesIn;

  Caps->wFirstDigitalBusOut = eg->FirstDigitalBusOut;
  Caps->wFirstDigitalBusIn = eg->FirstDigitalBusIn;

  Caps->wNumMidiOut = eg->NumMidiOut;
  Caps->wNumMidiIn = eg->NumMidiIn;

  Caps->dwOutClockTypes = 0;
  Caps->dwInClockTypes = ECHO_CLOCK_BIT_INTERNAL;

  // Add the controls to the output pipes that are the same for all cards
  for (i=0; i<Caps->wNumPipesOut; i++)
    Caps->dwPipeOutCaps[i] = ECHOCAPS_GAIN | ECHOCAPS_MUTE;

  // Add controls to the input busses that are the same for all cards
  for (i=0; i<Caps->wNumBussesIn; i++) {
    Caps->dwBusInCaps[i] = ECHOCAPS_PEAK_METER | ECHOCAPS_VU_METER;
    if (i>=Caps->wFirstDigitalBusIn)
      Caps->dwBusInCaps[i] |= ECHOCAPS_DIGITAL;
  }

  // And the output busses
  for (i=0; i<Caps->wNumBussesOut; i++) {
    Caps->dwBusOutCaps[i] = ECHOCAPS_PEAK_METER | ECHOCAPS_VU_METER | ECHOCAPS_GAIN | ECHOCAPS_MUTE;
    if (i>=Caps->wFirstDigitalBusOut)
      Caps->dwBusOutCaps[i] |= ECHOCAPS_DIGITAL;
  }

  // Digital modes & vmixer flag
  Caps->dwDigitalModes = eg->DigitalModes;
  Caps->fHasVmixer = eg->HasVmixer;

  // If this is not a vmixer card, output pipes are hard-wired to output busses.
  // Mark those pipes that are hard-wired to digital busses as digital pipes
  if (eg->HasVmixer) {
    Caps->wFirstDigitalPipeOut = Caps->wNumPipesOut;
  } else {
    Caps->wFirstDigitalPipeOut = Caps->wFirstDigitalBusOut;
    for (i=Caps->wFirstDigitalPipeOut; i<Caps->wNumPipesOut; i++)
      Caps->dwPipeOutCaps[i] |= ECHOCAPS_DIGITAL;
  }

  // Input pipes are the same, vmixer or no vmixer
  Caps->wFirstDigitalPipeIn = Caps->wFirstDigitalBusIn;

  for (i=Caps->wFirstDigitalPipeIn; i<Caps->wNumPipesIn; i++)
    Caps->dwPipeInCaps[i] |= ECHOCAPS_DIGITAL;

  return(ECHOSTATUS_OK);
}




/****************************************************************************

	Character strings for the ECHOSTATUS return values -
	useful for debugging.

 ****************************************************************************/

//
// pStatusStrs is used if you want to print out a friendlier version of
// the various ECHOSTATUS codes.
//
char *pStatusStrs[ECHOSTATUS_LAST] = {
  "ECHOSTATUS_OK",
  "ECHOSTATUS_BAD_FORMAT",
  "ECHOSTATUS_BAD_BUFFER_SIZE",
  "ECHOSTATUS_CANT_OPEN",
  "ECHOSTATUS_CANT_CLOSE",
  "ECHOSTATUS_CHANNEL_NOT_OPEN",
  "ECHOSTATUS_BUSY",
  "ECHOSTATUS_BAD_LEVEL",
  "ECHOSTATUS_NO_MIDI",
  "ECHOSTATUS_CLOCK_NOT_SUPPORTED",
  "ECHOSTATUS_CLOCK_NOT_AVAILABLE",
  "ECHOSTATUS_BAD_CARDID",
  "ECHOSTATUS_NOT_SUPPORTED",
  "ECHOSTATUS_BAD_NOTIFY_SIZE",
  "ECHOSTATUS_INVALID_PARAM",
  "ECHOSTATUS_NO_MEM",
  "ECHOSTATUS_NOT_SHAREABLE",
  "ECHOSTATUS_FIRMWARE_LOADED",
  "ECHOSTATUS_DSP_DEAD",
  "ECHOSTATUS_DSP_TIMEOUT",
  "ECHOSTATUS_INVALID_CHANNEL",
  "ECHOSTATUS_CHANNEL_ALREADY_OPEN",
  "ECHOSTATUS_DUCK_FULL",
  "ECHOSTATUS_INVALID_INDEX",
  "ECHOSTATUS_BAD_CARD_NAME",
  "ECHOSTATUS_IRQ_NOT_OURS",
  "",
  "",
  "",
  "",
  "",
  "ECHOSTATUS_BUFFER_OVERFLOW",
  "ECHOSTATUS_OPERATION_CANCELED",
  "ECHOSTATUS_EVENT_NOT_OPEN",
  "ECHOSTATUS_ASIC_NOT_LOADED" "ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED",
  "ECHOSTATUS_RESERVED",
  "ECHOSTATUS_BAD_COOKIE",
  "ECHOSTATUS_MIXER_DISABLED",
  "ECHOSTATUS_NO_SUPER_INTERLEAVE",
  "ECHOSTATUS_DUCK_NOT_WRAPPED"
};



// *** CEchoGals.cpp ***
