// ****************************************************************************
//
//              CEchoGals_transport.cpp
//
//              Audio transport methods for the CEchoGals driver class.
//              Set editor tabs to 3 for your viewing pleasure.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#ifndef ECHO24_FAMILY
#define ECHOGALS_FAMILY
#endif

#include "Echoaudio.h"



/******************************************************************************

 Functions for opening and closing pipes

 ******************************************************************************/

//===========================================================================
//
// AllocateAudio is used to reserve audio pipes for your exclusive use.  The call
// will fail if someone else has already opened the pipes.  Calling AllocateAudio
// is the first step if you want to play or record. Note that you can eventually
// set an interleave smaller than the number of channel you have allocated.
//
//===========================================================================

int AllocateAudio(struct CEchoGals *eg, u16 Channel, u16 Interleave, char IsInput, u16 *pipeindex) {
  int i, PipeIndex, max, Status;
  u32 ChannelMask;
  char IsCyclic;

  ECHO_DEBUGPRINTF(("OpenAudio: ch=%d int=%d\n", Channel, Interleave));

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  IsCyclic=1;
  if (IsInput) {
    PipeIndex=Channel+eg->NumPipesOut;
    max=eg->NumPipesOut+eg->NumBussesIn;
  } else {
    max=eg->NumBussesOut;
    PipeIndex=Channel;
  }

  if (PipeIndex+Interleave>max) {
    ECHO_DEBUGPRINTF(("OpenAudio: invalid channel\n"));
    return(ECHOSTATUS_INVALID_CHANNEL);
  }

  // If the width is more than two, make sure that this card can handle super interleave
  if (Interleave>2 && !(eg->Flags & ECHOGALS_ROFLAG_SUPER_INTERLEAVE_OK)) {
    ECHO_DEBUGPRINTF(("OpenAudio: no super interleave\n"));
    return(ECHOSTATUS_NO_SUPER_INTERLEAVE);
  }

  // Compute channel mask for this substream
  for (ChannelMask=i=0; i<Interleave; i++)
    ChannelMask|=1<<(PipeIndex+i);

  // See if the specified pipes are already open
  if (eg->AudioOpenMask&ChannelMask) {
    ECHO_DEBUGPRINTF(("OpenAudio: channel already open\n"));
    return(ECHOSTATUS_CHANNEL_ALREADY_OPEN);
  }

  if ((Status=AllocateSGList(eg, PipeIndex)))
    return(Status);

  // ResetPipePosition()
  eg->DspCommPage->dwPosition[PipeIndex]=0;

  // Prep stuff
  eg->AudioOpenMask|=ChannelMask;
  if (IsCyclic)
    eg->AudioCyclicMask|=ChannelMask;
  eg->Pipes[PipeIndex].Channel=Channel;
  eg->Pipes[PipeIndex].Interleave=Interleave;
  eg->Pipes[PipeIndex].IsInput=IsInput;
  eg->Pipes[PipeIndex].IsCyclic=IsCyclic;
  ResetAudio(eg, PipeIndex);
  *pipeindex=PipeIndex;
  ECHO_DEBUGPRINTF(("OpenAudio: ok\n"));
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// FreeAudio is, naturally, the inverse of AllocateAudio.
//
//===========================================================================

int FreeAudio(struct CEchoGals *eg, u16 PipeIndex) {
  u32 ChannelMask;
  int i, Status;

  ECHO_DEBUGPRINTF(("CloseAudio: Pipe %u\n", PipeIndex));

  if ((Status=VerifyAudioOpen(eg, PipeIndex)))
    return(Status);

  // Compute channel mask for this substream
  for (ChannelMask=i=0; i<eg->Pipes[PipeIndex].Interleave; i++)
    ChannelMask|=1<<(PipeIndex+i);

  ResetAudio(eg, PipeIndex);
  FreeSGList(eg, PipeIndex);

  eg->AudioOpenMask&=~ChannelMask;
  eg->AudioCyclicMask&=~ChannelMask;
  eg->Pipes[PipeIndex].Interleave=0;
  return(ECHOSTATUS_OK);
}



//===========================================================================
//
// VerifyAudioOpen is a utility function; it tells you if
// a pipe is open or not.
//
//===========================================================================

int VerifyAudioOpen(struct CEchoGals *eg, u16 PipeIndex) {

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  if (eg->AudioOpenMask&(1<<PipeIndex))
    return(ECHOSTATUS_OK);
  else
    return(ECHOSTATUS_CHANNEL_NOT_OPEN);
}



//===========================================================================
//
// This function does exactly what you think it does.
//
// Note that if the card is not set to internal clock (that is, the hardware
// is synced to word clock or some such), this call has no effect.
//
// Note that all of the inputs and outputs on a single card share the same clock.
//
//===========================================================================

int SetAudioSampleRate(struct CEchoGals *eg, u32 Rate) {
  int Status;

  ECHO_DEBUGPRINTF(("SetAudioSampleRate: to %d Hz\n", Rate));

  Status=QueryAudioSampleRate(eg, Rate);
  if (ECHOSTATUS_OK != Status)
    return(Status);

  if (Rate==SetSampleRate(eg, Rate)) {
    eg->SampleRate=Rate;
    return(ECHOSTATUS_OK);
  }
  return(ECHOSTATUS_BAD_FORMAT);
}



/******************************************************************************

 Functions for setting audio formats and the sample rate

 ******************************************************************************/

//===========================================================================
//
// Validate an audio format.
//
// For comments on audio formats, refer to the definition of Audioformat_t
//
//===========================================================================

int QueryAudioFormat(struct CEchoGals *eg, u16 PipeIndex, Audioformat_t *AudioFormat) {
  u16 Interleave, StoredPipeWidth, MaxPipe;
  int Status;

  ECHO_DEBUGPRINTF(("QueryAudioFormat\n"));
  Status=ECHOSTATUS_OK;

  // If this pipe is open, make sure that this audio format does not exceed the stored pipe width
  Interleave=AudioFormat->DataInterleave;
  StoredPipeWidth=eg->Pipes[PipeIndex].Interleave;

  if (StoredPipeWidth) {
    if (Interleave>StoredPipeWidth) {
      ECHO_DEBUGPRINTF(("QueryAudioFormat: pipe was opened with a width of %d; interleave of %d invalid.\n", StoredPipeWidth, AudioFormat->DataInterleave));
      return(ECHOSTATUS_BAD_FORMAT);
    }
  }

  // Check for super interleave (i.e. interleave > 2)
  if (Interleave>2) {
    // Make sure the card is capable of super interleave
    if (!(eg->Flags & ECHOGALS_ROFLAG_SUPER_INTERLEAVE_OK))
      return(ECHOSTATUS_NO_SUPER_INTERLEAVE);

    // Allow super interleave for 32 bit little endian data;
    // the interleave factor must be even.
    if (AudioFormat->BitsPerSample!=32 || AudioFormat->DataAreBigEndian || (Interleave & 1))
      return(ECHOSTATUS_BAD_FORMAT);

    // Make sure that this interleave factor on this pipe
    // does not exceed the number of pipes for the card
    if (PipeIndex>=eg->NumPipesOut) {
      MaxPipe=eg->NumPipesIn;
      PipeIndex-=eg->NumPipesOut;
    } else {
      MaxPipe=eg->NumPipesOut;
    }

    if (PipeIndex+Interleave > MaxPipe)
      return(ECHOSTATUS_BAD_FORMAT);

    return(ECHOSTATUS_OK);
  }

  // Check the interleave, which must be 1 or 2
  if (AudioFormat->DataInterleave!=1 && AudioFormat->DataInterleave!=2) {
    ECHO_DEBUGPRINTF(("QueryAudioFormat: interleave %d not allowed\n", AudioFormat->DataInterleave));
    return(ECHOSTATUS_BAD_FORMAT);
  }

  // If the big endian flag is set, the data must be mono or stereo interleave,
  // 32 bits wide, left justified data.  Only the upper 24 bits are used.
  if (AudioFormat->DataAreBigEndian) {
    // Must have 32 bits per sample
    if (AudioFormat->BitsPerSample!=32) {
      ECHO_DEBUGPRINTF(("QueryAudioFormat: Only 32 bits per sample supported for big-endian data\n"));
      return(ECHOSTATUS_BAD_FORMAT);
    }

    // Mono or stereo only
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
    if (AudioFormat->DataInterleave!=1 && AudioFormat->DataInterleave!=2) {
      ECHO_DEBUGPRINTF(("QueryAudioFormat: Interleave of %d not allowed for big-endian data\n", AudioFormat->DataInterleave));
      return(ECHOSTATUS_BAD_FORMAT);
    }
#else
    if (AudioFormat->DataInterleave!=1) {
      ECHO_DEBUGPRINTF(("QueryAudioFormat: Interleave of %d not allowed for big-endian data\n", AudioFormat->DataInterleave));
      return(ECHOSTATUS_BAD_FORMAT);
    }
#endif

    return(ECHOSTATUS_OK);
  }

  // Check bits per sample
  switch (AudioFormat->BitsPerSample) {
    case 8:          //@@ 8 and 16 bits: Mono->Mono is not supported. Mono->Stereo and Stereo->Stereo only
    case 16:         //@@
    case 32:
      break;
    default:
      ECHO_DEBUGPRINTF(("QueryAudioFormat: No valid format specified, bits per sample %d\n", AudioFormat->BitsPerSample));
      Status=ECHOSTATUS_BAD_FORMAT;
      break;
  }

  return(Status);
}



//===========================================================================
//
// SetAudioFormat sets the format of the audio data in host memory
// for this pipe.
//
//===========================================================================

int SetAudioFormat(struct CEchoGals *eg, u16 PipeIndex, Audioformat_t *AudioFormat) {
  int Status;

  ECHO_DEBUGPRINTF(("SetAudioFormat: for pipe %d\n", PipeIndex));

  // Check the pipe number
  if (PipeIndex >= eg->NumPipesOut+eg->NumPipesIn) {
    ECHO_DEBUGPRINTF(("SetAudioFormat: Invalid pipe %d\n", PipeIndex));
    return(ECHOSTATUS_INVALID_CHANNEL);
  }

  // Make sure this pipe is open
  Status = VerifyAudioOpen(eg, PipeIndex);
  if (ECHOSTATUS_OK != Status)
    return Status;

  // Check the format
  Status = QueryAudioFormat(eg, PipeIndex, AudioFormat);
  if (ECHOSTATUS_OK != Status)
    return(Status);

  // Set the format
  Status = DspSetFormat(eg, PipeIndex, AudioFormat, eg->Flags&ECHOGALS_FLAG_SPDIF_NODITHER);
  return(Status);
}



/******************************************************************************

 Functions for starting and stopping transport

******************************************************************************/


//===========================================================================
//
//      Start transport for a single pipe
//
//===========================================================================

int StartAudio(struct CEchoGals *eg, u16 PipeIndex) {

  return(StartAudioMask(eg, ((u32)1)<<PipeIndex));
}



//===========================================================================
//
//      Start transport for a group of pipes
//
//===========================================================================

int StartAudioMask(struct CEchoGals *eg, u32 Mask) {
  u16 PipeIndex;
  u32 StartMask;

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  StartMask=0;

  //--------------------------------------------------------
  // Process each pipe in the mask
  //--------------------------------------------------------
  for (PipeIndex=0; PipeIndex<eg->NumPipesOut+eg->NumBussesIn; PipeIndex++) {
    // Skip this pipe if it's not in the mask
    if (!(Mask&(1<<PipeIndex)))
      continue;

    // This pipe must have a scatter-gather list
    if (!eg->Pipes[PipeIndex].sglist) {
      ECHO_DEBUGPRINTF(("StartAudioMask: trying to start pipe index %d but there is no s-g list !\n", PipeIndex));
      return(ECHOSTATUS_CHANNEL_NOT_OPEN);
    }

    /*// If this pipe was opened in cyclic mode, make sure that the corresponding CDaffyDuck has been wrapped
    if ((0 != m_cmAudioCyclic.TestIndexInMask(wPipe)) && (FALSE == m_DaffyDucks[wPipe]->Wrapped())) {
      ECHO_DEBUGPRINTF(("CEchoGals::Start called for pipe index %d - " "pipe was opened in cyclic mode, but the duck " "has not been wrapped\n", wPipe));
      return ECHOSTATUS_DUCK_NOT_WRAPPED;
    }*/

    // Do different things to this pipe depending on the state
    switch (eg->Pipes[PipeIndex].State) {
      case PIPE_STATE_RESET:
        eg->DspCommPage->dwPosition[PipeIndex]=0;            // Clean up the DMA position stuff
      case PIPE_STATE_STOPPED:
        eg->Pipes[PipeIndex].State = PIPE_STATE_STARTED;
        StartMask |= 1<<PipeIndex;
        break;
      case PIPE_STATE_PENDING:
      case PIPE_STATE_STARTED:
        break;
    }
  }

  // Send the result of all this to the DSP
  if (!StartMask)
    return(ECHOSTATUS_OK);

  return(StartTransport(eg, StartMask, eg->AudioCyclicMask));
}



//===========================================================================
//
// Stop a single pipe
//
//===========================================================================

int StopAudio(struct CEchoGals *eg, u16 PipeIndex) {

  return(StopAudioMask(eg, ((u32)1)<<PipeIndex));
}



//===========================================================================
//
// Stop several pipes simultaneously
//
//===========================================================================
int StopAudioMask(struct CEchoGals *eg, u32 Mask) {
  int i, Status;

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  if ((Status=StopTransport(eg, Mask)))
    return(Status);

  for (i=eg->NumPipesOut+eg->NumPipesIn; i>=0; i--)
    if (Mask&(1<<i))
      eg->Pipes[i].State=PIPE_STATE_STOPPED;

  return(Status);
}



//===========================================================================
//
// Reset transport for a single pipe
//
//===========================================================================

int ResetAudio(struct CEchoGals *eg, u16 PipeIndex) {

  return(ResetAudioMask(eg, ((u32)1)<<PipeIndex));
}



//===========================================================================
//
// Reset transport for a group of pipes simultaneously
//
//===========================================================================

int ResetAudioMask(struct CEchoGals *eg, u32 Mask) {
  int i, Status;

  if (eg->BadBoard)
    return(ECHOSTATUS_DSP_DEAD);

  if ((Status=ResetTransport(eg, Mask)))
    return(Status);

  for (i=eg->NumPipesOut+eg->NumPipesIn; i>=0; i--)
    if (Mask&(1<<i))
      eg->Pipes[i].State=PIPE_STATE_RESET;

  return(Status);
}



//===========================================================================
//
// This is a very powerful feature; calling this function gives you a pointer
// to the memory location where the DSP writes the 32 bit DMA position for
// a pipe.  The DSP is constantly updating this value as it moves data.
// The DMA counter is in units of bytes, not samples.
//
// Since the DSP is constantly updating it, you can dereference this pointer
// from anywhere and read the DMA position without calling the generic driver.
// This means that with some adroit mapping, you could read the DMA position
// from user mode without calling the kernel.
//
// Remember, Peter - with great power comes great responsibility; you should
// only read this pointer and never write to it or to anywhere around it.  You
// could easily lock up your computer.
//
// Note that the DSP writes the position in little endian format; if you are
// on a big endian machine, you will need to byte swap the postion
// before you use it.
//
//===========================================================================

volatile u32 *GetAudioPositionPtr(struct CEchoGals *eg, u16 PipeIndex) {

  if (PipeIndex>=eg->NumPipesOut+eg->NumPipesIn)
    return(0);

  return(&eg->DspCommPage->dwPosition[PipeIndex]);
}

