/*
 *  Implementation file for Linux support services to the CEchoGals generic driver class.
 *  Copyright (C) 2003 Giuliano Pochini <pochini@shiny.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <sound/driver.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/info.h>
#include <sound/control.h>
#include <sound/pcm.h>

#include "OsSupport.h"
#define ECHO24_FAMILY           // No card-specific stuff is used here. This define is only for compile.
#include "Echoaudio.h"


//===========================================================================
//
// OsAllocateNonPaged is used to allocate a block of memory that will always
// be resident; that is, this particular block of memory won't be swapped
// out as virtual memory and can always be accessed at interrupt time.
//
// gdwAllocNonPagedCount tracks the number of times this has been
// successfully called.
//
// This is primarily used for overloading the new operator for the various
// generic classes.
//
//===========================================================================

void *OsAllocateNonPaged(size_t Bytes) {
  void *addr;

//ECHO_DEBUGPRINTF(("offs=%x\n", &((DspCommPage *)0)->dwControlReg));
  addr=snd_kcalloc(Bytes, GFP_KERNEL);
//  printk("Allocato a %x\n", (int)addr);
  if (!addr)
    ECHO_DEBUGPRINTF(("OsAllocateNonPaged: Failed on %d bytes\n", Bytes));
  return(addr);
};



//===========================================================================
//
// OsFreeNonPaged is used to free memory allocated by OsAllocateNonPaged.
//
// gdwAllocNonPagedCount tracks the number of times this has been
// successfully called.
//
//===========================================================================

int OsFreeNonPaged(void *MemAddr) {

  kfree(MemAddr);
  return(ECHOSTATUS_OK);
}



//***********************************************************************

//
// This class is uniquely defined for each OS.  It provides
//	information that other components may require.
//
// For example, in Windows NT it contains a device object used by various
//	memory management methods.
//
// An instance of this class must be constructed and initialized prior to
//	constructing the CEchoGals derived object.  The CEchoGals and
//	CDspCommObject classes must have access to it during their respective
// construction times.
//
//***********************************************************************


//===========================================================================
//
//	Timer methods
//
//===========================================================================

//---------------------------------------------------------------------------
//
// Return the system time in microseconds.
// Return error status if the OS doesn't support this function.
//
//---------------------------------------------------------------------------

int OsGetSystemTime(u64 *Time) {
  struct timeval tv;

  do_gettimeofday(&tv);
  *Time=1000000ULL*tv.tv_sec+tv.tv_usec;
  return(ECHOSTATUS_OK);
}



//---------------------------------------------------------------------------
//
// Stall execution for dwTime microseconds.
// Return error status if the OS doesn't support this
// function.
//
//---------------------------------------------------------------------------

int OsSnooze(u32 Time) {

  mdelay((Time+999)/1000);
  return(ECHOSTATUS_OK);
};



//===========================================================================
//
//	More memory management
//
//===========================================================================

//---------------------------------------------------------------------------
//
// Allocate locked, non-pageable, physically contiguous memory pages
// in the drivers address space.  Used to allocate memory for the DSP
// communications area.
//
// Currently, none of the generic code ever sets dwPageCt to higher than
// one.  On the platforms we've developed for so far, a page is 4096
// bytes and is aligned on a 4096 byte boundary.  None of the structures
// allocated by OsPageAllocate are more than 4096 bytes; memory
// allocated by this routine should be at least aligned on a 32 byte
// boundary and be physically contiguous.
//
// Note that this code is not 64 bit ready, since it's only using
// the low 32 bits of the physical address.
//
//---------------------------------------------------------------------------

int OsPageAllocate(const void *pci, u32 Pages, void **PageAddr, dma_addr_t *PhysicalPageAddr) {
/* Pagea             How many pages to allocate
   PageAddr          Where to return the memory ptr
   PhysicalPageAddr  Where to return the physical PCI addr */
  *PageAddr=snd_malloc_pci_pages((struct pci_dev *)pci, Pages*4096, PhysicalPageAddr);
  if (*PageAddr) {
    memset(*PageAddr, 0, Pages*4096);
    return(ECHOSTATUS_OK);
  } else {
    return(ECHOSTATUS_NO_MEM);
  }
}



//---------------------------------------------------------------------------
//
// Unlock and free non-pageable, physically contiguous memory pages in
//	the drivers address space; the inverse of OsPageAllocate
//
//---------------------------------------------------------------------------

int OsPageFree(const void *pci, u32 Pages, void *PageAddr, dma_addr_t PhysicalPageAddr) {

  snd_free_pci_pages((struct pci_dev *)pci, Pages*4096, PageAddr, PhysicalPageAddr);
  return(ECHOSTATUS_OK);
}



