/*
 *  ALSA driver for Echoaudio soundcards.
 *  Copyright (C) 2003 Giuliano Pochini <pochini@shiny.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Enable MIDI support */
/*#define EMIDI*/

#include <sound/driver.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/info.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/rawmidi.h>
#define SNDRV_GET_ID
#include <sound/initval.h>

#define DSP_MAXPIPES 32         /* Global limit of number of subtreams (input + output) */

#include "EchoGalsXface.h"
#include "MixerXface.h"
#define ECHO24_FAMILY           /* Not used, only for compile */
#include "Echoaudio.h"



/* Debug initialization and objects creation/destruction */
#if (1)
#define DE_INIT(x) snd_printk x;
#else
#define DE_INIT(x)
#endif

/* Debug hw_params callbacks */
#if (1)
#define DE_HWP(x) snd_printk x;
#else
#define DE_HWP(x)
#endif

/* Debug normal activity (open, start, stop...) */
#if (1)
#define DE_ACT(x) snd_printk x;
#else
#define DE_ACT(x)
#endif


/*@@@ Unconditionally disable power management support for now. */
#undef CONFIG_PM

MODULE_AUTHOR("Giuliano Pochini <pochini@shiny.it>");
MODULE_DESCRIPTION("Echoaudio soundcards driver");
MODULE_LICENSE("GPL");
MODULE_CLASSES("{sound}");
MODULE_DEVICES("{{Echoaudio series}}");



struct EchoChip {
  snd_card_t *card;
  struct pci_dev *pci;
  unsigned long iophys, iovirt;
  struct resource *iores;
  int irq;
  spinlock_t lock;
  struct CEchoGals *eg;			/* EchoGals structure */
  snd_pcm_t *AnPcm, *DiPcm;
  snd_pcm_substream_t *substream[DSP_MAXPIPES];
  snd_rawmidi_t *rmidi;
  snd_rawmidi_substream_t *MidiIn, *MidiOut;
  u8 Midibuf[ECHO_MIDI_QUEUE_SZ];
  spinlock_t Midilock;
  char tinuse;				/* Timer in use */
  struct timer_list timer;
  u16 DigitalModes, DigitalModeIndex[6];
  u16 ClockSources, ClockSourceIndex[10];
  int LastPeriod[DSP_MAXPIPES];
  ECHOGALS_CAPS Caps;			/* Card capabilities */
};

typedef struct EchoChip chip_t;		/* required by many ALSA macros */

struct subsdata {
  volatile u32 *CounterRegister;	/* Hw register that contains the current dma position (lower 32 bits only) */
  u32 LastCount;			/* The last position, which is used to calculate... */
  u32 DMACount;				/* ...the number of bytes tranferred by the DMA engine, modulo the size of the buffer */
  volatile int Configured;
  u16 PipeIndex;			/* Index of the channel */
  snd_pcm_hardware_t hw;
  snd_pcm_hw_constraint_list_t constr;
};

static unsigned int ChannelsList[17]={1, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32};


/* will go in /sndmagic.h */
#define chip_t_magic 0xEC0A0D10

char *cardname[]={
  [1]="Darla20",
  [2]="Gina20",
  [3]="Layla20",
  [4]="Darla24",
  [5]="Gina24",
  [6]="Layla24",
  [7]="Mona",
  [8]="Mia",
  [9]="Indigo"
};

struct {
  int frq;
  int bit;
} SampleRates[] = {
  { 8000, SNDRV_PCM_RATE_8000},
  {11025, SNDRV_PCM_RATE_11025},
/*  {12000, SNDRV_PCM_RATE_KNOT},    @ Mia */
  {16000, SNDRV_PCM_RATE_16000},
  {22050, SNDRV_PCM_RATE_22050},
/*  {24000, SNDRV_PCM_RATE_KNOT},    @ Mia */
  {32000, SNDRV_PCM_RATE_32000},
  {44100, SNDRV_PCM_RATE_44100},
  {48000, SNDRV_PCM_RATE_48000},
  {88200, SNDRV_PCM_RATE_88200},
  {96000, SNDRV_PCM_RATE_96000},
/*  {12345, SNDRV_PCM_RATE_CONTINUOUS},    Layla20(8-50KHz) and Layla24(25-100KHz) */
  {0, 0}
};

static struct pci_device_id snd_Echo_ids[] __devinitdata={
  {0x1057, 0x1801, 0xECC0, 0x0010, 0, 0, 0},  /* DSP 56301 Darla20 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0020, 0, 0, 0},  /* DSP 56301 Gina20 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0030, 0, 0, 0},  /* DSP 56301 Layla20 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0031, 0, 0, 0},  /* DSP 56301 Layla20 rev.1 */
  {0x1057, 0x1801, 0xECC0, 0x0040, 0, 0, 0},  /* DSP 56301 Darla24 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0041, 0, 0, 0},  /* DSP 56301 Darla24 rev.1 */
  {0x1057, 0x1801, 0xECC0, 0x0050, 0, 0, 0},  /* DSP 56301 Gina24 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0051, 0, 0, 0},  /* DSP 56301 Gina24 rev.1 */
  {0x1057, 0x3410, 0xECC0, 0x0050, 0, 0, 0},  /* DSP 56361 Gina24 rev.0 */
  {0x1057, 0x3410, 0xECC0, 0x0051, 0, 0, 0},  /* DSP 56361 Gina24 rev.1 */
  {0x1057, 0x3410, 0xECC0, 0x0060, 0, 0, 0},  /* DSP 56361 Layla24 rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0070, 0, 0, 0},  /* DSP 56301 Mona rev.0 */
  {0x1057, 0x1801, 0xECC0, 0x0071, 0, 0, 0},  /* DSP 56301 Mona rev.1 */
  {0x1057, 0x1801, 0xECC0, 0x0072, 0, 0, 0},  /* DSP 56301 Mona rev.2 */
  {0x1057, 0x3410, 0xECC0, 0x0070, 0, 0, 0},  /* DSP 56361 Mona rev.0 */
  {0x1057, 0x3410, 0xECC0, 0x0071, 0, 0, 0},  /* DSP 56361 Mona rev.1 */
  {0x1057, 0x3410, 0xECC0, 0x0072, 0, 0, 0},  /* DSP 56361 Mona rev.2 */
  {0x1057, 0x3410, 0xECC0, 0x0080, 0, 0, 0},  /* DSP 56361 Mia rev.0 */
  {0x1057, 0x3410, 0xECC0, 0x0081, 0, 0, 0},  /* DSP 56361 Mia rev.1 */
/*  {0x1057, 0x3410, 0xECC0, 0x0090, 0, 0, 0},  DSP 56361 Indigo rev.0 */
  {0, }
};

MODULE_DEVICE_TABLE(pci, snd_Echo_ids);

// Module parameters
static int index[SNDRV_CARDS]=SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS]=SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS]=SNDRV_DEFAULT_ENABLE_PNP;






/***************************************************************************************************
*** PCM interface
***************************************************************************************************/

static snd_pcm_hardware_t Skel_hw={
#ifdef CONFIG_PM
  .info=             SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED | SNDRV_PCM_INFO_BLOCK_TRANSFER | SNDRV_PCM_INFO_MMAP_VALID | SNDRV_PCM_INFO_PAUSE | SNDRV_PCM_INFO_RESUME,
#else
  .info=             SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED | SNDRV_PCM_INFO_BLOCK_TRANSFER | SNDRV_PCM_INFO_MMAP_VALID | SNDRV_PCM_INFO_PAUSE,
#endif
  .formats=          SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S32_LE | SNDRV_PCM_FMTBIT_S32_BE,
  .rates=            SNDRV_PCM_RATE_8000_48000,
  .rate_min=          8000,
  .rate_max=         48000,
  .channels_min=         1,
  .channels_max=         2,
  .buffer_bytes_max=262144,
  .period_bytes_min=    32,
  .period_bytes_max=131072,
  .periods_min=          1,
  .periods_max=        200,
  /* periods_max is limited is 500 in the generic driver. POINTER, IRQ and
     GOTO commands use 1 slot each, so 200 is a quite safe limit.  I don't
     know if >512 (one page) is supported by the hardware. */
};



static void ssdata_free(snd_pcm_runtime_t *runtime) {
  kfree(runtime->private_data);
}


/*@ The rules should check if STEREO_BIG_ENDIAN32_SUPPORT is defined. */


static int hw_rule_capture_format_by_channels(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule) {
  snd_interval_t *c=hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
  snd_mask_t *f=hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
  snd_mask_t fmt;

  snd_mask_any(&fmt);

  /* >=2 channels cannot be S32_BE */
  if (c->min==2) {
    fmt.bits[0]&=~SNDRV_PCM_FMTBIT_S32_BE;
    return(snd_mask_refine(f, &fmt));
  }
  /* >2 channels must be S32_LE */
  if (c->min>2) {
    fmt.bits[0]&=SNDRV_PCM_FMTBIT_S32_LE;
    return(snd_mask_refine(f, &fmt));
  }
  /* Mono is ok with any format */
  return(0);
}



static int hw_rule_capture_channels_by_format(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule) {
  snd_interval_t *c=hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
  snd_mask_t *f=hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
  snd_interval_t ch;

  snd_interval_any(&ch);

  /* S32_BE is mono only */
  if (f->bits[0]==SNDRV_PCM_FMTBIT_S32_BE) {
    ch.min=ch.max=1;
    ch.integer=1;
    return(snd_interval_refine(c, &ch));
  }
  /* U8 amd S16_LE can be only mono or stereo */
  if (f->bits[0]==SNDRV_PCM_FMTBIT_U8 || f->bits[0]==SNDRV_PCM_FMTBIT_S16_LE) {
    ch.min=1;
    ch.max=2;
    ch.integer=1;
    return(snd_interval_refine(c, &ch));
  }
  /* S32_LE supports any number of channels. */
  return(0);
}



static int hw_rule_playback_format_by_channels(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule) {
  snd_interval_t *c=hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
  snd_mask_t *f=hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
  snd_mask_t fmt;

  snd_mask_any(&fmt);

  /* >2 channels must be S32_LE */
  if (c->min>2) {
    fmt.bits[0]&=SNDRV_PCM_FMTBIT_S32_LE;
    return(snd_mask_refine(f, &fmt));
  }
  /* 1 channel must be S32_BE or S32_LE */
  if (c->max<2) {
    fmt.bits[0]&=SNDRV_PCM_FMTBIT_S32_LE|SNDRV_PCM_FMTBIT_S32_BE;
    return(snd_mask_refine(f, &fmt));
  }
  /* 2 channels cannot be S32_BE */
  if (c->min==2 && c->max==2) {
    fmt.bits[0]&=~SNDRV_PCM_FMTBIT_S32_BE;
    return(snd_mask_refine(f, &fmt));
  }
  return(0);
}



static int hw_rule_playback_channels_by_format(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule) {
  snd_interval_t *c=hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
  snd_mask_t *f=hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
  snd_interval_t ch;

  snd_interval_any(&ch);

  /* S32_BE is mono only */
  if (f->bits[0]==SNDRV_PCM_FMTBIT_S32_BE) {
    ch.min=ch.max=1;
    ch.integer=1;
    return(snd_interval_refine(c, &ch));
  }
  /* U8 and S16_LE are stereo only */
  if (f->bits[0]==SNDRV_PCM_FMTBIT_U8 || f->bits[0]==SNDRV_PCM_FMTBIT_S16_LE) {
    ch.min=ch.max=2;
    ch.integer=1;
    return(snd_interval_refine(c, &ch));
  }
  /* S32_LE supports any number of channels. */
  return(0);
}



static int pcm_open(snd_pcm_substream_t *substream, int MaxChannels, int adat) {
  struct EchoChip *chip;
  snd_pcm_runtime_t *runtime;
  struct subsdata *ssdata;
  int err, i;

  if (MaxChannels<=0)
    return(-EINVAL);    /*@@ Is -EINVAL correct ? */
/*
  if (!CheckAsicStatus(chip->eg)) {
    DE_HWP(("HW error. Trying to reload the firmware...\n"))
    err = LoadFirmware(chip->eg);
    if (err || chip->eg->BadBoard)
      return -EIO;
  }
*/
  chip=snd_pcm_substream_chip(substream);
  runtime=substream->runtime;

  if (!(ssdata=kmalloc(sizeof(struct subsdata), GFP_KERNEL)))
    return(-ENOMEM);
  ssdata->Configured=0;

  /* Set up hw capabilities and contraints */
  memcpy(&ssdata->hw, &Skel_hw, sizeof(snd_pcm_hardware_t));
  DE_HWP(("MaxChannels=%d\n", MaxChannels))
  ssdata->constr.list=ChannelsList;
  ssdata->constr.mask=0;
  for (i=0; i<17 && ChannelsList[i]<=MaxChannels; i++);
  ssdata->constr.count=i;
  /* Se it only if the card can handle more than 2 channels */
  if (chip->eg->Flags & ECHOGALS_ROFLAG_SUPER_INTERLEAVE_OK)
    ssdata->hw.channels_max=MaxChannels;
  if (adat) {
    ssdata->hw.rate_max=48000;
    ssdata->hw.rates&=SNDRV_PCM_RATE_8000_48000;
  }
  runtime->hw=ssdata->hw;
  runtime->private_data=ssdata;
  runtime->private_free=ssdata_free;

  if ((err=snd_pcm_hw_constraint_list(runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, &ssdata->constr))<0)
    return(err);

  return(0);
}



static int pcmain_open(snd_pcm_substream_t *substream) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  int err;

  DE_ACT(("pcmain_open\n"))
  if ((err=pcm_open(substream, chip->eg->FirstDigitalBusIn-substream->number, 0))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, hw_rule_capture_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT, hw_rule_capture_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1))<0)
    return(err);
  return(0);
}



static int pcmaout_open(snd_pcm_substream_t *substream) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  int err;

  DE_ACT(("pcmaout_open\n"))
  if ((err=pcm_open(substream, chip->eg->FirstDigitalBusOut-substream->number, 0))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, hw_rule_playback_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT, hw_rule_playback_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1))<0)
    return(err);
  return(0);
}



static int pcmdin_open(snd_pcm_substream_t *substream) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  int err, NumDigitalBussesIn;

  DE_ACT(("pcmdin_open\n"))
  NumDigitalBussesIn=chip->eg->NumBussesIn-chip->eg->FirstDigitalBusIn;
  if (chip->eg->DigitalMode==DIGITAL_MODE_ADAT) {
    if ((err=pcm_open(substream, NumDigitalBussesIn-substream->number, 1))<0)
      return(err);
  } else if (chip->eg->DigitalModes&ECHOCAPS_HAS_DIGITAL_MODE_ADAT) {	/* and mode==S/PDIF */
    if ((err=pcm_open(substream, NumDigitalBussesIn-substream->number-6, 0))<0);	/* Subtract the 6 channels that S/PDIF doesn't have */
      return(err);
  } else {		/* It hasn't ADAT */
    if ((err=pcm_open(substream, NumDigitalBussesIn-substream->number, 0))<0)
      return(err);
  }
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, hw_rule_capture_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT, hw_rule_capture_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1))<0)
    return(err);
  return(0);
}



static int pcmdout_open(snd_pcm_substream_t *substream) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  int err, NumDigitalBussesOut;

  DE_ACT(("pcmdout_open\n"))
  NumDigitalBussesOut=chip->eg->NumBussesOut-chip->eg->FirstDigitalBusOut;
  if (chip->eg->DigitalMode==DIGITAL_MODE_ADAT) {
    if ((err=pcm_open(substream, NumDigitalBussesOut-substream->number, 1))<0)
      return(err);
  } else if (chip->eg->DigitalModes&ECHOCAPS_HAS_DIGITAL_MODE_ADAT) {	/* and mode==S/PDIF */
    if ((err=pcm_open(substream, NumDigitalBussesOut-substream->number-6, 0))<0)	/* Subtract the 6 channels that S/PDIF doesn't have */
      return(err);
  } else {		/* It hasn't ADAT */
    if ((err=pcm_open(substream, NumDigitalBussesOut-substream->number, 0))<0)
      return(err);
  }
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, hw_rule_playback_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1))<0)
    return(err);
  if ((err=snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT, hw_rule_playback_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1))<0)
    return(err);
  return(0);
}



static int pcm_close(snd_pcm_substream_t *substream) {
  /* Nothing to do here. Audio is already off and ssdata will be freed by its callback */
  DE_ACT(("pcm_close\n"))
  return(0);
}



/* Opens a new substream and sets up the scatter-gather list */
int InitEngine(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params, u16 Channel, u16 Interleave, char IsInput) {
  struct EchoChip *chip;
  int err, per, rest, page, edge, offs;
  struct snd_sg_buf *sgbuf;
  struct subsdata *ssdata;

  chip=snd_pcm_substream_chip(substream);
  ssdata=(struct subsdata *)substream->runtime->private_data;

  /*@ TODO: The hw preferably accesses memory in chunks 32 frames long. If this condition is accomplished
      the hw works at its maximum efficiency. The code should try to do that, if possible. */

  /* Sets up che hardware. If it's already initialized, reset and redo, bacause the client changed its mind (alsaplayer does it every time hmpf!!). */
  spin_lock(&chip->lock);
  if (ssdata->Configured) {
    DE_HWP(("hwp_closeaudio(%d)\n", ssdata->PipeIndex))
    err=CloseAudio(chip->eg, ssdata->PipeIndex);
    if (err) {          /* should never happen */
      spin_unlock(&chip->lock);
      printk("hwpclose err=0x%x\n", (int)err);
      return(-1);
    }
    chip->substream[ssdata->PipeIndex]=0;
    ssdata->Configured=0;
  }

  err=OpenAudio(chip->eg, Channel, Interleave, IsInput, &ssdata->PipeIndex);
  if (err) {
    spin_unlock(&chip->lock);
    printk("OpenAudio() err=0x%x\n", err);
    return(-EAGAIN);
  }
  DE_HWP(("OpenAudio()=%d\n", ssdata->PipeIndex))

  if ((err=GetAudioPositionPtr(chip->eg, ssdata->PipeIndex, &ssdata->CounterRegister))) {
    printk("pcm_open no getpos err=0x%x\n", err);
    err=-1;
    goto IEcloseandexit;
  }

  DE_HWP(("pcm_hw_params (bufsize=%dB periods=%d persize=%dB)\n", params_buffer_bytes(hw_params), params_periods(hw_params), params_period_bytes(hw_params)))
  err=snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
  if (err<0) {
    printk("malloc_pages err=%d\n", err);
    goto IEcloseandexit;
  }

  chip->substream[ssdata->PipeIndex]=substream;
  chip->LastPeriod[ssdata->PipeIndex]=0;
  ssdata->LastCount=0;
  ssdata->DMACount=0;
  ssdata->Configured=1;
  spin_unlock(&chip->lock);

  sgbuf=snd_pcm_substream_sgbuf(substream);

  DE_HWP(("pcm_hw_params table size=%d pages=%d\n",sgbuf->size,sgbuf->pages))
  edge=PAGE_SIZE;
  for (offs=page=per=0; offs<params_buffer_bytes(hw_params); per++) {
    rest=params_period_bytes(hw_params);
    if (offs+rest>params_buffer_bytes(hw_params))
      rest=params_buffer_bytes(hw_params)-offs;

    while(rest) {
      if (rest<=edge-offs) {
        /*printk("pcm_hw_params Add%d (off=%d pa=%x phy=%x sz=%d rest=%d)i\n",per,offs,sgbuf->table[page].addr,snd_sgbuf_get_addr(sgbuf, offs),rest,rest); */
        SGListAddMapping(chip->eg, ssdata->PipeIndex, snd_sgbuf_get_addr(sgbuf, offs), rest, 1);    /*@ AddMapping() should never fail... */
        offs+=rest;
        rest=0;
      } else {
        /*printk("pcm_hw_params Add%d (off=%d pa=%x phy=%x sz=%d rest=%d)\n",per,offs,sgbuf->table[page].addr,snd_sgbuf_get_addr(sgbuf, offs),edge-offs,rest); */
        SGListAddMapping(chip->eg, ssdata->PipeIndex, snd_sgbuf_get_addr(sgbuf, offs), edge-offs, 0);
        rest-=edge-offs;
        offs=edge;
      }
      if (offs==edge) {
        edge+=PAGE_SIZE;
        page++;
      }
    }
  }
/*printk("pcm_hw_params End (off=%d rest=%d)\n",offs,rest); */

  /* Close the ring buffer */
  SGListWrap(chip->eg, ssdata->PipeIndex);

#if (0)
  /* dump sg list */
  {
    sg_entry *list=chip->eg->Pipes[ssdata->PipeIndex].sg_entry;

    for (per=0; per<80; per++) {
      printk("sglist[%d]=%x, %d\n", per, le32_to_cpu(list[per].PhysAddr), le32_to_cpu(chip->eg->Pipes[ssdata->PipeIndex].gs_entry[per].Size));
    }
  }
#endif

  DE_HWP(("pcm_hw_params ok\n"))
  return(0);

IEcloseandexit:
  CloseAudio(chip->eg, ssdata->PipeIndex);
  spin_unlock(&chip->lock);
  return(err);
}



static int pcmain_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params) {

  return(InitEngine(substream, hw_params, substream->number, params_channels(hw_params), 1));
}



static int pcmaout_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params) {

  return(InitEngine(substream, hw_params, substream->number, params_channels(hw_params), 0));
}



static int pcmdin_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);

  return(InitEngine(substream, hw_params, chip->eg->FirstDigitalBusIn+substream->number, params_channels(hw_params), 1));
}



static int pcmdout_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);

  return(InitEngine(substream, hw_params, chip->eg->FirstDigitalBusOut+substream->number, params_channels(hw_params), 0));
}



static int pcm_hw_free(snd_pcm_substream_t *substream) {
  struct EchoChip *chip;
  struct subsdata *ssdata;
  int st;

  chip=snd_pcm_substream_chip(substream);
  ssdata=(struct subsdata *)substream->runtime->private_data;

  spin_lock(&chip->lock);
  if (ssdata->Configured) {
    st=CloseAudio(chip->eg, ssdata->PipeIndex);
    if (st) {
      spin_unlock(&chip->lock);
      printk("no close err=0x%x\n", (int)st);
      return(-1);
    }
    DE_HWP(("pcm_hw_free closeaudio(%d)\n", ssdata->PipeIndex))
    chip->substream[ssdata->PipeIndex]=0;
    ssdata->Configured=0;
  }
  spin_unlock(&chip->lock);

  DE_HWP(("pcm_hw_freed\n"))
  snd_pcm_lib_free_pages(substream);
  return(0);
}



static int pcm_prepare(snd_pcm_substream_t *substream) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  snd_pcm_runtime_t *runtime=substream->runtime;
  Audioformat_t format;
  int st;

  DE_HWP(("Prepare rate=%d format=%d channels=%d\n", runtime->rate, runtime->format, runtime->channels))
  SetAudioSampleRate(chip->eg, runtime->rate);
  format.DataInterleave=runtime->channels;
  format.DataAreBigEndian=0;
  format.MonoToStereo=0;
  switch (runtime->format) {
    default:
      goto Preparefailed;
    case SNDRV_PCM_FORMAT_U8:
      format.BitsPerSample=8;
      break;
    case SNDRV_PCM_FORMAT_S16_LE:
      format.BitsPerSample=16;
      break;
    case SNDRV_PCM_FORMAT_S32_BE:
      format.DataAreBigEndian=1;
    case SNDRV_PCM_FORMAT_S32_LE:
      format.BitsPerSample=32;
      break;
  }
  if ((st=SetAudioFormat(chip->eg, ((struct subsdata *)runtime->private_data)->PipeIndex, &format))) {
    printk("Prepare: setformat error 0x%x\n", (int)st);
    return(-EINVAL);
  }
  DE_HWP(("Prepare ok\n"))
  return(0);
Preparefailed:
  printk("Prepare error: unsupported format %d\n", runtime->format);
  return(-EINVAL);
}



static int pcm_trigger(snd_pcm_substream_t *substream, int cmd) {
  struct EchoChip *chip=snd_pcm_substream_chip(substream);
  snd_pcm_runtime_t *runtime=substream->runtime;
  struct subsdata *ssdata=runtime->private_data;

  switch(cmd) {
#ifdef CONFIG_PM
    case SNDRV_PCM_TRIGGER_SUSPEND:
      DE_ACT(("pcm_trigger suspend: %d\n", chip->card->power_state))
      if (chip->card->power_state!=SNDRV_CTL_POWER_D0) {
        PowerDown(chip->eg);
        snd_power_change_state(chip->card, SNDRV_CTL_POWER_D2); /* @@ D2 ? */
      }
      break;
    case SNDRV_PCM_TRIGGER_RESUME:
      DE_ACT(("pcm_trigger resume\n"))
      if (chip->card->power_state!=SNDRV_CTL_POWER_D0) {
        PowerUp(chip->eg);
        snd_power_change_state(chip->card, SNDRV_CTL_POWER_D0);
      }
      /* no break */
#endif
    case SNDRV_PCM_TRIGGER_START:
    case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
      DE_ACT(("pcm_trigger start\n"))
      StartAudio(chip->eg, ssdata->PipeIndex);
      break;
    case SNDRV_PCM_TRIGGER_STOP:
      DE_ACT(("pcm_trigger stop\n"))
      ResetAudio(chip->eg, ssdata->PipeIndex);
      chip->LastPeriod[ssdata->PipeIndex]=0;
      ssdata->LastCount=0;
      ssdata->DMACount=0;
      break;
    case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
      DE_ACT(("pcm_trigger pause\n"))
      StopAudio(chip->eg, ssdata->PipeIndex);
      break;
    default:
      return(-EINVAL);
  }
  return(0);
}



static snd_pcm_uframes_t pcm_pointer(snd_pcm_substream_t *substream) {
  snd_pcm_runtime_t *runtime=substream->runtime;
  struct subsdata *ssdata=runtime->private_data;
  size_t Cnt, bufsize, pos;

  Cnt=le32_to_cpu(*ssdata->CounterRegister);
  ssdata->DMACount+=Cnt-ssdata->LastCount;
  ssdata->LastCount=Cnt;
  bufsize=substream->runtime->buffer_size;
  pos=bytes_to_frames(substream->runtime, ssdata->DMACount);
  if (pos>bufsize) {
    ssdata->DMACount-=frames_to_bytes(substream->runtime, bufsize);
    pos-=bufsize;
  }
/*printk("pcm_hw_pointer=%d fr=%d sz=%d dmc=%d\n",Cnt,pos,bufsize,ssdata->DMACount); */
  return(pos);
}




/* pcm *_ops structures */
static snd_pcm_ops_t AnPlayback_ops = {
  .open=pcmaout_open,
  .close=pcm_close,
  .ioctl=snd_pcm_lib_ioctl,
  .hw_params=pcmaout_hw_params,
  .hw_free=pcm_hw_free,
  .prepare=pcm_prepare,
  .trigger=pcm_trigger,
  .pointer=pcm_pointer,
  .page=snd_pcm_sgbuf_ops_page,
};
static snd_pcm_ops_t AnCapture_ops = {
  .open=pcmain_open,
  .close=pcm_close,
  .ioctl=snd_pcm_lib_ioctl,
  .hw_params=pcmain_hw_params,
  .hw_free=pcm_hw_free,
  .prepare=pcm_prepare,
  .trigger=pcm_trigger,
  .pointer=pcm_pointer,
  .page=snd_pcm_sgbuf_ops_page,
};
static snd_pcm_ops_t DiPlayback_ops = {
  .open=pcmdout_open,
  .close=pcm_close,
  .ioctl=snd_pcm_lib_ioctl,
  .hw_params=pcmdout_hw_params,
  .hw_free=pcm_hw_free,
  .prepare=pcm_prepare,
  .trigger=pcm_trigger,
  .pointer=pcm_pointer,
  .page=snd_pcm_sgbuf_ops_page,
};
static snd_pcm_ops_t DiCapture_ops = {
  .open=pcmdin_open,
  .close=pcm_close,
  .ioctl=snd_pcm_lib_ioctl,
  .hw_params=pcmdin_hw_params,
  .hw_free=pcm_hw_free,
  .prepare=pcm_prepare,
  .trigger=pcm_trigger,
  .pointer=pcm_pointer,
  .page=snd_pcm_sgbuf_ops_page,
};



/*<--snd_Echo_probe */
static int __devinit snd_Echo_new_pcm(struct EchoChip *chip, int HasDigitalPcm) {
  snd_pcm_t *pcm;
  int err;

  if (chip->eg->HasVmixer) {
    /* This card has a Vmixer, that is there is no direct mapping from PCM streams to physical outputs.
       The user is allowed to mix the streams as he wishes via control interface, so it makes no sense
       to have separate analog/digital outputs */

    /* PCM#0 Virtual outputs and analog inputs */
    if ((err=snd_pcm_new(chip->card, "Analog PCM", 0, chip->eg->NumPipesOut, chip->eg->FirstDigitalBusIn, &pcm))<0)
      return(err);
    pcm->private_data=chip;
    chip->AnPcm=pcm;
    strcpy(pcm->name, chip->card->shortname);
    snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &AnPlayback_ops);
    snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &AnCapture_ops);
    if ((err=snd_pcm_lib_preallocate_sg_pages_for_all(chip->pci, pcm, 128*1024, 256*1024))<0)
      return err;
    DE_INIT(("Analog PCM ok\n"))

    /* PCM#1 Digital inputs, no outputs */
    if (chip->eg->FirstDigitalBusIn!=chip->eg->NumBussesIn) {     /* If is has digital inputs */
      if ((err=snd_pcm_new(chip->card, "Digital PCM", 1, 0, chip->eg->NumBussesIn-chip->eg->FirstDigitalBusIn, &pcm))<0)
        return(err);
      pcm->private_data=chip;
      chip->DiPcm=pcm;
      strcpy(pcm->name, chip->card->shortname);
      snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &DiCapture_ops);
      if ((err=snd_pcm_lib_preallocate_sg_pages_for_all(chip->pci, pcm, 128*1024, 256*1024))<0)
        return err;
      DE_INIT(("Digital PCM ok\n"))
    }
  } else {
    /* The card can manage substreams composed by analog and digital channels at the
       same time, but I prefer to keep analog and digital channels separated, because
       that mixed thing is confusing and useless. So we register two PCM devices: */

    /* PCM#0 Analog i/o   (@fails if the card has no analog channels) */
    if ((err=snd_pcm_new(chip->card, "Analog PCM", 0, chip->eg->FirstDigitalBusOut, chip->eg->FirstDigitalBusIn, &pcm))<0)
      return(err);
    pcm->private_data=chip;
    chip->AnPcm=pcm;
    strcpy(pcm->name, chip->card->shortname);
    snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &AnPlayback_ops);
    snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &AnCapture_ops);
    if ((err=snd_pcm_lib_preallocate_sg_pages_for_all(chip->pci, pcm, 128*1024, 256*1024))<0)
      return err;
    DE_INIT(("Analog PCM ok\n"))

    /* PCM#1 Digital i/o */
    if (chip->eg->FirstDigitalBusOut!=chip->eg->NumBussesOut) {   /* If is has digital channels */
      if ((err=snd_pcm_new(chip->card, "Digital PCM", 1, chip->eg->NumBussesOut-chip->eg->FirstDigitalBusOut, chip->eg->NumBussesIn-chip->eg->FirstDigitalBusIn, &pcm))<0)
        return(err);
      pcm->private_data=chip;
      chip->DiPcm=pcm;
      strcpy(pcm->name, chip->card->shortname);
      snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &DiPlayback_ops);
      snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &DiCapture_ops);
      if ((err=snd_pcm_lib_preallocate_sg_pages_for_all(chip->pci, pcm, 128*1024, 256*1024))<0)
        return err;
      DE_INIT(("Digital PCM ok\n"))
    }
  }

  return(0);
}




/***************************************************************************************************
*** MIDI interface
***************************************************************************************************/

#ifdef EMIDI

static int snd_Echo_midi_input_open(snd_rawmidi_substream_t *substream) {
  struct EchoChip *chip=substream->rmidi->private_data;
  unsigned long flags;

  spin_lock_irqsave(&chip->Midilock, flags);
  chip->MidiIn=substream;
  spin_unlock_irqrestore(&chip->Midilock, flags);
printk("rawmidi_iopen\n");
  return(0);
}



static void snd_Echo_midi_input_trigger(snd_rawmidi_substream_t *substream, int up) {
  struct EchoChip *chip=substream->rmidi->private_data;
  unsigned long flags;
  int st;

  spin_lock_irqsave(&chip->Midilock, flags);
  if (up)
    st=OpenMidiInput(chip->eg);
  else
    st=CloseMidiInput(chip->eg);
  spin_unlock_irqrestore(&chip->Midilock, flags);
printk("rawmidi_itrigger(%d)=%d\n", up, st);
}



/* <-- snd_Echo_interrupt() */
static int snd_Echo_midi_input_read(struct EchoChip *chip) {
  unsigned long flags;
  int n;

  spin_lock_irqsave(&chip->Midilock, flags);    /*@@not sure if this is needed */
  if (chip->MidiIn) {
    n=0;
    while(ReadMidiByte(chip->eg, &chip->Midibuf[n])==ECHOSTATUS_OK)  n++;
    /*ECHOSTATUS_NO_DATA */
    if (n)
      snd_rawmidi_receive(chip->MidiIn, chip->Midibuf, n);
printk("rawmidi_read=%d\n", n);
  }
  spin_unlock_irqrestore(&chip->Midilock, flags);
  return(0);
}



static int snd_Echo_midi_input_close(snd_rawmidi_substream_t *substream) {
  struct EchoChip *chip=substream->rmidi->private_data;
  unsigned long flags;

  spin_lock_irqsave(&chip->Midilock, flags);
  chip->MidiIn=0;
  spin_unlock_irqrestore(&chip->Midilock, flags);
printk("rawmidi_iclose\n");
  return(0);
}



static int snd_Echo_midi_output_open(snd_rawmidi_substream_t *substream) {
  struct EchoChip *chip=substream->rmidi->private_data;

  chip->MidiOut=substream;
printk("rawmidi_oopen\n");
  return(0);
}



static void snd_Echo_midi_output_write(unsigned long data) {
  struct EchoChip *chip=(struct EchoChip *)data;
  unsigned long flags;
  int nbytes, st;
  unsigned char buf[16];
  DWORD sent;

  /* No interrupts are involved: I have to check at regular intervals if the card's buffer has room for new data. */
  sent=nbytes=0;
  spin_lock_irqsave(&chip->Midilock, flags);
  if (chip->MidiOut) {
    if (!snd_rawmidi_transmit_empty(chip->MidiOut)) {
      nbytes=snd_rawmidi_transmit_peek(chip->MidiOut, buf, 16);
      if (nbytes>0) {
        st=WriteMidi(chip->eg, nbytes, buf, &sent);
        snd_rawmidi_transmit_ack(chip->MidiOut, sent);
      }
    }
  }
  /* The kernel removes the timer when it expires. */
  if (chip->tinuse) {
    chip->timer.expires=((HZ+199)/200)+jiffies;
    add_timer(&chip->timer);
  }

  spin_unlock_irqrestore(&chip->Midilock, flags);
printk("rawmidi_owrite(%d)=%d\n", nbytes ,(int)sent);
}



static void snd_Echo_midi_output_trigger(snd_rawmidi_substream_t *substream, int up) {
  struct EchoChip *chip=substream->rmidi->private_data;
  unsigned long flags;

  spin_lock_irqsave(&chip->Midilock, flags);
  if (up) {
    if (!chip->tinuse) {
      init_timer(&chip->timer);
      chip->timer.function=snd_Echo_midi_output_write;
      chip->timer.data=(unsigned long)chip;
      chip->timer.expires=((HZ+199)/200)+jiffies;       /* 1/200s == 5ms */
      add_timer(&chip->timer);
      chip->tinuse=1;
    }
  } else {
    del_timer(&chip->timer);
    chip->tinuse=0;
  }
  spin_unlock_irqrestore(&chip->Midilock, flags);

  if (up)
    snd_Echo_midi_output_write((unsigned long)chip);
printk("rawmidi_otrigger\n");
}



static int snd_Echo_midi_output_close(snd_rawmidi_substream_t *substream) {
  struct EchoChip *chip=substream->rmidi->private_data;

  chip->MidiOut=0;
printk("rawmidi_oclose\n");
  return(0);
}



snd_rawmidi_ops_t snd_Echo_midi_input = {
  .open=snd_Echo_midi_input_open,
  .close=snd_Echo_midi_input_close,
  .trigger=snd_Echo_midi_input_trigger,
};

snd_rawmidi_ops_t snd_Echo_midi_output = {
  .open=snd_Echo_midi_output_open,
  .close=snd_Echo_midi_output_close,
  .trigger=snd_Echo_midi_output_trigger,
};



static int __devinit snd_Echo_new_midi(snd_card_t *card, struct EchoChip *chip) {
  char buf[32];

  spin_lock_init(&chip->Midilock);

  sprintf(buf, "%s MIDI 0", card->shortname);
  if (snd_rawmidi_new(card, buf, 0, 1, 1, &chip->rmidi) < 0) {
    return(-1);
  }

  sprintf (chip->rmidi->name, "%s MIDI 0", card->id);
  chip->rmidi->private_data=&chip;

  snd_rawmidi_set_ops(chip->rmidi, SNDRV_RAWMIDI_STREAM_INPUT, &snd_Echo_midi_input);
  snd_rawmidi_set_ops(chip->rmidi, SNDRV_RAWMIDI_STREAM_OUTPUT, &snd_Echo_midi_output);

  chip->rmidi->info_flags |= SNDRV_RAWMIDI_INFO_OUTPUT | SNDRV_RAWMIDI_INFO_INPUT | SNDRV_RAWMIDI_INFO_DUPLEX;
  DE_INIT(("MIDI ok\n"))
  return(0);
}


#endif




/***************************************************************************************************
*** Control interface
***************************************************************************************************/

/******************* PCM output volume *******************/

static int pcmout_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=chip->eg->NumBussesOut;
  uinfo->value.integer.min=ECHOGAIN_MINOUT;
  uinfo->value.integer.max=ECHOGAIN_MAXOUT;
  return(0);
}

static int pcmout_getvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c;

  chip=snd_kcontrol_chip(kcontrol);
  for (c=chip->eg->NumBussesOut-1; c>=0; c--)
    ucontrol->value.integer.value[c]=chip->eg->OutLevel[c];
  return(0);
}

static int pcmout_setvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c, changed;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  for (c=chip->eg->NumBussesOut-1; c>=0; c--) {
    if (chip->eg->OutLevel[c]!=ucontrol->value.integer.value[c]) {
      changed=1;
      SetOutputGain(chip->eg, c, ucontrol->value.integer.value[c]);
    }
  }
  if (changed)
    UpdateOutputLineLevel(chip->eg);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcmout_volume_control __devinitdata = {
  .name="PCM Playback Volume",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .info=pcmout_volume_info,
  .get=pcmout_getvolume,
  .put=pcmout_setvolume,
};

/* On Vmixer cards this one controls the line-out volume */
static snd_kcontrol_new_t line_volume_control __devinitdata = {
  .name="Line Playback Volume",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .info=pcmout_volume_info,
  .get=pcmout_getvolume,
  .put=pcmout_setvolume,
};



/******************* Analog output nominal level (+4dBu / -10dBV) *******************/

static int pcmout_nominal_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_BOOLEAN;
  uinfo->count=chip->eg->FirstDigitalBusOut;
  uinfo->value.integer.min=0;
  uinfo->value.integer.max=1;
  return(0);
}

static int pcmout_getnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c;

  chip=snd_kcontrol_chip(kcontrol);
  for (c=chip->eg->FirstDigitalBusOut-1; c>=0; c--)
    ucontrol->value.integer.value[c]=chip->eg->NominalLevel[c];
  return(0);
}

static int pcmout_setnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c, changed, st;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  for (c=chip->eg->FirstDigitalBusOut-1; c>=0; c--) {
    if (chip->eg->NominalLevel[c]!=ucontrol->value.integer.value[c]) {
      changed=1;
      SetNominalLevel(chip->eg, c, ucontrol->value.integer.value[c]);
    }
  }
  if (changed)
    UpdateOutputLineLevel(chip->eg);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcmout_nominal_control __devinitdata = {
  .name="Line Playback Switch (-10dBV)",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .info=pcmout_nominal_info,
  .get=pcmout_getnominal,
  .put=pcmout_setnominal,
};



/******************* Analog input volume *******************/

static int pcmin_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=chip->eg->FirstDigitalBusIn;
  uinfo->value.integer.min=ECHOGAIN_MININP;
  uinfo->value.integer.max=ECHOGAIN_MAXINP;
  return(0);
}

static int pcmin_getvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c;

  chip=snd_kcontrol_chip(kcontrol);
  for (c=chip->eg->FirstDigitalBusIn-1; c>=0; c--)
    ucontrol->value.integer.value[c]=chip->eg->InLevel[c];
  return(0);
}

static int pcmin_setvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c, changed, st;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  for (c=chip->eg->FirstDigitalBusIn-1; c>=0; c--) {
    if (chip->eg->InLevel[c]!=ucontrol->value.integer.value[c]) {
      changed=1;
      SetInputGain(chip->eg, c, ucontrol->value.integer.value[c]);
    }
  }
  if (changed)
    UpdateInputLineLevel(chip->eg);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcmin_volume_control __devinitdata = {
  .name="Line Capture Volume",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .info=pcmin_volume_info,
  .get=pcmin_getvolume,
  .put=pcmin_setvolume,
};



/******************* Analog input nominal level (+4dBu / -10dBV) *******************/

static int pcmin_nominal_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_BOOLEAN;
  uinfo->count=chip->eg->FirstDigitalBusIn;
  uinfo->value.integer.min=0;
  uinfo->value.integer.max=1;
  return(0);
}

static int pcmin_getnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c;

  chip=snd_kcontrol_chip(kcontrol);
  for (c=chip->eg->FirstDigitalBusIn-1; c>=0; c--)
    ucontrol->value.integer.value[c]=chip->eg->NominalLevel[c+chip->eg->NumBussesOut];
  return(0);
}

static int pcmin_setnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int c, changed, st;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  for (c=chip->eg->FirstDigitalBusIn-1; c>=0; c--) {
    if (chip->eg->NominalLevel[c+chip->eg->NumBussesOut]!=ucontrol->value.integer.value[c]) {
      changed=1;
      SetNominalLevel(chip->eg, c+chip->eg->NumBussesOut, ucontrol->value.integer.value[c]);
    }
  }
  if (changed)
    UpdateOutputLineLevel(chip->eg);    /* "Output" is not a mistake here. */
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcmin_nominal_control __devinitdata = {
  .name="Line Capture Switch (-10dBV)",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .info=pcmin_nominal_info,
  .get=pcmin_getnominal,
  .put=pcmin_setnominal,
};



/******************* Monitor mixer *******************/

static int pcm_mixervolume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=1;
  uinfo->value.integer.min=ECHOGAIN_MINOUT;
  uinfo->value.integer.max=ECHOGAIN_MAXOUT;
  uinfo->dimen.d[0]=chip->eg->NumBussesOut;
  uinfo->dimen.d[1]=chip->eg->NumBussesIn;
  return(0);
}

static int pcm_mixergetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  ucontrol->value.integer.value[0]=chip->eg->MonitorLevel[ucontrol->id.index/chip->eg->NumBussesIn][ucontrol->id.index%chip->eg->NumBussesIn];
  return(0);
}

static int pcm_mixersetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int changed, st;
  u16 out, in;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  out=ucontrol->id.index/chip->eg->NumBussesIn;
  in=ucontrol->id.index%chip->eg->NumBussesIn;
  spin_lock_irqsave(&chip->lock, flags);
  if (chip->eg->MonitorLevel[out][in]!=ucontrol->value.integer.value[0]) {
    SetMonitorGain(chip->eg, out, in, ucontrol->value.integer.value[0]);
    changed=1;
  }
  if (changed)
    UpdateOutputLineLevel(chip->eg);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcm_mixervolume_control __devinitdata = {
  .name="Monitor Mixer Volume",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .count=1,
  .info=pcm_mixervolume_info,
  .get=pcm_mixergetvolume,
  .put=pcm_mixersetvolume,
};



/******************* Vmixer *******************/

static int pcm_vmixervolume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=1;
  uinfo->value.integer.min=ECHOGAIN_MINOUT;
  uinfo->value.integer.max=ECHOGAIN_MAXOUT;
  uinfo->dimen.d[0]=chip->eg->NumBussesOut;
  uinfo->dimen.d[1]=chip->eg->NumPipesOut;
  return(0);
}

static int pcm_vmixergetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  ucontrol->value.integer.value[0]=chip->eg->VmixerLevel[ucontrol->id.index/chip->eg->NumPipesOut][ucontrol->id.index%chip->eg->NumPipesOut];
  return(0);
}

static int pcm_vmixersetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int changed, st;
  u16 vch, out;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  out=ucontrol->id.index/chip->eg->NumPipesOut;
  vch=ucontrol->id.index%chip->eg->NumPipesOut;
  spin_lock_irqsave(&chip->lock, flags);
  if (chip->eg->VmixerLevel[out][vch]!=ucontrol->value.integer.value[0]) {
    SetVmixerGain(chip->eg, out, vch, ucontrol->value.integer.value[0]);
    changed=1;
  }
  if (changed)
    UpdateOutputLineLevel(chip->eg);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcm_vmixervolume_control __devinitdata = {
  .name="VMixer Volume",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .count=1,
  .info=pcm_vmixervolume_info,
  .get=pcm_vmixergetvolume,
  .put=pcm_vmixersetvolume,
};



/******************* Digital mode switch *******************/

static int pcm_digitalmode_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  static char *DigitalModeName[4]={"S/PDIF Coaxial", "S/PDIF Optical", "ADAT Optical", "S/PDIF Cdrom"};
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_ENUMERATED;
  uinfo->value.enumerated.items=chip->DigitalModes;
  uinfo->count=1;
  if (uinfo->value.enumerated.item>=chip->DigitalModes)
    uinfo->value.enumerated.item=chip->DigitalModes-1;
  strcpy(uinfo->value.enumerated.name, DigitalModeName[chip->DigitalModeIndex[uinfo->value.enumerated.item]]);
  return(0);
}

static int pcm_getdigitalmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int i, mode;

  chip=snd_kcontrol_chip(kcontrol);
  mode=chip->eg->DigitalMode;
  for (i=chip->DigitalModes-1; i>=0; i--)
    if (mode==chip->DigitalModeIndex[i]) {
      ucontrol->value.enumerated.item[0]=i;
      break;
    }
  return(0);
}

static int pcm_setdigitalmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int changed, mode, st;
  unsigned long flags;

  changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  if ((mode=chip->DigitalModeIndex[ucontrol->value.enumerated.item[0]]) != chip->eg->DigitalMode) {
    changed=1;
    st=SetDigitalMode(chip->eg, mode);
    if (st) {
      printk("setmode val%d err 0x%x\n", mode, (int)st);
      changed = -EAGAIN;	/* Usually it occurs when the user tries to change digital mode while audio is open() */
    }
  }
  spin_unlock_irqrestore(&chip->lock, flags);
  return(changed);
}

static snd_kcontrol_new_t pcmin_digitalmode_control __devinitdata = {
  .name="Digital mode Switch",
  .iface=SNDRV_CTL_ELEM_IFACE_CARD,
  .info=pcm_digitalmode_info,
  .get=pcm_getdigitalmode,
  .put=pcm_setdigitalmode,
};



/******************* S/PDIF mode switch *******************/

static int pcm_spdifmode_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  static char *Modes[2]={"Consumer", "Professional"};

  uinfo->type=SNDRV_CTL_ELEM_TYPE_ENUMERATED;
  uinfo->value.enumerated.items=2;
  uinfo->count=1;
  if (uinfo->value.enumerated.item)
    uinfo->value.enumerated.item=1;
  strcpy(uinfo->value.enumerated.name, Modes[uinfo->value.enumerated.item]);
  return(0);
}

static int pcm_getspdifmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  ucontrol->value.enumerated.item[0]=chip->eg->ProfessionalSpdif;
  return(0);
}

static int pcm_setspdifmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int mode;
  unsigned long flags;

  chip=snd_kcontrol_chip(kcontrol);
  mode=ucontrol->value.enumerated.item[0];
  if (mode!=chip->eg->ProfessionalSpdif) {
    spin_lock_irqsave(&chip->lock, flags);
    SetProfessionalSpdif(chip->eg, mode);
    spin_unlock_irqrestore(&chip->lock, flags);
    return(1);
  }
  return(0);
}

static snd_kcontrol_new_t pcmin_spdifmode_control __devinitdata = {
  .name="S/PDIF mode Switch",
  .iface=SNDRV_CTL_ELEM_IFACE_CARD,
  .info=pcm_spdifmode_info,
  .get=pcm_getspdifmode,
  .put=pcm_setspdifmode,
};



/******************* Select input clock source *******************/

static int pcm_inclock_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  static char *Source[7]={"Internal", "Word", "Super", "S/PDIF", "ADAT", "ESync", "ESync96"};
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_ENUMERATED;
  uinfo->value.enumerated.items=chip->ClockSources;
  uinfo->count=1;
  if (uinfo->value.enumerated.item>chip->ClockSources)
    uinfo->value.enumerated.item=chip->ClockSources;
  strcpy(uinfo->value.enumerated.name, Source[chip->ClockSourceIndex[uinfo->value.enumerated.item]]);
  return(0);
}

static int pcm_getinclock(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int i, clock;

  chip=snd_kcontrol_chip(kcontrol);
  clock=chip->eg->InputClock;
  for (i=0; i<chip->ClockSources; i++)
    if (clock==chip->ClockSourceIndex[i])
      ucontrol->value.enumerated.item[0]=i;
  return(0);
}

static int pcm_setinclock(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  int changed, clock, st;
  unsigned long flags;

  st=changed=0;
  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  clock=chip->ClockSourceIndex[ucontrol->value.enumerated.item[0]];
  if (chip->eg->InputClock!=clock) {
    changed=1;
    st=SetInputClock(chip->eg, clock);
  }
  spin_unlock_irqrestore(&chip->lock, flags);

  if (st) {
    printk("seticlk val%d err 0x%x\n", clock, (int)st);
    switch (st) {
      case ECHOSTATUS_CLOCK_NOT_SUPPORTED:      /* Not supported by the card */
        changed = -EINVAL;
      case ECHOSTATUS_CLOCK_NOT_AVAILABLE:      /* This clock mode is not compatible with the current digital mode */
        changed = -EAGAIN;
      default:
        changed = -1;
    }
  }
  return(changed);
}

static snd_kcontrol_new_t pcm_inclock_control __devinitdata = {
  .name="Sample Clock Source",
  .iface=SNDRV_CTL_ELEM_IFACE_PCM,
  .info=pcm_inclock_info,
  .get=pcm_getinclock,
  .put=pcm_setinclock,
};



/******************* 16 bits dithering switch *******************/

static int ditherswitch_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_BOOLEAN;
  uinfo->count=1;
  uinfo->value.integer.min=0;
  uinfo->value.integer.max=1;
  return(0);
}

static int getditherswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  ucontrol->value.integer.value[0]=!(chip->eg->Flags&ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
  return(0);
}

// @@ This is a per-substream bit, not a global one
static int setditherswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);

  if ((chip->eg->Flags&ECHOGALS_FLAG_DIGITAL_IN_NODITHER) == ucontrol->value.integer.value[0]) {
    /* When value==1 we *clear* the bit */
    spin_lock(&chip->lock);
    if (ucontrol->value.integer.value[0])
      ClearFlags(chip->eg, ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
    else
      SetFlags(chip->eg, ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
    return(1);
    spin_unlock(&chip->lock);
  }
  return(0);
}

static snd_kcontrol_new_t ditherswitch_control __devinitdata = {
  .name="Dithering Switch",
  .iface=SNDRV_CTL_ELEM_IFACE_CARD,
  .info=ditherswitch_info,
  .get=getditherswitch,
  .put=setditherswitch,
};



/******************* VU-meters on/off switch *******************/

static int vumeterswitch_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_BOOLEAN;
  uinfo->count=1;
  uinfo->value.integer.min=0;
  uinfo->value.integer.max=1;
  return(0);
}

static int setvumeterswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;
  unsigned long flags;

  chip=snd_kcontrol_chip(kcontrol);
  spin_lock_irqsave(&chip->lock, flags);
  SetMetersOn(chip->eg, ucontrol->value.integer.value[0]);
  spin_unlock_irqrestore(&chip->lock, flags);
  return(1);
}

static snd_kcontrol_new_t vumeterswitch_control __devinitdata = {
  .name="VU-meters Switch",
  .iface=SNDRV_CTL_ELEM_IFACE_CARD,
  .access=SNDRV_CTL_ELEM_ACCESS_WRITE,
  .info=vumeterswitch_info,
  .put=setvumeterswitch,
};



/******************* Read VU-meters (input, output, analog and digital together) *******************/

static int vumeter_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=96;
  uinfo->value.integer.min=ECHOGAIN_MINOUT;
  uinfo->value.integer.max=0;
  uinfo->dimen.d[0]=chip->eg->HasVmixer ? 3 : 2;     /* Vmixer cards also have PCM levels */
  uinfo->dimen.d[1]=16;                                 /* Index */
  uinfo->dimen.d[2]=2;                                  /* 0=VU, 1=peak */
  return(0);
}

static int getvumeters(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  GetAudioMeters(chip->eg, ucontrol->value.integer.value);
  return(0);
}

static snd_kcontrol_new_t vumeters_control __devinitdata = {
  .name="VU-meters",
  .iface=SNDRV_CTL_ELEM_IFACE_MIXER,
  .access=SNDRV_CTL_ELEM_ACCESS_READ|SNDRV_CTL_ELEM_ACCESS_VOLATILE,
  .info=vumeter_info,
  .get=getvumeters,
};



/******************* Channels info - it exports informations about the number of channels *******************/

static int channelsinfo_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  uinfo->type=SNDRV_CTL_ELEM_TYPE_INTEGER;
  uinfo->count=5;/*+chip->eg->NumBussesIn+chip->eg->NumBussesOut; */
  uinfo->value.integer.min=0;
  uinfo->value.integer.max=1024;
  return(0);
}

static int getchannelsinfo(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
  struct EchoChip *chip;

  chip=snd_kcontrol_chip(kcontrol);
  ucontrol->value.integer.value[0]=chip->eg->NumBussesIn;
  ucontrol->value.integer.value[1]=chip->eg->FirstDigitalBusIn;
  ucontrol->value.integer.value[2]=chip->eg->NumBussesOut;
  ucontrol->value.integer.value[3]=chip->eg->FirstDigitalBusOut;
  ucontrol->value.integer.value[4]=chip->eg->NumPipesOut;
/*  for (c=4+chip->eg->NumBussesIn+chip->eg->NumBussesOut-1; c<=4 ; c--)
    ucontrol->value.integer.value[c]=0;*/
  return(0);
}

static snd_kcontrol_new_t channelsinfo_control __devinitdata = {
  .name="Channels info",
  .iface=SNDRV_CTL_ELEM_IFACE_HWDEP,
  .access=SNDRV_CTL_ELEM_ACCESS_READ,/*|SNDRV_CTL_ELEM_ACCESS_VOLATILE, */
  .info=channelsinfo_info,
  .get=getchannelsinfo,
};




/***************************************************************************************************
*** IRQ Handler
***************************************************************************************************/

static irqreturn_t snd_Echo_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
  struct EchoChip *chip;
  snd_pcm_substream_t *substream;
  int MidiReceived, per, ss;

  chip=snd_magic_cast(chip_t, dev_id, return IRQ_NONE);

  if (ServiceIrq(chip->eg, &MidiReceived)==ECHOSTATUS_OK) {
    for (ss=0; ss<DSP_MAXPIPES; ss++) {         /* The hardware doesn't tell me which substream caused the irq, so I have to check all running substreams. */
      if ((substream=chip->substream[ss])) {    /*@@ Have I to lock ->substream ? */
        per=pcm_pointer(substream)/substream->runtime->period_size;
        if (per!=chip->LastPeriod[ss]) {
          chip->LastPeriod[ss]=per;
          snd_pcm_period_elapsed(chip->substream[ss]);
        }
      }
    }
#ifdef EMIDI
    snd_Echo_midi_input_read(chip);
#endif
    return IRQ_HANDLED;
  }
  return IRQ_NONE;
}




/***************************************************************************************************
*** Module construction / destruction
***************************************************************************************************/

static int snd_Echo_free(struct EchoChip *chip) {

  DE_INIT(("DelEG...\n"))
  if (chip->eg)
    DelEchoGals(chip->eg);
  DE_INIT(("Deleted.\n"))

  if (chip->iovirt)
    iounmap((void *)chip->iovirt);
  if (chip->iores) {
    release_resource(chip->iores);
    kfree_nocheck(chip->iores);
  }
  DE_INIT(("MMIO freed.\n"))

  /* release the irq */
  if (chip->irq>=0)
    free_irq(chip->irq, (void *)chip);
  /* release chip data */
  snd_magic_kfree(chip);
  DE_INIT(("Chip freed.\n"))
  return(0);
}



static int snd_Echo_dev_free(snd_device_t *device) {
  struct EchoChip *chip;

  DE_INIT(("Device free()...\n"))
  chip=snd_magic_cast(chip_t, device->device_data, return -ENXIO);
  return(snd_Echo_free(chip));
}



/* <--snd_Echo_probe */
static __devinit int snd_Echo_create(snd_card_t *card, struct pci_dev *pci, struct EchoChip **rchip) {
  struct EchoChip *chip;
  int err;
  size_t sz;
  static snd_device_ops_t ops={
    .dev_free=snd_Echo_dev_free,
  };

  *rchip=NULL;

  pci_write_config_byte(pci, PCI_LATENCY_TIMER, 0xC0);

  if ((err=pci_enable_device(pci))<0)
    return(err);

  /* allocate a chip-specific data with magic-alloc */
  chip=snd_magic_kcalloc(chip_t, 0, GFP_KERNEL);
  if (chip==NULL)
    return(-ENOMEM);

  spin_lock_init(&chip->lock);
  chip->card=card;
  chip->pci=pci;
  chip->irq=-1;

  /* PCI resource allocation */
  chip->iophys=pci_resource_start(pci, 0);
  sz=pci_resource_len(pci, 0);
  chip->iovirt=(unsigned long)ioremap_nocache(chip->iophys, sz);
  if ((chip->iores=request_mem_region(chip->iophys, sz, "Echoaudio"))==NULL) {
    snd_Echo_free(chip);
    printk(KERN_ERR "cannot get memory region\n");
    return(-EBUSY);
  }

  if (request_irq(pci->irq, snd_Echo_interrupt, SA_INTERRUPT|SA_SHIRQ, "Echoaudio", (void *)chip)) {
    snd_Echo_free(chip);
    printk(KERN_ERR "cannot grab irq\n");
    return(-EBUSY);
  }
  chip->irq=pci->irq;
  DE_INIT(("pci=%x irq=%d\n", (int)chip->pci, chip->irq))
  DE_INIT(("New EGs %04x\n", chip->pci->subsystem_device))
  chip->eg=NewEchoGals(chip->pci->device, chip->pci->subsystem_device);
  if (!chip->eg)
    goto InitError;

  chip->eg->ossprivate_data=pci;                        /* Used by OsPage*() functions to allocate memory in the DMA space */
  chip->eg->DspRegBase=(u32 *)chip->iovirt;        /* Virtual memory addr assigned to the card's registers */

  DE_INIT(("Init hardware...\n"))
  err=InitHw(chip->eg);
  if (err) {
    DE_INIT(("InitHw err=%x\n", err))
    goto InitError;
  }

  if (GetCapabilities(chip->eg, &chip->Caps)!=ECHOSTATUS_OK) {
    printk("No GetCaps\n");
    goto InitError;
  }
  DE_INIT(("Card init OK\n"))

  if ((err=snd_device_new(card, SNDRV_DEV_LOWLEVEL, chip, &ops))<0) {
    snd_Echo_free(chip);
    return(err);
  }
  *rchip=chip;
  return(0);            /* Init done ! */

  /* An error occurred. Clean up and exit. */
InitError:
  DE_INIT(("Init error\n"))
  snd_Echo_free(chip);
  return(-1);
}



/* constructor */
static int __devinit snd_Echo_probe(struct pci_dev *pci, const struct pci_device_id *pci_id) {
  static int dev;
  snd_card_t *card;
  struct EchoChip *chip;
  char *dsp;
  int i, err, rates, minrate, maxrate;

  if (dev>=SNDRV_CARDS)
    return(-ENODEV);
  if (!enable[dev]) {
    dev++;
    return(-ENOENT);
  }

  DE_INIT(("Echoaudio soundcards driver.\n"))
  card=snd_card_new(index[dev], id[dev], THIS_MODULE, 0);
  if (card==NULL)
    return(-ENOMEM);

  if ((err=snd_Echo_create(card, pci, &chip))<0) {
    snd_card_free(card);
    return(err);
  }

  strcpy(card->driver, "Echoaudio");
  strcpy(card->shortname, cardname[pci_id->subdevice>>4]);
  switch (pci_id->device) {
    case 0x1801:
      dsp="56301";
      break;
    case 0x3410:
      dsp="56361";
      break;
    default:
      dsp=" unknown";
  }
  sprintf(card->longname, "%s rev.%d (DSP%s) at 0x%lx irq %i", card->shortname, pci_id->subdevice&0x000f, dsp, chip->iophys, chip->irq);

  /* Sets up pcm structures according to capabilities of the card */
  //@@ it's static, do a copy.
  pcm_mixervolume_control.count=chip->eg->NumBussesIn*chip->eg->NumBussesOut;

  /* Check for supported sample rates   //@add support for contiguous ranges (Layla's) */
  i=rates=minrate=maxrate=0;
  while (SampleRates[i].frq) {
    if (QueryAudioSampleRate(chip->eg, SampleRates[i].frq)==ECHOSTATUS_OK) {
      rates|=SampleRates[i].bit;
      maxrate=SampleRates[i].frq;
      if (!minrate)
        minrate=SampleRates[i].frq;
    }
    i++;
  }
  Skel_hw.rates=rates;
  Skel_hw.rate_min=minrate;
  Skel_hw.rate_max=maxrate;

  DE_HWP(("BussesIn=%d  AnalogBussesIn=%d BussesOut=%d  AnalogBussesOut=%d\n", chip->eg->NumBussesIn, chip->eg->FirstDigitalBusIn, chip->eg->NumBussesOut, chip->eg->FirstDigitalBusOut))

  if ((err=snd_Echo_new_pcm(chip, chip->eg->FirstDigitalBusOut!=chip->eg->NumBussesOut))<0) {
    printk("Echoaudio - new pcm error %d\n", err);
    snd_card_free(card);
    return(err);
  }
/*
#ifdef EMIDI
  if ((pci_id->subdevice&0xfff0)==LAYLA24 || (pci_id->subdevice&0xfff0)==LAYLA)
    if ((err=snd_Echo_new_midi(card, chip))<0) {
      printk("Echoaudio - new midi error %d\n", err);
      snd_card_free(card);
      return(err);
    }
#endif
*/
  /*@@ These ones should check all channels */

  if (chip->eg->HasVmixer) {
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&line_volume_control, chip)))<0)
      return(err);
    //@@@@@  statica, serve una copia
    pcm_vmixervolume_control.count=chip->eg->NumPipesOut*chip->eg->NumBussesOut;
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcm_vmixervolume_control, chip)))<0)
      return(err);
  } else {
    if (chip->Caps.dwBusOutCaps[0]&ECHOCAPS_GAIN)
      if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmout_volume_control, chip)))<0)
        return(err);
  }

  if (chip->Caps.dwBusInCaps[0]&ECHOCAPS_GAIN)
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_volume_control, chip)))<0)
      return(err);

  if (chip->Caps.dwBusInCaps[0]&ECHOCAPS_NOMINAL_LEVEL)
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_nominal_control, chip)))<0)
      return(err);

  if (chip->Caps.dwBusOutCaps[0]&ECHOCAPS_NOMINAL_LEVEL)
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmout_nominal_control, chip)))<0)
      return(err);

  if ((chip->Caps.dwBusOutCaps[0]&(ECHOCAPS_VU_METER|ECHOCAPS_PEAK_METER)) || (chip->Caps.dwBusInCaps[0]&(ECHOCAPS_VU_METER|ECHOCAPS_PEAK_METER))) {
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&vumeterswitch_control, chip)))<0)
      return(err);
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&vumeters_control, chip)))<0)
      return(err);
  }

  if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcm_mixervolume_control, chip)))<0)
    return(err);

  /* Kludge to make the Mia work until I finish the Vmixer */
  if (chip->eg->HasVmixer) {
    SetVmixerGain(chip->eg, 0, 0, 0);
    SetVmixerGain(chip->eg, 1, 1, 0);
    SetVmixerGain(chip->eg, 0, 2, 0);
    SetVmixerGain(chip->eg, 1, 3, 0);
    SetVmixerGain(chip->eg, 2, 4, 0);
    SetVmixerGain(chip->eg, 3, 5, 0);
    SetVmixerGain(chip->eg, 2, 6, 0);
    SetVmixerGain(chip->eg, 3, 7, 0);
    UpdateVmixerLevel(chip->eg);
  }

  if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&channelsinfo_control, chip)))<0)
    return(err);

  /* Creates a list of available digital modes */
  chip->DigitalModes=chip->ClockSources=0;
  for (i=0; i<6; i++)
    if (chip->eg->DigitalModes&(1<<i))
      chip->DigitalModeIndex[chip->DigitalModes++]=i;

  if (chip->DigitalModes>1) {
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_digitalmode_control, chip)))<0)
      return(err);

    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&ditherswitch_control, chip)))<0)
      return(err);
  }

  if (chip->eg->DigitalModes&(ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA|ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL|ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_CDROM))
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_spdifmode_control, chip)))<0)
      return(err);

  /* Creates a list of available clock sources */
  for (i=0; i<10; i++)
    if (chip->eg->InClockTypes&(1<<i))
      chip->ClockSourceIndex[chip->ClockSources++]=i;

  if (chip->ClockSources>1)
    if ((err=snd_ctl_add(chip->card, snd_ctl_new1(&pcm_inclock_control, chip)))<0)
      return(err);

  for (i=0; i<DSP_MAXPIPES; i++) {
    chip->substream[i]=0;
    chip->LastPeriod[i]=0;
  }

  if ((err=snd_card_register(card))<0) {
    snd_card_free(card);
    return(err);
  }
  printk("Card registered: %s\n", card->longname);

  pci_set_drvdata(pci, chip);
  dev++;
  return(0);
}



static void __devexit snd_Echo_remove(struct pci_dev *pci) {
  struct EchoChip *chip;

  chip=snd_magic_cast(chip_t, pci_get_drvdata(pci), return);
  if (chip)
    snd_card_free(chip->card);
  pci_set_drvdata(pci, NULL);
}



/***************************************************************************************************
*** Everything starts and ends here
***************************************************************************************************/

/* pci_driver definition */
static struct pci_driver driver={
  .name="Echoaudio",
  .id_table=snd_Echo_ids,
  .probe=snd_Echo_probe,
  .remove=__devexit_p(snd_Echo_remove),
};



/* initialization of the module */
static int __init alsa_card_Echo_init(void) {
  int err;

  if ((err=pci_module_init(&driver))<0) {
#ifdef MODULE
    printk(KERN_ERR "Echoaudio soundcard not found or device busy\n");
#endif
    return(err);
  }
  return(0);
}



/* clean up the module */
static void __exit alsa_card_Echo_exit(void) {
  pci_unregister_driver(&driver);
}



module_init(alsa_card_Echo_init)
module_exit(alsa_card_Echo_exit)

EXPORT_NO_SYMBOLS;







/* Debug and to-do notes
16/64
aplay -v -B92880 -F1451  ~/track1.wav
16/32
aplay -v -B92880 -F2902  ~/track1.wav
16/16
aplay -v -B92880 -F5805  ~/track1.wav
16/2
aplay -v -B92880 -F46440  ~/track1.wav
24/2
aplay -v -B139320 -F69660  ~/track1.wav
24/4
aplay -v -B139320 -F34830  ~/track1.wav

64/4
aplay -v -B371519 -F92880  ~/track1.wav
128/4
aplay -v -B743038 -F185759  ~/track1.wav
64/2
aplay -v -B113770


controllare ECHOCAPS_MUTE
clock out layla20
valore rest da setsamplerate
do a copy of pcm_mixervolume_control for correctness, even if there are no problems
better handling of n of channels according to digital mode
sistemare DspMask_*()
da update*() si pu togliere waithandsh



SetSampleRate		/d20 g20 l20 d24 g24 l24 mon mia	H
SetInputClock		/    g20 l20				H
SetOutputClock		/        l20
DspSetDigitalMode	/                g24 l24 mon		H
UpdateVmixerLevel	/                            mia
SetProfessionalSpdif	/					H

ClearFlags
SetFlags

CheckAsicStatus
LoadDSP...
GoComatose
UpdateOutputLineLevel	H
UpdateInputLineLevel	H
UpdateFlags		H
SetMetersOn
StartTransport		H
StopTransport		H
ResetTransport		H


WriteControlReg GML



*/
