/*
 * SPCA50x based usb camera driver (currently supports
 * yuv native stream spca501a, spca501c, spca505, spca508, and untested spca506
 * jpeg native stream spca500, spca504a, spca504b, spca533a
 * bayer native stream spca561a ).
 * Cameras using this series of briges are Intel PC Camera Pro, Intel Create and Share
 * and Viewquest Technologies camera line
 * Cameras using jpeg Aiptek Mustek ...
 * Based on :
 * OmniVision OV511 Camera-to-USB Bridge Driver
 *
 * Copyright (c) 1999-2000 Mark W. McClelland
 * Many improvements by Bret Wallach <bwallac1@san.rr.com>
 * Color fixes by by Orion Sky Lawlor <olawlor@acm.org> (2/26/2000)
 * Snapshot code by Kevin Moore
 * OV7620 fixes by Charl P. Botha <cpbotha@ieee.org>
 * Changes by Claudio Matsuoka <claudio@conectiva.com>
 * SPCA50x version by Joel Crisp <cydergoth@users.sourceforge.net>
 *
 * Based on the Linux CPiA driver written by Peter Pregler,
 * Scott J. Bertin and Johannes Erdfelt.
 *
 * Please see the README file in this package
 * Please see the file: linux/Documentation/usb/ov511.txt
 * and the website at:  http://alpha.dyndns.org/ov511
 * for more info.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static const char version[] = SPCA50X_VERSION;
//static const char SPCA50X_H_CVS_VERSION[]="$Id: spca50x.h,v 1.28 2004/01/10 21:37:40 mxhaard Exp $";

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#ifdef CONFIG_MODVERSIONS
#include <linux/modversions.h>
#endif /* CONFIG_MODVERSIONS */
#endif 


#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/usb.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#include <linux/wrapper.h>
#endif
#include <linux/param.h>
#ifdef SPCA50X_ENABLE_OSD
#ifdef CONFIG_FB
#include <video/font.h>
#endif /* CONFIG_FB */
#endif /* SPCA50X_ENABLE_OSD */

#include "spca50x.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#  undef CONFIG_VIDEO_PROC_FS
#	 undef CONFIG_PROC_FS
#endif
/* V4L API extension for raw JPEG (=JPEG without header)   
 */
#define VIDEO_PALETTE_RAW_JPEG  20
//#define RH9_REMAP 1

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#include "spcaCompat.h"
#endif
/*
 * Let's include the initialization data for each camera type
 */

#include "spca500_init.h"
#include "spca501_init.h"
#include "spca504_init.h"
#include "spca505_init.h"
#include "spca506_init.h"
#include "spca508_init.h"
#include "spca561_init.h"
#include "spcai2c_init.h"


#include "spcadecoder.h"
#include "jpeg_qtables.h"

#define PROC_NAME_LEN 10 //length of the proc name

#ifndef CONFIG_FB
/* Force OSD off if required dependencies not enabled */
/* Turning OSD on requires a kernel patch to the fbcon__ routines */
#undef SPCA50X_ENABLE_OSD
#endif /* CONFIG_FB */

/* Video Size 640 x 480 x 4 bytes for RGB */
#define MAX_FRAME_SIZE (640 * 480 * 4)
#define MAX_DATA_SIZE (MAX_FRAME_SIZE + sizeof(struct timeval))

#ifdef SPCA50X_ENABLE_OSD
struct fbcon_font_desc *osd_font = NULL;
#endif /* SPCA50X_ENABLE_OSD */

/* PARAMETER VARIABLES: */
/* autoadjust = 1 schedule a BH and break the jpeg cam with tasklet */ 
static int autoadjust = 0;    /* CCD dynamically changes exposure, etc... */
/* Hardware auto exposure / whiteness (PC-CAM 600) */
static int autoexpo = 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,5)
/* Video device number (-1 is first available) */
static int video_nr = -1;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,5) */

/* 0=no debug messages
 * 1=init/detection/unload and other significant messages,
 * 2=some warning messages
 * 3=config/control function calls
 * 4=most function calls and data parsing messages
 * 5=highly repetitive mesgs
 * NOTE: This should be changed to 0, 1, or 2 for production kernels
 */
static int debug = 0;

/* Snapshot mode enabled flag */
static int snapshot = 0;

/* Force image to be read in RGB instead of BGR. This option allow
 * programs that expect RGB data (e.g. gqcam) to work with this driver. */
static int force_rgb = 0;

/* Number of seconds before inactive buffers are deallocated */
static int buf_timeout = 5;

#ifdef SPCA50X_ENABLE_CAMS
/* Number of cameras to stream from simultaneously */
static int cams = 1;
#endif /* SPCA50X_ENABLE_CAMS */



#ifdef SPCA50X_ENABLE_COMPRESSION
/* Enable compression. This is for experimentation only; compressed images
 * still cannot be decoded yet. */
static int compress = 0;
#endif /* SPCA50X_ENABLE_COMPRESSION */

#ifdef SPCA50X_ENABLE_TESTPATTERN
/* Display test pattern - doesn't work yet either */
static int testpat = 0;
#endif /* SPCA50X_ENABLE_TESTPATTERN */

/* Dump raw pixel data, in one of many formats. */
static int dumppix = 0;

/* Initial brightness & contrast (for debug purposes) */
static int bright = 0x80;
static int contrast = 0x60;

/* Internal/external CCD flag, initially internal CCD */
static int ccd = 0;

/* Parameter that enables you to set the minimal suitable bpp */
static int min_bpp = 0;

/* Parameter defines the average luminance that should be kept */
static int lum_level = 0x2d;

#ifdef SPCA50X_ENABLE_OSD
/* Enable On-screen display of some data (frame sequence, time) */
static int osd = 1;
#endif /* SPCA50X_ENABLE_OSD */

MODULE_PARM(autoadjust, "i");
MODULE_PARM_DESC(autoadjust, "CCD dynamically changes exposure (spca501x only)");
MODULE_PARM(autoexpo, "i");
MODULE_PARM_DESC(autoexpo, "Enable/Disable hardware auto exposure / whiteness (default: enabled) (PC-CAM 600 only)");
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level: 0=none, 1=init/detection, 2=warning, 3=config/control, 4=function call, 5=max");
MODULE_PARM(snapshot, "i");
MODULE_PARM_DESC(snapshot, "Enable snapshot mode");
MODULE_PARM(force_rgb, "i");
MODULE_PARM_DESC(force_rgb, "Read RGB instead of BGR");
MODULE_PARM(buf_timeout, "i");
MODULE_PARM_DESC(buf_timeout, "Number of seconds before buffer deallocation");
#ifdef SPCA50X_ENABLE_CAMS
MODULE_PARM(cams, "i");
MODULE_PARM_DESC(cams, "Number of simultaneous cameras (currently only 1 supported)");
#endif /* SPCA50X_ENABLE_CAMS */

#ifdef SPCA50X_ENABLE_COMPRESSION
MODULE_PARM(compress, "i");
MODULE_PARM_DESC(compress, "Turn on/off compression (not functional yet)");
#endif /* SPCA50X_ENABLE_COMPRESSION */
MODULE_PARM(dumppix, "i");
MODULE_PARM_DESC(dumppix, "Dump raw pixel data, in one of 3 formats. See spca50x_dumppix() for details");
MODULE_PARM(bright, "i");
MODULE_PARM_DESC(bright, "Initial brightness factor (0-255)");
MODULE_PARM(contrast, "i");
MODULE_PARM_DESC(contrast, "Initial contrast factor (0-255)");
MODULE_PARM(ccd, "i");
MODULE_PARM_DESC(ccd, "If zero, default to the internal CCD, otherwise use the external video input");
#ifdef SPCA50X_ENABLE_OSD
MODULE_PARM(osd, "i");
MODULE_PARM_DESC(osd, "If non-zero, enable various types of on-screen display");
#endif /* SPCA50X_ENABLE_OSD */
MODULE_PARM(min_bpp, "i");
MODULE_PARM_DESC(min_bpp, "The minimal color depth that may be set (default 0)");
MODULE_PARM(lum_level, "0-255i");
MODULE_PARM_DESC(lum_level, "Luminance level for brightness autoadjustment (default 32)");

MODULE_AUTHOR("Joel Crisp <cydergoth@users.sourceforge.net>, based on ov511 driver by Mark McClelland <mwm@i.am> & Bret Wallach & Orion Sky Lawlor <olawlor@acm.org> & Kevin Moore & Charl P. Botha <cpbotha@ieee.org> & Claudio Matsuoka <claudio@conectiva.com>");
MODULE_DESCRIPTION("SPCA50X USB Camera Driver");

MODULE_LICENSE("GPL");


static int spca50x_move_data(struct usb_spca50x *spca50x, struct urb *urb);
static void auto_bh(void* data);
static void spca504_acknowledged_command(struct usb_spca50x *spca50x, 
                                         __u16               reg,
                                         __u16               idx,
                                         __u16               val);
static void spca504A_acknowledged_command(struct usb_spca50x *spca50x, 
				 __u16 reg,
				 __u16 idx,
				 __u16 val,
				 __u8 stat,
				 __u8 count);
static struct usb_driver spca50x_driver;

#ifndef max
static inline int max(int a, int b)
{
	return (a > b)? a: b;
}
#endif /* max */

/**********************************************************************
 * List of known SPCA50X-based cameras
 **********************************************************************/

enum
{
	UnknownCamera = 0,
	IntelPCCameraPro,
	IntelCreateAndShare,
	GrandtecVcap,
	ViewQuestM318B,
	ViewQuestVQ110,
	KodakDVC325,
	MustekGsmartMini2,
	MustekGsmartMini3,
	CreativePCCam300,
	DLinkDSC350,
	CreativePCCam600,
	IntelPocketPCCamera,
	IntelEasyPCCamera,
	ThreeComHomeConnectLite,
	KodakEZ200,
	MaxellMaxPocket,
	AiptekMiniPenCam2,
	AiptekPocketDVII,
	AiptekPenCamSD,
	AiptekMiniPenCam13,
	MustekGsmartLCD3,
	MustekMDC5500Z,
	MegapixV4,
	AiptekPocketDV,
	HamaUSBSightcam,
	Arowana300KCMOSCamera,
	MystFromOriUnknownCamera,	
	AiptekPocketDV3100,
	AiptekPocketCam3M,
	GeniusVideoCAMExpressV2,
	Flexcam100Camera,
	MustekGsmartLCD2,
	PureDigitalDakota,
	PetCam,
	BenqDC1500,
	LogitechClickSmart420,
	LogitechClickSmart510,	
	BenqDC1300,
	HamaUSBSightcam2, 
	MustekDV3000, /* 40 */
	LastCamera
};
static struct cam_list clist[] = {
	{UnknownCamera, "Unknown"},
	{IntelPCCameraPro, "Intel PC Camera Pro"},
	{IntelCreateAndShare, "Intel Create and Share"},
	{GrandtecVcap, "Grandtec V.cap"},
	{ViewQuestM318B, "ViewQuest M318B"},
	{ViewQuestVQ110, "ViewQuest VQ110"},
	{KodakDVC325, "Kodak DVC-325"},
	{MustekGsmartMini2, "Mustek gSmart mini 2"},
	{MustekGsmartMini3, "Mustek gSmart mini 3"},
	{CreativePCCam300, "Creative PC-CAM 300"},
	{DLinkDSC350, "D-Link DSC-350"},
	{CreativePCCam600, "Creative PC-CAM 600"},
	{IntelPocketPCCamera, "Intel Pocket PC Camera"},
	{IntelEasyPCCamera, "Intel Easy PC Camera"},
	{ThreeComHomeConnectLite, "3Com Home Connect Lite"},
	{KodakEZ200, "Kodak EZ200"},
	{MaxellMaxPocket, "Maxell Max Pocket LEdit. 1.3 MPixels"},
	{AiptekMiniPenCam2, "Aiptek Mini PenCam  2 MPixels"},
	{AiptekPocketDVII, "Aiptek PocketDVII  1.3 MPixels"},
	{AiptekPenCamSD, "Aiptek Pencam SD  2 MPixels"},
	{AiptekMiniPenCam13, "Aiptek mini PenCam 1.3 MPixels"},
	{MustekGsmartLCD3, "Mustek Gsmart LCD 3"},
	{MustekMDC5500Z, "Mustek MDC5500Z"},
	{MegapixV4, "Megapix V4"},
	{AiptekPocketDV, "Aiptek PocketDV "},
	{HamaUSBSightcam, "Hama USB Sightcam 100"},
	{Arowana300KCMOSCamera, "Arowana 300K CMOS Camera"},
	{MystFromOriUnknownCamera,"Unknow Ori Camera"},
	{AiptekPocketDV3100, "Aiptek PocketDV3100+ "},
	{AiptekPocketCam3M, "Aiptek PocketCam  3 M "},
	{GeniusVideoCAMExpressV2, "Genius VideoCAM Express V2"},
	{Flexcam100Camera, "Flexcam 100 Camera"},
	{MustekGsmartLCD2, "Mustek Gsmart LCD 2"},
	{PureDigitalDakota, "Pure Digital Dakota"},
	{PetCam, "PetCam"},
	{BenqDC1500, "Benq DC1500"},
	{LogitechClickSmart420, "Logitech Inc. ClickSmart 420"},
	{LogitechClickSmart510, "Logitech Inc. ClickSmart 510"},
	{BenqDC1300, "Benq DC1300"},
	{HamaUSBSightcam2, "Hama USB Sightcam 100 (2)"},
	{MustekDV3000, "Mustek DV 3000"},
	{-1, NULL}
};

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
static __devinitdata struct usb_device_id device_table[] = {
	{USB_DEVICE (0x0733, 0x0430)},	/* Intel PC Camera Pro */
	{USB_DEVICE (0x0733, 0x0401)},	/* Intel Create and Share */
	{USB_DEVICE (0x99FA, 0x8988)},	/* Grandtec V.cap */
	{USB_DEVICE (0x0733, 0x0402)},	/* ViewQuest M318B */
	{USB_DEVICE (0x0733, 0x0110)},	/* ViewQuest VQ110 */
	{USB_DEVICE (0x040A, 0x0002)},	/* Kodak DVC-325 */
	{USB_DEVICE (0x055f, 0xc420)},	/* Mustek gSmart Mini 2 */
	{USB_DEVICE (0x055f, 0xc520)},	/* Mustek gSmart Mini 3 */
	{USB_DEVICE (0x041E, 0x400A)},	/* Creative PC-CAM 300 */
	{USB_DEVICE (0x084D, 0x0003)},	/* D-Link DSC-350 */
	{USB_DEVICE (0x041E, 0x400B)},	/* Creative PC-CAM 600 */
	{USB_DEVICE (0x8086, 0x0630)},	/* Intel Pocket PC Camera */
	{USB_DEVICE (0x8086, 0x0110)},	/* Intel Easy PC Camera */
	{USB_DEVICE (0x0506, 0x00df)},	/* 3Com HomeConnect Lite */
	{USB_DEVICE (0x040a, 0x0300)},	/* Kodak EZ200 */
	{USB_DEVICE (0x04fc, 0x504b)},	/* Maxell MaxPocket LE 1.3 */
	{USB_DEVICE (0x08ca, 0x2008)},	/* Aiptek Mini PenCam 2 M */
	{USB_DEVICE (0x08ca, 0x0104)},	/* Aiptek PocketDVII 1.3 */
	{USB_DEVICE (0x08ca, 0x2018)},	/* Aiptek Pencam SD 2M */
	{USB_DEVICE (0x04fc, 0x504a)},	/* Aiptek Mini PenCam 1.3 */
	{USB_DEVICE (0x055f, 0xc530)},	/* Mustek Gsmart LCD 3 */
	{USB_DEVICE (0x055f, 0xc650)},	/* Mustek MDC5500Z */
	{USB_DEVICE (0x052b, 0x1513)},	/* Megapix V4 */
	{USB_DEVICE (0x08ca, 0x0103)},	/* Aiptek PocketDV */
	{USB_DEVICE (0x0af9, 0x0010)},	/* Hama USB Sightcam 100 */
	{USB_DEVICE (0x1776, 0x501c)},	/* Arowana 300K CMOS Camera */
	{USB_DEVICE (0x08ca, 0x0106)},	/* Aiptek Pocket DV3100+ */
	{USB_DEVICE (0x08ca, 0x2010)},	/* Aiptek PocketCam 3M */
	{USB_DEVICE (0x0458, 0x7004)},	/* Genius VideoCAM Express V2 */
	{USB_DEVICE (0x04fc, 0x0561)},	/* Flexcam 100 */
	{USB_DEVICE (0x055f, 0xc430)},	/* Mustek Gsmart LCD 2 */
	{USB_DEVICE (0x04fc, 0xffff)},	/* Pure DigitalDakota */
	{USB_DEVICE (0xabcd, 0xcdee)},	/* Petcam */
	{USB_DEVICE (0x04a5, 0x3008)},	/* Benq DC 1500 */
	{USB_DEVICE (0x046d, 0x0960)},	/* Logitech Inc. ClickSmart 420 */
	{USB_DEVICE (0x046d, 0x0901)},	/* Logitech Inc. ClickSmart 510 */
	{USB_DEVICE (0x04a5, 0x3003)},	/* Benq DC 1300 */
	{USB_DEVICE (0x0af9, 0x0011)},	/* Hama USB Sightcam 100 */
	{USB_DEVICE (0x055f, 0xc440)},  /* Mustek DV 3000 */
	{USB_DEVICE (0x0000, 0x0000)},	/* MystFromOri Unknow Camera */
	{}			/* Terminating entry */
};


MODULE_DEVICE_TABLE (usb, device_table);
/* 
 We also setup the function for getting 
 page number from the virtual address 
*/
#define VIRT_TO_PAGE virt_to_page 
#else /* LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0) */
#define VIRT_TO_PAGE MAP_NR 
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0) */

/* Mode list for spca50x on external input */
/* Must be in descending order as the closest match which is equal or smaller than
 * the requested size is returned */

#define GET_EXT_MODES(bridge) (\
	(bridge) == BRIDGE_SPCA500 ? spca500_ext_modes : \
	(bridge) == BRIDGE_SPCA501 ? spca501_ext_modes : \
	(bridge) == BRIDGE_SPCA504 ? spca504_ext_modes : \
	(bridge) == BRIDGE_SPCA504B ? spca504_ext_modes : \
	(bridge) == BRIDGE_SPCA504_PCCAM600 ? spca504_pccam600_ext_modes : \
	(bridge) == BRIDGE_SPCA508 ? spca508_ext_modes : \
	(bridge) == BRIDGE_SPCA533 ? spca533_ext_modes : \
	(bridge) == BRIDGE_SPCA561 ? spca561_ext_modes : \
	spca50x_ext_modes)
/* code is used for harware low nibble and software hight nibble */

static __u16 spca500_ext_modes[][6] = {
	/* x , y , Code, x multiplier, y multiplier, pipe */
	{640, 480, 0x00, 40, 30, 1023},
	{352, 288, 0x00, 22, 18, 1023},
	{320, 240, 0x01, 40, 30, 1023},
	{176, 144, 0x01, 22, 18, 1023},
	/* 160x120 disable jpeg %16 */
	{0, 0, 0, 0, 0, 0}
};

static __u16 spca501_ext_modes[][6] = {
	/* x , y , Code, Value (6), Value (7), pipe */
	{640, 480, 0, 0x94, 0x004A, 1023},
	{352, 288, 0x10, 0x94, 0x004A, 1023},
	{320, 240, 0, 0x94, 0x104A, 896},
	{176, 144, 0x10, 0x94, 0x104A, 896},
	{160, 120, 0, 0x94, 0x204A, 384},
	{0, 0, 0, 0, 0}
};

static __u16 spca533_ext_modes[][6] =
{
	/* x , y , Code, Value (6), Value (7), pipe */
	{ 640, 480, 0x41, 1, 0, 1023 },
	{ 464, 480, 0x01, 0, 0, 1023 },//PocketDVII unscaled resolution aspect ratio need to expand x axis
	{ 464, 352, 0x01, 0, 0, 1023 },//Gsmart LCD3 feature good aspect ratio 
	{ 384, 288, 0x11, 5, 4, 1023 },
	{ 352, 288, 0x11, 7, 4, 1023 },
	{ 320, 240, 0x02, 0, 0, 1023 },
	{ 192, 144, 0x12, 8, 6, 1023 },
	{ 176, 144, 0x12, 9, 6, 1023 },
	//{ 160, 120, 0x22, 1, 1, 1023 },
	//{ 320, 128, 0x03, 0, 0, 1023 }, /* don't work with megapix V4*/
	{ 0, 0, 0, 0, 0 }
};

static __u16 spca504_ext_modes[][6] =
{
	/* x , y , Code, Value (6), Value (7), pipe */
	{ 640, 480, 0x01, 0, 0, 1023 },
	{ 384, 288, 0x11, 16, 12, 1023 },
	{ 352, 288, 0x11, 18, 12, 1023 },
	{ 320, 240, 0x02, 0, 0, 896 },
	{ 192, 144, 0x12, 8, 6, 896 },
	{ 176, 144, 0x12, 9, 6, 896 },
	//{ 160, 120, 0x22, 1, 1, 896 },
	{ 0, 0, 0, 0, 0 }
};

 
static __u16 spca504_pccam600_ext_modes[][6] =
{
	/* x , y , Code, Value (6), Value (7), pipe */
	{ 1024, 768, 0x00, 0, 0, 1023 }, 
	{ 640,  480, 0x01, 1, 0, 1023 },
	{ 352,  288, 0x02, 2, 0, 896 },
	{ 320,  240, 0x03, 3, 0, 896 },
	{ 176,  144, 0x04, 4, 0, 768 },
/*	{ 160,  120, 0x05, 5, 0, 768 },  */
	{ 0, 0, 0, 0, 0 }
};

static __u16 spca561_ext_modes[][6] =
{
	{ 352, 288, 0x00, 0x27, 0x00, 1023},
	{ 320, 240, 0x01, 0x27, 0x00, 1023},
	{ 176, 144, 0x02, 0x23, 0x00, 1023},
	{ 160, 120, 0x03, 0x23, 0x00, 1023},
	{ 0, 0, 0, 0, 0 }
};

static __u16 spca50x_ext_modes[][6] =
{
	/* x , y , Code, Value (6), Value (7), pipe */
	{ 352, 288, 1, 0x1a, 0x1a, 1023 }, /* Tested */
	{ 320, 240, 2, 0x1c, 0x1d, 896 }, /* Tested 20:01:2004*/
#if 0	
	{ 240, 180, 3, 0x20, 0x20, 640 }, /* Guessed values */
#endif
	{ 176, 144, 4, 0x34, 0x34, 512 }, /* Tested */
	{ 160, 120, 5, 0x40, 0x40, 384 }, /* Tested */
	{ 0, 0, 0, 0, 0 }
};

static __u16 spca508_ext_modes[][6] =
{
	/* In this table, element 3 (clk) controls the
	 * clock, and gets written to 0x8700.
	 */
	/* x , y , Code, clk, n/a,  pipe */
	{ 352, 288, 0x00, 0x28, 0x00, 1023},
	{ 320, 240, 0x01, 0x28, 0x00, 1023},
	{ 176, 144, 0x02, 0x23, 0x00, 1023}, 
	{ 160, 120, 0x03, 0x23, 0x00, 1023},
	{ 0, 0, 0, 0, 0 }
};

#ifdef CONFIG_PROC_FS
/* Not sure what we should do with this. I think it is V4L level 2 stuff */
/* Currently only use RGB24 */
static struct palette_list plist[] = {
	{ VIDEO_PALETTE_GREY,      "GREY" },
	{ VIDEO_PALETTE_HI240,     "HI240" },
	{ VIDEO_PALETTE_RGB565,    "RGB565" },
	{ VIDEO_PALETTE_RGB24,     "RGB24" },
	{ VIDEO_PALETTE_RGB32,     "RGB32" },
	{ VIDEO_PALETTE_RGB555,    "RGB555" },
	{ VIDEO_PALETTE_YUV422,    "YUV422" },
	{ VIDEO_PALETTE_YUYV,      "YUYV" },
	{ VIDEO_PALETTE_UYVY,      "UYVY" },
	{ VIDEO_PALETTE_YUV420,    "YUV420" },
	{ VIDEO_PALETTE_YUV411,    "YUV411" },
	{ VIDEO_PALETTE_RAW,       "RAW" },
	{ VIDEO_PALETTE_YUV422P,   "YUV422P" },
	{ VIDEO_PALETTE_YUV411P,   "YUV411P" },
	{ VIDEO_PALETTE_YUV420P,   "YUV420P" },
	{ VIDEO_PALETTE_YUV410P,   "YUV410P" },
	{ VIDEO_PALETTE_RAW_JPEG , "RJPG" },
	{ -1, NULL }
};
#endif /* CONFIG_PROC_FS */

/* struct for the tasklets */
struct spca50x_frame *taskletframe;
void outpict_do_tasklet (unsigned long ptr); 
DECLARE_TASKLET (outpicttasklet,outpict_do_tasklet,(unsigned long)&taskletframe);

/* Common functions prototype */
static int spca50x_reg_read(struct usb_device *dev,
                            __u16              reg,
                            __u16              index,
                            __u16              length);
static int spca50x_reg_write(struct usb_device *dev,
                             __u16              reg,
                             __u16              index,
                             __u16              value);
void spca500_reinit(struct usb_spca50x *spca50x );
static int spca500_initialise(struct usb_spca50x *spca50x);
static int spca500_full_reset(struct usb_spca50x *spca50x);

static int spca50x_setup_qtable(struct usb_spca50x *spca50x,
                                unsigned int        request,
                                unsigned int        ybase,
                                unsigned int        cbase,
                                unsigned char       qtable[2][64]);
				
static void spca504_pccam600_initialize(struct usb_spca50x *spca50x);

static void spca504_initialize(struct usb_spca50x *spca50x);
static void spca504_wait_status(struct usb_spca50x *spca50x);

/*
 * Simple function to wait for a given 8-bit value to be returned from
 * a spca50x_reg_read call.
 * Returns: negative is error or timeout, zero is success.
 */
static int spca50x_reg_readwait(struct usb_device *dev,
                                __u16              reg,
                                __u16              index,
                                __u16              value)
{
	int count = 0;
	int result = 0;

	while (count < 20)
	{
		result = spca50x_reg_read(dev, reg, index, 1);
		if (result == value) return 0;

		wait_ms(50);

		count++;
	}

	PDEBUG(2, "spca50x_reg_readwait failed");

	return -EIO;
}

/*********************** Specific spca504b *********************************/

static void spca504B_init(struct usb_spca50x *spca50x);
static void spca504B_start(struct usb_spca50x *spca50x);
static void spca504B_stop (struct usb_spca50x *spca50x);
static void spca504B_SetSizeType(struct usb_spca50x *spca50x,
			 __u16 extSize,
			 __u16 extType);
static void spca504B_reg_read(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) ;			 
/*********************** Specific spca533 MegapixV4 ************************/

static void spca533_Megapix(struct usb_spca50x *spca50x );
		 
/**********************************************************************
 *
 * Memory management
 *
 * This is a shameless copy from the USB-cpia driver (linux kernel
 * version 2.3.29 or so, I have no idea what this code actually does ;).
 * Actually it seems to be a copy of a shameless copy of the bttv-driver.
 * Or that is a copy of a shameless copy of ... (To the powers: is there
 * no generic kernel-function to do this sort of stuff?)
 *
 * Yes, it was a shameless copy from the bttv-driver. IIRC, Alan says
 * there will be one, but apparentely not yet -jerdfelt
 *
 * So I copied it again for the ov511 driver -claudio
 * And again for the spca50x driver -jcrisp
 **********************************************************************/

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
#ifndef RH9_REMAP
static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
	unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;

	if (!pgd_none(*pgd)) {
		pmd = pmd_offset(pgd, adr);
		if (!pmd_none(*pmd)) {
			/* WOLT kernels appear to have changed pte_offset to pte_offset_kernel */
#ifdef pte_offset		
			ptep = pte_offset(pmd, adr);
#else /* pte_offset */
			ptep = pte_offset_kernel(pmd, adr);
#endif /* pte_offset */
			pte = *ptep;
			if (pte_present(pte)) {
				ret = (unsigned long) page_address(pte_page(pte));
				ret |= (adr & (PAGE_SIZE - 1));
			}
		}
	}

	return ret;
}
#endif /* RH9_REMAP */
/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
#ifdef RH9_REMAP
static inline unsigned long
kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1); /* restore the offset */
	ret = __pa(kva);
	return ret;
}

#else /* RH9_REMAP */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *) adr));
	kva |= adr & (PAGE_SIZE - 1);
	ret = __pa(kva);
	return ret;
}
#else 
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
	return ret;
}
#endif
#endif /* RH9_REMAP */


static void *
rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
	return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while ((long) size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		SetPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_reserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
																	  
	return mem;
}

static void
rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		ClearPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_unreserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}


/**********************************************************************
 * Get average luminance
 **********************************************************************/
static inline __u8 get_avg_lum(struct usb_spca50x *spca50x)
{
	__u8 luminance; //The average luminance from camera
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA501:
			luminance = spca50x_reg_read(spca50x->dev, SPCA501_REG_CCDSP, 
				0x26, 2) >> 8;
			break;
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
                case BRIDGE_SPCA508:
                        luminance = spca50x_reg_read(spca50x->dev, 0, 0x8621, 1);
                        break;
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */
		default:
			luminance = 0;
			break;
	}
	return luminance;
}

/**********************************************************************
 * Get average R-G and B-G
 **********************************************************************/
static inline __u8 get_avg_RG(struct usb_spca50x *spca50x)
{
	__u8 rg; //The average R-G for window5
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA501:
			rg = spca50x_reg_read(spca50x->dev, SPCA501_REG_CCDSP, 0x30, 
				2) >> 8;
			break;
		default:
			rg = 0;
			break;
	}
	return rg;
}

/**********************************************************************
 * Get average B-G
 **********************************************************************/
static inline __u8 get_avg_BG(struct usb_spca50x *spca50x)
{
	__u8 bg; //The average B-G for window5
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA501:
			bg = spca50x_reg_read(spca50x->dev, SPCA501_REG_CCDSP, 0x2f, 
				2) >> 8;
			break;
		default:
			bg = 0;
			break;
	}
	return bg;
}

/**********************************************************************
 * /proc interface
 * Based on the CPiA driver version 0.7.4 -claudio
 * ..and again copied from the ov511 driver for the SPCA50x driver - jac
 **********************************************************************/

#ifdef CONFIG_PROC_FS

static struct proc_dir_entry *spca50x_proc_entry = NULL;
#ifdef CONFIG_VIDEO_PROC_FS
extern struct proc_dir_entry *video_proc_entry;
#endif /* CONFIG_VIDEO_PROC_FS */

#define YES_NO(x) ((x) ? "yes" : "no")

static int spca50x_read_proc(char *page, char **start, off_t off,
			  int count, int *eof, void *data)
{
	char *out = page;
	int i, j, len;
	struct usb_spca50x *spca50x = data;

	/* IMPORTANT: This output MUST be kept under PAGE_SIZE
	 *            or we need to get more sophisticated. */

	out += sprintf (out, "driver          : SPCA50X USB Camera\n");
	out += sprintf (out, "driver_version  : %s\n", version);
	out += sprintf (out, "model           : %s\n", (spca50x->desc) ?
	                     clist[spca50x->desc].description : "unknown");
	out += sprintf (out, "streaming       : %s\n", YES_NO (spca50x->streaming));
	out += sprintf (out, "grabbing        : %s\n", YES_NO (spca50x->grabbing));
	out += sprintf (out, "compress        : %s\n", YES_NO (spca50x->compress));
	out += sprintf (out, "data_format     : %s\n", spca50x->force_rgb ? "RGB" : "BGR");
	out += sprintf (out, "brightness      : %d\n", spca50x->brightness >> 8);
	out += sprintf (out, "colour          : %d\n", spca50x->colour >> 8);
	out += sprintf (out, "hue             : %d\n", spca50x->hue >> 8);
	out += sprintf (out, "contrast        : %d\n", spca50x->contrast);
	out += sprintf (out, "num_frames      : %d\n", SPCA50X_NUMFRAMES);
	out += sprintf (out, "curframe        : %d\n", spca50x->curframe);
	out += sprintf (out, "lastFrameRead   : %d\n", spca50x->lastFrameRead);
	spca50x->avg_lum = get_avg_lum(spca50x);
	out += sprintf (out, "Avg. luminance  : 0x%X %d\n", 
	                     spca50x->avg_lum, spca50x->avg_lum);
	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		out += sprintf (out, "frame           : %d\n", i);
		out += sprintf (out, "  sequence      : %d\n", spca50x->frame[i].seq);
		out += sprintf (out, "  grabstate     : %d\n", spca50x->frame[i].grabstate);
		out += sprintf (out, "  depth         : %d\n",
		                     spca50x->frame[i].depth);
		out += sprintf (out, "  size          : %d %d\n",
		                     spca50x->frame[i].width, spca50x->frame[i].height);
		out += sprintf (out, "  format        : ");
		for (j = 0; plist[j].num >= 0; j++) {
			if (plist[j].num == spca50x->frame[i].format) {
				out += sprintf (out, "%s\n", plist[j].name);
				break;
			}
		}
		if (plist[j].num < 0)
			out += sprintf (out, "unknown\n");
		out += sprintf (out, "  data_buffer   : 0x%p\n",
		                     spca50x->frame[i].data);
	}
	out += sprintf (out, "snap_enabled    : %s\n", YES_NO (spca50x->snap_enabled));
	out += sprintf (out, "packet_size     : %d\n", spca50x->packet_size);
	out += sprintf (out, "internal ccd    : %s\n", YES_NO (!spca50x->ccd));
	out += sprintf (out, "framebuffer     : 0x%p\n", spca50x->fbuf);
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	out += sprintf (out, "stable          : %d\n", spca50x->nstable);
	out += sprintf (out, "unstable        : %d\n", spca50x->nunstable);
	out += sprintf (out, "whiteness       : %d\n", spca50x->whiteness >> 12);
	spca50x->avg_rg = get_avg_RG(spca50x);
	spca50x->avg_bg = get_avg_BG(spca50x);
	out += sprintf (out, "Avg. R-G/B-G  : 0x%X/0x%X %d/%d\n", 
	                     spca50x->avg_rg, spca50x->avg_bg, 
	                     (char)spca50x->avg_rg, (char)spca50x->avg_bg);
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) 
			return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

static int spca50x_write_proc(struct file *file, const char *buffer,
			   unsigned long count, void *data)
{
	return -EINVAL;
}

/*
 * Function services read requests to control proc entry
 * and prints all the static variables
 */
static int spca50x_ctlread_proc(char *page, char **start, off_t off,
			  int count, int *eof, void *data)
{
	char *out = page;
	int len = 0;
	struct usb_spca50x *spca50x = data;

	out += sprintf(out, "force_rgb = %d\n", spca50x->force_rgb);
	out += sprintf(out, "min_bpp = %d\n", spca50x->min_bpp);
	out += sprintf(out, "lum_level = %d\n", spca50x->lum_level);
	out += sprintf(out, "debug = %d\n", debug);
#ifdef SPCA50X_ENABLE_OSD
	out += sprintf(out, "osd = %d\n", osd);
#endif /* SPCA50X_ENABLE_OSD */

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) 
			return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

/*
 * Function compares two strings.
 * Return offset in pussy where prick ends if "prick" may penetrate 
 * int "pussy" like prick into pussy, -1 otherwise.
 */
static inline int match(const char* prick, const char* pussy, int len2)
{
	int len1 = strlen(prick); //length of male string
	int i; //just an index variable
	const char* tmp; //temporary pointer for my own pleasure

	// We skip all spaces and tabs
	for (i = 0; i < len2 && (pussy[i] == ' ' || pussy[i] == '\t'); i++)
	{
	} 

	tmp = pussy + i; // pointer to pussy with skipped shit (spaces and tabs)
	len2 = strlen(tmp); //calculate length again

	if (len1 > len2)
		return -1; //Fuck off, no fucking

	if (!strncmp(prick, tmp, len1))
		return i + len1;

	return -1;
}

#ifdef SPCA50X_ENABLE_RAWPROCENTRY
static int spca50x_rawread_proc(char  *page,
                                char **start,
                                off_t  off,
                                int    count,
                                int   *eof,
                                void  *data)
{
	struct usb_spca50x *spca50x = data;

	*start = page;

	/* check whether the buffer exists */
	if (spca50x->rawBuffer == NULL)
	{
		*eof = 1;
		return 0;
	}

	/* check offset is valid */
	if (off > spca50x->rawBufferSize)
	{
		*eof = 1;
		return 0;
	}

	/* can't read more than exists in the buffer, either */
	if ((count+off) > spca50x->rawBufferSize)
	{
		count = spca50x->rawBufferSize - off;
	}

	/* can't read more than is available in the output buffer */
	if (count > PAGE_SIZE)
	{
		count = PAGE_SIZE;
	}

	if (count == 0)
	{
		*eof = 1;
		return 0;
	}

	/* populate the output buffer */
	memcpy(page, spca50x->rawBuffer + off, count);

	/* return read count */
	return count;
}

static int spca50x_rawwrite_proc(struct file  *file,
                                 const char   *buffer,
                                 unsigned long count,
                                 void         *data)
{
	struct usb_spca50x *spca50x = data;

	/* if anything is written, flush the buffer */
	PDEBUG(3, "flushed raw proc entry buffer");
	spca50x->rawBufferSize = 0;

	return count;
}
#endif /* SPCA50X_ENABLE_RAWPROCENTRY */

/*
 * Try to calculate value from string (atoi). Converts  
 * decimal integer
 */
static inline int atoi(const char* str)
{
	int result = 0; //result of the function
	int i; //just an index variable

	for (i = 0; str[i] >= '0' && str[i] <= '9'; i++)
	{
		result *= 10;
		result += str[i] - '0';
	}
	return result;
}


static int spca50x_ctlwrite_proc(struct file *file, const char *buffer,
			   unsigned long count, void *data)
{
	int off; //where look for a value
	struct usb_spca50x *spca50x = data;

	if ((off = match("lum_level=", buffer, count)) >= 0)
		spca50x->lum_level = atoi(buffer + off);
	if ((off = match("min_bpp=", buffer, count)) >= 0)
		spca50x->min_bpp = atoi(buffer + off);
	if ((off = match("force_rgb=", buffer, count)) >= 0)
		spca50x->force_rgb = atoi(buffer + off);
	if ((off = match("debug=", buffer, count)) >= 0)
		debug = atoi(buffer + off);

	return count;
}

static void create_proc_spca50x_cam (struct usb_spca50x *spca50x)
{
	char name[PROC_NAME_LEN];
	struct proc_dir_entry *ent;

	if (!spca50x_proc_entry || !spca50x)
		return;

//Create videoxx proc entry
	sprintf(name, "video%d", spca50x->vdev->minor);
	PDEBUG (4, "creating /proc/video/spca50x/%s", name);

	ent = create_proc_entry(name, S_IFREG|S_IRUGO|S_IWUSR, spca50x_proc_entry);

	if (!ent)
		return;

	ent->data = spca50x;
	ent->read_proc = spca50x_read_proc;
	ent->write_proc = spca50x_write_proc;
	spca50x->proc_entry = ent;

// Create the controlxx proc entry
	sprintf(name, "control%d", spca50x->vdev->minor);
	PDEBUG (4, "creating /proc/video/spca50x/%s", name);
	ent = create_proc_entry(name, S_IFREG|S_IRUGO|S_IWUSR, 
		spca50x_proc_entry);

	if (!ent)
		return;

	ent->data = spca50x;
	ent->read_proc = spca50x_ctlread_proc;
	ent->write_proc = spca50x_ctlwrite_proc;
	spca50x->ctl_proc_entry = ent;

#ifdef SPCA50X_ENABLE_RAWPROCENTRY
// Create the rawxx proc entry
	sprintf(name, "raw%d", spca50x->vdev.minor);
	PDEBUG (4, "creating /proc/video/spca50x/%s", name);
	ent = create_proc_entry(name, S_IFREG|S_IRUGO|S_IWUSR, 
		spca50x_proc_entry);

	if (!ent)
		return;

	ent->data = spca50x;
	ent->read_proc = spca50x_rawread_proc;
	ent->write_proc = spca50x_rawwrite_proc;
	spca50x->raw_proc_entry = ent;
	spca50x->rawBufferSize = 0;
	spca50x->rawBuffer = vmalloc(10*1024*1024);
	if (spca50x->rawBuffer != NULL)
	{
		spca50x->rawBufferMax = 10*1024*1024;
		PDEBUG(3, "allocated 10Mb raw proc entry buffer");
	}
	else
	{
		PDEBUG(3, "vmalloc of raw proc entry buffer failed");
		spca50x->rawBufferMax  = 0;
	}
#endif /* SPCA50X_ENABLE_RAWPROCENTRY */
}

static void destroy_proc_spca50x_cam (struct usb_spca50x *spca50x)
{
	char name[PROC_NAME_LEN];

	if (!spca50x || !spca50x_proc_entry)
		return;

	/* destroy videoxx proc entry */
	if (spca50x->proc_entry != NULL)
	{
		sprintf(name, "video%d", spca50x->vdev->minor);
		PDEBUG (4, "destroying %s", name);
		remove_proc_entry(name, spca50x_proc_entry);
		spca50x->proc_entry = NULL;
	}

	/* destroy controlxx proc entry */
	if (spca50x->ctl_proc_entry != NULL)
	{
		sprintf(name, "control%d", spca50x->vdev->minor);
		PDEBUG (4, "destroying %s", name);
		remove_proc_entry(name, spca50x_proc_entry);
		spca50x->ctl_proc_entry = NULL;
	}

#ifdef SPCA50X_ENABLE_RAWPROCENTRY
	/* destroy rawxx proc entry */
	if (spca50x->raw_proc_entry != NULL)
	{
		sprintf(name, "raw%d", spca50x->vdev.minor);
		remove_proc_entry(name, spca50x_proc_entry);
		spca50x->raw_proc_entry = NULL;
		vfree(spca50x->rawBuffer);
	}
#endif /* SPCA50X_ENABLE_RAWPROCENTRY */

}

static void proc_spca50x_create(void)
{
	/* No current standard here. Alan prefers /proc/video/ as it keeps
	 * /proc "less cluttered than /proc/randomcardifoundintheshed/"
	 * -claudio
	 */

#ifdef CONFIG_VIDEO_PROC_FS
	if (video_proc_entry == NULL) {
		err("Unable to initialise /proc/video/spca50x");
		return;
	}
	spca50x_proc_entry = create_proc_entry("spca50x", S_IFDIR, video_proc_entry);
#else /* CONFIG_VIDEO_PROC_FS */
	spca50x_proc_entry = create_proc_entry("spca50x", S_IFDIR, 0);
#endif /* CONFIG_VIDEO_PROC_FS */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	if (spca50x_proc_entry)
		spca50x_proc_entry->owner = THIS_MODULE;
	else
#ifdef CONFIG_VIDEO_PROC_FS
		err("Unable to initialise /proc/video/spca50x");
#else /* CONFIG_VIDEO_PROC_FS */
		err("Unable to initialise /proc/spca50x");
#endif /* CONFIG_VIDEO_PROC_FS */
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0) */
}

static void proc_spca50x_destroy(void)
{
#ifdef CONFIG_VIDEO_PROC_FS
	PDEBUG (3, "removing /proc/video/spca50x");
#else /* CONFIG_VIDEO_PROC_FS */
	PDEBUG (3, "removing /proc/spca50x");
#endif /* CONFIG_VIDEO_PROC_FS */

	if (spca50x_proc_entry == NULL)
		return;

#ifdef CONFIG_VIDEO_PROC_FS
	remove_proc_entry("spca50x", video_proc_entry);
#else /* CONFIG_VIDEO_PROC_FS */
	remove_proc_entry("spca50x", 0);
#endif /* CONFIG_VIDEO_PROC_FS */
}
#endif /* CONFIG_PROC_FS */

/**********************************************************************
 *
 * Camera interface
 *
 **********************************************************************/

static int spca50x_reg_write(struct usb_device *dev,
			   __u16 reg,
			   __u16 index,
			   __u16 value)
{
	int rc;

	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev, 0),
		reg,
		USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		value, index, NULL, 0, HZ);

	PDEBUG(5, "reg write: 0x%02X,0x%02X:0x%02X, 0x%x", reg, index, value, rc);

	if (rc < 0)
		err("reg write: error %d", rc);

	return rc;
}

static int spca50x_reg_read_with_value(struct usb_device *dev,
                                       __u16              reg, // bRequest
                                       __u16              value, // wValue
                                       __u16              index, // wIndex
                                       __u16              length) // wLength
{
	int rc;
	unsigned char buffer[4]={0,0,0,0};

	rc = usb_control_msg(dev,
		usb_rcvctrlpipe(dev, 0),
		reg,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		(__u16)value, (__u16)index, buffer, length, HZ);

	PDEBUG(5, "reg read: 0x%02X,0x%02X:0x%04X", reg, index, *(int *)&buffer[0]);

	if (rc < 0) {
		err("reg read: error %d", rc);
		return rc;
	} else {
		return *(int *)&buffer[0];
	}
}

/* returns: negative is error, pos or zero is data */
static int spca50x_reg_read(struct usb_device *dev, 
                            __u16              reg, // bRequest
                            __u16              index, // wIndex
                            __u16              length) // wLength
{
	return spca50x_reg_read_with_value(dev, reg, 0, index, length);
}

/* Read a value from the I2C bus. Returns the value read */
static int spca50x_read_i2c(struct usb_spca50x *spca50x,__u16 device,__u16 address)
{
	struct usb_device *dev = spca50x->dev;
	int err_code;
	int retry;
	int ctrl = spca50x->i2c_ctrl_reg; //The I2C control register
	int base = spca50x->i2c_base; //The I2C base address

	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_DEVICE, device);
	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_SUBADDR, address);
	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_TRIGGER, 
		SPCA50X_I2C_TRIGGER_BIT);
	/* Hmm. 506 docs imply we should poll the ready register before reading the return value */
	/* Poll the status register for a ready status*/
	/* Doesn't look like the windows driver does tho' */
	retry = 60;
	while(--retry)
	{
		err_code = spca50x_reg_read(dev, ctrl, base + SPCA50X_I2C_STATUS, 1);
		if (err_code < 0) 
			PDEBUG(1, "Error reading I2C status register");
		if (!err_code) 
			break;
	}
	if (!retry) 
		PDEBUG(1, "Too many retries polling I2C status after write to register");
	err_code = spca50x_reg_read(dev, ctrl, base + SPCA50X_I2C_READ, 1);
	if (err_code < 0) 
		PDEBUG(1, "Failed to read I2C register at %d:%d",device,address);
	PDEBUG(3, "Read %d from %d:%d",err_code,device,address);
	return err_code;
}

static int spca50x_read_SAA7113_status(struct usb_spca50x *spca50x)
{
	int value=0;

	value=spca50x_read_i2c(spca50x,SAA7113_I2C_BASE_READ,SAA7113_REG_STATUS);
	if(value<0) PDEBUG(1,"Failed to read SAA7113 status");

	PDEBUG(1,"7113 status : ");
	PDEBUG(1,"  READY %s",(SAA7113_STATUS_READY(value)?"YES":"NO"));
	PDEBUG(1,"  COPRO %s",(SAA7113_STATUS_COPRO(value)?"YES":"NO"));
	/*PDEBUG(1,"  SLTCA %s",(SAA7113_STATUS_SLTCA(value)?"YES":"NO"));*/
	PDEBUG(1,"  WIPA  %s",(SAA7113_STATUS_WIPA(value)?"YES":"NO"));
	PDEBUG(1,"  GLIMB %s",(SAA7113_STATUS_GLIMB(value)?"YES":"NO"));
	PDEBUG(1,"  GLIMT %s",(SAA7113_STATUS_GLIMT(value)?"YES":"NO"));
	PDEBUG(1,"  FIDT  %s",(SAA7113_STATUS_FIDT(value)?"YES":"NO"));
	PDEBUG(1,"  HLVLN %s",(SAA7113_STATUS_HLVLN(value)?"YES":"NO"));
	PDEBUG(1,"  INTL  %s",(SAA7113_STATUS_INTL(value)?"YES":"NO"));

	return value;
}

static int spca50x_write_i2c(struct usb_spca50x *spca50x,__u16 device,__u16 subaddress,__u16 data)
{
	struct usb_device *dev=spca50x->dev;
	int err_code;
	int retry;
	int ctrl = spca50x->i2c_ctrl_reg; //The I2C control register
	int base = spca50x->i2c_base; //The I2C base address

	/* Tell the SPCA50x i2c subsystem the device address of the i2c device */
	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_DEVICE,device);

	/* Poll the status register for a ready status*/
	retry = 60; // Arbitrary
	while(--retry)
	{
		err_code = spca50x_reg_read(dev, ctrl, base + SPCA50X_I2C_STATUS, 1);
		if (err_code < 0) 
			PDEBUG(1,"Error reading I2C status register");
		if (!err_code) 
			break;
	}
	if (!retry) 
		PDEBUG(1, "Too many retries polling I2C status");

	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_SUBADDR, subaddress);
	err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_VALUE,data);
	if (spca50x->i2c_trigger_on_write)
		err_code = spca50x_reg_write(dev, ctrl, base + SPCA50X_I2C_TRIGGER,
			SPCA50X_I2C_TRIGGER_BIT);

	/* Poll the status register for a ready status*/
	retry = 60;
	while(--retry)
	{
		err_code=spca50x_reg_read(dev, ctrl, SPCA50X_I2C_STATUS, 2);
		if (err_code < 0) 
			PDEBUG(1,"Error reading I2C status register");
		if (!err_code) 
			break;
	}
	if (!retry) 
		PDEBUG(1, "Too many retries polling I2C status after write to register");

	if (debug > 2)
	{
		err_code = spca50x_read_i2c(spca50x, device, subaddress);
		if (err_code < 0) 
		{
			PDEBUG(3, "Can't read back I2C register value for %d:%d",
				device, subaddress); 
		}
		else if ((err_code & 0xff) != (data & 0xff)) 
			PDEBUG(3, "Read back %x should be %x at subaddr %x",
				err_code, data, subaddress);
	}
	return 0;
}

static int spca50x_write_i2c_vector(struct usb_spca50x *spca50x,__u16 device,__u16 data[][2])
{
	int I=0;

	while(data[I][0])
	{
		spca50x_write_i2c(spca50x,device,data[I][0],data[I][1]);
		I++;
	}
	return 0;
}

static int spca50x_write_vector(struct usb_spca50x *spca50x,__u16 data[][3])
{
	struct usb_device *dev=spca50x->dev;
	int err_code;

	int I=0;
	while((data[I][0])!=(__u16)0 || (data[I][1])!=(__u16)0 || (data[I][2])!=(__u16)0)
	{
		err_code = spca50x_reg_write(dev, data[I][0], (__u16)(data[I][2]), 
			(__u16)(data[I][1]));
		if(err_code < 0) 
		{ 
			PDEBUG(1, "Register write failed for 0x%x,0x%x,0x%x", 
				data[I][0],data[I][1], data[I][2]); 
			return -1; 
		}
		I++;
	}
	return 0;
}

static int spca50x_reset(struct usb_device *dev, unsigned char reset_type)
{
	PDEBUG(4, "Reset: type=0x%X (Currently unimplemented)", reset_type);

	return 0;
}

/* Temporarily stops SPCA50X from functioning. Must do this before changing
 * registers while the camera is streaming */
static inline int spca50x_stop(struct usb_device *dev)
{
	PDEBUG(4, "stopping (unimplemented)");
	return 0;
}

/* Restarts SPCA50X after spca50x_stop() is called */
static inline int spca50x_restart(struct usb_device *dev)
{
	PDEBUG(4, "restarting (unimplemented)");
	return 0;
}

static int spca50x_set_packet_size(struct usb_spca50x *spca50x, int size)
{
	int alt, mult;

	if (spca50x_stop(spca50x->dev) < 0)
		return -EIO;

	mult = size >> 5;

	if      (size == 0)    alt = SPCA50X_ALT_SIZE_0;
	else if (size == 128)  alt = SPCA50X_ALT_SIZE_128;
	else if (size == 256)  alt = SPCA50X_ALT_SIZE_256;
	else if (size == 384)  alt = SPCA50X_ALT_SIZE_384;
	else if (size == 512)  alt = SPCA50X_ALT_SIZE_512;
	else if (size == 640)  alt = SPCA50X_ALT_SIZE_640;
	else if (size == 768)  alt = SPCA50X_ALT_SIZE_768;
	else if (size == 896)  alt = SPCA50X_ALT_SIZE_896;
	else if (size == 1023) alt = SPCA50X_ALT_SIZE_1023;
	else
	{
		/* if an unrecognised size, default to the minimum */
		PDEBUG(5,"Set packet size: invalid size (%d), defaulting to %d",
		         size, SPCA50X_ALT_SIZE_128);
		alt = SPCA50X_ALT_SIZE_128;
	}

	PDEBUG(3, "set packet size: %d, mult=%d, alt=%d", size, mult, alt);

	/* FIXME: Should do this. Don't know how to on SPCA50x.
	 *        if (spca50x_reg_write(spca50x->dev, SPCA50X_REG_FIFO_PACKET_SIZE, mult) < 0)
	 *            return -ENOMEM;
	 */
	PDEBUG(5,"iface alt: %d %d ",spca50x->iface, alt);
	if (usb_set_interface(spca50x->dev, spca50x->iface, alt) < 0) {
		err("Set packet size: set interface error");
		return -EBUSY;
	}

	/* FIXME: Should we only reset the FIFO?
	 *        if (spca50x_reset(spca50x->dev, SPCA50X_RESET_NOREGS) < 0)
	 *            return -ENOMEM;
	 */

	spca50x->packet_size = size;

	if (spca50x_restart(spca50x->dev) < 0)
		return -EIO;

	return 0;
}

/* Returns number of bits per pixel (regardless of where they are located; planar or
 * not), or zero for unsupported format.
 */
static int spca50x_get_depth(int palette)
{
	switch (palette)
	{
//	case VIDEO_PALETTE_GREY:     return 8;
	case VIDEO_PALETTE_RGB565:   return 16;
	case VIDEO_PALETTE_RGB24:    return 24;

//	case VIDEO_PALETTE_YUV422:   return 16;
//	case VIDEO_PALETTE_YUYV:     return 16;
//	case VIDEO_PALETTE_YUV420:   return 24;
	case VIDEO_PALETTE_YUV420P:  return 12; /* strange need 12 this break the read method for this planar mode (6*8/4) */
//	case VIDEO_PALETTE_YUV422P:  return 24; /* Planar */

	case VIDEO_PALETTE_RGB32:    return 32;
	case VIDEO_PALETTE_RAW_JPEG: return 24; /* raw jpeg. what should we return ?? */
	default:                     return 0;  /* Invalid format */
	}
}

/**********************************************************************
* spca50x_isoc_irq
* Function processes the finish of the USB transfer by calling 
* spca50x_move_data function to move data from USB buffer to internal
* driver structures 
***********************************************************************/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void spca50x_isoc_irq(struct urb *urb, struct pt_regs *regs)
#else
static void spca50x_isoc_irq(struct urb *urb)
#endif
{	int i;
	int len;
	struct usb_spca50x *spca50x;

	if (!urb->context) 
	{
		PDEBUG(4, "no context");
		return;
	}

	spca50x = (struct usb_spca50x *) urb->context;

	if (!spca50x->dev || !spca50x->user) {
		PDEBUG(4, "no device, or not open");
		return;
	}

	if (!spca50x->streaming) {
	  /* Always get some of these after close but before packet engine stops */
		PDEBUG(4, "hmmm... not streaming, but got interrupt");
		return;
	}

	/* Copy the data received into our scratch buffer */
	if (spca50x->curframe >= 0) {
		len = spca50x_move_data(spca50x, urb);
	}
	else if (waitqueue_active(&spca50x->wq)) {
		wake_up_interruptible(&spca50x->wq);
	}

	/* Move to the next sbuf */
	spca50x->cursbuf = (spca50x->cursbuf + 1) % SPCA50X_NUMSBUF;

	urb->dev = spca50x->dev;
	urb->status = 0;
	spin_lock(&spca50x->lock_urb);
/***************************************************************/	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 4)
	if ((i = usb_submit_urb(urb, GFP_ATOMIC)) != 0)
#else
	if ((i = usb_submit_urb(urb)) != 0)
#endif
		err("usb_submit_urb() ret %d", i);
/*****************************************************************/
	spin_unlock(&spca50x->lock_urb);
	/* let's try to set up brightness for the next frame*/
	if (autoadjust && !spca50x->bh_requested)
	{
		spca50x->task.data = spca50x;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
		spca50x->task.func = auto_bh;
		schedule_work(&spca50x->task);
#else
		spca50x->task.sync = 0;
		spca50x->task.routine = auto_bh;
		SCHEDULE_TASK(&spca50x->task);
#endif
		spca50x->bh_requested = 1;
	}
	
	return;
}


/**********************************************************************
* spca50x_init_isoc
* Function starts the ISO USB transfer by enabling this process
* from USB side and enabling ISO machine from the chip side
***********************************************************************/
static void spca50x_hardware_cmd (struct usb_spca50x *spca50x)
{	int err_code;
	struct usb_device *dev=spca50x->dev;
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA500:
		{
			err_code = spca500_initialise(spca50x);
			
			init_jpeg_decoder();
			break;
		}

		case BRIDGE_SPCA501:
		{
			/* Enable ISO packet machine CTRL reg=2,
			 * index=1 bitmask=0x2 (bit ordinal 1)
			 */
			err_code = spca50x_reg_write(dev, SPCA501_REG_CTLRL, 
			                             (__u16)0x1, (__u16)0x2);
			break;
		}

		case BRIDGE_SPCA504:
			spca504_initialize(spca50x);
			init_jpeg_decoder();
			break;
		case BRIDGE_SPCA504_PCCAM600:
		{	/* specific code for clicsmart is in pccam600 */
			spca504_pccam600_initialize(spca50x);
			
			init_jpeg_decoder();
			break;
		}
		case BRIDGE_SPCA533:	
		case BRIDGE_SPCA504B:
		{
			spca504B_start(spca50x);
			init_jpeg_decoder();
			break;
		}

		case BRIDGE_SPCA505:
		case BRIDGE_SPCA506:
		{
			/* Enable ISO packet machine - should we do this here or in ISOC init ? */
			err_code = spca50x_reg_write(dev, SPCA50X_REG_USB,
				SPCA50X_USB_CTRL, SPCA50X_CUSB_ENABLE);
			
			spca50x_reg_write(dev, 0x5, 0x0, 0x0);
			spca50x_reg_write(dev, 0x5, 0x0, 0x1);
			spca50x_reg_write(dev, 0x5, 0x11, 0x2);
			break;
		}

		case BRIDGE_SPCA508:
		case BRIDGE_SPCA561:
		{
			/* Video ISO enable, Video Drop Packet enable: */
			err_code = spca50x_reg_write(dev, 0, 0x8112, 0x10 | 0x20); 
			break;
		}
	}

}
static int spca50x_init_isoc(struct usb_spca50x *spca50x)
{
	
	struct urb *urb;
	int fx, err, n;

	PDEBUG(3, "*** Initializing capture ***");

	spca50x->compress = 0;
	spca50x->curframe = 0;
	spca50x->cursbuf = 0;
	spca50x->scratchlen = 0;

	spca50x_set_packet_size(spca50x, spca50x->pipe_size);
	PDEBUG(2, "setpacketsize %d",spca50x->pipe_size);

	for (n = 0; n < SPCA50X_NUMSBUF; n++) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
		urb = usb_alloc_urb(FRAMES_PER_DESC,GFP_KERNEL);
#else
		urb = usb_alloc_urb(FRAMES_PER_DESC);
#endif
		if (!urb) {
			err("init isoc: usb_alloc_urb ret. NULL");
			return -ENOMEM;
		}
		spca50x->sbuf[n].urb = urb;
		urb->dev = spca50x->dev;
		urb->context = spca50x;
		urb->pipe = usb_rcvisocpipe(spca50x->dev, SPCA50X_ENDPOINT_ADDRESS);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		urb->transfer_flags = URB_ISO_ASAP;
		urb->interval = 1;
#else
		urb->transfer_flags = USB_ISO_ASAP;
#endif
		urb->transfer_buffer = spca50x->sbuf[n].data;
		urb->complete = spca50x_isoc_irq;
		urb->number_of_packets = FRAMES_PER_DESC;
		urb->transfer_buffer_length = spca50x->packet_size * FRAMES_PER_DESC;
		for (fx = 0; fx < FRAMES_PER_DESC; fx++) {
			urb->iso_frame_desc[fx].offset = spca50x->packet_size * fx;
			urb->iso_frame_desc[fx].length = spca50x->packet_size;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,41)
	spca50x->sbuf[SPCA50X_NUMSBUF - 1].urb->next = spca50x->sbuf[0].urb;
	for (n = 0; n < SPCA50X_NUMSBUF - 1; n++)
		spca50x->sbuf[n].urb->next = spca50x->sbuf[n + 1].urb;
#endif
	spca50x_hardware_cmd (spca50x);
	for (n = 0; n < SPCA50X_NUMSBUF; n++) {
		spca50x->sbuf[n].urb->dev = spca50x->dev;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
		err = usb_submit_urb(spca50x->sbuf[n].urb,GFP_KERNEL);
#else
		err = usb_submit_urb(spca50x->sbuf[n].urb);
#endif
		if (err)
			err("init isoc: usb_submit_urb(%d) ret %d", n, err);
	}

	for (n = 0; n < SPCA50X_NUMFRAMES; n++)
	{
		spca50x->frame[n].grabstate = FRAME_UNUSED;
		spca50x->frame[n].scanstate = STATE_SCANNING;
	}
	

// Here will be find a pipe ?
/* strange that will do nothing MX 30012004*/
	//usb_clear_halt(dev, 0);
	//usb_clear_halt(dev, urb->pipe);
	spca50x->streaming = 1;
	
	return 0;

}


/**********************************************************************
* spca50x_stop_isoc
* Function stops the USB ISO pipe by stopping the chip ISO machine
* and stopping USB transfer
***********************************************************************/
static void spca50x_stop_isoc(struct usb_spca50x *spca50x)
{
	int n;

	if (!spca50x->streaming || !spca50x->dev)
		return;

	PDEBUG (3, "*** Stopping capture ***");

	/* For some reason this seems to lock the driver on close as
	 * still in use
	 */
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA500:
		{
			/* stop the transfer */
			// Tha's the code need for familycam 300
		 	//spca50x_reg_write(spca50x->dev, 0x00, 0x8000, 0x0000);
		/*
			spca50x_reg_write(spca50x->dev, 0xe1, 0x0001, 0x0000);
			if (spca50x_reg_readwait(spca50x->dev, 0x06, 0, 0) != 0)
			{
				PDEBUG(2, "spca50x_reg_readwait() failed");
			}

			spca50x_reg_write(spca50x->dev, 0x0c, 0x0000, 0x0000);
			spca50x_reg_write(spca50x->dev, 0x30, 0x0000, 0x0004);
			if (spca500_full_reset(spca50x) < 0)
			{
				PDEBUG(2, "spca50x_full_reset() failed");
			}
		*/
			break;
		}
		case BRIDGE_SPCA501:
		{
			/* Disable ISO packet machine CTRL reg=2, index=1 bitmask=0x0 (bit ordinal 1)*/
			spca50x_reg_write(spca50x->dev, SPCA501_REG_CTLRL, 
				(__u16)0x1, (__u16)0x0);
			break;
		}
		case BRIDGE_SPCA504_PCCAM600:
		case BRIDGE_SPCA504:
			spca50x_reg_write(spca50x->dev, 0x00, 0x2000, 0x0000);
			
			if (spca50x->desc == AiptekMiniPenCam13){
					/* spca504a aiptek */
			
			spca504A_acknowledged_command(spca50x, 0x08, 6, 0, 0x86, 1);
			spca504A_acknowledged_command(spca50x, 0x24, 0x0000, 0x0000, 0x9d,1);
			spca504A_acknowledged_command(spca50x, 0x01, 0x000f, 0x0000, 0xFF,1);
			} else {
			spca504_acknowledged_command(spca50x, 0x24, 0x0000, 0x0000);
			spca50x_reg_write(spca50x->dev, 0x01, 0x000f, 0x0);
			}
			break;
			
		case BRIDGE_SPCA533:	
		case BRIDGE_SPCA504B:
		{
			spca504B_stop ( spca50x );
			break;
		}	
		case BRIDGE_SPCA505:
		case BRIDGE_SPCA506:
		{
			spca50x_reg_write(spca50x->dev,0x2,0x0,0x0); //Disable ISO packet machine
			break;
		}
		case BRIDGE_SPCA508:
		case BRIDGE_SPCA561:
		{
			// Video ISO disable, Video Drop Packet enable:
			spca50x_reg_write(spca50x->dev, 0, 0x8112, 0x20); 
			break;
		}
	}

	spca50x_set_packet_size(spca50x, 0);

	spca50x->streaming = 0;
/*************************************************************/
spin_lock(&spca50x->lock_urb);
	/* Unschedule all of the iso td's */
	for (n = SPCA50X_NUMSBUF - 1; n >= 0; n--) {
		if (spca50x->sbuf[n].urb) {
			usb_unlink_urb(spca50x->sbuf[n].urb);
			usb_free_urb(spca50x->sbuf[n].urb);
			spca50x->sbuf[n].urb = NULL;
		}
	}
spin_unlock(&spca50x->lock_urb);
/************************************************************/
	PDEBUG (3, "*** Capture stopped ***");
}

/**********************************************************************
* spca50x_find_mode_index
* Function finds the mode index in the modes table
***********************************************************************/
static inline int spca50x_find_mode_index(int   width,
                                          int   height, 
                                          __u16 ext_modes [][6])
{
	int i = 0;
	int j = 0;
	/* search through the mode list until we hit one which has
	 * a smaller or equal resolution, or we run out of modes.
	 */
	for (i = 0; ext_modes[i][0] && 
		(ext_modes[i][0] >= width && ext_modes[i][1] >= height);
		i++)
	{ if (((ext_modes[i][2] & 0xF0) >> 4) == 0 ) j = (i & 0x0F) << 4;
	}
	/* search to found the first hardware available mode */
	i--; /* FIXME */
	PDEBUG(2, "find Hardware mode %d soft mode %d", j >> 4 ,i);
	/* check whether we found the requested mode, or ran out of
	 * modes in the list.
	 */
	if (!ext_modes[i][0]         ||
	    ext_modes[i][0] != width ||
	    ext_modes[i][1] != height )
	{
		PDEBUG(3, "Failed to find mode %d x %d", width, height);
		return -EINVAL;
	}
	i = (i & 0x0F) | j;
	PDEBUG(2, "Pass Hardware mode %d soft mode %d", (i >> 4) ,(i & 0x0F));
	/* return the index of the requested mode */
	return i;
}

/**********************************************************************
* spca50x_smallest_mode_index
* Function finds the mode index in the modes table of the smallest
* available mode.
***********************************************************************/
static inline int spca50x_smallest_mode_index(__u16 ext_modes [][6],
                                              int  *cols,
                                              int  *rows,
					      int  *pipe)
{
	int i;
	int width  = INT_MAX;
	int height = INT_MAX;
	int intpipe = 1023 ;
	int index  = -EINVAL;

	/* search through the mode list until we run out of modes */
	/* and take the smallest hardware mode and pipe size*/
	for (i = 0; ext_modes[i][0]; i++)
	//i=0;
	{
		if (ext_modes[i][0] < width || ext_modes[i][1] < height)
		{ if (!(ext_modes[i][2] & 0xF0)){
			width  = ext_modes[i][0];
			height = ext_modes[i][1];
			intpipe = ext_modes[i][5];
			index  = i;
			}
		}
	}

	/* return the index of the smallest mode */
	if (index != -EINVAL)
	{
		*pipe = intpipe;
		if (cols != NULL)
			*cols = width;
		if (rows != NULL)
			*rows = height;
	}
	return i;
}

/**********************************************************************
* spca50x_set_mode
* Function sets up the resolution directly. 
* Attention!!! index, the index in modes array is NOT checked. 
***********************************************************************/
static inline int spca50x_set_mode(struct usb_spca50x *spca50x,
                                   int                 index,
                                   __u16               ext_modes [][6])
{
	struct usb_device *dev = spca50x->dev;
	int i;
	int mode = 0;
	int method = 0;
	int crop = 0, cropx1 = 0, cropx2 = 0, cropy1 = 0, cropy2 = 0, x = 0, y =
		0;

	if (spca50x->bridge == BRIDGE_SPCA505
	    || spca50x->bridge == BRIDGE_SPCA506) {
		spca50x_reg_write (dev, SPCA50X_REG_COMPRESS, 0x0,
				   ext_modes[index][2]);
		if (spca50x->bridge == BRIDGE_SPCA505) {
			spca50x_reg_write (dev, SPCA50X_REG_COMPRESS, 0x6,
					   ext_modes[index][3]);
			spca50x_reg_write (dev, SPCA50X_REG_COMPRESS, 0x7,
					   ext_modes[index][4]);
		}
	} else if (spca50x->bridge == BRIDGE_SPCA501) {
		mode = (ext_modes[index][2]) & 0x0F;
		method = (ext_modes[index][2] & 0xF0) >> 4;
		spca50x_reg_write (dev, SPCA50X_REG_USB, 0x6,
				   ext_modes[index][3]);
		spca50x_reg_write (dev, SPCA50X_REG_USB, 0x7,
				   ext_modes[index][4]);
	} else if (spca50x->bridge == BRIDGE_SPCA508) {
		spca50x_reg_write (dev, 0, 0x8500, ext_modes[index][2]);	// mode
		spca50x_reg_write (dev, 0, 0x8700, ext_modes[index][3]);	// clock
	} else if (spca50x->bridge == BRIDGE_SPCA561) {
		spca50x_reg_write (dev, 0, 0x8500, ext_modes[index][2]);	// mode
		spca50x_reg_write (dev, 0, 0x8700, ext_modes[index][3]);	// clock
	} else if (spca50x->bridge == BRIDGE_SPCA500)
		
	{
		mode = (ext_modes[index][2]) & 0x0F;
		method = (ext_modes[index][2] & 0xF0) >> 4;
		x = ext_modes[index][3] & 0x0F;
		y = ext_modes[index][4] & 0x0F;
		/* Experimental Wait until Marek Test 
		// set x multiplier
		spca50x_reg_write (dev, 0, 0x8001,ext_modes[index][3]);
		// set y multiplier 
		spca50x_reg_write (dev, 0, 0x8002,ext_modes[index][4]);
		// use compressed mode, VGA, with mode specific subsample 
		spca50x_reg_write (dev, 0, 0x8003,((ext_modes[index][2] & 0x0f) << 4));
		*/
	} else if (spca50x->bridge == BRIDGE_SPCA504) {
		mode = (ext_modes[index][2] + 3) & 0x0F;
		method = (ext_modes[index][2] & 0xF0) >> 4;
		x = ext_modes[index][3] & 0x0F;
		y = ext_modes[index][4] & 0x0F;
		if (spca50x->desc == AiptekMiniPenCam13) {
			/* spca504a aiptek */
			spca504A_acknowledged_command (spca50x, 0x8, mode, 0,
						       (0x80 | (mode & 0x0F)),
						       1);
			spca504A_acknowledged_command (spca50x, 1, 3, 0, 0x9F,
						       0);
		} else {
			spca504_acknowledged_command (spca50x, 0x8, mode, 0);
		}
	} else if (spca50x->bridge == BRIDGE_SPCA504_PCCAM600) {
		spca50x_reg_write (spca50x->dev, 0xa0, (0x0500 | (ext_modes[index][2] & 0x0F)), 0x0);	// capture mode
		spca50x_reg_write (spca50x->dev, 0x20, 0x1, (0x0500 | (ext_modes[index][2] & 0x0F)));	// snapshot mode
	} else if (spca50x->bridge == BRIDGE_SPCA504B) {
		mode = ext_modes[index][2] & 0x0F;
		method = (ext_modes[index][2] & 0xF0) >> 4;
		x = ext_modes[index][3] & 0x0F;
		y = ext_modes[index][4] & 0x0F;
		spca504B_SetSizeType (spca50x, mode, 6);
	} else if (spca50x->bridge == BRIDGE_SPCA533) {
		if (spca50x->desc == MegapixV4) {
			// megapix specific stuff
			spca533_Megapix (spca50x);
		} else {
			mode = ext_modes[index][2] & 0x0F;
			method = (ext_modes[index][2] & 0xF0) >> 4;
			x = ext_modes[index][3] & 0x0F;
			y = ext_modes[index][4] & 0x0F;
			spca504B_SetSizeType (spca50x, mode, 6);
		}
	}
	spca50x->pipe_size = ext_modes[index][5];
	spca50x->method = method;
	/* initialize the hdrwidth and hdrheight for the first init_source */
	if (!(spca50x->hdrwidth) || !(spca50x->hdrheight)) {

		spca50x->hdrwidth = spca50x->width;
		spca50x->hdrheight = spca50x->height;
	}
	/* precompute the crop x y value for each frame */
	if (!method) {
		/* nothing todo hardware found stream */
		cropx1 = cropx2 = cropy1 = cropy2 = x = y = 0;
	}
	if (method & 0x01) {
		/* cropping method */
		if (spca50x->hdrwidth > spca50x->width) {
			
			crop = (spca50x->hdrwidth - spca50x->width);
			if(spca50x->cameratype == JPEG) crop = crop >> 4 ;
			cropx1 = crop >> 1;
			cropx2 = cropx1 + (crop % 2);
		} else {
			cropx1 = cropx2 = 0;
		}
		if (spca50x->hdrheight > spca50x->height) {
			crop = (spca50x->hdrheight - spca50x->height);
			if(spca50x->cameratype == JPEG) crop = crop >> 4 ;
			cropy1 = crop >> 1;
			cropy2 = cropy1 + (crop % 2);
		} else {
			cropy1 = cropy2 = 0;
		}
	}
	if (method & 0x02) {
		/* what can put here for div method */
	}
	if (method & 0x04) {
		/* and here for mult */
	}
	PDEBUG (2, "Found code %d method %d", mode, method);
	PDEBUG (2, "Soft Win width height %d x %d", spca50x->width,
		spca50x->height);
	PDEBUG (2, "Hard Win width height %d x %d", spca50x->hdrwidth,
		spca50x->hdrheight);

	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		spca50x->frame[i].method = spca50x->method;
		spca50x->frame[i].cameratype = spca50x->cameratype;
		spca50x->frame[i].cropx1 = cropx1;
		spca50x->frame[i].cropx2 = cropx2;
		spca50x->frame[i].cropy1 = cropy1;
		spca50x->frame[i].cropy2 = cropy2;
		spca50x->frame[i].x = x;
		spca50x->frame[i].y = y;
		spca50x->frame[i].hdrwidth = spca50x->hdrwidth;
		spca50x->frame[i].hdrheight = spca50x->hdrheight;
		spca50x->frame[i].width = spca50x->width;
		spca50x->frame[i].height = spca50x->height;
		spca50x->frame[i].scanlength =
			spca50x->width * spca50x->height * 3 / 2;
		// ?? assumes 4:2:0 data
	}

	return 0;
}

/**********************************************************************
* spca50x_mode_init_regs
* Function sets up the resolution with checking if it's necessary 
***********************************************************************/
static int
spca50x_mode_init_regs(struct usb_spca50x *spca50x,
		       int width, int height, int mode, __u16 ext_modes [][6])
{
	int i ,j;
	int r;
	int was_streaming;

	/* Avoid changing to already selected mode */
	if(spca50x->width == width && spca50x->height == height) 
		return 0;
	
	/* Can't do this yet on the CCD .. or can we ? It appears we can ! */
	/*        if(spca50x->ccd) return -ECHRNG; */

	/* find a suitable mode */
	if ((i = spca50x_find_mode_index(width, height, ext_modes)) == -EINVAL)
		return -EINVAL;
		/* keep the hardware index from the hight nibble */
	j = (i & 0xF0) >> 4;
		/* the software too */
	i = i & 0x0F;
	/* make sure that the suitable mode isn't the current mode */
	if (spca50x->width  == ext_modes[i][0] &&
	    spca50x->height == ext_modes[i][1]) 
		return 0;
	
	PDEBUG(1, "Mode changing to %d,%d", width, height);

	down(&spca50x->lock);

	was_streaming = spca50x->streaming;
	
	if (was_streaming) 
	 if (spca50x->bridge != BRIDGE_SPCA500)
		spca50x_stop_isoc(spca50x);
		
	spca50x->hdrwidth = ext_modes[j][0];
	spca50x->hdrheight = ext_modes[j][1];
	spca50x->width  = ext_modes[i][0];
	spca50x->height = ext_modes[i][1];

	spca50x_set_mode(spca50x, i, ext_modes);

	r = 0;
	if (was_streaming) {
	 if (spca50x->bridge != BRIDGE_SPCA500){
			r = spca50x_init_isoc(spca50x);
	 	} else {
	 		spca50x_hardware_cmd (spca50x);
	 }
	}
	up(&spca50x->lock);

	return r;
}

/**********************************************************************
* spca50x_get_brightness
* Function reads the brightness from the camera
* Receives the pointer to the description structure
* returns the value of prightness
**********************************************************************/
static inline __u16 spca50x_get_brightness(struct usb_spca50x *spca50x)
{
	__u16 brightness; // value of the brightness
	if(spca50x->ccd)
		brightness = spca50x_read_i2c(spca50x, SAA7113_I2C_BASE_READ, 0xa);
	else	
	{
		switch(spca50x->bridge)
		{
			case BRIDGE_SPCA500:
			{
				/* brightness is stored by the bridge as
				 * 129 -> 128, with 0 being the midpoint.
				 * v4l always keeps these values unsigned
				 * 16 bit, that is 0 to 65535.
				 * scale accordingly.
				 */
				brightness = spca50x_reg_read(spca50x->dev, 0x00, 0x8167, 1);
				brightness = brightness << 8;
				break;
			}
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
			case BRIDGE_SPCA505:
				brightness = spca50x_reg_read(spca50x->dev, 6, 0x51, 2);
				break;
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */
			case BRIDGE_SPCA501:
			{
				brightness = spca50x_reg_read(spca50x->dev, 0x0, 0x00, 2) & 0xFF;
				brightness -= 125;
				brightness <<= 1;
				break;
			}
			case BRIDGE_SPCA533:
			case BRIDGE_SPCA504:
			case BRIDGE_SPCA504B:
			case BRIDGE_SPCA504_PCCAM600:
			{
				/* The chip keeps the brightness as
				 * 129 -> 128 with 0 being the midpoint.
				 * v4l always keeps these values unsigned
				 * 16 bit, that is 0 to 65535.
				 * Scale accordingly.
				 */
				brightness = spca50x_reg_read(spca50x->dev, 0x0, 0x21a7, 2);

				brightness = (((brightness&&0xFF)-128)%255)<<8;
				 break;
			}
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS 
                        case BRIDGE_SPCA508:
                                brightness = spca50x_reg_read(spca50x->dev, 0, 0x8651, 1);
                                brightness = brightness << 8;
                                break;
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */ 
			default:
			{
				brightness = 0;
				break;
			}
		}
	}
	return brightness;
}

/**********************************************************************
* spca50x_set_brightness
* Function sets the brightness to the camera
* Receives the pointer to the description structure 
* and brightness value
**********************************************************************/
static inline void spca50x_set_brightness(struct usb_spca50x *spca50x, 
	__u8 brightness)
{
	if(spca50x->ccd)
		spca50x_write_i2c(spca50x, SAA7113_I2C_BASE_WRITE, 0xa, brightness);
	else	
	{
		switch(spca50x->bridge)
		{
			case BRIDGE_SPCA500:
			{
				/* 0 - 255 with zero at 128 */
				spca50x_reg_write(spca50x->dev,
				                  0x00, 0x8167, brightness);
				break;
			}
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
			case BRIDGE_SPCA505:
			{
				spca50x_reg_write(spca50x->dev, 6, 0x51, brightness);
				spca50x_reg_write(spca50x->dev, 6, 0x52, brightness);
				spca50x_reg_write(spca50x->dev, 6, 0x53, brightness);
				break;
			}
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */
			case BRIDGE_SPCA501:
			{
				spca50x_reg_write(spca50x->dev, 0x0, 0x00, 
					(brightness >> 1) + 125);
				break;
			}
			case BRIDGE_SPCA533:
			case BRIDGE_SPCA504:
			case BRIDGE_SPCA504B:
			case BRIDGE_SPCA504_PCCAM600:
			{
				/* 0 - 255 with zero at 128 */
				spca50x_reg_write(spca50x->dev, 0x0, 0x21a7,
				                  brightness); 
				break;
			}
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS 
                        case BRIDGE_SPCA508:
                        {
                                spca50x_reg_write(spca50x->dev, 0, 0x8651, brightness);
                                spca50x_reg_write(spca50x->dev, 0, 0x8652, brightness);
                                spca50x_reg_write(spca50x->dev, 0, 0x8653, brightness);
                                spca50x_reg_write(spca50x->dev, 0, 0x8654, brightness);
				/* autoadjust is set to 0 to avoid doing repeated brightness
				   calculation on the same frame. autoadjust is set to 0
				   when a new frame is ready. */
				autoadjust = 0;
                                break;
                        }
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */ 
			default:
			{
				break;
			}
		}
	}
}

/**********************************************************************
* spca50x_get_whiteness
* Function reads the "whiteness parameter" from the camera
* Actually "whiteness" is a parameter that close to brighness
* but currently ununderstanded
* Receives the pointer to the description structure
* returns the value of whiteness
**********************************************************************/
static inline __u16 spca50x_get_whiteness(struct usb_spca50x *spca50x)
{
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	__u16 whiteness = 0; //the "whiteness" parameter
	__u16 br = 0; //the brightness value
	int i;   // just an index variable
	__u16 mask; //the bit mask to set bit in p.contrast
	__u8 br_table[] = { 128, 64, 32, 16, 0 }; //the brightness table

	//this parameter is known for spca501 only
	if (spca50x->bridge != BRIDGE_SPCA501) 
		return 0;

	/* Getting "whiteness" (strange parameters in TG 0x0D */
	br = spca50x_reg_read(spca50x->dev, 0x0, 0x0d, 2);
	for (i = 0, mask = 1; br_table[i]; i++, mask <<= 1)
	{
		if (br & br_table[i])
			whiteness |= mask;
	}
	whiteness <<= 4;
	return whiteness;
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
	return 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
}

/**********************************************************************
* spca50x_set_whiteness
* Function sets the "whiteness parameter" from the camera
* Actually "whiteness" is a parameter that close to brighness
* but currently ununderstanded
* Receives the pointer to the description structure 
* and whiteness value
**********************************************************************/
static inline void spca50x_set_whiteness(struct usb_spca50x *spca50x, 
	__u8 whiteness)
{
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	__u8 br_high = 0; //the contrast value
	__u8 br_out = 0; // contrast output value
	int i;   // just an index variable
	__u16 mask; //the bit mask to set bit in contrast
	__u8 br_table[] = { 128, 64, 32, 16, 0 }; //the brightness table


	//this parameter is known for spca501 only
	if (spca50x->bridge != BRIDGE_SPCA501) 
		return;

	PDEBUG(3, "Setting whiteness %d", whiteness);
	br_high = whiteness >> 4;
	for (i = 0, mask = 1; br_table[i]; i++, mask <<= 1)
	{
		if (br_high & mask)
			br_out |= br_table[i];
	}
	PDEBUG(3, "Writing whiteness %d", br_out);
	spca50x_reg_write(spca50x->dev, 0x0, 0x0d, br_out);
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
	return;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
}

/**********************************************************************
* autobrightness
* Function tries to control brightness
* Receives the pointer to camera descriptor and 
* target average luminance
***********************************************************************/
static inline void autobrightness(struct usb_spca50x *spca50x)
{
	__u8 cur_lum = 0; //The current luminance
	__u8 brightness = 0;
	char br_set = 0; //If the brightness should be set
	__u8 lum = spca50x->lum_level; //The desired luminance level
	__u8 delta = lum >> 3; //overshooting that we shouldn't care
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	__u8 whiteness = 0; //whiteness level
	char wh_set = 0; //If whiteness must be set
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

	if (in_interrupt())
	{
		PDEBUG(2, "In intr");
		return;
	}
	////
	/*
	* This algorithm may be dangerous. We MUST be sure that the 
	* !spca50x->streaming will be set to 0 BEFORE kfree(spca50x)
	*/
	////
	if (!spca50x->streaming) //No isoc
		return;
	cur_lum = get_avg_lum(spca50x);
	brightness = spca50x->brightness >> 8;
	if (cur_lum < (lum - delta))
	{
		if (brightness < 255)
		{
			brightness++;
			br_set = 1;
		}
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
		else 
		{
			whiteness = spca50x->whiteness >> 12;
			if (whiteness < 15)
			{   
				whiteness++;
				wh_set = 1;
			}
		}
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
	}
	if (cur_lum > (lum + delta) && brightness > 0)
	{
		brightness--;
		br_set = 1;
	}
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	if ((spca50x->nstable == NSTABLE_MAX || 
		spca50x->nunstable == NUNSTABLE_MAX) && 
		brightness < MIN_BRIGHTNESS)
	{
		whiteness = spca50x->whiteness >> 12;
		if (whiteness > 0)
		{   
			whiteness--;
			wh_set = 1;
		}
	}        
	if (wh_set)
	{
		spca50x_set_whiteness(spca50x, whiteness << 4);
		spca50x->whiteness = whiteness << 12;
		spca50x->nstable = 0;
		spca50x->nunstable = 0;
		return; //if we have set whiteness, we needn't care about brightness
	}
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
	if (br_set)
	{
		spca50x_set_brightness(spca50x, brightness);
		spca50x->brightness = brightness << 8;
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
		if (spca50x->nunstable < NUNSTABLE_MAX)
			spca50x->nunstable++;
		spca50x->nstable = 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
	}
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	else
	{
		if (spca50x->nstable < NSTABLE_MAX)
			spca50x->nstable++;
		else
			spca50x->nunstable = 0;
	}
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
}

static void auto_bh(void* data)
{
	struct usb_spca50x *spca50x = (struct usb_spca50x*)data;
	
	autobrightness(spca50x);
	spca50x->bh_requested = 0; /* bottom half process finished */
}


/**********************************************************************
 *
 * SPCA50X data transfer, IRQ handler
 *
 **********************************************************************/

#ifdef SPCA50X_ENABLE_OSD

static void spca50x_osd_text(struct spca50x_frame *frame,int x,int y,char *text)
{
	int I;
	int J;
	int K;
	int limit;
	limit=strlen(text);

	for(I=0;I<limit;I++)
	{
		/* Offest into font table for first byte of character data */
		int charBase=8*text[I];
		/* Frame pointer */
		char *fp=frame->data+I*3*8+y*8*frame->width*3+x*3;
		/* pointer to character data */
		char *font_char=((char *)(osd_font->data))+charBase;
		for(J=0;J<8;J++)
		{
			int line=font_char[J];
			for(K=0;K<8;K++)
			{
				int level=(line & 0x80)?255:0;
				*(fp+3*K)=level; // R
				*(fp+3*K+1)=level; // G
				*(fp+3*K+2)=level; // B
				line<<=1;
			}
			fp+=frame->width*3; // skip a line of RGB
		}
	}
}

/* On screen display */
static int spca50x_osd_render(struct spca50x_frame *frame,struct timeval *ts)
{
	char seq[40];
	int frameSeq;

	if(!osd_font)
	{
		osd_font=fbcon_find_font("VGA8x8");
	}

	/* No font available */
	if(!osd_font || osd_font->data==NULL) return -EBFONT;

	frameSeq=frame->seq;
	if(frameSeq<0) frameSeq+=256;
	sprintf(seq,"Frame: %3.3d ",frameSeq);

	spca50x_osd_text(frame,0,0,seq);

	sprintf(seq,"%dx%d",frame->width,frame->height);

	spca50x_osd_text(frame,0,1,seq);

	return 0;
}

#endif /* SPCA50X_ENABLE_OSD */

static struct spca50x_frame *spca50x_next_frame(struct usb_spca50x *spca50x,unsigned char *cdata)
{
	int iFrameNext;
	struct spca50x_frame *frame=NULL;
	if( spca50x->frame[spca50x->curframe].grabstate == FRAME_ERROR) {
	/* recycle the bad one */
		frame=&spca50x->frame[spca50x->curframe];
	} else {
		/* Cycle through the frame buffer looking for a free frame to overwrite */
		iFrameNext = (spca50x->curframe + 1) % SPCA50X_NUMFRAMES;
		while(frame==NULL && iFrameNext!=(spca50x->curframe))
		{
		if (spca50x->frame[iFrameNext].grabstate == FRAME_READY  ||
		    spca50x->frame[iFrameNext].grabstate == FRAME_UNUSED ||
		    spca50x->frame[iFrameNext].grabstate == FRAME_ERROR)
		{
			spca50x->curframe = iFrameNext;
			frame=&spca50x->frame[iFrameNext];
			break;
		}
		else
		{
			iFrameNext=(iFrameNext+1)%SPCA50X_NUMFRAMES;
		}
		}
	}
	if(frame==NULL)
	{
		PDEBUG(3,"Can't find a free frame to grab into...using next. "
		         "This is caused by the application not reading fast enough.");
		spca50x->curframe=(spca50x->curframe+1)%SPCA50X_NUMFRAMES;
		frame=&spca50x->frame[spca50x->curframe];
	}

	frame->grabstate=FRAME_GRABBING;

	/* Record the frame sequence number the camera has told us */
	if(cdata)
	{
		switch(spca50x->bridge)
		{
			case BRIDGE_SPCA500: frame->seq=cdata[SPCA500_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA501: frame->seq=cdata[SPCA501_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA504: frame->seq=cdata[SPCA50X_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA504_PCCAM600: frame->seq=cdata[SPCA50X_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA504B: frame->seq=cdata[SPCA50X_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA505: frame->seq=cdata[SPCA50X_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA508: frame->seq=cdata[SPCA508_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA533: frame->seq=cdata[SPCA533_OFFSET_FRAMSEQ]; break;
			case BRIDGE_SPCA561: frame->seq=cdata[SPCA561_OFFSET_FRAMSEQ]; break;
			
		}
	}

	/* Reset some per-frame variables */
	frame->highwater=frame->data;
	frame->scanstate = STATE_LINES;
	frame->segment = 0;
	frame->curpix=0;
	frame->curline=0;
	frame->last_packet=-1;
	frame->mcu = 0;

	return frame;
}
/* Tasklet function to decode */

void outpict_do_tasklet (unsigned long ptr)
{ 	int err;
	memcpy(taskletframe->tmpbuffer, taskletframe->data, 
		taskletframe->highwater - taskletframe->data);
	PDEBUG(2, "Tasklet ask jpeg decoder hdrwidth %d hdrheight %d method %d", taskletframe->hdrwidth,
		taskletframe->hdrheight,taskletframe->method);
	
	err = spca50x_outpicture( taskletframe,force_rgb);						
							
	if (err != 0)
	{
		PDEBUG(2, "frame decoder failed (%d)", err);
	}
	taskletframe->grabstate = FRAME_DONE;
	if (waitqueue_active(&taskletframe->wq)) 
					wake_up_interruptible(&taskletframe->wq);
	
}

/* ******************************************************************
* spca50x_move_data
* Function serves for moving data from USB transfer buffers
* to internal driver frame buffers.
******************************************************************* */
static int spca50x_move_data(struct usb_spca50x *spca50x, struct urb *urb)
{
	unsigned char *cdata;  //Pointer to buffer where we do store next packet
	unsigned char *pData;  //Pointer to buffer where we do store next packet
	int i, totlen = 0;
	struct spca50x_frame *frame; //Pointer to frame data
	
	
	int iPix; //Offset of pixel data in the ISO packet
	
	
	static int dropframe = 1;
	
	// PDEBUG (4, "Moving %d packets", urb->number_of_packets);

	// Lock the buffers to ensure that they don't vanish while we are 
	// writing to them !
	if(down_trylock(&spca50x->buf_lock)) return -ENOBUFS;

	// Verify the buffers still exist
	if(!spca50x->fbuf)
	{
		up(&spca50x->buf_lock);
		return -ENOBUFS;
	}

	for (i = 0; i < urb->number_of_packets; i++)
	{
		int datalength = urb->iso_frame_desc[i].actual_length;
		int st = urb->iso_frame_desc[i].status;
		int sequenceNumber;

		/* PDEBUG(5,"Packet data [%d,%d,%d]", datalength, st,
		         urb->iso_frame_desc[i].offset); */

		urb->iso_frame_desc[i].actual_length = 0;
		urb->iso_frame_desc[i].status = 0;

		cdata = ((unsigned char *) urb->transfer_buffer) +
		        urb->iso_frame_desc[i].offset;

		/* Check for zero length block or no selected frame buffer */
		if (!datalength || spca50x->curframe == -1) 
			continue;

		PDEBUG(5,"Packet data [%d,%d,%d] Status: %d", datalength, st,
		         urb->iso_frame_desc[i].offset, st);

		if (st) 
			PDEBUG(2, "data error: [%d] len=%d, status=%d", i, datalength, st);

		frame = &spca50x->frame[spca50x->curframe];

#ifdef SPCA50X_ENABLE_RAWPROCENTRY
		if (spca50x->rawBuffer != NULL)
		{
			/* check that there's enough space in the buffer */
			if ((spca50x->rawBufferMax - spca50x->rawBufferSize) < (datalength + 5))
			{
				PDEBUG(3, "raw proc entry buffer is full");
			}
			else
			{
				strncpy(spca50x->rawBuffer + spca50x->rawBufferSize, "FRAME", 5);
				memcpy(spca50x->rawBuffer + spca50x->rawBufferSize + 5,
				       cdata,
				       datalength);
				spca50x->rawBufferSize += datalength + 5;
			}
		}
#endif /* SPCA50X_ENABLE_RAWPROCENTRY */

		/* read the sequence number */
		sequenceNumber = cdata[SPCA50X_OFFSET_SEQUENCE];

		/* check if this is a SPCA500 based device */
		if ((spca50x->bridge == BRIDGE_SPCA500) || (spca50x->bridge == BRIDGE_SPCA533))
		{
			if (sequenceNumber == SPCA50X_SEQUENCE_DROP)
			{
				if (cdata[1] == 0x01)
				{
					sequenceNumber = 0;
				}
				else
				{
					/* drop packet */
					PDEBUG(5, "Dropped packet (expected seq 0x%02x)",
					          frame->last_packet + 1);
					continue;
				}
			}
			else
			{
				sequenceNumber ++;
			}
		}
		else if ((spca50x->bridge == BRIDGE_SPCA504) 
		      || (spca50x->bridge == BRIDGE_SPCA504B)
		      || (spca50x->bridge == BRIDGE_SPCA504_PCCAM600))
		{
			switch (sequenceNumber)
			{
				case 0xfe:
					sequenceNumber = 0;
					break;
				case SPCA50X_SEQUENCE_DROP:
					/* drop packet */
					PDEBUG(5, "Dropped packet (expected seq 0x%02x)",
					          frame->last_packet + 1);
					continue;
				default:
					sequenceNumber ++;
					break;
			}
		}
		else
		{
			/* check if this is a drop packet */
			if (sequenceNumber == SPCA50X_SEQUENCE_DROP)
			{
				PDEBUG(3, "Dropped packet (expected seq 0x%02x)",
					  frame->last_packet + 1);
				continue;
			}
		}

		PDEBUG(3, "spca50x: Packet seqnum = 0x%02x.  curframe=%2d",
		          sequenceNumber, spca50x->curframe);

		pData = cdata;
		// Offset of pixel data in iso packet, after sequence byte at offset 0
		iPix = 1; 

		/* Can we find a frame start */
		if (sequenceNumber == 0)
		{
			iPix = spca50x->header_len;
			PDEBUG(3, "spca50x: Found Frame Start!, framenum = %d", spca50x->curframe);
			switch(spca50x->bridge)
			{
				case BRIDGE_SPCA500:
				{
					PDEBUG(3, "Padding = 0x%02x%02x",
					          cdata[SPCA500_OFFSET_PADDINGHB],
					          cdata[SPCA500_OFFSET_PADDINGLB]);
					PDEBUG(3, "Camera operation mode = %u", cdata[SPCA500_OFFSET_MODE]);
					PDEBUG(3, "Image width = %u, height = %u",
					          cdata[SPCA500_OFFSET_IMGWIDTH],
					          cdata[SPCA500_OFFSET_IMGHEIGHT]);
					PDEBUG(3, "Image mode = %u", cdata[SPCA500_OFFSET_IMGMODE]);
					PDEBUG(3, "Q-table index = %u", cdata[SPCA500_OFFSET_QTBLINDEX]);
					PDEBUG(3, "Frame sequence = %d", cdata[SPCA500_OFFSET_FRAMSEQ]);
					PDEBUG(3, "CDSP information = %d", cdata[SPCA500_OFFSET_CDSPINFO]);
					PDEBUG(3, "GPIO = %d", cdata[SPCA500_OFFSET_GPIO]);
					PDEBUG(3, "AUGPIO = %d", cdata[SPCA500_OFFSET_AUGPIO]);
					break;
				}
				case BRIDGE_SPCA501:
				{
					PDEBUG(3, "GPIO: 0x%02X", cdata[SPCA501_OFFSET_GPIO]);
					PDEBUG(3, "Image type: %d with snapbit %d and snap control %d",
						(cdata[SPCA501_OFFSET_TYPE] & 0xF),
						(cdata[SPCA501_OFFSET_TYPE] & SPCA501_SNAPBIT), 
						(cdata[SPCA501_OFFSET_TYPE] & SPCA501_SNAPCTRL));
					PDEBUG(3,"Frame sequence %d",cdata[SPCA501_OFFSET_FRAMSEQ]);
					PDEBUG(3,"Compression status = %d, Threshold = %d, Quant = %d",
						SPCA501_PROP_COMP_ENABLE(cdata[SPCA501_OFFSET_COMPRESS]), 
						SPCA501_PROP_COMP_THRESH(cdata[SPCA501_OFFSET_COMPRESS]),
						SPCA501_PROP_COMP_QUANT(cdata[SPCA501_OFFSET_COMPRESS]) );
					break;
				}
				case BRIDGE_SPCA505:
				case BRIDGE_SPCA506:
				{
					PDEBUG(3, "Image type: %d with snapbit %d and snap control %d",
						(cdata[SPCA50X_OFFSET_TYPE] & 0xF),
						(cdata[SPCA50X_OFFSET_TYPE] & SPCA50X_SNAPBIT), 
						(cdata[SPCA50X_OFFSET_TYPE] & SPCA50X_SNAPCTRL));
					PDEBUG(3,"Frame sequence %d",cdata[SPCA50X_OFFSET_FRAMSEQ]);
					PDEBUG(3,"Compression status = %d, ThreeBandNormalizarion = %d, Threshold 3D = %d",
						SPCA50X_PROP_COMP_ENABLE(cdata[SPCA50X_OFFSET_COMPRESS]), 
						SPCA50X_PROP_COMP_T3A(cdata[SPCA50X_OFFSET_COMPRESS]), 
						SPCA50X_PROP_COMP_T3D(cdata[SPCA50X_OFFSET_COMPRESS]) );
					break;
				}
				case BRIDGE_SPCA508:
				{
					PDEBUG(3, "Image type: %d with snapbit %d and snap control %d",(cdata[SPCA508_OFFSET_TYPE] & 0xF),
						(cdata[SPCA508_OFFSET_TYPE] & 0x20),
						((cdata[SPCA508_OFFSET_TYPE] & 0x40) != 0));
					PDEBUG(3,"Frame sequence %d", cdata[SPCA508_OFFSET_FRAMSEQ]);
					PDEBUG(3,"Compression status = %d",
						cdata[SPCA508_OFFSET_COMPRESS] & 0x02);
					PDEBUG(3,"Average luminance = %3d\n", cdata[SPCA508_OFFSET_WIN1LUM]);
					break;
				}
			}
			// Start of frame is implicit end of previous frame
			// Check for a previous frame and finish it off if one exists
			if(frame->scanstate == STATE_LINES)
			{
#ifdef SPCA50X_ENABLE_OSD
			 struct timeval* ts;
				PDEBUG(4, "Frame end, curframe = %d, has %d bytes",
					spca50x->curframe,(frame->highwater - frame->data));
				ts = (struct timeval*)(frame->data + MAX_FRAME_SIZE);
				do_gettimeofday(ts);
				if(osd) 
					spca50x_osd_render(frame, ts);
#endif /* SPCA50X_ENABLE_OSD */

				/* Compressed raw jpeg data is already present in frame->data.
				   So for VIDEO_PALETTE_RAW_JPEG we have nothing to do.
				 */
				 
				if (  dumppix!=7 &&
				     ((frame->format == VIDEO_PALETTE_RGB32) ||
				      (frame->format == VIDEO_PALETTE_RGB24) ||
				      (frame->format == VIDEO_PALETTE_RGB565)||
				      (frame->format == VIDEO_PALETTE_YUV420P) ))
					{
					if (dropframe){
					/* Schedule the tasklet */

							taskletframe = frame;
							tasklet_schedule(&outpicttasklet);

						}

				}
					
				if ( dropframe){
					frame->grabstate = FRAME_DONE;
				} else {

					dropframe = 1;
					frame->grabstate = FRAME_ERROR;
				}
				
				
				if (waitqueue_active(&frame->wq)) 
					wake_up_interruptible(&frame->wq);
				// If someone decided to wait for ANY frame - wake him up 
				if (waitqueue_active(&spca50x->wq))
					wake_up_interruptible(&spca50x->wq);
				
				frame = spca50x_next_frame(spca50x, cdata);
			}
			else 
				frame->scanstate=STATE_LINES;
		}

		/* Are we in a frame? */
		if (frame == NULL || frame->scanstate != STATE_LINES) 
			continue;

		PDEBUG(5,"Pkt sequence number %d", sequenceNumber);
		if(sequenceNumber != frame->last_packet + 1 && 
		   frame->last_packet != -1)
		{
			/* Note, may get one of these for the first packet after opening */
			PDEBUG(2,"Out of order packet, last = %d, this = %d",
			         frame->last_packet, sequenceNumber);
			//dropframe = 0; // out of order or corrupt data error 
			dropframe = 1; // Why this feature block 640x480 spca501 so disable it
		}
		frame->last_packet = sequenceNumber;

		


		/* This is the real conversion of the raw camera data to BGR for V4L */
		datalength -= iPix; // correct length for packet header
		pData = cdata + iPix; // Skip packet header (1 or 10 bytes)
		
		
		
		// Consume data
		PDEBUG(5, "Processing packet starting at %d, %d, %d",
		          frame->curpix, frame->segment, frame->width);

		
		
		/* this copy consume input data from the isoc stream */
		
		
			memcpy(frame->highwater, pData, datalength);
			frame->highwater += datalength;
			totlen += datalength;
			
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
		/* autoadjust is set to 1 to so now autobrightness will be
		calculated frome this frame. autoadjust will be set to 0 when
		autobrightness has been corrected if needed. */
		autoadjust = 1;
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */			
		
	
	}
	up(&spca50x->buf_lock);
	return totlen;
}

/****************************************************************************
 *
 * Buffer management
 *
 ***************************************************************************/
static int spca50x_alloc(struct usb_spca50x *spca50x)
{
	int i;

	PDEBUG(4, "entered");

	

	down(&spca50x->buf_lock);
	
	spca50x->tmpBuffer = rvmalloc(MAX_FRAME_SIZE);
	if (spca50x->buf_state == BUF_PEND_DEALLOC) {
		spca50x->buf_state = BUF_ALLOCATED;
		del_timer(&spca50x->buf_timer);
	}

	if (spca50x->buf_state == BUF_ALLOCATED)
		goto out;

	spca50x->fbuf = rvmalloc(SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
	if (!spca50x->fbuf)
		goto error;

	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		spca50x->frame[i].tmpbuffer = spca50x->tmpBuffer;
		spca50x->frame[i].grabstate = FRAME_UNUSED;
		spca50x->frame[i].scanstate = STATE_SCANNING;
		spca50x->frame[i].data = spca50x->fbuf + i * MAX_DATA_SIZE;
		spca50x->frame[i].highwater=spca50x->frame[i].data;
		PDEBUG(4, "frame[%d] @ %p", i, spca50x->frame[i].data);
	}
	
	for (i = 0; i < SPCA50X_NUMSBUF; i++) {
		spca50x->sbuf[i].data = kmalloc(FRAMES_PER_DESC *
			MAX_FRAME_SIZE_PER_DESC, GFP_KERNEL);
		if (!spca50x->sbuf[i].data)
			goto error; 
		PDEBUG(4, "sbuf[%d] @ %p", i, spca50x->sbuf[i].data);
	}
	spca50x->buf_state = BUF_ALLOCATED;
out:
	up(&spca50x->buf_lock);
	PDEBUG(4, "leaving");
	return 0;
error:
	/* FIXME: IMHO, it's better to move error deallocation code here. */

	for (i = 0; i < SPCA50X_NUMSBUF; i++) {

		if (spca50x->sbuf[i].data) {
			kfree(spca50x->sbuf[i].data);
			spca50x->sbuf[i].data = NULL;
		}
	}
	if (spca50x->fbuf)
	{
		rvfree(spca50x->fbuf, SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
		spca50x->fbuf = NULL;
	}
	if (spca50x->tmpBuffer)
	{
		//vfree(spca50x->tmpBuffer);
		rvfree(spca50x->tmpBuffer,MAX_FRAME_SIZE);
		spca50x->tmpBuffer = NULL;
	}

	spca50x->buf_state = BUF_NOT_ALLOCATED;
	up(&spca50x->buf_lock);
	PDEBUG(1, "errored");
	return -ENOMEM;
}

/*
 * - You must acquire buf_lock before entering this function.
 * - Because this code will free any non-null pointer, you must be sure to null
 *   them if you explicitly free them somewhere else!
 */
static void spca50x_do_dealloc(struct usb_spca50x *spca50x)
{
	int i;
	PDEBUG(2, "entered");

	if (spca50x->fbuf) {
		rvfree(spca50x->fbuf, SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
		spca50x->fbuf = NULL;
		for(i = 0; i < SPCA50X_NUMFRAMES; i++)
			spca50x->frame[i].data = NULL;
	}

	if (spca50x->tmpBuffer)
	{
		//vfree(spca50x->tmpBuffer);
		rvfree(spca50x->tmpBuffer,MAX_FRAME_SIZE);
		spca50x->tmpBuffer = NULL;
	}

	for (i = 0; i < SPCA50X_NUMSBUF; i++) {
		if (spca50x->sbuf[i].data) {
			kfree(spca50x->sbuf[i].data);
			spca50x->sbuf[i].data = NULL;
		}
	}

	PDEBUG(2, "buffer memory deallocated");
	spca50x->buf_state = BUF_NOT_ALLOCATED;
	PDEBUG(2, "leaving");
}

static void spca50x_buf_callback(unsigned long data)
{
	struct usb_spca50x *spca50x = (struct usb_spca50x *)data;
	PDEBUG(4, "entered");
	down(&spca50x->buf_lock);

	if (spca50x->buf_state == BUF_PEND_DEALLOC)
		spca50x_do_dealloc(spca50x);

	up(&spca50x->buf_lock);
	PDEBUG(4, "leaving");
}

static void spca50x_dealloc(struct usb_spca50x *spca50x, int now)
{
	struct timer_list *bt = &(spca50x->buf_timer);
	PDEBUG(2, "entered");
	down(&spca50x->buf_lock);

	PDEBUG(2, "deallocating buffer memory %s", now ? "now" : "later");

	if (spca50x->buf_state == BUF_PEND_DEALLOC) {
		spca50x->buf_state = BUF_ALLOCATED;
		del_timer(bt);
	}

	if (now)
		spca50x_do_dealloc(spca50x);
	else {
		spca50x->buf_state = BUF_PEND_DEALLOC;
		init_timer(bt);
		bt->function = spca50x_buf_callback;
		bt->data = (unsigned long)spca50x;
		bt->expires = jiffies + buf_timeout * HZ;
		add_timer(bt);
	}
	up(&spca50x->buf_lock);
	PDEBUG(2, "leaving");
}

/**
 * Reset the camera and send the correct initialization sequence for the
 * currently selected source
 */
static int spca50x_init_source(struct usb_spca50x *spca50x)
{
	int err_code;
	struct usb_device *dev=spca50x->dev;
	int index; // index in the modes table

	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA500:
		{
			/* initialisation of spca500 based cameras is deferred */
			spca500_initialise(spca50x);
			break;
		}
		case BRIDGE_SPCA501:
		{
			PDEBUG(2, "Initializing SPCA501 started");
			if (spca50x->dev->descriptor.idVendor == 0x0506
				    && spca50x->dev->descriptor.idProduct ==
				    0x00df) {
					/* Special handling for 3com data */
					spca50x_write_vector (spca50x,
							      spca501_3com_open_data);
				} else if ( spca50x->desc == Arowana300KCMOSCamera ) {
					/* Arowana 300k CMOS Camera data */
					spca50x_write_vector (spca50x,
							      spca501c_arowana_open_data);
				} else if ( spca50x->desc == MystFromOriUnknownCamera ) {
					/* UnKnow  CMOS Camera data */
					spca50x_write_vector (spca50x,
							      spca501c_mysterious_init_data);
				} else {
					/* Generic 501 open data */
					spca50x_write_vector (spca50x,
							      spca501_open_data);
				}
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
			spca50x->a_blue  = spca50x_reg_read(spca50x->dev, 
				SPCA501_REG_CCDSP, SPCA501_A11, 2) & 0xFF;
			spca50x->a_green = spca50x_reg_read(spca50x->dev, 
				SPCA501_REG_CCDSP, SPCA501_A21, 2) & 0xFF;
			spca50x->a_red   = spca50x_reg_read(spca50x->dev, 
				SPCA501_REG_CCDSP, SPCA501_A31, 2) & 0xFF;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
			PDEBUG(2, "Initializing SPCA501 finished");
			break;
		}
		case BRIDGE_SPCA504:

		{
			PDEBUG(2,"Opening SPCA504");
			
				if (spca50x->desc == AiptekMiniPenCam13) {
					/* spca504a aiptek */
					spca50x_reg_write (spca50x->dev, 0,
							   0x2000, 0);
					spca50x_reg_write (spca50x->dev, 0,
							   0x2883, 1);
					spca504A_acknowledged_command (spca50x,
								       0x08, 6,
								       0, 0x86,
								       1);
					spca504A_acknowledged_command (spca50x,
								       0x24, 0,
								       0, 0x9D,
								       1);
					spca504A_acknowledged_command (spca50x,
								       1, 0x0f,
								       0, 0xFF,
								       0);
				}
				/* setup qtable */
				err_code = spca50x_setup_qtable (spca50x,
								 0x00, 0x2800,
								 0x2840,
								 qtable_spca504_default);
				if (err_code < 0) {
					PDEBUG (2,
						"spca50x_setup_qtable failed");
					return err_code;
				}



			
			break;
		}
		case BRIDGE_SPCA504_PCCAM600:
		{
			PDEBUG(2,"Opening SPCA504 (PC-CAM 600)");
			
				spca50x_reg_write (spca50x->dev, 0xe0, 0x0000,
						   0x0000);
				spca50x_reg_write (spca50x->dev, 0xe0, 0x0000, 0x0001);	// reset
				spca504_wait_status (spca50x);
				if (spca50x->desc == LogitechClickSmart420) {	/* clicksmart 420 */
					spca50x_write_vector (spca50x,
							      spca504A_clicksmart420_open_data);
				} else {
					spca50x_write_vector (spca50x,
							      spca504_pccam600_open_data);
				}
				
				err_code = spca50x_setup_qtable (spca50x,
								 0x00, 0x2800,
								 0x2840,
								 qtable_creative_pccam);
				if (err_code < 0) {
					PDEBUG (2,
						"spca50x_setup_qtable failed");
					return err_code;
				}

			break;
		}
		case BRIDGE_SPCA533:
		case BRIDGE_SPCA504B:
		{
			PDEBUG(2,"Opening SPCA504B or SPCA533");  
			spca504B_init (spca50x); 
			break;
		}
		case BRIDGE_SPCA505:
		{
			PDEBUG(2,"Initializing SPCA505 for sensor type %s",
				((spca50x->ccd == 0)? "CCD": "Video In"));
			if(!spca50x->ccd)
			{
				spca50x_write_vector(spca50x, spca505_open_data_ccd);
			}
			else
			{
				spca50x_write_vector(spca50x, spca505_open_data_ext);
				spca50x_write_i2c_vector(spca50x, SAA7113_I2C_BASE_WRITE, 
					spca50x_i2c_data);
				spca50x_write_vector(spca50x, spca505_open_data2);
			}

			err_code = 0;
			err_code = spca50x_reg_read(dev, 6, (__u16)0x16, 2);
			
			if(err_code < 0)
			{
				PDEBUG(1, "register read failed for after vector read err = %d",
					err_code);
				return -EIO;
			}
			else 
			{
				PDEBUG(3, "After vector read returns : 0x%x should be 0x0101",
					err_code & 0xFFFF);
			}
			
			err_code = spca50x_reg_write(dev, 6, (__u16)0x16, (__u16)0xa);
			if(err_code < 0)
			{
				PDEBUG(1, "register write failed for (6,0xa,0x16) err=%d", err_code);
				return -EIO;
			}
			
			if(spca50x->ccd)
			{
				spca50x_write_i2c(spca50x, SAA7113_I2C_BASE_WRITE, 0xa, bright & 0xff);
				spca50x_write_i2c(spca50x, SAA7113_I2C_BASE_WRITE, 0xb, contrast & 0xff);
			
				spca50x_read_SAA7113_status(spca50x);
			}
			else
			{
				spca50x_reg_write(spca50x->dev, 5, 0xc2, 18);
			}
			
			break;
		}
		case BRIDGE_SPCA506:
		{
			// Untested */
			spca50x_write_vector(spca50x, spca506_open_data_ext);
			spca50x_write_i2c_vector(spca50x, SAA7113_I2C_BASE_WRITE, 
				spca50x_i2c_data);
			spca50x_write_vector(spca50x, spca506_open_data2);
			
			spca50x_write_i2c(spca50x, SAA7113_I2C_BASE_WRITE, 0xa, bright & 0xff);
			spca50x_write_i2c(spca50x, SAA7113_I2C_BASE_WRITE, 0xb, contrast & 0xff);
			
			spca50x_read_SAA7113_status(spca50x);
			
			break;
		}
		case BRIDGE_SPCA561:
		{
			spca50x_write_vector(spca50x,spca561_open_data) ;
			break ;
		}
		case BRIDGE_SPCA508:
		{
			spca50x_write_vector(spca50x,spca508_open_data);
			break;
		}
		default:
		{
			err("Unimplemented bridge type");
			return -EINVAL;
		}
	}

	if ((index = spca50x_find_mode_index(spca50x->width, 
	    spca50x->height, GET_EXT_MODES(spca50x->bridge))) == -EINVAL)
	{
		err("Can't find proper mode?!!");
	}
	else
	{
		err_code = spca50x_set_mode(spca50x, (index & 0x0F),
		                            GET_EXT_MODES(spca50x->bridge));
		if (err_code)
		{
			err("Can't set proper camera mode");
			return -EINVAL;
		}
	}

	return 0;
}


/****************************************************************************
 *
 * V4L API
 *
 ***************************************************************************/


static int spca50x_open(struct inode *inode, struct file *file)

{
	struct video_device *vdev = video_devdata(file);
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	int err;
	int i;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#endif
	PDEBUG(2, "opening");
	

	down(&spca50x->lock);

	err = -EBUSY;
	if (spca50x->user) goto out;

	err = -ENOMEM;
	if (spca50x_alloc(spca50x)) goto out;

	err = spca50x_init_source(spca50x);
	if (err != 0) goto out;

	spca50x->curframe = 0;
	PDEBUG(4," frame[0]=0x%p frame[0].grabstata=Frameready\n", &(spca50x->frame[0]));
	spca50x->frame[0].grabstate = FRAME_READY;	
	spca50x->frame[0].seq = -1;
	spca50x->lastFrameRead = -1;

	/* Set default sizes in case IOCTL (VIDIOCMCAPTURE) is not used
	 * (using read() instead). */
	for (i = 0; i < SPCA50X_NUMFRAMES; i++)
	{
		spca50x->frame[i].width      = spca50x->width;
		spca50x->frame[i].height     = spca50x->height;
		spca50x->frame[i].cameratype = spca50x->cameratype;
		spca50x->frame[i].scanlength = spca50x->width * spca50x->height * 3/2; // assumes 4:2:0 data
		spca50x->frame[i].depth      = 24;
		spca50x->frame[i].bytes_read = 0;
		spca50x->frame[i].segment    = 0;

		/* Note: format reflects format of data as returned to
		 * a process, not as read from camera hardware.
		 * This might be a good idea for dumb programs always
		 * assuming the following settings.
		 */
		spca50x->frame[i].format = VIDEO_PALETTE_RGB24;
	}

	err = spca50x_init_isoc(spca50x);
	if (err)
	{ 	
		PDEBUG(4,"4 DEALLOC error on spca50x_init\n");
		spca50x_dealloc(spca50x, 0);
		goto out;
	}

	/* Now, let's get brightness from the camera */
	spca50x->brightness = spca50x_get_brightness(spca50x) << 8;
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	spca50x->whiteness = spca50x_get_whiteness(spca50x) << 8;
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
	spca50x->whiteness = 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

	spca50x->user++;
	file->private_data = vdev;
	
out:
	up(&spca50x->lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (err)
		MOD_DEC_USE_COUNT;
#endif	


#ifdef SPCA50X_ENABLE_EXPERIMENTAL
	spca50x->nstable = 0;
	spca50x->nunstable = 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

	if (err)
	{
		PDEBUG(2, "Open failed");
	}
	else
	{
		PDEBUG(2, "Open done");
	}

	return err;
}

static void spca50x_close_callback(unsigned long data)
{
  struct usb_spca50x *spca50x = (struct usb_spca50x *)data;

  PDEBUG(4, "entered close CB");

  del_timer(&spca50x->buf_timer);

  video_unregister_device(spca50x->vdev);
  kfree(spca50x);
  spca50x = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif

  PDEBUG(4, "leaving close CB");
}



static int spca50x_close(struct inode *inode, struct file *file)

{

 	struct video_device *vdev = file->private_data;
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	int i;

	PDEBUG(2, "spca50x_close");

	down(&spca50x->lock);

	spca50x->user--;
	spca50x->curframe = -1;

	spca50x_stop_isoc(spca50x);
	if (spca50x->dev)
	{
		if (spca50x->bridge == BRIDGE_SPCA505 ||
			spca50x->bridge == BRIDGE_SPCA506)
		{
			spca50x_reg_write(spca50x->dev, 0x3, 0x3, 0x20); // This maybe reset or power control
			spca50x_reg_write(spca50x->dev, 0x3, 0x1, 0x0);
			spca50x_reg_write(spca50x->dev, 0x3, 0x0, 0x1);
			spca50x_reg_write(spca50x->dev, 0x5, 0x10, 0x1);
			spca50x_reg_write(spca50x->dev, 0x5, 0x11,0xF);
		}
		else if (spca50x->bridge == BRIDGE_SPCA501)
		{
			// This maybe reset or power control
			spca50x_reg_write(spca50x->dev, SPCA501_REG_CTLRL, 0x5, 0x0);
		}
	}


	for(i = 0; i < SPCA50X_NUMFRAMES; i++)
	{
		if (waitqueue_active(&spca50x->frame[i].wq)) 
			wake_up_interruptible(&spca50x->frame[i].wq);
	}
	if (waitqueue_active(&spca50x->wq))
			wake_up_interruptible(&spca50x->wq);

	spca50x->streaming = 0;
/*************************************************************/
spin_lock(&spca50x->lock_urb);
	/* Unschedule all of the iso td's */
	for (i = SPCA50X_NUMSBUF - 1; i >= 0; i--) {
		if (spca50x->sbuf[i].urb) {
			usb_unlink_urb(spca50x->sbuf[i].urb);
			usb_free_urb(spca50x->sbuf[i].urb);
			spca50x->sbuf[i].urb = NULL;
		}
	}
spin_unlock(&spca50x->lock_urb);
/************************************************************/

	up(&spca50x->lock);
	spca50x_dealloc(spca50x, 1);
	if (!spca50x->dev)
	{
          /* The camera is already disconnected, but cleanup delayed
             bcz the video device was open by an user. FEB: 2002-2-2.
          */
	  struct timer_list *bt = &(spca50x->buf_timer);
	  
	  PDEBUG(3,"spca50x_close close on disconnected cam");


          /* Don't call video_unregister_device here or you will get a
             deadlock. Using a timer to avoid that problem.
           */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		MOD_INC_USE_COUNT;
#endif
          init_timer(bt);
          bt->function = spca50x_close_callback;
          bt->data = (unsigned long)spca50x;
          bt->expires = jiffies + buf_timeout * HZ;
          add_timer(bt);
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif


	file->private_data = NULL;
	return 0;

}
#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int spca50x_init_done(struct video_device *dev)
{
#ifdef CONFIG_PROC_FS
	create_proc_spca50x_cam((struct usb_spca50x *)dev);
#endif /* CONFIG_PROC_FS */

	return 0;
}

static long spca50x_write(struct video_device *dev, const char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}
#endif
#endif

static int spca50x_do_ioctl(struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
  struct video_device *vdev = file->private_data;
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	PDEBUG(2, "do_IOCtl: 0x%X", cmd);
	PDEBUG(2, "spca50x pointer: %p Video pointer: %p ",spca50x, vdev);
	if (!spca50x->dev)
		return -EIO;

	switch (cmd) {
	case VIDIOCGCAP:
	{
	 struct video_capability *b = arg;

		PDEBUG (2, "VIDIOCGCAP %p :",b);

		memset(b, 0, sizeof(struct video_capability));
		strcpy(b->name, "SPCA5XX USB Camera");
		b->type = VID_TYPE_CAPTURE;
		b->channels = ((spca50x->bridge == BRIDGE_SPCA506)? 2: 1);
		b->audios = 0;
		b->maxwidth = spca50x->maxwidth;
		b->maxheight = spca50x->maxheight;

		b->minwidth = spca50x->minwidth;
		b->minheight = spca50x->minheight;


		return 0;
	}
	case VIDIOCGCHAN:
	{
	 struct video_channel *v = arg;

		if (v->channel > ((spca50x->bridge == BRIDGE_SPCA505)? 1: 0) 
			|| v->channel < 0 )
			return -EINVAL;

		switch(spca50x->bridge)
		{
			case BRIDGE_SPCA500:
			{
				strncpy(v->name, "CCD", 32);
				break;
			}
			case BRIDGE_SPCA501:
			{
				strncpy(v->name, "CCD", 32);
				break;
			}
			case BRIDGE_SPCA505:
			{
				strncpy(v->name, ((v->channel == 0)? "CCD": "Video In"), 32);
				break;
			}
			case BRIDGE_SPCA506:
			{
				strncpy(v->name, "Video In", 32);
				break;
			}
			case BRIDGE_SPCA508:
			case BRIDGE_SPCA561:
			{
				strncpy(v->name, "CCD", 32);
				break;
			}
		}

		v->flags = 0;
		v->tuners = 0;
		v->type = VIDEO_TYPE_CAMERA;


		return 0;
	}
	case VIDIOCSCHAN:
	{
		/* 
		There seems to be some confusion as to whether this is a struct 
		video_channel or an int. This should be safe as the first element 
		of video_channel is an int channel number
		and we just ignore the rest 
		*/
		struct video_channel *v = arg;


		if ((v->channel > 0 && spca50x->bridge != BRIDGE_SPCA505) || v->channel > 1 || v->channel < 0)
			return -EINVAL;

		if(v->channel != spca50x->ccd)
		{
			/* Haven't tested this yet */
			spca50x->ccd = v->channel;
			if (spca50x_init_source(spca50x) != 0)
				return -EIO;
		}

		return 0;
	}
	case VIDIOCGPICT:
	{
		struct video_picture *p = arg;

		p->depth = spca50x->frame[0].depth;
		p->palette = spca50x->frame[0].format;
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
		if (!autoadjust)
			p->brightness = spca50x_get_brightness(spca50x) << 8;
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
	 	p->brightness = spca50x_get_brightness(spca50x) << 8;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

		PDEBUG (4, "VIDIOCGPICT: depth=%d, palette=%d", p->depth, p->palette);

		if(spca50x->ccd)
		{
			p->colour = spca50x_read_i2c(spca50x,SAA7113_I2C_BASE_READ,0xc)<<8;
			p->whiteness = 0;
			p->contrast = spca50x_read_i2c(spca50x,SAA7113_I2C_BASE_READ,0xb)<<8;
			p->hue = spca50x_read_i2c(spca50x,SAA7113_I2C_BASE_READ,0xd)<<8;
		}
		else
		{
			switch(spca50x->bridge)
			{
				case BRIDGE_SPCA500:
				{
					int tmp;

					p->colour     = 0;
					p->whiteness  = 0;
					p->contrast   = spca50x_reg_read(spca50x->dev, 0x0, 0x8168, 1) << 8;
					p->brightness = spca50x_get_brightness(spca50x);

					/* get hue */
					tmp   = (spca50x_reg_read(spca50x->dev, 0x0, 0x816b, 1) & 12) << 6;
					tmp  += spca50x_reg_read(spca50x->dev, 0x0, 0x816a, 1);
					p->hue = tmp << 6;
					PDEBUG(2, "brightness = %d",
					       p->brightness);
					break;
				}
				case BRIDGE_SPCA501:
				{
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
					p->whiteness = spca50x_get_whiteness(spca50x) << 8;
					if (autoadjust)
						p.brightness = p.whiteness;
					p->contrast = ((spca50x_reg_read(spca50x->dev, 
						SPCA501_REG_CCDSP, SPCA501_A11, 2) & 0xFF) 
						- spca50x->a_blue) << 8;
					printk("Contrast = %d\n", p.contrast >> 8);
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
					p->whiteness = 0;
					p->contrast  = 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
					p->colour = (spca50x_reg_read(spca50x->dev, 
						SPCA501_REG_CCDSP, 0x11, 2) & 0xFF) << 8;
					p->hue = (spca50x_reg_read(spca50x->dev, 
						SPCA501_REG_CCDSP, 0x13, 2) & 0xFF) << 8;
					break;
				}
				case BRIDGE_SPCA533:
				case BRIDGE_SPCA504:
				case BRIDGE_SPCA504B:
				case BRIDGE_SPCA504_PCCAM600:
				{
					p->colour = spca50x_reg_read(spca50x->dev, 0x0, 0x21ae, 2)<<8;
					p->whiteness = 0;
					p->contrast = spca50x_reg_read(spca50x->dev, 0x0, 0x21a8, 2)<<8;
					p->brightness = spca50x_get_brightness(spca50x);
					p->hue = spca50x_reg_read(spca50x->dev, 0x0, 0x21ad, 2)<<8;
					/* FIXME: how do we deal with gamma? the chip knows it, v4l doesnt */
					PDEBUG(0, "color: 0x%02x contrast: 0x%02x brightness: 0x%02x hue: 0x%02x",
					          p->colour, p->contrast, p->brightness, p->hue);
					break;
				}
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
				case BRIDGE_SPCA505:
				{
					p->whiteness = (spca50x_reg_read(spca50x->dev,6,0x59,2)*10)<<8;
					p->contrast = spca50x_reg_read(spca50x->dev,5,0xc2,2)<<8;
					p->colour = spca50x_reg_read(spca50x->dev,5,0xc2,2)<<8;
					p->hue = spca50x_reg_read(spca50x->dev,5,0xc1,2)<<8;
					break;
				}
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */
				default:
				{
					/* default to not allowing any settings to change */
					p->colour     = spca50x->colour;
					p->whiteness  = spca50x->whiteness;
					p->contrast   = spca50x->contrast;
					p->brightness = spca50x->brightness;
					p->hue        = spca50x->hue;
					break;
				}
			}
		}

		return 0;
	}
	case VIDIOCSPICT:
	{
	 struct video_picture *p = arg;
	 int i;

		PDEBUG (4, "VIDIOCSPICT");


		if (!spca50x_get_depth(p->palette))
			return -EINVAL;

		if (p->depth < spca50x->min_bpp)
			return -EINVAL;

		PDEBUG(4, "Setting depth=%d, palette=%d", p->depth, p->palette);
		for (i = 0; i < SPCA50X_NUMFRAMES; i++) 
		{
			spca50x->frame[i].depth = p->depth;
			spca50x->frame[i].format = p->palette;
		}

		if(spca50x->ccd)
		{
			spca50x_write_i2c(spca50x,SAA7113_I2C_BASE_WRITE,0xa,p->brightness>>8);
			spca50x_write_i2c(spca50x,SAA7113_I2C_BASE_WRITE,0xb,p->contrast>>8);
			spca50x_write_i2c(spca50x,SAA7113_I2C_BASE_WRITE,0xc,p->colour>>8);
			spca50x_write_i2c(spca50x,SAA7113_I2C_BASE_WRITE,0xd,p->hue>>8);
		}
		else
		{
			switch(spca50x->bridge)
			{
				case BRIDGE_SPCA500:
				{
					int tmp;

					spca50x_set_brightness(spca50x, p->brightness >> 8);  //((p.brightness >> 8) + 128) % 255);
					spca50x_reg_write(spca50x->dev, 0x00, 0x8168, p->contrast>>8);

					/* set hue */
					tmp  = spca50x_reg_read(spca50x->dev, 0x00, 0x816b, 1);
					tmp  = tmp & 0xf3;
					tmp += (p->hue & 0xc000) >> 12;
					spca50x_reg_write(spca50x->dev, 0x00, 0x816b, tmp);
					spca50x_reg_write(spca50x->dev, 0x00, 0x816a, (p->hue & 0x3fff)>>8);
				}
				case BRIDGE_SPCA501:
				{
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
				 __u16 kred, kgreen, kblue; //coeffs of color correction

					if (autoadjust)
						p->whiteness = p->brightness;
					spca50x_set_whiteness(spca50x, p->whiteness >> 8);
					
					/* Setting "contrast" (offset to color correction parameters) */
					p->contrast >>= 8;
					kred = spca50x->a_red + p->contrast;
					kgreen = spca50x->a_green + p->contrast;
					kblue = spca50x->a_blue + p->contrast;
					PDEBUG(4, "Setting red = %d, green = %d, blue = %d", 
						(__u8)(kred & 0xFF), (__u8)(kgreen & 0xFF), (__u8)(kblue & 0xFF));
					if (kred < 256 && kgreen < 256 && kblue < 256)
					{
						spca50x_reg_write(spca50x->dev, 
							SPCA501_REG_CCDSP, SPCA501_A11, (__u8)(kblue & 0xFF));
						spca50x_reg_write(spca50x->dev, 
							SPCA501_REG_CCDSP, SPCA501_A21, (__u8)(kgreen & 0xFF));
						spca50x_reg_write(spca50x->dev, 
							SPCA501_REG_CCDSP, SPCA501_A31, (__u8)(kred & 0xFF));
					}
					spca50x_reg_write(spca50x->dev, SPCA501_REG_CCDSP, 0x11, 
						(__u8)((p->colour >> 8) & 0xFF));
					spca50x_reg_write(spca50x->dev, SPCA501_REG_CCDSP, 0x13, 
						(__u8)((p->hue >> 8) & 0xFF));
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */
					/* Setting brightness */
					PDEBUG(4, "Setting brightness = %d", p->brightness);
//// This call must be moved to better place
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
					if (!autoadjust)
						spca50x_set_brightness(spca50x, p->brightness >> 8);
#else /* SPCA50X_ENABLE_EXPERIMENTAL */
					spca50x_set_brightness(spca50x, p->brightness >> 8);
#endif  /* SPCA50X_ENABLE_EXPERIMENTAL */
					break;
				}
				case BRIDGE_SPCA533:
				case BRIDGE_SPCA504:
				case BRIDGE_SPCA504B:
				case BRIDGE_SPCA504_PCCAM600:
				{
					/* v4l docs tell me that all parameters are scaled 0-65535 
					 * This means we keep them at that scale internally and adjust
					 * to the actual hardware capabilites on write. */

					/* 0 - 8 default on win: 1 */
					spca50x_reg_write(spca50x->dev, 0x0, 0x21ae, p->colour >> 8);
					p->whiteness = 0;
					/* 1 - 8 default on win: 1 */
					spca50x_reg_write(spca50x->dev, 0x0, 0x21a8, p->contrast >> 8);

					/* The chip keeps the brightness as 129 -> 128 with 0 being the midpoint.
					 * v4l always keeps these values unsigned 16 bit, that is 0 to 65535. 
					 * Scale accordingly. */

					spca50x_set_brightness(spca50x, ((p->brightness>>8)+128)%255);
								
					/* 0 - 360 default on win: 0 */
					//spca50x_reg_write(spca50x->dev, 0x0, 0x21ad, p.hue >> 8);
;;;
					break;
				}
				case BRIDGE_SPCA505:
				{
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS

					spca50x_reg_write(spca50x->dev,6,0x1f,0x16);
					spca50x_reg_write(spca50x->dev,6,0x20,0x0);
					
					spca50x_reg_write(spca50x->dev,5,0xc1,p->hue>>8);
					spca50x_reg_write(spca50x->dev,5,0xc2,p->contrast>>8);
					spca50x_reg_write(spca50x->dev,5,0xca,0x0);
////I have no this type of camera. Please, check this and move this
// call to better place
					spca50x_set_brightness(spca50x, p->brightness >> 8);
////
					spca50x_reg_write(spca50x->dev,6,0x59,max((p->whiteness>>8)/10,20));
#endif /* SPCA50X_ENABLE_EXP_BRIGHTNESS */
					break;
				}
				default:
				{
					/* default to not allowing any settings to change */
					p->colour     = spca50x->colour;
					p->whiteness  = spca50x->whiteness;
					p->contrast   = spca50x->contrast;
					p->brightness = spca50x->brightness;
					p->hue        = spca50x->hue;
					break;
				}
			}
		}
		spca50x->contrast = p->contrast;
		spca50x->brightness = p->brightness;
		spca50x->colour = p->colour;
		spca50x->hue = p->hue;
		spca50x->whiteness = p->whiteness;
#ifdef SPCA50X_ENABLE_EXPERIMENTAL
		spca50x->nstable = 0;
		spca50x->nunstable = 0;
#endif /* SPCA50X_ENABLE_EXPERIMENTAL */

		return 0;
	}
	case VIDIOCGCAPTURE:
	{
		int *vf = arg;

		PDEBUG (4, "VIDIOCGCAPTURE");
		*vf = 0;
	// no subcapture
		return -EINVAL;
	}
	case VIDIOCSCAPTURE:
	{
		struct video_capture *vc = arg;

		if (vc->flags)
			return -EINVAL;
		if (vc->decimation)
			return -EINVAL;

		return -EINVAL;
	}
	case VIDIOCSWIN:
	{
		struct video_window *vw = arg;
		int result;


		PDEBUG (3, "VIDIOCSWIN: width=%d, height=%d, flags=%d",
			vw->width, vw->height,vw->flags);
#if 0
		if (vw->flags) return -EINVAL;
		if (vw->clipcount) return -EINVAL;
#endif
		if (vw->x) return -EINVAL;
		if (vw->y) return -EINVAL;
		if (vw->width  > (unsigned int) spca50x->maxwidth)  return -EINVAL;
		if (vw->height > (unsigned int) spca50x->maxheight) return -EINVAL;
		if (vw->width  < (unsigned int) spca50x->minwidth)  return -EINVAL;
		if (vw->height < (unsigned int) spca50x->minheight) return -EINVAL;
		
#if 0
		if (vw->chromakey !=0) return -EINVAL;
#endif

		/* If we're collecting previous frame, wait before changing
		 * modes
		 *
		 * This seems to be blocking forever, dangerous not using it tho'
		 */
		
			/**** FIX NEED ***/
		//interruptible_sleep_on(&spca50x->wq);
		//if (signal_pending(current)) return -EINTR;

		result = spca50x_mode_init_regs(spca50x, vw->width, vw->height,
		                                spca50x->frame[0].format, 
		                                GET_EXT_MODES(spca50x->bridge));

		if (result == 0) {
			spca50x->frame[0].width = vw->width;
			spca50x->frame[0].height = vw->height;
		}

		return result;
	}
	case VIDIOCGWIN:
	{
		struct video_window *vw = arg;

		memset(vw, 0, sizeof(struct video_window));
		vw->x      = 0;
		vw->y      = 0;
		vw->width  = spca50x->frame[0].width;
		vw->height = spca50x->frame[0].height;
		vw->flags  = 0;

		PDEBUG (4, "VIDIOCGWIN: %dx%d", vw->width, vw->height);


		return 0;
	}
	case VIDIOCGMBUF:
	{
		struct video_mbuf *vm=arg;
		int               i;
		PDEBUG (2, "VIDIOCGMBUF: %p ", vm);
		memset(vm, 0, sizeof(struct video_mbuf));
		vm->size = SPCA50X_NUMFRAMES * MAX_DATA_SIZE;
		vm->frames = SPCA50X_NUMFRAMES;

		for(i=0;i<SPCA50X_NUMFRAMES;i++)
		{
			vm->offsets[i] = MAX_DATA_SIZE*i;
		}


		return 0;
	}
	case VIDIOCMCAPTURE:
	{
		struct video_mmap *vm = arg;
		int ret, depth;

		PDEBUG(2, "CMCAPTURE");
		PDEBUG(2, "frame: %d, size: %dx%d, format: %d",
			vm->frame, vm->width, vm->height, vm->format);
		depth = spca50x_get_depth(vm->format);
		if (!depth || depth < spca50x->min_bpp) {
			err("VIDIOCMCAPTURE: invalid format (%d)", vm->format);
			return -EINVAL;
		}

		if ((vm->frame < 0) || (vm->frame > 3)) {
			err("VIDIOCMCAPTURE: invalid frame (%d)", vm->frame);
			return -EINVAL;
		}

		if (vm->width > spca50x->maxwidth || vm->height > spca50x->maxheight) {
			err("VIDIOCMCAPTURE: requested dimensions too big");
			return -EINVAL;
		}
		if (vm->width < spca50x->minwidth || vm->height < spca50x->minheight) {
			err("VIDIOCMCAPTURE: requested dimensions too small");
			return -EINVAL;
		}
		/* 
		 * If we are grabbing the current frame, let it pass
		 */ 
		 if (spca50x->frame[vm->frame].grabstate == FRAME_GRABBING){
		 	PDEBUG(4, "MCAPTURE: already grabbing");
			ret = wait_event_interruptible(spca50x->wq,
			(spca50x->frame[vm->frame].grabstate != FRAME_GRABBING));
		 if (ret)
			return -EINTR;
			
			}
		if ((spca50x->frame[vm->frame].width != vm->width)         ||
		    (spca50x->frame[vm->frame].height != vm->height)       ||
		    (spca50x->frame[vm->frame].format != vm->format))
		{

		ret = spca50x_mode_init_regs(spca50x, vm->width, vm->height,
			                             vm->format, 
			                             GET_EXT_MODES(spca50x->bridge));
		if (ret < 0)
			return ret;
		spca50x->frame[vm->frame].width  = vm->width;
		spca50x->frame[vm->frame].height = vm->height;
		spca50x->frame[vm->frame].format = vm->format;
		spca50x->frame[vm->frame].depth  = depth;
		
		}

		/* Mark it as ready */
		spca50x->frame[vm->frame].grabstate = FRAME_READY;
		return 0;
	}
	case VIDIOCSYNC:
	{
	    unsigned int frame = *((unsigned int *) arg);
		int ret;

		PDEBUG(3, "syncing to frame %d, grabstate = %d", frame,
		       spca50x->frame[frame].grabstate);

		switch (spca50x->frame[frame].grabstate) {
		case FRAME_UNUSED:
			return -EINVAL;
    case FRAME_ABORTING:
      return -ENODEV;
		case FRAME_READY:
		case FRAME_GRABBING:
		case FRAME_ERROR:
redo:
			if (!spca50x->dev)
				return -EIO;

		
		ret = wait_event_interruptible(spca50x->wq,
		(spca50x->frame[frame].grabstate == FRAME_DONE));
		if (ret)
			return -EINTR;
		
			PDEBUG(5,"Synched on frame %d, grabstate = %d",
			         frame, spca50x->frame[frame].grabstate);

			if (spca50x->frame[frame].grabstate == FRAME_ERROR) {
				goto redo;
			}
			/* Fallthrough.
			 * We have waited in state FRAME_GRABBING until it
			 * becomes FRAME_DONE, so now we can move along.
			 */
		case FRAME_DONE:
			if (spca50x->snap_enabled && !spca50x->frame[frame].snapshot) {
				goto redo;
			}
			PDEBUG(3,"Already synched %d\n",frame);

			/* Release the current frame. This means that it
			 * will be reused as soon as all other frames are
			 * full, so the app better be done with it quickly.
			 * Can this be avoided somehow?
			 */
			spca50x->frame[frame].grabstate = FRAME_UNUSED;

			break;
		} /* end switch */

		return 0;
	}
	case VIDIOCGFBUF:
	{
		struct video_buffer *vb=arg;

		memset(vb, 0, sizeof(struct video_buffer));
		vb->base = NULL;	/* frame buffer not supported, not used */


		return 0;
	}
	case VIDIOCKEY:
		return 0;
	case VIDIOCCAPTURE:
		return -EINVAL;
	case VIDIOCSFBUF:
		return -EINVAL;
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
		return -EINVAL;
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		return -EINVAL;
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		return -EINVAL;
	default:
		return -ENOIOCTLCMD;
	} /* end switch */

	return 0;
}


static int spca50x_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg)
{
  int rc ;
		 
	rc = video_usercopy(inode, file, cmd, arg, spca50x_do_ioctl);
				  
	
	return rc;
}



static ssize_t spca50x_read(struct file *file, char *buf, size_t cnt, loff_t *ppos)

{

  struct video_device *dev = file->private_data;
	int noblock = file->f_flags&O_NONBLOCK;
	unsigned long count = cnt;
	struct usb_spca50x *spca50x = video_get_drvdata(dev);

	int i;
	int frmx = -1;
	
	volatile struct spca50x_frame *frame;

	PDEBUG(4, "%ld bytes, noblock=%d", count, noblock);

// restart:
	if (signal_pending(current)) 
		return -EINTR;

	if (!dev || !buf) 
		return -EFAULT;

	if (!spca50x->dev) 
		return -EIO;

	if(!spca50x->streaming) 
		return -EIO;

/*FIXME*/	/* Wait for the end of the frame. This is the waitq for all frames,
	 * so we get triggered as soon as any frame is DONE.
	 */
	interruptible_sleep_on(&spca50x->wq);
	if (signal_pending(current)) 
		return -EINTR;
	
	
		/* 
		 * Ok, for the jpeg cams, we know we have received a whole
		 * frame now. The rest of the code in this routine confuses 
		 * frames and sequence numbers, so we do our own thing here.
		 * */

		/* One frame has just been set to DONE. Find it. */
		for(i = 0; i < SPCA50X_NUMFRAMES; i++)
			if (spca50x->frame[i].grabstate == FRAME_DONE)
				frmx = i;
		
		PDEBUG(4, "Frame number: %d", frmx);
		if (frmx < 0)
		{
			/* We havent found a frame that is DONE. Damn. Should
			 * not happen. */
			PDEBUG(2, "Couldnt find a frame ready to be read.");
			return -EFAULT;
		}
		frame = &spca50x->frame[frmx];
		// PDEBUG(4, "count asked: %d",(int ) count);
		// count is ask by users why compute here !!!!
		// best way:
		// if (count > frame->scanlength ) count = frame->scanlength;
		// but frame->scanlength seem to be not used
		//count = frame->width * frame->height * (frame->depth >> 3);
		if ((i = copy_to_user(buf, frame->data, count))) {
			PDEBUG(2, "Copy failed! %d bytes not copied", i);
			return -EFAULT;
		}
		/* Release the frame */
		frame->grabstate = FRAME_READY;
		
	return count;
	

}



static int spca50x_mmap(struct file *file, struct vm_area_struct *vma)

{

  struct video_device *dev = file->private_data;
  unsigned long start = vma->vm_start;
  unsigned long size  = vma->vm_end - vma->vm_start;
  struct usb_spca50x *spca50x = video_get_drvdata(dev);

	unsigned long page, pos;

	if (spca50x->dev == NULL)
		return -EIO;

	PDEBUG(4, "mmap: %ld (%lX) bytes", size, size);

	if (size > (((SPCA50X_NUMFRAMES * MAX_DATA_SIZE) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1)))
		return -EINVAL;

	pos = (unsigned long)spca50x->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) || defined (RH9_REMAP)
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
#else /* RH9_REMAP */
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
#endif /* RH9_REMAP */
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return 0;
}


static struct file_operations spca50x_fops = {
  	.owner =  THIS_MODULE,
	.open =   spca50x_open,
	.release =  spca50x_close,
  	.read =   spca50x_read,
  	.mmap =   spca50x_mmap,
  	.ioctl =  spca50x_ioctl,
  	.llseek = no_llseek,
};
static struct video_device spca50x_template = {
  .owner =  THIS_MODULE,
	.name=		"SPCA50X USB Camera",
	.type=		VID_TYPE_CAPTURE,
	.hardware=	VID_HARDWARE_SPCA50X,
	.fops =   &spca50x_fops,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	.release =  video_device_release,
#endif
	.minor =  -1,
};

/****************************************************************************
 *
 * SPCA50X configuration
 *
 ***************************************************************************/

static int spca50x_configure_sensor(struct usb_spca50x *spca50x)
{
	struct usb_device *dev = spca50x->dev;

	PDEBUG (4, "starting configuration");

	if (spca50x_reset(dev,0) < 0)
		return -1;

	/* Set sensor-specific vars */
	switch(spca50x->bridge)
	{	
		case BRIDGE_SPCA500:
		case BRIDGE_SPCA504_PCCAM600:
		case BRIDGE_SPCA504:
		case BRIDGE_SPCA504B:
		case BRIDGE_SPCA533:
		{
			spca50x->maxwidth=640;
			spca50x->maxheight=480;
			spca50x->minwidth=176;
			spca50x->minheight=144;
			break;
		}
		case BRIDGE_SPCA501:
		{
			spca50x->maxwidth=640;
			spca50x->maxheight=480;
			spca50x->minwidth=160;
			spca50x->minheight=120;
			break;
		}
		case BRIDGE_SPCA506:
		case BRIDGE_SPCA505:
		{
			spca50x->maxwidth = 352;
			spca50x->maxheight = 288;
			spca50x->minwidth = 160;
			spca50x->minheight = 120;
			break;
		}
		case BRIDGE_SPCA508:
		case BRIDGE_SPCA561:
		{
			spca50x->maxwidth=352;
			spca50x->maxheight=288;
			spca50x->minwidth=160;
			spca50x->minheight=120;
			break;
		}
		default:
		{
			spca50x->maxwidth=640;
			spca50x->maxheight=480;
			spca50x->minwidth=160;
			spca50x->minheight=120;
			break;
		}
	}
	return 0;
}

static int config_spca561(struct usb_spca50x *spca50x)
{
	struct usb_device *dev = spca50x->dev;
	int data1, data2;
	// Read frm global register the USB product and vendor IDs, just to	
	// prove that we can communicate with the device.  This works, which
	// confirms at we are communicating properly and that the device
	// is a 561.
	data1 = spca50x_reg_read(dev, 0, 0x8104, 1);
	if (data1 < 0) PDEBUG(1,"Error reading USB Vendor ID from Global register");
	data2 = spca50x_reg_read(dev, 0, 0x8105, 1);
	if (data2 < 0)
		PDEBUG(1,"Error reading USB Vendor ID from Global register");
	PDEBUG(1, "Read from GLOBAL: USB Vendor ID 0x%02x%02x", data2, data1);

	data1=spca50x_reg_read(dev, 0, 0x8106, 1);
	if (data1 < 0)
		PDEBUG(1, "Error reading USB Product ID from Global register");
	data2 = spca50x_reg_read(dev, 0, 0x8107, 1);
	if (data2 < 0)
		PDEBUG(1, "Error reading USB Product ID from Global register");
	PDEBUG(1, "Read from GLOBAL: USB Product ID 0x%02x%02x", data2, data1);
	
	if (spca50x_write_vector(spca50x, spca561_init_data))
			return -1;
	
	return 0; // success
}

static int config_spca508(struct usb_spca50x *spca50x)
{
	struct usb_device *dev = spca50x->dev;
	int data1, data2;
	
	// Read frm global register the USB product and vendor IDs, just to
	// prove that we can communicate with the device.  This works, which
	// confirms at we are communicating properly and that the device
	// is a 508.
	data1 = spca50x_reg_read(dev, 0, 0x8104, 1);
	if (data1 < 0) PDEBUG(1,"Error reading USB Vendor ID from Global register");
	data2 = spca50x_reg_read(dev, 0, 0x8105, 1);
	if (data2 < 0) 
		PDEBUG(1,"Error reading USB Vendor ID from Global register");
	PDEBUG(1, "Read from GLOBAL: USB Vendor ID 0x%02x%02x", data2, data1);
	
	data1=spca50x_reg_read(dev, 0, 0x8106, 1);
	if (data1 < 0) 
		PDEBUG(1, "Error reading USB Product ID from Global register");
	data2 = spca50x_reg_read(dev, 0, 0x8107, 1);
	if (data2 < 0) 
		PDEBUG(1, "Error reading USB Product ID from Global register");
	PDEBUG(1, "Read from GLOBAL: USB Product ID 0x%02x%02x", data2, data1);
	
	data1=spca50x_reg_read(dev, 0, 0x8621, 1);
	if (data1 < 0) 
		PDEBUG(1, "Error reading Window 1 Average Luminance from Global register");
	PDEBUG(1, "Read from GLOBAL: Window 1 average luminance %3d", data1);
	
	switch (spca50x->desc) {
		case ViewQuestVQ110:
			{
				if (spca50x_write_vector
				    (spca50x, spca508_init_data))
					return -1;
				break;
			}
		case IntelEasyPCCamera:
			{
				if (spca50x_write_vector
				    (spca50x, spca508cs110_init_data))
					return -1;
				break;
			}
		case HamaUSBSightcam:
			{
				if (spca50x_write_vector
				    (spca50x, spca508_sightcam_init_data))
					return -1;
				break;
			}
		case HamaUSBSightcam2:
			{
				if (spca50x_write_vector
				    (spca50x, spca508_sightcam2_init_data))
					return -1;
				break;
			}
		default:
			return -1;
	}
	return 0; // success
}

static int spca50x_configure(struct usb_spca50x *spca50x)
{
	int             i;
	int             defaultcols;
	int             defaultrows;
	int		defaultpipe;
	struct usb_device *dev = spca50x->dev;

	/* dynamically look up the smallest available frame size */
	/* That is not really true with Hardware size need smallest and hardware */
	/* Also give the pipe size for the hardware resolution */
	i = spca50x_smallest_mode_index(GET_EXT_MODES(spca50x->bridge),
	                                &defaultcols,
	                                &defaultrows,
					&defaultpipe);
	if (i < 0)
		return i;

	PDEBUG(2,"video_register_device succeeded");
	/* Initialise the camera bridge */
	switch(spca50x->bridge)
	{
		case BRIDGE_SPCA500:
		{
			if(ccd) 
				PDEBUG(1, "Forcing CCD to zero for SPCA500 which doesn't have external input !");
			spca50x->ccd = 0;
			/*
			if (spca500_full_reset(spca50x) < 0)
			{
				PDEBUG(2, "spca500_full_reset failed");
				goto error;
			}
			*/
			break;
		}
		case BRIDGE_SPCA501:
		{
			/* Determine the sensor type for cameras with an
			 * internal CCD or external input
			 */
			if(ccd)
				PDEBUG(1, "Forcing CCD to zero for SPCA501? which doesn't have external input !");
			spca50x->ccd = 0;
			if (spca50x->desc == Arowana300KCMOSCamera) {
					/* Arowana 300k CMOS Camera data */
					if (spca50x_write_vector
					    (spca50x,
					     spca501c_arowana_init_data))
						goto error;
				} else if (spca50x->desc ==
					   MystFromOriUnknownCamera) {
					/* UnKnow Ori CMOS Camera data */
					if (spca50x_write_vector
					    (spca50x,
					     spca501c_mysterious_open_data))
						goto error;
				} else {
					/* generic spca501 init data */
					if (spca50x_write_vector
					    (spca50x, spca501_init_data))
						goto error;
				}
			break;
		}
		case BRIDGE_SPCA533:
		case BRIDGE_SPCA504:
		case BRIDGE_SPCA504B:
		case BRIDGE_SPCA504_PCCAM600:
		{
			spca50x->ccd = 0;
			PDEBUG(2, "Forcing CCD to 0 for the spca504x !");
			/* FIXME: What can sensibly be moved here from open? */
			break;
		}
		case BRIDGE_SPCA505:
		{
			/* Determine the sensor type for cameras with an internal CCD or external input */
			spca50x->ccd = ccd;
			if (spca50x_write_vector(spca50x, spca505_init_data))
				goto error;
			break;
		}
		case BRIDGE_SPCA506:
		{
			if(ccd) 
				PDEBUG(1, "Forcing CCD to one for SPCA506 which can't use one !");
			/* Determine the sensor type for cameras with an internal CCD or external input */
			spca50x->ccd = 1;
			if (spca50x_write_vector(spca50x, spca506_init_data)) 
				goto error;
			break;
		}
		case BRIDGE_SPCA508:
		{
			/* FIXME: I've changed the following check to force ccd to be zero, instead of one.
			 *        Since I don't have one of these cameras, I don't know whether this is
			 *        the right thing to do.
			 *        NWG: Mon 12th August 2002.
			 */
			if (ccd)
				PDEBUG(1, "Forcing CCD to zero for SPCA508 which doesn't have external input !");
			spca50x->ccd = 0;
			if (config_spca508(spca50x)) 
				goto error;
			break;
		}
		case BRIDGE_SPCA561:
		{
			if (ccd)
				PDEBUG(1, "Forcing CCD to zero for SPCA561 which doesn't have external input !");
			spca50x->ccd = 0;
			if (config_spca561(spca50x))
				goto error;
 			break;
 		}

		default:
		{
			err("Unknown bridge type");
			goto error;
		}
	}

	spca50x_set_packet_size(spca50x, 0);
	/* Set an initial pipe size; this will be overridden by
	 * spca50x_set_mode(), called indirectly by the open routine.
	 
	if (spca50x->bridge == BRIDGE_SPCA500){
		spca50x->pipe_size    = 1023; 
	} else {
	 	spca50x->pipe_size    = 384; }
	*/
	spca50x->pipe_size    = defaultpipe; 
	spca50x->snap_enabled = snapshot;
	spca50x->force_rgb    = force_rgb;
	spca50x->min_bpp      = min_bpp;
	spca50x->lum_level    = lum_level;
	spca50x->width        = defaultcols;
	spca50x->height       = defaultrows;

	if (spca50x_reset(dev, 0) < 0)
		goto error;

	if(spca50x_configure_sensor(spca50x) < 0) {
		err("failed to configure");
		goto error;
	}

	/* Set default sizes in case IOCTL (VIDIOCMCAPTURE) is not used
	 * (using read() instead). */
	for (i = 0; i < SPCA50X_NUMFRAMES; i++)
	{
		spca50x->frame[i].width      = spca50x->width;
		spca50x->frame[i].height     = spca50x->height;
		spca50x->frame[i].cameratype = spca50x->cameratype;
		spca50x->frame[i].scanlength = spca50x->width * spca50x->height * 3/2; // assumes 4:2:0 data
		spca50x->frame[i].depth      = 24;
		spca50x->frame[i].bytes_read = 0;
		spca50x->frame[i].segment    = 0;

		/* Note: format reflects format of data as returned to
		 * a process, not as read from camera hardware.
		 * This might be a good idea for dumb programs always
		 * assuming the following settings.
		 */
		spca50x->frame[i].format = VIDEO_PALETTE_RGB24;
	}
	PDEBUG(2,"Spcaxx Configure done !!");
	return 0;

error:

	return -EBUSY;
}

/***************************************************************************
 *
 * Model specific helper functions
 *
 * *************************************************************************/
void spca500_reinit(struct usb_spca50x *spca50x )
{	int err;
	__u8 Data;
		
	// some unknow command from Aiptek pocket dv and family300
	
	spca50x_reg_write(spca50x->dev, 0x00, 0x0d01, 0x01);
	spca50x_reg_write(spca50x->dev, 0x00, 0x0d03, 0x00);
	spca50x_reg_write(spca50x->dev, 0x00, 0x0d02, 0x01);
	
	/* enable drop packet */
	spca50x_reg_write(spca50x->dev, 0x00, 0x850a, 0x0001);
			
	err = spca50x_setup_qtable(spca50x,
			            0x00, 0x8800, 0x8840,
			            qtable_pocketdv);
	if (err < 0)
		{
		PDEBUG(2, "spca50x_setup_qtable failed on init");				
		}
	/* set qtable index */
	spca50x_reg_write(spca50x->dev, 0x00, 0x8880, 2);		
	/* family cam Quicksmart stuff */
	spca50x_reg_write(spca50x->dev, 0x00, 0x800a, 0x00);
	//Set agc transfer: synced inbetween frames
	spca50x_reg_write(spca50x->dev, 0x00, 0x820f, 0x01);
	//Init SDRAM - needed for SDRAM access
	spca50x_reg_write(spca50x->dev, 0x00, 0x870a, 0x04);
	/*Start init sequence or stream*/
	
	spca50x_reg_write(spca50x->dev, 0, 0x8003, 0x00);
	/* switch to video camera mode */
	err = spca50x_reg_write(spca50x->dev, 0x00, 0x8000, 0x0004);
wait_ms(2000);
	if (spca50x_reg_readwait(spca50x->dev, 0, 0x8000, 0x44) != 0)
	
	spca504B_reg_read(spca50x->dev,0 ,0 ,0x816b , &Data ,1);
	spca50x_reg_write(spca50x->dev, 0x00, 0x816b, Data);

}

static int spca500_initialise(struct usb_spca50x *spca50x)
{

	int index;
	int err;
	__u8 Data;
	/* read mode index */
	index = spca50x_find_mode_index(spca50x->width, spca50x->height,
	                                GET_EXT_MODES(spca50x->bridge));
	if (index == -EINVAL)
	{
		PDEBUG(2, "Can't find proper mode?!!");
		return -EINVAL;
	}
	/* EXPERIMENTAL Wait until test Marek */
	index = index & 0x0F;
	spca50x_set_mode( spca50x,index,GET_EXT_MODES(spca50x->bridge));
	
	// spca50x_write_vector(spca50x, spca500_read_stats);
	
	/* setup qtable */
	switch (spca50x->desc)
	{
		case CreativePCCam300:	/* Creative PC-CAM 300 */
		case IntelPocketPCCamera:	/* FIXME: Temporary fix for Intel Pocket PC Camera - NWG (Sat 29th March 2003) */

			/* do a full reset */
			if ((err = spca500_full_reset (spca50x)) < 0) {
				PDEBUG (2, "spca500_full_reset failed");
				return err;
			}

			if ((err =
			     spca50x_reg_write (spca50x->dev, 0xe0, 0x0000,
						0x0000)) < 0) {
				PDEBUG (2, "spca50x_reg_write() failed");
				return err;
			}
			if ((err =
			     spca50x_reg_readwait (spca50x->dev, 0x06, 0,
						   0)) < 0) {
				PDEBUG (2, "spca50x_reg_readwait() failed");
				return err;
			}

			/* enable drop packet */
			if ((err =
			     spca50x_reg_write (spca50x->dev, 0x00, 0x850a,
						0x0001)) < 0) {
				PDEBUG (2, "failed to enable drop packet");
				return err;
			}
			err = spca50x_reg_write (spca50x->dev, 0x00, 0x8880, 3);
			if (err < 0) {
				PDEBUG (2, "spca50x_reg_write failed");
				return err;
			}
			err = spca50x_setup_qtable (spca50x,
						    0x00, 0x8800, 0x8840,
						    qtable_creative_pccam);
			if (err < 0) {
				PDEBUG (2, "spca50x_setup_qtable failed");
				return err;
			}
			
			/* set x multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8001,
	                  	GET_EXT_MODES(spca50x->bridge)[index][3]);

			/* set y multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8002,
	                  	GET_EXT_MODES(spca50x->bridge)[index][4]);

			/* use compressed mode, VGA, with mode specific subsample */
			spca50x_reg_write(spca50x->dev, 0, 0x8003, ((GET_EXT_MODES(spca50x->bridge)[index][2] & 0x0f)<<4));

			spca50x_reg_write(spca50x->dev, 0x20, 0x0001, 0x0004);

			/* switch to video camera mode */
			err = spca50x_reg_write(spca50x->dev, 0x00, 0x8000, 0x0004);
			if (err < 0)
			{
				PDEBUG(2, "spca50x_reg_write camera mode failed");
				return err;
			}

			if (spca50x_reg_readwait(spca50x->dev, 0, 0x8000, 0x44) != 0)
			{
				PDEBUG(2, "spca50x_reg_readwait() failed");
				return -EIO;
			}

			spca504B_reg_read(spca50x->dev,0 ,0 ,0x816b , &Data ,1);
			spca50x_reg_write(spca50x->dev, 0x00, 0x816b, Data);
	
			spca50x_write_vector(spca50x, spca500_visual_defaults);

			break;
		case KodakEZ200:	/* Kodak EZ200 */

			/* do a full reset */
			if ((err = spca500_full_reset (spca50x)) < 0) {
				PDEBUG (2, "spca500_full_reset failed");
				return err;
			}

			if ((err =
			     spca50x_reg_write (spca50x->dev, 0xe0, 0x0000,
						0x0000)) < 0) {
				PDEBUG (2, "spca50x_reg_write() failed");
				return err;
			}
			if ((err =
			     spca50x_reg_readwait (spca50x->dev, 0x06, 0,
						   0)) < 0) {
				PDEBUG (2, "spca50x_reg_readwait() failed");
				return err;
			}

			/* enable drop packet */
			if ((err =
			     spca50x_reg_write (spca50x->dev, 0x00, 0x850a,
						0x0001)) < 0) {
				PDEBUG (2, "failed to enable drop packet");
				return err;
			}
			err = spca50x_reg_write (spca50x->dev, 0x00, 0x8880, 0);
			if (err < 0) {
				PDEBUG (2, "spca50x_reg_write failed");
				return err;
			}
			err = spca50x_setup_qtable (spca50x,
						    0x00, 0x8800, 0x8840,
						    qtable_kodak_ez200);
			if (err < 0) {
				PDEBUG (2, "spca50x_setup_qtable failed");
				return err;
			}
			/* set x multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8001,
	                  		GET_EXT_MODES(spca50x->bridge)[index][3]);

			/* set y multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8002,
	                  		GET_EXT_MODES(spca50x->bridge)[index][4]);

			/* use compressed mode, VGA, with mode specific subsample */
			spca50x_reg_write(spca50x->dev, 0, 0x8003, ((GET_EXT_MODES(spca50x->bridge)[index][2] & 0x0f)<<4));

			spca50x_reg_write(spca50x->dev, 0x20, 0x0001, 0x0004);

			/* switch to video camera mode */
			err = spca50x_reg_write(spca50x->dev, 0x00, 0x8000, 0x0004);
			if (err < 0)
			{
				PDEBUG(2, "spca50x_reg_write camera mode failed");
				return err;
			}

			if (spca50x_reg_readwait(spca50x->dev, 0, 0x8000, 0x44) != 0)
			{
				PDEBUG(2, "spca50x_reg_readwait() failed");
				return -EIO;
			}

			spca504B_reg_read(spca50x->dev,0 ,0 ,0x816b , &Data ,1);
			spca50x_reg_write(spca50x->dev, 0x00, 0x816b, Data);
	
			spca50x_write_vector(spca50x, spca500_visual_defaults);

			break;
		
		case DLinkDSC350:	/* FamilyCam 300 */
		case AiptekPocketDV:	/* Aiptek PocketDV */


			spca500_reinit (spca50x);
			spca50x_reg_write (spca50x->dev, 0x00, 0x0d01, 0x01);
			/* enable drop packet */
			spca50x_reg_write (spca50x->dev, 0x00, 0x850a, 0x0001);

			err = spca50x_setup_qtable (spca50x,
						    0x00, 0x8800, 0x8840,
						    qtable_pocketdv);
			if (err < 0) {
				PDEBUG (2, "spca50x_setup_qtable failed");
				return err;
			}
			spca50x_reg_write (spca50x->dev, 0x00, 0x8880, 2);

			/* familycam Quicksmart pocketDV stuff */
			spca50x_reg_write (spca50x->dev, 0x00, 0x800a, 0x00);
			//Set agc transfer: synced inbetween frames
			spca50x_reg_write (spca50x->dev, 0x00, 0x820f, 0x01);
			//Init SDRAM - needed for SDRAM access
			spca50x_reg_write (spca50x->dev, 0x00, 0x870a, 0x04);
			
			
			/* set x multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8001,
	                  	GET_EXT_MODES(spca50x->bridge)[index][3]);

			/* set y multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8002,
	                  	GET_EXT_MODES(spca50x->bridge)[index][4]);

			/* use compressed mode, VGA, with mode specific subsample */
			spca50x_reg_write(spca50x->dev, 0, 0x8003, ((GET_EXT_MODES(spca50x->bridge)[index][2] & 0x0f)<<4));

			/* switch to video camera mode */
			err = spca50x_reg_write(spca50x->dev, 0x00, 0x8000, 0x0004);
		

			spca50x_reg_readwait(spca50x->dev, 0, 0x8000, 0x44);
		
			spca504B_reg_read(spca50x->dev,0 ,0 ,0x816b , &Data ,1);
			spca50x_reg_write(spca50x->dev, 0x00, 0x816b, Data);

			break;

		case LogitechClickSmart510:{

				spca50x_reg_write (spca50x->dev, 0x02, 0x00,
						   0x00);
				/* enable drop packet */
				if ((err =
				     spca50x_reg_write (spca50x->dev, 0x00,
							0x850a, 0x0001)) < 0) {
					PDEBUG (2,
						"failed to enable drop packet");
					return err;
				}

				err = spca50x_setup_qtable (spca50x,
							    0x00, 0x8800,
							    0x8840,
							    qtable_creative_pccam);
				if (err < 0) {
					PDEBUG (2,
						"spca50x_setup_qtable failed");
					return err;
				}
				err = spca50x_reg_write (spca50x->dev, 0x00,
							 0x8880, 3);
				if (err < 0) {
					PDEBUG (2, "spca50x_reg_write failed");
					return err;
				}
				spca50x_reg_write (spca50x->dev, 0x00, 0x800a,
						   0x00);
				//Init SDRAM - needed for SDRAM access
				spca50x_reg_write (spca50x->dev, 0x00, 0x870a,
						   0x04);
			/* set x multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8001,
	                  	GET_EXT_MODES(spca50x->bridge)[index][3]);

			/* set y multiplier */
			spca50x_reg_write(spca50x->dev, 0, 0x8002,
	                  	GET_EXT_MODES(spca50x->bridge)[index][4]);

			/* use compressed mode, VGA, with mode specific subsample */
			spca50x_reg_write(spca50x->dev, 0, 0x8003, ((GET_EXT_MODES(spca50x->bridge)[index][2] & 0x0f)<<4));


				/* switch to video camera mode */
				err = spca50x_reg_write (spca50x->dev, 0x00,
							 0x8000, 0x0004);


				spca50x_reg_readwait (spca50x->dev, 0, 0x8000,
						      0x44);

				spca504B_reg_read (spca50x->dev, 0, 0, 0x816b,
						   &Data, 1);
				spca50x_reg_write (spca50x->dev, 0x00, 0x816b,
						   Data);
				spca50x_write_vector (spca50x,
						      Clicksmart510_defaults);
			}

			break;

		default:
			return -EINVAL;
	}

	/* all ok */
	return 0;
}

static int spca500_full_reset(struct usb_spca50x *spca50x)
{
	int err;

	/* send the reset command */
	err = spca50x_reg_write(spca50x->dev, 0xe0, 0x0001, 0x0000);
	if (err < 0)
	{
		return err;
	}

	/* wait for the reset to complete */
	err = spca50x_reg_readwait(spca50x->dev, 0x06, 0x0000, 0x0000);
	if (err < 0)
	{
		return err;
	}

	/* all ok */
	return 0;
}

static int spca50x_setup_qtable(struct usb_spca50x *spca50x,
                                unsigned int        request,
                                unsigned int        ybase,
                                unsigned int        cbase,
                                unsigned char       qtable[2][64])
{
	int i;
	int err;

	/* loop over y components */
	for (i = 0; i < 64; i++)
	{
		err = spca50x_reg_write(spca50x->dev, request, ybase + i, qtable[0][i]);
		if (err < 0)
		{
			PDEBUG(2, "spca50x_reg_write failed");
			return err;
		}
	}

	/* loop over c components */
	for (i = 0; i < 64; i++)
	{
		err = spca50x_reg_write(spca50x->dev, request, cbase + i, qtable[1][i]);
		if (err < 0)
		{
			PDEBUG(2, "spca50x_reg_write failed");
			return err;
		}
	}

	/* all ok */
	return 0;
}

static void spca504_acknowledged_command(struct usb_spca50x *spca50x, 
				 __u16 reg,
				 __u16 idx,
				 __u16 val) 
{ 
	__u8 notdone = 0;

	spca50x_reg_write(spca50x->dev,reg,idx,val);
	notdone=spca50x_reg_read(spca50x->dev, 0x01, 0x0001, 1);


	//spca50x_reg_write(spca50x->dev,reg,idx,val);

	PDEBUG(5,"before wait 0x%x",notdone);
/* 
 * 	
 * 	set_current_state(TASK_INTERRUPTIBLE);
 * 	schedule_timeout (2*HZ/10);
 * 	
 */
	wait_ms(200);
	notdone=spca50x_reg_read(spca50x->dev, 0x01, 0x0001, 1);
	PDEBUG(5,"after wait 0x%x",notdone);
	
	return;
}

static void spca504A_acknowledged_command(struct usb_spca50x *spca50x, 
				 __u16 reg,
				 __u16 idx,
				 __u16 val,
				 __u8 stat,
				 __u8 count) 
{ 
	__u8  status ;
	__u8  endcode;
	

	spca50x_reg_write(spca50x->dev,reg,idx,val);
	status=spca50x_reg_read(spca50x->dev, 0x01, 0x0001, 1);
	endcode = stat ;
	PDEBUG(5,"Status 0x%x Need 0x%x",status, stat);
	if (count) {
		while (1){
			wait_ms(10);
			status=spca50x_reg_read(spca50x->dev, 0x01, 0x0001, 1);
			if (status == endcode) {
				 PDEBUG(5,"status 0x%x after wait 0x%x",status,count);
				break;
			}
			count++;
			if(count > 200) break;
		
		}
	}
	return;
}

static void spca504_pccam600_initialize(struct usb_spca50x *spca50x)
{
	int enable;
	if (spca50x->desc == LogitechClickSmart420) {
	spca50x_write_vector(spca50x, spca504A_clicksmart420_init_data);
	} else {
	spca50x_write_vector(spca50x, spca504_pccam600_init_data);
	}
	enable = (autoexpo ? 0x4 : 0x1);
	spca50x_reg_write(spca50x->dev, 0x0c, 0x0000, enable); // auto exposure
	spca50x_reg_write(spca50x->dev, 0xb0, 0x0000, enable); // auto whiteness
		
	/* set default exposure compensation and whiteness balance */
	spca50x_reg_write(spca50x->dev, 0x30, 0x0001, 800); // ~ 20 fps
	spca50x_reg_write(spca50x->dev, 0x30, 0x0002, 1600);
}


/* version based on the snoops of gsmart mini3 */
static void spca504_initialize(struct usb_spca50x *spca50x)
{
	__u8 i;
	__u8 info[6];

		for (i=0; i<6; i++)
		{
			info[i]=spca50x_reg_read_with_value(spca50x->dev,
			                                    0x20, i, 0x0000, 1);
		}
		PDEBUG(0, "Read info: %d %d %d %d %d %d . Should be 1,0,2,2,0,0\n",
		       info[0], info[1], info[2], info[3], info[4], info[5]);
		if (spca50x->desc == AiptekMiniPenCam13){
					/* spca504a aiptek */
			// Set AE AWB Banding Type 3-> 50Hz 2-> 60Hz 	       
			spca504A_acknowledged_command(spca50x, 0x24, 8, 3,0x9e,1);
			// Twice sequencial need status 0xff->0x9e->0x9d 
			spca504A_acknowledged_command(spca50x, 0x24, 8, 3,0x9e,0);

			spca504A_acknowledged_command(spca50x, 0x24, 0, 0,0x9d,1);
		} else {
			spca504_acknowledged_command(spca50x, 0x24, 8, 3);
			spca504_acknowledged_command(spca50x, 0x24, 8, 3);

			spca504_acknowledged_command(spca50x, 0x24, 0, 0);

		
		}
		
		
		spca50x_reg_write(spca50x->dev, 0x0, 0x270c, 0x5); // L92 sno1t.txt 

		spca50x_reg_write(spca50x->dev, 0x0, 0x2310, 0x5);

		spca50x_reg_write(spca50x->dev, 0x0, 0x21a7, 0x0000); /* brightness */
		spca50x_reg_write(spca50x->dev, 0x0, 0x21a8, 0x0020); /* contrast */
		spca50x_reg_write(spca50x->dev, 0x0, 0x21ad, 0x0000); /* hue */
		spca50x_reg_write(spca50x->dev, 0x0, 0x21ac, 0x0001); /* sat/hue */
		spca50x_reg_write(spca50x->dev, 0x0, 0x21ae, 0x0020); /* saturation */
		spca50x_reg_write(spca50x->dev, 0x0, 0x21a3, 0x0000); /* gamma */
	
}

static void spca504_wait_status(struct usb_spca50x *spca50x)
{
	int ret;

	/* FIXME: This is a potential infinite loop.
	 *        NWG: 18/05/2003.
	 */
	do {
		/* With this we get the status, when return 0 it's all ok */
		ret = spca50x_reg_read(spca50x->dev, 0x06, 0x00, 1);
	} while(ret != 0);
}


static void spca504B_reg_read(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) 
{	
	int rc;
	__u16 RegType;
	RegType = USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
		usb_rcvctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, buffer, length, HZ);
	if (buffer) { 
		PDEBUG(5, "reg read: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else { 
		PDEBUG(5, "reg read: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}


static void spca504B_reg_write(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) 
{	int rc;
	__u16 RegType;
	RegType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, buffer, length, HZ);
	if (buffer) { 
		PDEBUG(5, "reg write: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else { 
		PDEBUG(5, "reg write: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}


static void spca50x_GetFirmware(struct usb_spca50x *spca50x)
{	__u8 FW[5] = {0,0,0,0,0};
	__u8 ProductInfo[64];
	
	spca504B_reg_read(spca50x->dev,0x20 ,0 ,0 ,FW ,5);
	PDEBUG(0, "FirmWare : %d %d %d %d %d ", FW[0],FW[1],FW[2],FW[3],FW[4] );
	spca504B_reg_read(spca50x->dev,0x23 ,0 ,0 ,ProductInfo ,64);
	spca504B_reg_read(spca50x->dev,0x23 ,0 ,1 ,ProductInfo ,64);
	return ;
}


static int spca504B_PollingDataReady(struct usb_device *dev)		            
{	
	__u8 DataReady = 0 ;
	int count = 0;
	while(1) {
		spca504B_reg_read(dev,0x21,0, 0, &DataReady,1);
		if ( (DataReady & 0x01) == 0) break;
		wait_ms(10);
		count++;
		if (count > 10) break;
		
	}
	return DataReady;
}


static void spca504B_WaitCmdStatus(struct usb_spca50x *spca50x)
{	
	__u8 DataReady = 0;
	int ReqDone;
	int count = 0;
	while (1) {
		spca504B_reg_read(spca50x->dev,0x21,0, 1, &DataReady,1);

		if ( DataReady ) {
			DataReady = 0;
			spca504B_reg_write(spca50x->dev,0x21,0, 1, &DataReady, 1);		
			spca504B_reg_read(spca50x->dev,0x21,0, 1, &DataReady, 1);
			ReqDone = spca504B_PollingDataReady(spca50x->dev);
			break ;           
		}
		wait_ms (10);	
		count++;
		if (count > 50) break;
		
	} 
	return ;
}



static void spca504B_SetSizeType(struct usb_spca50x *spca50x , __u16 extSize ,__u16 extType)
{
	__u8 Size ;
	__u8 Type ;
	int rc;
	Size = (__u8) extSize;
	Type = (__u8) extType;
	
	if ( spca50x->bridge == BRIDGE_SPCA533) {
		spca504B_reg_write(spca50x->dev,0x31,0 ,0 ,NULL ,0 );
		spca504B_WaitCmdStatus(spca50x);
		rc = spca504B_PollingDataReady(spca50x->dev);
	
		spca50x_GetFirmware( spca50x );
		
		Type = 2;
		spca504B_reg_write(spca50x->dev,0x24,0 ,8 ,&Type ,1 );
		spca504B_reg_read(spca50x->dev,0x24,0 ,8 ,&Type ,1 );
	
		spca504B_reg_write(spca50x->dev,0x25,0 ,4 ,&Size ,1 );
		spca504B_reg_read(spca50x->dev,0x25,0 ,4 ,&Size ,1 );
		rc = spca504B_PollingDataReady(spca50x->dev );	
	
		/* Init the cam width height with some values get on init ?*/
		spca504B_reg_write(spca50x->dev,0x31,0 ,4 ,NULL ,0 );
		spca504B_WaitCmdStatus( spca50x );
		
				rc = spca504B_PollingDataReady(spca50x->dev );	
	
	}
	
	
	if ( spca50x->bridge == BRIDGE_SPCA504B){
		spca504B_reg_write(spca50x->dev,0x25,0 ,4 ,&Size ,1 );
		spca504B_reg_read(spca50x->dev,0x25,0 ,4 ,&Size ,1 );
		spca504B_reg_write(spca50x->dev,0x27,0 ,0 ,&Type ,1 );
		spca504B_reg_read(spca50x->dev,0x27,0 ,0 ,&Type ,1 );
		
		rc = spca504B_PollingDataReady ( spca50x->dev );
	}
	
	return ;
}


static void spca533_Megapix(struct usb_spca50x *spca50x )
{
	
	__u8 Type ;
	__u8 DataReady;
	
	
	if ( spca50x->bridge == BRIDGE_SPCA533) {
		
		spca50x_GetFirmware( spca50x );
		Type = 2;
		spca504B_reg_write(spca50x->dev,0x24,0 ,8 ,&Type ,1 );
		spca504B_reg_read(spca50x->dev,0x24,0 ,8 ,&Type ,1 );
		spca504B_reg_read(spca50x->dev,0x21,0, 0, &DataReady,1);
		spca504B_reg_write(spca50x->dev,0xF0 ,0 ,0 ,NULL ,0 );
		spca504B_WaitCmdStatus( spca50x );
		spca504B_reg_read(spca50x->dev,0xF0 ,0 ,4 ,NULL ,0 );
		spca504B_WaitCmdStatus( spca50x );	
	}
	
	return ;
}


static void spca504B_setQtable ( struct usb_spca50x *spca50x )
{
	__u8 Data = 3 ;
	int rc ;
	spca504B_reg_write(spca50x->dev,0x26,0 ,0 ,&Data ,1 );
	spca504B_reg_read(spca50x->dev,0x26 ,0 ,0 , &Data ,1);
	rc = spca504B_PollingDataReady ( spca50x->dev );
	return ;
}

static void spca504B_SetInitRegs(struct usb_spca50x *spca50x)
{	
	int rc ;	
	
	if (spca50x->bridge == BRIDGE_SPCA504B){
	/* Like snoops ? */
	// --------------------- perhaps not necessary -----------------------
	spca504B_reg_write(spca50x->dev,0x1d ,0 ,0 ,NULL,0 );	
	spca504B_reg_write(spca50x->dev,0,1 ,0x2306 ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0,0 ,0x0d04 ,NULL ,0 );

	// --------------------- but sometimes lost init-----------------------

	spca504B_reg_write(spca50x->dev,0,0 ,0x2000 ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0,0x13 ,0x2301 ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0,0 ,0x2306 ,NULL ,0 );
	
	}
	
	rc = spca504B_PollingDataReady ( spca50x->dev );
	spca50x_GetFirmware( spca50x );

	return ;
}


static void spca504B_SetPcCamStart (struct usb_spca50x *spca50x)
{
	int rc ;		
	spca504B_reg_write(spca50x->dev,0x31,0 ,4 ,NULL ,0 );
	spca504B_WaitCmdStatus( spca50x );
	rc = spca504B_PollingDataReady(spca50x->dev );
	return ;
}

static void spca504B_stop (struct usb_spca50x *spca50x)
{
	int rc ;
	spca504B_reg_write(spca50x->dev,0x31,0 ,0 ,NULL ,0 );
	spca504B_WaitCmdStatus(spca50x);
	rc = spca504B_PollingDataReady(spca50x->dev);
	return ;
}


static void spca504B_setContBrigHueRegisters (struct usb_spca50x *spca50x)
{	
	int rc ;
	spca504B_reg_write(spca50x->dev,0 ,0 ,0x21a7 ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0 ,0x20 ,0x21a8 ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0 ,0 ,0x21ad ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0 ,1 ,0x21ac ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0 ,0x20 ,0x21ae ,NULL ,0 );
	spca504B_reg_write(spca50x->dev,0 ,0 ,0x21a3 ,NULL ,0 );
	rc = spca504B_PollingDataReady( spca50x->dev );
	return ;
}

static void spca504B_init (struct usb_spca50x *spca50x)
{	
	spca504B_SetInitRegs( spca50x );
	if (spca50x->bridge == BRIDGE_SPCA504B){
		
	}
	
	return;
}

static void spca504B_start (struct usb_spca50x *spca50x)
{	
	spca504B_setContBrigHueRegisters ( spca50x );
	if (spca50x->bridge == BRIDGE_SPCA504B){
		spca504B_setQtable ( spca50x );
		spca504B_SetPcCamStart ( spca50x );
	}
	
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/****************************************************************************
 *  sysfs
 ***************************************************************************/

static inline struct usb_spca50x *cd_to_spca50x(struct class_device *cd)
{
	struct video_device *vdev = to_video_device(cd);
	return video_get_drvdata(vdev);
}

static ssize_t show_stream_id(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return sprintf(buf, "%d\n", spca50x->cameratype);
} 
static CLASS_DEVICE_ATTR(stream_id, S_IRUGO, show_stream_id, NULL);

static ssize_t show_model(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return sprintf(buf, "%s\n", (spca50x->desc) ?
	clist[spca50x->desc].description : " Unknow ");
} 
static CLASS_DEVICE_ATTR(model, S_IRUGO, show_model, NULL);
/*
static ssize_t show_bridge(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return sprintf(buf, "%s\n", symbolic(brglist, spca50x->bridge));
} 
static CLASS_DEVICE_ATTR(bridge, S_IRUGO, show_bridge, NULL);

static ssize_t show_sensor(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return sprintf(buf, "%s\n", symbolic(senlist, spca50x->sensor));
} 
static CLASS_DEVICE_ATTR(sensor, S_IRUGO, show_sensor, NULL);

static ssize_t show_brightness(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_brightness(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(brightness, S_IRUGO, show_brightness, NULL);

static ssize_t show_saturation(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_saturation(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(saturation, S_IRUGO, show_saturation, NULL);

static ssize_t show_contrast(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_contrast(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(contrast, S_IRUGO, show_contrast, NULL);

static ssize_t show_hue(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_hue(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(hue, S_IRUGO, show_hue, NULL);

static ssize_t show_exposure(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned char exp;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_exposure(spca50x, &exp);
	return sprintf(buf, "%d\n", exp >> 8);
} 
static CLASS_DEVICE_ATTR(exposure, S_IRUGO, show_exposure, NULL);
*/

static void spca50x_create_sysfs(struct video_device *vdev)
{ 

	video_device_create_file(vdev, &class_device_attr_stream_id);
	video_device_create_file(vdev, &class_device_attr_model);
/*
	video_device_create_file(vdev, &class_device_attr_bridge);
	video_device_create_file(vdev, &class_device_attr_sensor);

	video_device_create_file(vdev, &class_device_attr_brightness);
	video_device_create_file(vdev, &class_device_attr_saturation);
	video_device_create_file(vdev, &class_device_attr_contrast);
	video_device_create_file(vdev, &class_device_attr_hue);
	video_device_create_file(vdev, &class_device_attr_exposure);
*/
} 
#endif
/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/
static int
spcaDetectCamera (struct usb_spca50x *spca50x)
{
	struct usb_device *dev = spca50x->dev;

	/* Is it a recognised camera ? */
	switch (dev->descriptor.idVendor) {
		case 0x0733:	/* Rebadged ViewQuest (Intel) and ViewQuest cameras */
			switch (dev->descriptor.idProduct) {
				case 0x430:
					spca50x->desc = IntelPCCameraPro;
					spca50x->bridge = BRIDGE_SPCA505;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YYUV;
					info ("USB SPCA5XX camera found. Type Intel PC Camera Pro (SPCA505+SAA7113)");
					break;
				case 0x401:
					spca50x->desc = IntelCreateAndShare;
					spca50x->bridge = BRIDGE_SPCA501;	/* This is a guess. At least the chip looks closer to the 501 than the 505 */
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found. Type Intel Create and Share (SPCA501?+SAA7113?)");
					break;
				case 0x402:
					spca50x->desc = ViewQuestM318B;
					spca50x->bridge = BRIDGE_SPCA501;	/* This is a guess. At least the chip looks closer to the 501 than the 505 */
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found.  ViewQuest M318B (SPCA501a)");
					/*
					   Looks not perfectly but until we understand the difference
					   between spca501 and spca500 we'll treat them as one
					 */
					break;
				case 0x110:
					spca50x->desc = ViewQuestVQ110;
					spca50x->bridge = BRIDGE_SPCA508;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA508_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base =
						SPCA508_INDEX_I2C_BASE;
					spca50x->i2c_trigger_on_write = 1;
					spca50x->cameratype = YUVY;
					info ("USB SPCA5XX camera found. Type ViewQuest (SPCA508?+SAA7113?)");
					break;
				default:
					goto error;
			};
			break;
		case 0x99FA:	/* GrandTec cameras */
			switch (dev->descriptor.idProduct) {
				case 0x8988:
					spca50x->desc = GrandtecVcap;
					spca50x->bridge = BRIDGE_SPCA506;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YYUV;
					info ("USB SPCA5XX camera found. Grandtec V.cap (SPCA506+SAA7113?)");
					break;
				default:
					goto error;
			};
			break;
		case 0x0AF9:	/* Hama cameras */
			switch (dev->descriptor.idProduct) {
				case 0x0010:
					spca50x->desc = HamaUSBSightcam;
					spca50x->bridge = BRIDGE_SPCA508;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA508_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUVY;
					info ("USB SPCA5XX camera found. Hama Sightcam 100 (SPCA508A+PAS106B)");
					break;
				case 0x0011:
					spca50x->desc = HamaUSBSightcam2;
					spca50x->bridge = BRIDGE_SPCA508;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA508_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUVY;
					info ("USB SPCA5XX camera found. Hama Sightcam 100 (2) (SPCA508A+? )");
					break;
				default:
					goto error;
			};
			break;
		case 0x040A:	/* Kodak cameras */
			switch (dev->descriptor.idProduct) {
				case 0x0002:
					spca50x->desc = KodakDVC325;
					spca50x->bridge = BRIDGE_SPCA501;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found. Type Kodak DVC-325 (SPCA501A+SAA7113?)");
					break;
				case 0x0300:

					spca50x->desc = KodakEZ200;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Kodak EZ200 (SPCA500+unknown CCD)");
					break;


				default:
					goto error;
			};
			break;
		case 0x04a5:	/* Benq */
		case 0x08ca:	/* Aiptek */
		case 0x055f:	/* Mustek cameras */
		case 0x04fc:	/* SunPlus */
		case 0x052b:	/* ?? Megapix */
			switch (dev->descriptor.idProduct) {
				case 0xc520:
					spca50x->desc = MustekGsmartMini3;
					spca50x->bridge = BRIDGE_SPCA504;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Mustek gSmart Mini 3(SPCA504A)");
					break;
				case 0xc420:
					spca50x->desc = MustekGsmartMini2;
					spca50x->bridge = BRIDGE_SPCA504;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Mustek gSmart Mini 2(SPCA504A)");
					break;
				case 0x504a:

					spca50x->desc = AiptekMiniPenCam13;
					spca50x->bridge = BRIDGE_SPCA504;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Aiptek mini PenCam 1.3(SPCA504A)");
					break;

				case 0x2018:

					spca50x->desc = AiptekPenCamSD;
					spca50x->bridge = BRIDGE_SPCA504B;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PenCam SD(SPCA504A FW2)");
					break;

				case 0x2008:

					spca50x->desc = AiptekMiniPenCam2;
					spca50x->bridge = BRIDGE_SPCA504B;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PenCam 2M(SPCA504A FW2)");
					break;

				case 0x504b:

					spca50x->desc = MaxellMaxPocket;
					spca50x->bridge = BRIDGE_SPCA504B;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Maxell MaxPocket 1.3 (SPCA504A FW2)");
					break;

				case 0xffff:

					spca50x->desc = PureDigitalDakota;
					spca50x->bridge = BRIDGE_SPCA504B;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Pure Digital Dakota (SPCA504A FW2)");
					break;

				case 0x0103:

					spca50x->desc = AiptekPocketDV;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PocketDV");
					break;

				case 0x0104:

					spca50x->desc = AiptekPocketDVII;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PocketDVII 1.3Mp");
					break;

				case 0x0106:

					spca50x->desc = AiptekPocketDV3100;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PocketDV3100+");
					break;

				case 0x3008:

					spca50x->desc = BenqDC1500;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Benq DC 1500");
					break;
					
				case 0x3003:
				
					spca50x->desc = BenqDC1300;
					spca50x->bridge = BRIDGE_SPCA504B;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA50X_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Benq DC 1300");
					break;
					
				case 0x2010:

					spca50x->desc = AiptekPocketCam3M;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Aiptek PocketCam 3M");
					break;

				case 0x0561:
					spca50x->desc = Flexcam100Camera;
					spca50x->bridge = BRIDGE_SPCA561;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA561_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base =
						SPCA561_INDEX_I2C_BASE;
					spca50x->i2c_trigger_on_write = 1;
					spca50x->cameratype = GBRG;
					info ("USB SPCA5XX camera found. Type Flexcam 100 (SPCA561A)");
					break;
				case 0xc530:

					spca50x->desc = MustekGsmartLCD3;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Mustek Gsmart LCD 3");
					break;

				case 0xc430:

					spca50x->desc = MustekGsmartLCD2;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Mustek Gsmart LCD 2");
					break;
					
				case 0xc440:

					spca50x->desc = MustekDV3000;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. DV3000");
					break;

				case 0xc650:

					spca50x->desc = MustekMDC5500Z;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found Mustek MDC5500Z");
					break;

				case 0x1513:


					spca50x->desc = MegapixV4;
					spca50x->bridge = BRIDGE_SPCA533;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA533_OFFSET_DATA;
					spca50x->i2c_ctrl_reg = 0;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found Megapix V4");

					break;

				default:
					goto error;
			};
			break;
		case 0x046d:	/* Logitech */
		case 0x041E:	/* Creative cameras */
			switch (dev->descriptor.idProduct) {
				case 0x400A:

					spca50x->desc = CreativePCCam300;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Creative PC-CAM 300 (SPCA500+unknown CCD)");
					break;

				case 0x0901:

					spca50x->desc = LogitechClickSmart510;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Logitech ClickSmart 510 (SPCA500+unknown CCD)");
					break;

				case 0x400B:
					spca50x->desc = CreativePCCam600;
					spca50x->bridge =
						BRIDGE_SPCA504_PCCAM600;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA504_PCCAM600_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Creative PC-CAM 600 (SPCA504+unknown CCD)");
					break;
				case 0x0960:
					spca50x->desc = LogitechClickSmart420;
					spca50x->bridge =
						BRIDGE_SPCA504_PCCAM600;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA504_PCCAM600_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Logitech Clicksmart 420 (SPCA504+unknown CCD)");
					break;
				default:
					goto error;
			};
			break;
		case 0x084D:	/* D-Link / Minton */
			switch (dev->descriptor.idProduct) {
				case 0x0003:	/* DSC-350 / S-Cam F5 */

					spca50x->desc = DLinkDSC350;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type D-Link DSC-350 / Minton S-Cam F5 (SPCA500+unknown CCD)");
					break;

				default:
					goto error;
			};
			break;
		case 0x1776:	/* Arowana */
			switch (dev->descriptor.idProduct) {
				case 0x501c:	/* Arowana 300k CMOS Camera */
					spca50x->desc = Arowana300KCMOSCamera;
					spca50x->bridge = BRIDGE_SPCA501;
					spca50x->sensor = SENSOR_HV7131B;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found. Type Arowana 300k CMOS Camera (SPCA501C+HV7131B)");
					break;
				default:
					goto error;
			};
			break;
		case 0x0000:	/* Unknow Camera */
			switch (dev->descriptor.idProduct) {
				case 0x0000:	/* UnKnow from Ori CMOS Camera */
					spca50x->desc =
						MystFromOriUnknownCamera;
					spca50x->bridge = BRIDGE_SPCA501;
					spca50x->sensor = SENSOR_HV7131B;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found. UnKnow CMOS Camera (SPCA501C+HV7131B)");
					break;
				default:
					goto error;
			};
			break;
		case 0x8086:	/* Intel */
			switch (dev->descriptor.idProduct) {
				case 0x0110:
					spca50x->desc = IntelEasyPCCamera;
					spca50x->bridge = BRIDGE_SPCA508;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA508_OFFSET_DATA;
					/* FIXME: The following two settings are a guess.
					 *        Before, only i2c_ctrl_reg was specified,
					 *        and in fact was being set to
					 *        SPCA508_INDEX_I2C_BASE.
					 *        This setting needs to be checked.
					 *        NWG: Wed 12 Feb 2003.
					 */
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base =
						SPCA508_INDEX_I2C_BASE;
					spca50x->i2c_trigger_on_write = 1;
					spca50x->cameratype = YUVY;
					info ("USB SPCA5XX camera found. Type Intel Easy PC Camera CS110 (SPCA508+SAA7113?)");
					break;
				case 0x0630:	/* Pocket PC Camera */

					spca50x->desc = IntelPocketPCCamera;
					spca50x->bridge = BRIDGE_SPCA500;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA500_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = JPEG;
					info ("USB SPCA5XX camera found. Type Intel Pocket PC Camera (SPCA500+unknown CCD)");
					break;

				default:
					goto error;
			};
			break;
		case 0x0506:	/* 3COM cameras */
			switch (dev->descriptor.idProduct) {
				case 0x00DF:
					spca50x->desc = ThreeComHomeConnectLite;
					spca50x->bridge = BRIDGE_SPCA501;
					spca50x->sensor = SENSOR_SAA7113;
					spca50x->header_len =
						SPCA501_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base = 0;
					spca50x->i2c_trigger_on_write = 0;
					spca50x->cameratype = YUYV;
					info ("USB SPCA5XX camera found. Type 3Com HomeConnect Lite (SPCA501A+?)");
					break;
				default:
					goto error;
			};
			break;
		case 0x0458:	/* Genius KYE cameras */
			switch (dev->descriptor.idProduct) {
				case 0x7004:
					spca50x->desc = GeniusVideoCAMExpressV2;
					spca50x->bridge = BRIDGE_SPCA561;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA561_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base =
						SPCA561_INDEX_I2C_BASE;
					spca50x->i2c_trigger_on_write = 1;
					spca50x->cameratype = GBRG;
					info ("USB SPCA5XX camera found. Type Genius VideoCAM Express V2 (SPCA561A)");
					break;
				default:
					goto error;
			};
			break;
		case 0xabcd:	/* PetCam  */
			switch (dev->descriptor.idProduct) {
				case 0xcdee:
					spca50x->desc = PetCam;
					spca50x->bridge = BRIDGE_SPCA561;
					spca50x->sensor = SENSOR_INTERNAL;
					spca50x->header_len =
						SPCA561_OFFSET_DATA;
					spca50x->i2c_ctrl_reg =
						SPCA50X_REG_I2C_CTRL;
					spca50x->i2c_base =
						SPCA561_INDEX_I2C_BASE;
					spca50x->i2c_trigger_on_write = 1;
					spca50x->cameratype = GBRG;
					info ("USB SPCA5XX camera found. Type Petcam (SPCA561A)");
					break;
				default:
					goto error;
			};
			break;
		default:
			goto error;
	}
	return 0;
      error:
	return -ENODEV;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static int
spca50x_probe(struct usb_interface *intf, const struct usb_device_id *id)
#else
static void *
spca50x_probe (struct usb_device *dev, unsigned int ifnum,
	       const struct usb_device_id *id)
#endif
{
	struct usb_interface_descriptor *interface;
	struct usb_spca50x *spca50x;
	int err_probe;
	int i;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	struct usb_device *dev = interface_to_usbdev(intf);
#endif
	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		goto nodevice;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	/* Only work on Interface 0 */
	if (ifnum)
		goto nodevice;

	interface = &dev->actconfig->interface[ifnum].altsetting[0];
#else
	interface = &intf->altsetting[0].desc;
	if (interface->bInterfaceNumber)
		goto nodevice;
#endif
	/* Checking vendor/product should be enough, but what the hell */
	if (interface->bInterfaceClass != 0xFF)
		goto nodevice;
	if (interface->bInterfaceSubClass != 0x00)
		goto nodevice;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;
#endif
	if ((spca50x =
	     kmalloc (sizeof (struct usb_spca50x), GFP_KERNEL)) == NULL) {
		err ("couldn't kmalloc spca50x struct");
		goto error;
	}

	memset (spca50x, 0, sizeof (struct usb_spca50x));

	spca50x->dev = dev;
	spca50x->iface = interface->bInterfaceNumber;
	if ((err_probe = spcaDetectCamera (spca50x)) < 0) {
		err (" Devices not found !! ");
		goto error;
	}

	if (dumppix > 0)
		PDEBUG (1, "Using dumppix = %d", dumppix);
	for (i = 0; i < SPCA50X_NUMFRAMES; i++)
		init_waitqueue_head(&spca50x->frame[i].wq);
	init_waitqueue_head(&spca50x->wq);
	
	if (!spca50x_configure(spca50x)) {
		spca50x->user = 0;
		init_MUTEX (&spca50x->lock);	/* to 1 == available */
		init_MUTEX (&spca50x->buf_lock);
		spin_lock_init(&spca50x->lock_urb);
		spca50x->buf_state = BUF_NOT_ALLOCATED;
	} else {
		err ("Failed to configure camera");
		goto error;
	}
	/* Init video stuff */
 	 spca50x->vdev = video_device_alloc();
	if (!spca50x->vdev)
	   goto error;
	memcpy(spca50x->vdev, &spca50x_template, sizeof(spca50x_template));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	 spca50x->vdev->dev = &dev->dev;
#endif	
	video_set_drvdata(spca50x->vdev, spca50x);

	PDEBUG(2, "setting video device = %p, spca50x = %p",spca50x->vdev,spca50x);

	if (video_register_device(spca50x->vdev, VFL_TYPE_GRABBER,video_nr) < 0) {
		err("video_register_device failed");
		goto error;
	}

	/* Workaround for some applications that want data in RGB
	 * instead of BGR */
	if (spca50x->force_rgb)
		info ("data format set to RGB");
#ifdef CONFIG_PROC_FS
	create_proc_spca50x_cam ( spca50x );
#endif /* CONFIG_PROC_FS */ 
		
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	spca50x->task.func = auto_bh;
	usb_set_intfdata(intf, spca50x);
	spca50x_create_sysfs(spca50x->vdev);
	spca50x->present = 1;
#else
	spca50x->task.sync = 0;
	spca50x->task.routine = auto_bh;
#endif
	spca50x->task.data = spca50x;
	spca50x->bh_requested = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
	return spca50x;
#else
	return 0;
#endif

error:
	if(spca50x->vdev){
 		if (spca50x->vdev->minor == -1)
		video_device_release(spca50x->vdev);
	else
		video_unregister_device(spca50x->vdev);
 	spca50x->vdev = NULL;
	 }
	if (spca50x) {
		kfree (spca50x);
		spca50x = NULL;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
	return NULL;
#else
	return -EIO;
#endif

nodevice:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	return NULL;
#else
	return -ENODEV;
#endif
}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void
spca50x_disconnect(struct usb_interface *intf)

{
	struct usb_spca50x *spca50x = usb_get_intfdata(intf);
	dev_set_drvdata(&intf->dev, NULL);
#else
static void
spca50x_disconnect (struct usb_device *dev, void *ptr)
{
	struct usb_spca50x *spca50x = (struct usb_spca50x *) ptr;
#endif
	int n;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#endif

	spca50x->present = 0;
	
	PDEBUG(1, "Disconnect");
	
	
	if (!spca50x)
		return;
	/* We don't want people trying to open up the device */
	if (spca50x->vdev)
		video_unregister_device(spca50x->vdev);
	
	

	for (n = 0; n < SPCA50X_NUMFRAMES; n++)
		spca50x->frame[n].grabstate = FRAME_ABORTING;

	
	spca50x->curframe = -1;

	/* This will cause the process to request another frame */
	for (n = 0; n < SPCA50X_NUMFRAMES; n++)
		if (waitqueue_active(&spca50x->frame[n].wq))
			wake_up_interruptible(&spca50x->frame[n].wq);
	
	if (waitqueue_active(&spca50x->wq))
		wake_up_interruptible(&spca50x->wq);
	
	down(&spca50x->lock);
	
	spca50x->streaming = 0;

	/* Unschedule all of the iso td's */
	for (n = SPCA50X_NUMSBUF - 1; n >= 0; n--) {
		if (spca50x->sbuf[n].urb) {
			usb_unlink_urb(spca50x->sbuf[n].urb);
			usb_free_urb(spca50x->sbuf[n].urb);
			spca50x->sbuf[n].urb = NULL;
		}
	}
	PDEBUG(4, "Debug5");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	usb_driver_release_interface (&spca50x_driver,
				      &spca50x->dev->actconfig->
				      interface[spca50x->iface]);
#endif	
	spca50x->dev = NULL;
	up(&spca50x->lock);
#ifdef CONFIG_PROC_FS
	destroy_proc_spca50x_cam(spca50x);
#endif /* CONFIG_PROC_FS */

	/* Free the memory */
	if (spca50x && !spca50x->user) {
		spca50x_dealloc(spca50x, 1);
		kfree(spca50x);
		spca50x = NULL;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif
PDEBUG(3, "Disconnect complete");
}


static struct usb_driver spca50x_driver = {
	.owner =  THIS_MODULE,
	.name =   "spca50x",
	.id_table = device_table,
	.probe =  spca50x_probe,
	.disconnect = spca50x_disconnect
};



/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

static int __init usb_spca50x_init(void)
{
#ifdef CONFIG_PROC_FS
	proc_spca50x_create();
#endif /* CONFIG_PROC_FS */

	if (usb_register(&spca50x_driver) < 0)
		return -1;

	info("spca5xx kernel 2.6.x driver %s registered", version);

	return 0;
}

static void __exit usb_spca50x_exit(void)
{
	usb_deregister(&spca50x_driver);
	info("driver spca5xx 2.6 deregistered");

#ifdef CONFIG_PROC_FS
	proc_spca50x_destroy();
#endif /* CONFIG_PROC_FS */
}

module_init(usb_spca50x_init);
module_exit(usb_spca50x_exit);

//eof
